﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics; 

namespace Common.DataStructures
{
    public class CacheEntry<KEY> : IComparer<CacheEntry<KEY>>, IComparable, IEquatable<CacheEntry<KEY>>
    {
        /// <summary>
        /// Name of the file
        /// </summary>
        public readonly KEY Key;

        /// <summary>
        /// Last time file was accessed
        /// </summary>
        public DateTime LastAccessed;

        /// <summary>
        /// One way to clean the cache is to remove all entries that have not been used since a checkpoint.
        /// This flag can be set to indicate the entry was recently used.  When a checkpoint occurs we remove
        /// all entries that were not used and reset this flag to false for all entries that were used.
        /// </summary>
        public bool WasUsedSinceLastCheckpoint = true;

        /// <summary>
        /// Sometimes we want an entry to stay in the cache even if it is failing checkpoints.  Setting this 
        /// exempts the entry from failing checkpoints
        /// </summary>
        public bool CheckpointExempt = false; 

        /// <summary>
        /// Size of the file
        /// </summary>
        public Int64 Size;

        public CacheEntry(KEY key)
        {
            this.Key = key;
        }

        public CacheEntry(KEY key, DateTime lastAccessed, long size)
        {
            this.Key = key;
            LastAccessed = lastAccessed;
            Size = size; 
        }

        int? _FirstHashCode; 
        public override int GetHashCode()
        {
            if (_FirstHashCode.HasValue == false)
            {
                _FirstHashCode = Key.GetHashCode();
                return _FirstHashCode.Value;
            }
            else
            {
                Debug.Assert(Key.GetHashCode() == _FirstHashCode.Value); 
                return Key.GetHashCode();
            }
        }

        public override bool Equals(object obj)
        {
            CacheEntry<KEY> entry = obj as CacheEntry<KEY>;
            if (entry == null)
                return false;

            return entry.Key.Equals(this.Key);
        }

        #region IComparer<CacheEntry<KEY>> Members

        int IComparer<CacheEntry<KEY>>.Compare(CacheEntry<KEY> x, CacheEntry<KEY> y)
        {
            return x.LastAccessed.CompareTo(y.LastAccessed);
        }

        #endregion

        #region IComparable Members

        int IComparable.CompareTo(object obj)
        {
            CacheEntry<KEY> entry = obj as CacheEntry<KEY>;
            if (entry != null)
            {
                return LastAccessed.CompareTo(entry.LastAccessed);
            }

            return 0; 
        }

        #endregion

        #region IEquatable<CacheEntry<KEY>> Members

        bool IEquatable<CacheEntry<KEY>>.Equals(CacheEntry<KEY> other)
        {
            if (other == null)
                return false; 

            return this.Key.Equals(other.Key);
        }

        #endregion
    }
}
