﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent; 
using System.Linq;
using System.Text;
using System.Diagnostics; 

namespace Common.DataStructures
{
    /// <summary>
    /// A cache which prioritizes entries based upon last access time. 
    /// Users can call checkpoint which will call OnFailedCheckpoint for all entries not accessed since the last checkpoint
    /// </summary>
    /// <typeparam name="KEY">Key used to retrieve entries</typeparam>
    /// <typeparam name="CACHEENTRY">Data type of stores entries, derived from CacheEntry template</typeparam>
    /// <typeparam name="ADDTYPE">Type which is added to cache</typeparam>
    /// <typeparam name="FETCHTYPE">Type returned from cache</typeparam>
    abstract public class TimeQueueCache<KEY, CACHEENTRY, ADDTYPE, FETCHTYPE>
        where CACHEENTRY : CacheEntry<KEY>
        where FETCHTYPE : class
    {
        //        static public long MaxCacheSize = 17179869184;// = 2 << 34;
        public Int64 MaxCacheSize = 2 << 26; //128 MB

        protected Int64 TotalCacheSize = 0;
        
        protected ConcurrentDictionary<KEY, CACHEENTRY> dictEntries = new ConcurrentDictionary<KEY, CACHEENTRY>();

        abstract protected FETCHTYPE Fetch(CACHEENTRY key);

        /// <summary>
        /// The derived object should create a cache entry for the key/value pair.
        /// May return null if the entry cannot be created or if they derived object will update the cahce itself later,,
        /// perhaps if waiting for an asynch operation to complete
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        abstract protected CACHEENTRY CreateEntry(KEY key, ADDTYPE value);

        /// <summary>
        /// Retrieve an entry from the cache
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public virtual FETCHTYPE Fetch(KEY key)
        {
            FETCHTYPE value = default(FETCHTYPE);
            CACHEENTRY entry;

            bool success = dictEntries.TryGetValue(key, out entry);
            if(success == false)
                return default(FETCHTYPE); 
            
            //Record the fact that someone asked for this tile
            entry.WasUsedSinceLastCheckpoint = true;
            entry.LastAccessed = DateTime.Now;

            value = Fetch(entry);
            /*
            if(value != null)
            {
                //Update the cache entry
                dictEntries[key] = entry; 
            }
             */
            
            return value;
        }

        /// <summary>
        /// Creates a file for the texture passed.
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="textureStream"></param>
        public virtual bool Add(KEY key, ADDTYPE value)
        {

            CACHEENTRY entry = CreateEntry(key, value);
            if (entry == null)
                return false;

            return AddEntry(entry); 
        }

        /// <summary>
        /// This can be called by derived classes if they want to use asynch operations and wan't to delay adding an entry to 
        /// the cache after thier CreateEntry method is called.  I
        /// </summary>
        /// <param name="entry"></param>
        protected bool AddEntry(CACHEENTRY entry)
        {
            bool success = dictEntries.TryAdd(entry.Key, entry);

            if (success)
                ChangeCacheSize(entry.Size);

            return success; 
        }

        private void ChangeCacheSize(Int64 amount)
        {
            Int64 cacheSize;
            Int64 readcacheSize;

            do
            {
                cacheSize = this.TotalCacheSize;
                readcacheSize = System.Threading.Interlocked.CompareExchange(ref this.TotalCacheSize, cacheSize + amount, cacheSize);
            }
            while (readcacheSize != cacheSize); 
        }

        /// <summary>
        /// This should be called periodically to reduce the disk footprint
        /// </summary>
        public void ReduceCacheFootprint()
        {
            int RemoveCount = 0;
            int LostCount = 0;
            long FreedCount = 0; 
            if (dictEntries.IsEmpty)
                return; 

            List<CACHEENTRY> listEntries = dictEntries.Values.ToList<CACHEENTRY>();
            listEntries.Sort(); 

            while (TotalCacheSize > MaxCacheSize)
            {
                if (listEntries.Count == 0)
                    return;

                CACHEENTRY entry = listEntries[0];
                //Debug.Assert(dictEntries.ContainsKey(entry.Key));
                if (dictEntries.ContainsKey(entry.Key) == false)
                {
                    LostCount++; 
                }
                else
                {
                    FreedCount += entry.Size; 
                    RemoveEntry(entry);
                    RemoveCount++;
                }

                listEntries.RemoveAt(0); 
            }

            if (RemoveCount > 0 || LostCount > 0)
            {
                Trace.WriteLine(this.GetType().ToString() + " reduce cache size", "Cache");
                Trace.WriteLine("\tLost    " + LostCount.ToString() + " entries");
                Trace.WriteLine("\tRemoved " + RemoveCount.ToString() + " entries");
                Trace.WriteLine("\tFreed   " + FreedCount.ToString() + " bytes");
            }
        }

        /// <summary>
        /// Perform a checkpoint.  Call OnCheckpointFailed on each entry that does not pass
        /// </summary>
        public void Checkpoint()
        {
            CACHEENTRY[] EntryListCopy = dictEntries.Values.ToArray<CACHEENTRY>();
            
            int FailCount = 0; 

            //Walk the list and remove all entries who have not been accessed since the last checkpoint
            
            for (int iEntry = 0; iEntry < EntryListCopy.Length; iEntry++)
            {
                CACHEENTRY entry = EntryListCopy[iEntry];
                if (entry == null)
                    continue;

                if (entry == null)
                    continue;

                if (entry.WasUsedSinceLastCheckpoint)
                {
                    entry.WasUsedSinceLastCheckpoint = false;
                }
                else if(false == entry.CheckpointExempt)
                {
                    //Give derived classes a chance to handle this how they see fit
                    OnCheckpointFailed(entry);
                    FailCount++;
                }
                    
            }
        
            if(FailCount > 0)
                Trace.WriteLine(this.GetType().ToString() + " " + FailCount.ToString() + " entries failed checkpoint", "Cache"); 

            EntryListCopy = null;
        }

        /// <summary>
        /// Remove an entry from the cache, does not lock
        /// </summary>
        /// <param name="entry"></param>
        protected void RemoveEntry(CACHEENTRY entry)
        {
            long size = 0;
            bool CanRemoveEntry = true;
            if (entry != null)
            {
                size = entry.Size;
                CanRemoveEntry = OnRemoveEntry(entry);
            }

            if (CanRemoveEntry)
            {
                bool success = dictEntries.TryRemove(entry.Key, out entry);
                if (success)
                {
                    ChangeCacheSize(-size);
                    //TotalCacheSize -= size;
                }
            }
        }


        /// <summary>
        /// Called when the cache wants to remove an entry and provides an oppurtunity derived classes to cleanup
        /// Returns true if the entry was successfully cleaned up, otherwise false
        /// </summary>
        /// <param name="entry"></param>
        protected virtual bool OnRemoveEntry(CACHEENTRY entry)
        {
            return true;
        }

        /// <summary>
        /// Called when a cache entry has not been used since the last checkpoint
        /// </summary>
        /// <param name="entry"></param>
        protected virtual void OnCheckpointFailed(CACHEENTRY entry)
        {

        }
    }
}
