﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading; 
using System.Diagnostics; 

namespace Geometry
{
    class AddTransformThreadObj : IDisposable
    {
        public ManualResetEvent DoneEvent = new ManualResetEvent(false);

        /// <summary>
        /// The output returned after processing
        /// </summary>
        public MappingGridVector2[] newPoints;

        readonly int[] iPoints; 
        readonly GridTransform warpingTransform; 
        readonly GridTransform fixedTransform;
       
        public AddTransformThreadObj(int[] iMapPoints, GridTransform warpingT, GridTransform fixedT)
        {
            this.iPoints = iMapPoints; 
            this.warpingTransform = warpingT;
            this.fixedTransform = fixedT; 
        }

        public void ThreadPoolCallback(Object threadContext)
        {
            List<MappingGridVector2> newPointsList = new List<MappingGridVector2>(iPoints.Length); 

            foreach(int iPoint in iPoints)
            {
                MappingGridVector2 UnmappedPoint = warpingTransform.mapPoints[iPoint]; 
                MappingGridTriangle mapTri = fixedTransform.GetTransform(UnmappedPoint.ControlPoint);

                if (mapTri != null)
                {
                    GridVector2 newControl = mapTri.Transform(UnmappedPoint.ControlPoint);
                    newPointsList.Add( new MappingGridVector2(newControl, UnmappedPoint.MappedPoint) );
                }
                else
                {
                    //In this case we need to test each edge connecting this point to other points.
                    List<int> MovingEdgeIndicies = warpingTransform.GetEdges(iPoint);

                    
                    //Find out which of these edge points intersect triangles in the fixed warp.  If they are inside the control warp
                    //triangle mesh we find the point where the edge intersects the fixed warp mesh.
                    for (int iEdge = 0; iEdge < MovingEdgeIndicies.Count; iEdge++)
                    {
                        int iEdgePoint = MovingEdgeIndicies[iEdge]; 
                        
                        GridLineSegment ctrlLine = new GridLineSegment(UnmappedPoint.ControlPoint, this.warpingTransform.mapPoints[iEdgePoint].ControlPoint);
                        GridLineSegment mapLine = new GridLineSegment(UnmappedPoint.MappedPoint, this.warpingTransform.mapPoints[iEdgePoint].MappedPoint);

                        GridLineSegment foundCtrlLine; //Control line found in nearest line call
                        GridLineSegment foundMapLine; //Corresponding map line found in nearest line call
                        GridVector2 intersect;

                        //Find out if there is a line in the fixed transform we intersect with. 
                        double distance = fixedTransform.NearestLine(ctrlLine, out foundCtrlLine, out foundMapLine, out intersect);
                        if (distance == double.MaxValue)
                            continue;

                        if (fixedTransform.GetTransform(this.warpingTransform.mapPoints[iEdgePoint].ControlPoint) == null)
                            continue; 

                        //Find out our new position in control space
                        /*
                        GridVector2 newCtrlPoint = ctrlLine.Direction;
                        newCtrlPoint.Scale(distance);
                        newCtrlPoint = newCtrlPoint + ctrlLine.A; 
                        */

                        //Translate from the fixed transform map space into control space. 
                        
                        double mapLineFraction = GridVector2.Distance(foundMapLine.A, intersect) / foundMapLine.Length;
                        double ctrlLineDistance = foundCtrlLine.Length * mapLineFraction;
                        GridVector2 newCtrlPoint = foundCtrlLine.Direction; //Get unit vector describing direction and scale it
                        newCtrlPoint.Scale(ctrlLineDistance);
                        newCtrlPoint = newCtrlPoint + foundCtrlLine.A; 
                        
                        //Figure out where the transformed point lies in the moving transform mapped space. 
                        double fraction = distance / ctrlLine.Length;

                        //Debug.Assert(fraction <= 1.0 && fraction >= 0.0);
                        if (fraction > 1f)
                            fraction = 1f;
                        else if (fraction < 0f)
                            fraction = 0f; 

                        double mappedDistance = mapLine.Length * fraction;

                        GridVector2 newMapPoint = mapLine.Direction;
                        newMapPoint.Scale(mappedDistance);
                        newMapPoint = newMapPoint + mapLine.A;

                        newPointsList.Add(new MappingGridVector2(newCtrlPoint, newMapPoint));                      
                    }

                    
                }
            }

            newPoints = newPointsList.ToArray(); 
            DoneEvent.Set(); 
        }
        /*
        public void ThreadPoolCallback(Object threadContext)
        {
            /*
            Triangle control = UnmappedTri.Control;

            Vector2 p1 = control.p1;
            Vector2 p2 = control.p2;
            Vector2 p3 = control.p3;

            MappingTriangle m1 = transform.GetTransform(p1);
            MappingTriangle m2 = transform.GetTransform(p2);
            MappingTriangle m3 = transform.GetTransform(p3);

            int numNull = 0;
            if (m1 == null)
                numNull++;
            if (m2 == null)
                numNull++;
            if (m3 == null)
                numNull++;

            
            switch (numNull)
            {
                case 0:
                    control.p1 = m1.Transform(p1);
                    control.p2 = m2.Transform(p2);
                    control.p3 = m3.Transform(p3);

                    newMaps = new MappingTriangle[] { new MappingTriangle(control, UnmappedTri.Mapped)};
                    break;

                case 1:
                    newMaps = transform.FitTriangleOneExternalVertex(UnmappedTri);
 //                   mappings.RemoveAt(iTri);
 //                   mappings.InsertRange(iTri, newMaps);
 //                   iTri += newMaps.Length;
                    break;
                case 2:
                    newMaps = transform.FitTriangleTwoExternalVertex(UnmappedTri);
 //                   mappings.RemoveAt(iTri);
 //                   mappings.InsertRange(iTri, newMaps);
 //                   iTri += newMaps.Length;
                    break;
                case 3:
                    //Triangle is removed, so newMaps remains empty
                    newMaps = new MappingTriangle[0]; 
 //                   mappings.RemoveAt(iTri);
                    break;
            }
            
            DoneEvent.Set(); 
        }
    */

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            this.DoneEvent.Close();
        }

        #endregion
    }
}
