﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace Geometry
{
    /// <summary>
    /// The line search grid divides space into a regular grid.  Each line segment defines a rectangle.  If the rectangle intersects
    /// a grid cell a pointer is added to that cell.  When a point is passed in the lines in the cell containing the point and all adjacent cells
    /// are checked for intersection.  The matching line and object are returned
    /// </summary>
    public class LineSearchGrid<T>
    {
        struct Coord
        {
            public int iX;
            public int iY;

            public Coord(int ix, int iy)
            {
                this.iX = ix;
                this.iY = iy;
            }

            public override string ToString()
            {
                return iX.ToString() + "," + iY.ToString();
            }

            public override bool Equals(object obj)
            {
                Coord coord = (Coord)obj;
                return coord.iX == iX && coord.iY == iY;
            }
        }

        GridRectangle Bounds;
        List<GridLineSegment>[,] _LineGrid = new List<GridLineSegment>[0, 0];

        Dictionary<GridLineSegment, T> tableLineToValue;
        Dictionary<T, GridLineSegment> tableValueToLine;

        double GridWidth;
        double GridHeight;

        int NumGridsX;
        int NumGridsY;

        int EstimatedLinesPerCell;

        public int Count { get { return tableLineToValue.Count; } }

        System.Threading.ReaderWriterLockSlim rwLock = new System.Threading.ReaderWriterLockSlim();

        public LineSearchGrid(GridRectangle bounds, int EstimatedLineCount)
        {
            if (EstimatedLineCount <= 1)
            {
                EstimatedLineCount = 1000;
            }

            this.Bounds = bounds;
            tableLineToValue = new Dictionary<GridLineSegment, T>(EstimatedLineCount / 10);
            tableValueToLine = new Dictionary<T, GridLineSegment>(EstimatedLineCount / 10);

            //Calculate number of grid cells based on num points and boundaries
            double NumGrids = Math.Ceiling(System.Math.Sqrt(EstimatedLineCount));
            int NumGridsEachDimension = (int)Math.Ceiling(Math.Sqrt(NumGrids));

            this.GridWidth = bounds.Width / NumGridsEachDimension;
            this.GridHeight = bounds.Width / NumGridsEachDimension;

            if (GridWidth < GridHeight)
                GridWidth = GridHeight;
            else
                GridHeight = GridWidth;

            NumGridsX = (int)Math.Ceiling(bounds.Width / GridWidth);
            NumGridsY = (int)Math.Ceiling(bounds.Height / GridHeight);

            this.EstimatedLinesPerCell = (int)((double)EstimatedLineCount / NumGrids);

            _LineGrid = new List<GridLineSegment>[NumGridsX + 1, NumGridsY + 1];
            //Initialize the grid
            for (int iX = 0; iX < NumGridsX + 1; iX++)
            {
                for (int iY = 0; iY < NumGridsY + 1; iY++)
                {
                    _LineGrid[iX, iY] = new List<GridLineSegment>(this.EstimatedLinesPerCell);
                }
            }
        }

        #region LineSearchGridCoordListEnumerator

        /// <summary>
        /// Enumerates over a range of cells, only returning unique values.
        /// We use these because returning a list would copy a massive amount of memory
        /// </summary>
        /// <typeparam name="T"></typeparam>
        private class LineSearchGridCoordListEnumerator<T> : IEnumerable<GridLineSegment>, IEnumerator<GridLineSegment>
        {
            private LineSearchGrid<T> searchGrid;
            private List<GridLineSegment> currentCell;
            private IEnumerable<Coord> coords;
            private IEnumerator<Coord> coordEnum;
            private int iGridIndex = -1;

            //Only return unique values
            SortedSet<GridLineSegment> UniqueLines = new SortedSet<GridLineSegment>();

            public LineSearchGridCoordListEnumerator(LineSearchGrid<T> SearchGrid, IEnumerable<Coord> Coords)
            {
                this.searchGrid = SearchGrid;
                this.coords = Coords;
                Reset();
            }

            public GridLineSegment Current
            {
                get
                {
                    GridLineSegment segment = currentCell[iGridIndex];
                    Debug.Assert(UniqueLines.Contains(segment) == false);
                    UniqueLines.Add(segment);
                    return segment;
                }
            }

            object System.Collections.IEnumerator.Current
            {
                get
                {
                    GridLineSegment segment = currentCell[iGridIndex];
                    Debug.Assert(UniqueLines.Contains(segment) == false);
                    UniqueLines.Add(segment);
                    return segment;
                }
            }

            public bool MoveNext()
            {
                iGridIndex++;
                if (currentCell == null)
                {
                    bool success = coordEnum.MoveNext();
                    if (!success)
                        return false;

                    currentCell = searchGrid._LineGrid[coordEnum.Current.iX, coordEnum.Current.iY];
                }

                while (iGridIndex >= currentCell.Count || UniqueLines.Contains(currentCell[iGridIndex]))
                {
                    //Figure out if we are advancing because of a repeat value or grid index rollover
                    if (iGridIndex >= currentCell.Count)
                    {
                        iGridIndex = 0;
                        bool success = coordEnum.MoveNext();
                        if (!success)
                            return false;

                        currentCell = searchGrid._LineGrid[coordEnum.Current.iX, coordEnum.Current.iY];

                    }
                    else //we advanced because of a repeat value, increment iGridIndex
                    {
                        iGridIndex++;
                    }
                }

                return true;
            }

            public void Reset()
            {
                UniqueLines.Clear();
                iGridIndex = -1; //-1 because MoveNext is called before the first value is read
                coordEnum = coords.GetEnumerator();
            }

            public void Dispose()
            {
                return;
            }

            public IEnumerator<GridLineSegment> GetEnumerator()
            {
                this.Reset();
                return this;
            }

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                this.Reset();
                return this;
            }
        }

    #endregion

        #region LineSearchGridRectangleEnumerator

        /// <summary>
        /// Enumerates over a range of cells, only returning unique values.
        /// We use these because returning a list would copy a massive amount of memory
        /// </summary>
        /// <typeparam name="T"></typeparam>
        private class LineSearchGridRectangleEnumerator<T> : IEnumerable<GridLineSegment>, IEnumerator<GridLineSegment>
        {
            private LineSearchGrid<T> searchGrid;
            private List<GridLineSegment> currentCell;
            private Coord start;
            private Coord end;
            private int iX = 0;
            private int iY = 0;
            private int iGridIndex = -1;

            //Only return unique values
            SortedSet<GridLineSegment> UniqueLines = new SortedSet<GridLineSegment>();

            public LineSearchGridRectangleEnumerator(LineSearchGrid<T> SearchGrid, Coord Start, Coord End)
            {
                this.searchGrid = SearchGrid;
                this.start = Start;
                this.end = End;
                Reset();

            }

            public GridLineSegment Current
            {
                get
                {
                    GridLineSegment segment = currentCell[iGridIndex];
                    Debug.Assert(UniqueLines.Contains(segment) == false);
                    UniqueLines.Add(segment);
                    return segment;
                }
            }

            object System.Collections.IEnumerator.Current
            {
                get
                {
                    GridLineSegment segment = currentCell[iGridIndex];
                    Debug.Assert(UniqueLines.Contains(segment) == false);
                    UniqueLines.Add(segment);
                    return segment;
                }
            }

            public bool MoveNext()
            {
                iGridIndex++;
                while (iGridIndex >= currentCell.Count || UniqueLines.Contains(currentCell[iGridIndex]))
                {
                    //Figure out if we are advancing because of a repeat value or grid index rollover
                    if (iGridIndex >= currentCell.Count)
                    {
                        iGridIndex = 0;
                        iX++;
                        if (iX > end.iX)
                        {
                            iX = start.iX;
                            iY++;
                            if (iY > end.iY)
                            {
                                return false;
                            }
                            else
                            {
                                currentCell = searchGrid._LineGrid[iX, iY];
                            }
                        }
                        else
                        {
                            currentCell = searchGrid._LineGrid[iX, iY];
                        }
                    }
                    else //we advanced because of a repeat value, increment iGridIndex
                    {
                        iGridIndex++;
                    }
                }

                return true;
            }

            public void Reset()
            {
                UniqueLines.Clear();
                iX = start.iX;
                iY = start.iY;
                iGridIndex = -1; //-1 because MoveNext is called before the first value is read
                currentCell = searchGrid._LineGrid[iX, iY];
            }

            public void Dispose()
            {
                return;
            }

            public IEnumerator<GridLineSegment> GetEnumerator()
            {
                this.Reset();
                return this;
            }

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                this.Reset();
                return this;
            }
        }

#endregion

        public void Clear()
        {
            try
            {
                rwLock.EnterWriteLock();

                tableLineToValue.Clear();

                for (int iX = 0; iX < NumGridsX; iX++)
                {
                    for (int iY = 0; iY < NumGridsY; iY++)
                    {
                        _LineGrid[iX, iY].Clear();
                    }
                }
            }
            finally
            {
                rwLock.ExitWriteLock();
            }
        }

        public bool Contains(GridLineSegment line)
        {
            try
            {
                rwLock.EnterReadLock();
                return tableLineToValue.ContainsKey(line);
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }

        public bool Contains(T key)
        {
            try
            {
                rwLock.EnterReadLock();
                return tableValueToLine.ContainsKey(key);
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }

        public GridLineSegment[] Lines
        {
            get { return tableLineToValue.Keys.ToArray(); }
        }

        public T[] Values
        {
            get { return tableValueToLine.Keys.ToArray(); }
        }


        public void Add(GridLineSegment line, T value)
        {
            try
            {
                rwLock.EnterWriteLock();

                Debug.Assert(tableLineToValue.ContainsKey(line) == false);
                if (tableLineToValue.ContainsKey(line))
                {
                    tableLineToValue[line] = value;
                    return;
                }

                IEnumerable<Coord> coords = GetCoordsForLine(line);
                foreach (Coord coord in coords)
                {
                    List<GridLineSegment> lines = _LineGrid[coord.iX, coord.iY];
                    Debug.Assert(lines.Contains(line) == false);
                    lines.Add(line);
                }

                tableLineToValue.Add(line, value);
                tableValueToLine.Add(value, line);
            }
            finally
            {
                rwLock.ExitWriteLock();
            }
        }

        

        public bool TryAdd(GridLineSegment line, T value)
        {
            try
            {
                rwLock.EnterUpgradeableReadLock();

                if (tableValueToLine.ContainsKey(value))
                    return false;

                try
                {
                    rwLock.EnterWriteLock();

                    if (tableValueToLine.ContainsKey(value))
                        return false;

                    //Add the line to all cells it belongs in 
                    IEnumerable<Coord> coords = GetCoordsForLine(line);
                    foreach (Coord coord in coords)
                    {
                        List<GridLineSegment> lines = _LineGrid[coord.iX, coord.iY];
                        Debug.Assert(lines.Contains(line) == false);
                        lines.Add(line);
                    }

                    Debug.Assert(tableLineToValue.ContainsKey(line) == false);
                    tableLineToValue.Add(line, value);
                    tableValueToLine.Add(value, line);
                }
                finally
                {
                    rwLock.ExitWriteLock();
                }
            }
            finally
            {
                rwLock.ExitUpgradeableReadLock();
            }

            return true;

        }

        public void Remove(GridLineSegment line)
        {
            try
            {
                rwLock.EnterWriteLock();

                //Add the line to all cells it belongs in 
                IEnumerable<Coord> coords = GetCoordsForLine(line);
                foreach (Coord coord in coords)
                {
                    List<GridLineSegment> lines = _LineGrid[coord.iX, coord.iY];
                    Debug.Assert(lines.Contains(line));
                    lines.Remove(line);
                }

               T value = tableLineToValue[line];
                bool LineRemove = tableLineToValue.Remove(line);
                bool ValueRemove = tableValueToLine.Remove(value);
                Debug.Assert(LineRemove == ValueRemove);
            }
            finally
            {
                rwLock.ExitWriteLock();
            }
        }

        public bool TryRemove(GridLineSegment line, out T value)
        {
            value = default(T);
            try
            {
                rwLock.EnterUpgradeableReadLock();
                if (!tableLineToValue.ContainsKey(line))
                    return false;

                try
                {
                    rwLock.EnterWriteLock();

                    //Add the line to all cells it belongs in 
                    IEnumerable<Coord> coords = GetCoordsForLine(line);
                    foreach (Coord coord in coords)
                    {
                        List<GridLineSegment> lines = _LineGrid[coord.iX, coord.iY];
                        Debug.Assert(lines.Contains(line));
                        lines.Remove(line);
                    }

                    value = tableLineToValue[line];
                    bool LineRemove = tableLineToValue.Remove(line);
                    bool ValueRemove = tableValueToLine.Remove(value);
                    Debug.Assert(LineRemove == ValueRemove);
                    return true;
                }
                finally
                {
                    rwLock.ExitWriteLock();
                }
            }
            finally
            {
                rwLock.ExitUpgradeableReadLock();
            }
        }

        public bool TryRemove(T value, out GridLineSegment OldLine)
        {
            try
            {

                rwLock.EnterUpgradeableReadLock();

                if (!tableValueToLine.ContainsKey(value))
                {
                    OldLine = new GridLineSegment();
                    return false;
                }

                OldLine = tableValueToLine[value];
                try
                {
                    rwLock.EnterWriteLock();

                    if (!tableValueToLine.ContainsKey(value))
                        return false;

                    //Add the line to all cells it belongs in 
                    IEnumerable<Coord> coords = GetCoordsForLine(OldLine);
                    foreach (Coord coord in coords)
                    {
                        List<GridLineSegment> lines = _LineGrid[coord.iX, coord.iY];
                        Debug.Assert(lines.Contains(OldLine));
                        lines.Remove(OldLine);
                    }

                    bool LineRemove = tableLineToValue.Remove(OldLine);
                    bool ValueRemove = tableValueToLine.Remove(value);
                    Debug.Assert(LineRemove == ValueRemove);
                    return true;
                }
                finally
                {
                    rwLock.ExitWriteLock();
                }
            }
            finally
            {
                rwLock.ExitUpgradeableReadLock();
            }
        }

        private Coord GetCoord(GridVector2 position)
        {
            return GetCoord(position.X, position.Y);
        }

        private Coord GetCoord(double x, double y)
        {
            Coord coord;
            coord.iX = (int)((x - Bounds.Left) / this.GridWidth);
            coord.iY = (int)((y - Bounds.Bottom) / this.GridHeight);
            if (coord.iX < 0)
                coord.iX = 0;
            if (coord.iY < 0)
                coord.iY = 0;
            if (coord.iX > NumGridsX)
                coord.iX = NumGridsX;
            if (coord.iY > NumGridsY)
                coord.iY = NumGridsY;

            return coord;
        }

        private IEnumerable<Coord> GetCoordsForLine(GridLineSegment line)
        {
            Coord start;
            Coord end;
            if (line.A.X < line.B.X)
            {
                start = GetCoord(line.A);
                end = GetCoord(line.B);
            }
            else
            {
                start = GetCoord(line.B);
                end = GetCoord(line.A);
            }

            List<Coord> listCoords = new List<Coord>((end.iX - start.iX) * Math.Abs(end.iY - start.iY));
            if (start.iX == end.iX)
            {
                for (int iY = start.iY; iY <= end.iY; iY++)
                {
                    listCoords.Add(new Coord(start.iX, iY));
                }
            }
            else if (start.iY == end.iY)
            {
                for (int iX = start.iX; iX <= end.iX; iX++)
                {
                    listCoords.Add(new Coord(iX, start.iY));
                }
            }
            else
            {
                //Figure out the line function
                double m = line.slope;
                double b = line.intercept;

                listCoords.Add(start);

                int iLastY = start.iY;
                //Add the line to all cells it belongs in 
                for (int iX = start.iX; iX <= end.iX; iX++)
                {
                    double x = ((iX + 1) * this.GridWidth) + Bounds.Left;
                    double y = m * x + b;

                    Coord intersect = GetCoord(x, y);

                    for (int iY = iLastY; iY <= intersect.iY; iY++)
                    {
                        if (iX == end.iX)
                        {
                            if (m > 0)
                            {
                                if (iY > end.iY)
                                    break;
                            }
                            else if (iY < end.iY)
                            {
                                break;
                            }
                        }

                        if ((iX != start.iX || iY != start.iY) &&
                           (iX != end.iX || iY != end.iY))
                            listCoords.Add(new Coord(iX, iY));

                    }

                    iLastY = intersect.iY;
                }

                listCoords.Add(end);
            }

            return listCoords;
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        private List<GridLineSegment> GetPotentialIntersections(GridVector2 position, int SearchSize)
        {
            Coord coord = GetCoord(position);
            List<GridLineSegment> LineList = new List<GridLineSegment>(this.EstimatedLinesPerCell);
            Coord start;
            Coord end;

            start.iX = coord.iX - SearchSize;
            start.iY = coord.iY - SearchSize;
            end.iX = coord.iX + SearchSize;
            end.iY = coord.iY + SearchSize;

            if (start.iX < 0)
                start.iX = 0;
            if (start.iY < 0)
                start.iY = 0;
            if (end.iX > NumGridsX)
                end.iX = NumGridsX;
            if (end.iY > NumGridsY)
                end.iY = NumGridsY;


            //Add the line to all cells it belongs in 
            for (int iX = start.iX; iX <= end.iX; iX++)
            {
                for (int iY = start.iY; iY <= end.iY; iY++)
                {
                    LineList.AddRange(_LineGrid[iX, iY]);
                }
            }

            //Only return unique values
            List<GridLineSegment> UniqueLines = new List<GridLineSegment>(LineList.Count);
            LineList.Sort();
            GridLineSegment lastLine = default(GridLineSegment);
            foreach (GridLineSegment item in LineList)
            {
                if (item == lastLine)
                    continue;
                else
                {
                    UniqueLines.Add(item);
                    lastLine = item;
                }
            }

            return UniqueLines;
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        public IEnumerable<GridLineSegment> GetPotentialIntersections(GridLineSegment line)
        {
            return GetPotentialIntersections(line, true);
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// Called internally.  When the spinlock is held set TakeSpinLock to false
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        private IEnumerable<GridLineSegment> GetPotentialIntersections(GridLineSegment line, bool TakeSpinLock)
        {
            try
            {
                if (TakeSpinLock)
                    rwLock.EnterReadLock();

                List<GridLineSegment> LineList = new List<GridLineSegment>();
                Coord start = GetCoord(new GridVector2(line.MinX, line.MinY));
                Coord end = GetCoord(new GridVector2(line.MaxX, line.MaxY));

                IEnumerable<Coord> coords = GetCoordsForLine(line);
                return new LineSearchGridCoordListEnumerator<T>(this, coords);
            }
            finally
            {
                if (TakeSpinLock)
                    rwLock.ExitReadLock();
            }
        }

        public T this[GridLineSegment key]
        {
            get { return tableLineToValue[key]; }
        }

        public GridLineSegment this[T key]
        {
            get { return tableValueToLine[key]; }
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        public T[] GetValues(GridLineSegment line)
        {
            try
            {
                rwLock.EnterReadLock();

                IEnumerable<GridLineSegment> LineList = GetPotentialIntersections(line, false);
                List<T> values = new List<T>();
                foreach (GridLineSegment gridLine in LineList)
                {
                    values.Add(tableLineToValue[gridLine]);
                }

                // Trace.WriteLine("Enumerator: " + values.Count.ToString()); 
                return values.ToArray();
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        public T[] GetValues(GridRectangle rect)
        {
            try
            {
                Coord start = GetCoord(new GridVector2(rect.Left, rect.Bottom));
                Coord end = GetCoord(new GridVector2(rect.Right, rect.Top));

                rwLock.EnterReadLock();

                IEnumerable<GridLineSegment> LineList = new LineSearchGridRectangleEnumerator<T>(this, start, end);
                List<T> values = new List<T>();
                foreach (GridLineSegment gridLine in LineList)
                {
                    values.Add(tableLineToValue[gridLine]);
                }

                // Trace.WriteLine("Enumerator: " + values.Count.ToString()); 
                return values.ToArray();
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }

        /// <summary>
        /// Return the object nearest to the passed line
        /// </summary>
        /// <param name="TestLine"></param>
        /// <param name="intersection"></param>
        /// <param name="nearestIntersect"></param>
        /// <returns></returns>
        public T FindNearest(GridLineSegment TestLine, out GridVector2 intersection, out double nearestIntersect)
        {
            bool LockTaken = false;
            try
            {
                rwLock.EnterReadLock();

                intersection = default(GridVector2);
                nearestIntersect = double.MinValue;
                IEnumerable<GridLineSegment> potentialIntersections = GetPotentialIntersections(TestLine, !LockTaken);
                GridLineSegment BestLine = default(GridLineSegment);
                foreach (GridLineSegment l in potentialIntersections)
                {
                    //Build the edge and find out if it intersects
                    if (l.MinX > TestLine.MaxX)
                        continue;
                    if (l.MaxX < TestLine.MinX)
                        continue;
                    if (l.MinY > TestLine.MaxY)
                        continue;
                    if (l.MaxY < TestLine.MinY)
                        continue;

                    GridVector2 result;
                    bool bIntersected = l.Intersects(TestLine, out result);
                    double distance = GridVector2.Distance(TestLine.A, result);
                    if (distance < nearestIntersect && bIntersected)
                    {
                        nearestIntersect = distance;
                        intersection = result;
                        BestLine = l;
                    }
                }

                if (tableLineToValue.ContainsKey(BestLine))
                    return tableLineToValue[BestLine];

                return default(T);
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }


        /// <summary>
        /// Return the object nearest to the passed line
        /// </summary>
        /// <param name="TestLine"></param>
        /// <param name="intersection"></param>
        /// <param name="nearestIntersect"></param>
        /// <returns></returns>
        public T GetNearest(GridVector2 Position, out GridVector2 BestIntersection, out double ClosestDistance)
        {
            try
            {
                rwLock.EnterReadLock();

                BestIntersection = default(GridVector2);
                ClosestDistance = double.MaxValue;
                int SearchSize = 1;
                List<GridLineSegment> potentialIntersections = GetPotentialIntersections(Position, SearchSize);

                //Expand search until we've found a line to test
                while (potentialIntersections.Count == 0 &&
                    SearchSize < (GridWidth / 2) &&
                    SearchSize < (GridHeight / 2))
                {
                    SearchSize++;
                    potentialIntersections = GetPotentialIntersections(Position, SearchSize);
                }

                GridLineSegment BestLine = default(GridLineSegment);
                foreach (GridLineSegment l in potentialIntersections)
                {
                    GridVector2 thisIntersection;
                    double distance = l.DistanceToPoint(Position, out thisIntersection);
                    if (distance < ClosestDistance)
                    {
                        ClosestDistance = distance;
                        BestIntersection = thisIntersection;
                        BestLine = l;
                    }
                }

                if (tableLineToValue.ContainsKey(BestLine))
                    return tableLineToValue[BestLine];

                return default(T);
            }
            finally
            {
                rwLock.ExitReadLock();
            }
        }
    }
}
