﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics; 
using System.Text;

namespace Geometry
{
    public struct GridLineSegmentPair : IComparable
    {
        public GridLineSegment mapLine;
        public GridLineSegment ctrlLine;

        public override bool Equals(object obj)
        {
            GridLineSegmentPair linePair = (GridLineSegmentPair)obj;

            return (linePair.mapLine == mapLine) && (linePair.ctrlLine == ctrlLine); 
        }

        public override int GetHashCode()
        {
            return mapLine.GetHashCode(); 
        }

        public int CompareTo(object obj)
        {
            GridLineSegmentPair linePair = (GridLineSegmentPair)obj;

            int result = mapLine.CompareTo(linePair.mapLine);
            if (result == 0)
            {
                result = ctrlLine.CompareTo(linePair.ctrlLine);
            }
            
            return result;
        }
    }

    /// <summary>
    /// The line search grid divides space into a regular grid.  Each line segment defines a rectangle.  If the rectangle intersects
    /// a grid cell a pointer is added to that cell.  When a point is passed in the lines in the cell containing the point and all adjacent cells
    /// are checked for intersection.
    /// </summary>
    public class PairedLineSearchGrid
    {
        struct Coord
        {
            public int iX;
            public int iY;

            public Coord(int ix, int iy)
            {
                this.iX = ix;
                this.iY = iy;
            }

            public override string ToString()
            {
                return iX.ToString() + "," + iY.ToString();
            }

            public override bool Equals(object obj)
            {
                Coord coord = (Coord)obj;
                return coord.iX == iX && coord.iY == iY;
            }
        }


        #region PairedLineSearchGridCoordListEnumerator

        /// <summary>
        /// Enumerates over a range of cells, only returning unique values.
        /// We use these because returning a list would copy a massive amount of memory
        /// </summary>
        /// <typeparam name="T"></typeparam>
        private class PairedLineSearchGridCoordListEnumerator : IEnumerable<GridLineSegmentPair>, IEnumerator<GridLineSegmentPair>
        {
            private PairedLineSearchGrid searchGrid;
            private List<GridLineSegmentPair> currentCell;
            private IEnumerable<Coord> coords;
            private IEnumerator<Coord> coordEnum;
            private int iGridIndex = -1;

            //Only return unique values
 //           SortedSet<GridLineSegmentPair> UniqueLines = new SortedSet<GridLineSegmentPair>();

            public PairedLineSearchGridCoordListEnumerator(PairedLineSearchGrid SearchGrid, IEnumerable<Coord> Coords)
            {
                this.searchGrid = SearchGrid;
                this.coords = Coords;
                Reset();
            }

            public GridLineSegmentPair Current
            {
                get
                {
                    GridLineSegmentPair segment = currentCell[iGridIndex];
 //                   Debug.Assert(UniqueLines.Contains(segment) == false);
 //                   UniqueLines.Add(segment);
                    return segment;
                }
            }

            object System.Collections.IEnumerator.Current
            {
                get
                {
                    GridLineSegmentPair segment = currentCell[iGridIndex];
//                    Debug.Assert(UniqueLines.Contains(segment) == false);
//                    UniqueLines.Add(segment);
                    return segment;
                }
            }

            public bool MoveNext()
            {
                iGridIndex++;
                if (currentCell == null)
                {
                    bool success = coordEnum.MoveNext();
                    if (!success)
                        return false;

                    currentCell = searchGrid._LineGrid[coordEnum.Current.iX, coordEnum.Current.iY];
                }

                while (iGridIndex >= currentCell.Count)// || UniqueLines.Contains(currentCell[iGridIndex]))
                {
                    //Figure out if we are advancing because of a repeat value or grid index rollover
                    if (iGridIndex >= currentCell.Count)
                    {
                        iGridIndex = 0;
                        bool success = coordEnum.MoveNext();
                        if (!success)
                            return false;

                        currentCell = searchGrid._LineGrid[coordEnum.Current.iX, coordEnum.Current.iY];

                    }
                    else //we advanced because of a repeat value, increment iGridIndex
                    {
                        iGridIndex++;
                    }
                }

                return true;
            }

            public void Reset()
            {
  //              UniqueLines.Clear();
                iGridIndex = -1; //-1 because MoveNext is called before the first value is read
                coordEnum = coords.GetEnumerator();
            }

            public void Dispose()
            {
                return;
            }

            public IEnumerator<GridLineSegmentPair> GetEnumerator()
            {
                this.Reset();
                return this;
            }

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                this.Reset();
                return this;
            }
        }

        #endregion

        #region PairedLineSearchGridEnumerator

        private class PairedLineSearchGridRectangleEnumerator : IEnumerable<GridLineSegmentPair>, IEnumerator<GridLineSegmentPair>
        {
            private PairedLineSearchGrid searchGrid;
            private List<GridLineSegmentPair> currentCell;
            private Coord start;
            private Coord end;
            private int iX = 0;
            private int iY = 0; 
            private int iGridIndex = -1;

            //Only return unique values
 //           SortedSet<GridLineSegmentPair> UniqueLines = new SortedSet<GridLineSegmentPair>();

            public PairedLineSearchGridRectangleEnumerator(PairedLineSearchGrid SearchGrid, Coord Start, Coord End)
            {
                this.searchGrid = SearchGrid; 
                this.start = Start;
                this.end = End;
                Reset(); 
            }

            public GridLineSegmentPair Current
            {
                get {
                    GridLineSegmentPair segment = currentCell[iGridIndex]; 
//                    UniqueLines.Add(segment);
                    return segment;
                }
            }
            
            object System.Collections.IEnumerator.Current
            {
                get
                {
                    GridLineSegmentPair segment = currentCell[iGridIndex];
//                    UniqueLines.Add(segment);
                    return segment;
                }
            }
            
            public bool MoveNext()
            {
                iGridIndex++;
                while (iGridIndex >= currentCell.Count)// || UniqueLines.Contains(currentCell[iGridIndex]))
                {
                    iGridIndex = 0;
                    iX++;
                    if (iX > end.iX)
                    {
                        iX = start.iX;
                        iY++;
                        if (iY > end.iY)
                        {
                            return false;
                        }
                        else
                        {
                            currentCell = searchGrid._LineGrid[iX, iY];
                        }
                    }
                    else
                    {
                        currentCell = searchGrid._LineGrid[iX, iY];
                    }
                }

                return true; 
            }
            
            public void Reset()
            {
 //               UniqueLines.Clear();
                iX = start.iX;
                iY = start.iY;
                iGridIndex = -1; //-1 because MoveNext is called before the first value is read
                currentCell = searchGrid._LineGrid[iX, iY];
            }

            public void Dispose()
            {
                return; 
            }

            public IEnumerator<GridLineSegmentPair> GetEnumerator()
            {
                return this; 
            }

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return this; 
            }
        }

        #endregion

        GridRectangle Bounds;
        List<GridLineSegmentPair>[,] _LineGrid = new List<GridLineSegmentPair>[0, 0];

        double GridWidth;
        double GridHeight;

        int NumGridsX;
        int NumGridsY;

        int EstimatedLinesPerCell; 


        public PairedLineSearchGrid(MappingGridVector2[] _mapPoints, GridRectangle bounds, Dictionary<int, List<int>> edges)
        {
            this.Bounds = bounds;

            int numPoints = _mapPoints.Length;
            if (numPoints == 0)
                numPoints = 1; 

            //Calculate number of grid cells based on num points and boundaries
            double NumGrids = Math.Ceiling(System.Math.Sqrt(numPoints));
            
            int NumGridsEachDimension = (int)Math.Ceiling(Math.Sqrt(NumGrids));

            this.GridWidth = bounds.Width / NumGridsEachDimension;
            this.GridHeight = bounds.Width / NumGridsEachDimension;

            if (GridWidth < GridHeight)
                GridWidth = GridHeight;
            else
                GridHeight = GridWidth;

            if (GridWidth < 1)
                GridWidth = 1;
            if (GridHeight < 1)
                GridHeight = 1; 

            NumGridsX = (int)Math.Ceiling(bounds.Width / GridWidth);
            NumGridsY = (int)Math.Ceiling(bounds.Height / GridHeight);

            this.EstimatedLinesPerCell = (int)(numPoints / NumGrids);

            _LineGrid = new List<GridLineSegmentPair>[NumGridsX+1, NumGridsY+1]; 
            //Initialize the grid
            for (int iX = 0; iX < NumGridsX+1; iX++)
            {
                for (int iY = 0; iY < NumGridsY+1; iY++)
                {
                    _LineGrid[iX, iY] = new List<GridLineSegmentPair>(this.EstimatedLinesPerCell); 
                }
            }

            //Populate the grid by adding each intersecting GridLineSegment to the proper cells
            foreach (int iPoint in edges.Keys)
            {
                //Get the list of edges
                List<int> edgeList = edges[iPoint];

                for (int iEdge = 0; iEdge < edgeList.Count; iEdge++)
                {
                    int iEdgePoint = edgeList[iEdge];
                    if (iEdgePoint <= iPoint) //We would have tested this point already so skip it
                        continue;

                    //Build the edge and find out if it intersects
                    GridLineSegmentPair pair = new GridLineSegmentPair(); 
                    GridLineSegment mapLine = new GridLineSegment(_mapPoints[iPoint].MappedPoint, _mapPoints[iEdgePoint].MappedPoint);
                    GridLineSegment ctrlLine = new GridLineSegment(_mapPoints[iPoint].ControlPoint, _mapPoints[iEdgePoint].ControlPoint);

                    pair.mapLine = mapLine;
                    pair.ctrlLine = ctrlLine;

                    IEnumerable<Coord> Coords = GetCoordsForLine(mapLine);
                    foreach (Coord coord in Coords)
                    {
                        _LineGrid[coord.iX, coord.iY].Add(pair); 
                    }

                }
            }
        }

        private Coord GetCoord(GridVector2 position)
        {
            return GetCoord(position.X, position.Y); 
        }

        private Coord GetCoord(double x, double y)
        {
            Coord coord; 
            coord.iX = (int)((x - Bounds.Left) / this.GridWidth);
            coord.iY = (int)((y - Bounds.Bottom) / this.GridHeight);
            if (coord.iX < 0)
                coord.iX = 0;
            if (coord.iY < 0)
                coord.iY = 0;
            if (coord.iX > NumGridsX)
                coord.iX = NumGridsX;
            if (coord.iY > NumGridsY)
                coord.iY = NumGridsY; 

            return coord;
        }

        private IEnumerable<Coord> GetCoordsForLine(GridLineSegment line)
        {
            Coord start;
            Coord end;
            if (line.A.X < line.B.X)
            {
                start = GetCoord(line.A);
                end = GetCoord(line.B);
            }
            else
            {
                start = GetCoord(line.B);
                end = GetCoord(line.A);
            }

            List<Coord> listCoords; 
            if (start.iX == end.iX)
            {
                listCoords = new List<Coord>(Math.Abs(end.iY - start.iY));
                for (int iY = start.iY; iY <= end.iY; iY++)
                {
                    listCoords.Add(new Coord(start.iX, iY));
                }
            }
            else if (start.iY == end.iY)
            {
                listCoords = new List<Coord>((end.iX - start.iX));
                for (int iX = start.iX; iX <= end.iX; iX++)
                {
                    listCoords.Add(new Coord(iX, start.iY));
                }
            }
            else
            {
                listCoords = new List<Coord>((end.iX - start.iX) * Math.Abs(end.iY - start.iY));
                //Figure out the line function
                double m = line.slope;
                double b = line.intercept;

                listCoords.Add(start);

                int iLastY = start.iY;
                //Add the line to all cells it belongs in 
                for (int iX = start.iX; iX <= end.iX; iX++)
                {
                    double x = ((iX + 1) * this.GridWidth) + Bounds.Left;
                    double y = m * x + b;

                    Coord intersect = GetCoord(x, y);

                    for (int iY = iLastY; iY <= intersect.iY; iY++)
                    {
                        if (iX == end.iX)
                        {
                            if (m > 0)
                            {
                                if (iY > end.iY)
                                    break;
                            }
                            else if (iY < end.iY)
                            {
                                break;
                            }
                        }

                        if ((iX != start.iX || iY != start.iY) &&
                           (iX != end.iX || iY != end.iY))
                            listCoords.Add(new Coord(iX, iY));

                    }

                    iLastY = intersect.iY;
                }

                listCoords.Add(end);
            }

            return listCoords;
        }

        /// <summary>
        /// Returns a list of GridLineSegments that could possible intersect the passed line
        /// </summary>
        /// <param name="L"></param>
        /// <returns></returns>
        public IEnumerable<GridLineSegmentPair> GetPotentialIntersections(GridLineSegment line)
        {
       //     List<GridLineSegmentPair> LineList;
            //Coord start = GetCoord(new GridVector2(line.MinX, line.MinY));
            //Coord end = GetCoord(new GridVector2(line.MaxX, line.MaxY));

            return new PairedLineSearchGridCoordListEnumerator(this, GetCoordsForLine(line)); 
            
        }
    }
}
