﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using Geometry.Graphics; 

namespace Geometry
{
    /// <summary>
    /// Grid triangle uses pointers to nodes in the grid.  This means any alteration to nodes automatically affects the triangle
    /// </summary>
    /// 
    [Serializable]
    public struct GridTriangle : ICloneable
    {
        readonly GridVector2 _p1;
        readonly GridVector2 _p2;
        readonly GridVector2 _p3;

        Color color; 
        /*
        public GridTriangle(Vector3 p1, Vector3 p2, Vector3 p3)
            : this(new GridVector2(p1.X, p1.Y), new GridVector2(p2.X, p2.Y), new GridVector2(p3.X, p3.Y), Color.Blue)
        {

        }
        */
        public GridTriangle(GridVector2 p1, GridVector2 p2, GridVector2 p3)
            : this(p1, p2, p3, Color.Blue)
        {
            
        }

        public GridTriangle(GridVector2 p1, GridVector2 p2, GridVector2 p3, Color color)
        {
            if (GridVector2.Distance(p1, p2) <= Global.Epsilon ||
               GridVector2.Distance(p2, p3) <= Global.Epsilon ||
               GridVector2.Distance(p3, p1) <= Global.Epsilon)
            {
                
                throw new ArgumentException("This is not a triangle, it is a line");
            }

            this._p1 = p1; 
            this._p2 = p2;
            this._p3 = p3;
            /*
            this._p1.X = Math.Round(this._p1.X, 2);
            this._p2.X = Math.Round(this._p2.X, 2);
            this._p3.X = Math.Round(this._p3.X, 2);
            this._p1.Y = Math.Round(this._p1.Y, 2);
            this._p2.Y = Math.Round(this._p2.Y, 2);
            this._p3.Y = Math.Round(this._p3.Y, 2);

             */
            this.color = color; 

          //  _v1 = 
          //  _v2 = new VertexPositionColor(new Vector3(_p2, ZHeight), color);
          //  _v3 = new VertexPositionColor(new Vector3(_p3, ZHeight), color);
        }

        public GridCircle Circle
        {
            get
            {
                return GridCircle.CircleFromThreePoints(new GridVector2[] { _p1, _p2, _p3 });
            }
        }

        object ICloneable.Clone()
        {
            return this.MemberwiseClone();
        }

        public GridVector2 p1
        {
            get { return _p1; }
           
        }

        public GridVector2 p2
        {
            get { return _p2; }
            
        }

        public GridVector2 p3
        {
            get { return _p3; }
            
        }


        public Color Color
        {
            get { return this.color; }
            set
            {
                this.color = value; 
            }
        }

        public double VectorProducts
        {
            get
            {
                //z = x1 (y2 - y3) + x2 (y3 - y1) + x3 (y1 - y2) 
                double z = (_p1.X * (_p2.Y - _p3.Y)) + (_p2.X * (_p3.Y - _p1.Y)) + (_p3.X * (_p1.Y - _p1.Y));
                return z;
            }
        }

        public double Area
        {
            get{
                return Math.Abs(this.VectorProducts) / 2.0;
            }
        }

        /// <summary>
        /// Returns true if the Point is inside the triangle
        /// </summary>
        /// <param name="test"></param>
        /// <returns></returns>
        public bool Intersects(GridVector2 point)
        {                
            GridVector2 uv = Barycentric(point);

            if (uv.X >= 0 && uv.Y >= 0)
            {
                if (uv.X + uv.Y <= 1.0f)
                    return true; 
            }

            return false; 
        }

        /// <summary>
        /// Returns u,v coordinate of point in triangle.  Calculates areas and returns fractions of area.  This can return 0,0 if the point is well outside the 
        /// triangle because the math hits the limit of the double data-type
        /// </summary>
        /// <param name="point"></param>
        /// <returns></returns>
        public GridVector2 Barycentric(GridVector2 point)
        {
            GridVector2 vCA = _p3 - _p1;
            GridVector2 vBA = _p2 - _p1;
            GridVector2 vPA = point - _p1;

            double dotCACA = GridVector2.Dot(vCA, vCA);
            double dotCABA = GridVector2.Dot(vCA, vBA);
            double dotCAPA = GridVector2.Dot(vCA, vPA);
            double dotBABA = GridVector2.Dot(vBA, vBA);
            double dotBAPA = GridVector2.Dot(vBA, vPA);

            double invDenom = 1.0 / ((dotCACA * dotBABA) - (dotCABA * dotCABA));
            double u = ((dotBABA * dotCAPA) - (dotCABA * dotBAPA)) * invDenom;
            double v = ((dotCACA * dotBAPA) - (dotCABA * dotCAPA)) * invDenom;

            GridVector2 uv = new GridVector2(u, v);

            //There is always a little floating point error, so cut some slack
            if (uv.X < 0 && uv.X >= -Global.Epsilon)
                uv.X = 0.0f;

            if (uv.Y < 0 && uv.Y >= -Global.Epsilon)
                uv.Y = 0.0f;

            if (uv.X > 0 && uv.Y > 0 && uv.X + uv.Y > 1.0f && uv.X + uv.Y <= 1.0f + Global.Epsilon)
            {
                //TODO: This correction could probably by more precise
                double diff = ((u + v) - 1) + (Global.Epsilon / 100);
                uv.X -= diff * u;
                uv.Y -= diff * v;

                Debug.Assert(uv.X + uv.Y <= 1.0f, "Failed to correct for u+v near 1.0f + epsilon case in barycentric conversion"); 
            }

            return uv;
        }
    }
}
