﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Geometry
{

    public class MappingGridVector2SortByMapPoints : IComparer<MappingGridVector2>
    {

        #region IComparer<MappingGridVector2> Members

        int IComparer<MappingGridVector2>.Compare(MappingGridVector2 x, MappingGridVector2 y)
        {
            double diff = x.MappedPoint.X - y.MappedPoint.X;

            if (diff == 0.0)
            {
                diff = x.MappedPoint.Y - y.MappedPoint.Y;
            }

            if (diff > 0)
                return 1;
            if (diff < 0)
                return -1;

            return 0;  
        }

        #endregion
    }
    /// <summary>
    /// Records the position of a point in two different 2D planes
    /// </summary>
    [Serializable]
    public class MappingGridVector2 : ICloneable, IComparable
    {
        public GridVector2 MappedPoint;
        public GridVector2 ControlPoint;

        public MappingGridVector2(GridVector2 control, GridVector2 mapped)
        {
            MappedPoint = mapped;
            ControlPoint = control; 
        }

        public MappingGridVector2 Copy()
        {
            return ((ICloneable)this).Clone() as MappingGridVector2;
        }

        public override string ToString()
        {
            return "Ctrl: " + ControlPoint.ToString() + " Mapped: " + MappedPoint.ToString(); 
        }

        public static string ToMatlab(MappingGridVector2[] array)
        {
            string s = "[";
            for (int i = 0; i < array.Length; i++)
            {
                s += array[i].ControlPoint.X.ToString() + " " + array[i].ControlPoint.Y.ToString() + " " + array[i].MappedPoint.X.ToString() + " " + array[i].MappedPoint.Y.ToString() + ";" + System.Environment.NewLine;
            }
            s += "]";

            return s;
        }

        /// <summary>
        /// Sorted by X coordinante of control point, using Y coordinate as tie-breaker
        /// </summary>
        /// <param name="Obj"></param>
        /// <returns></returns>
        int IComparable.CompareTo(object Obj)
        {
            MappingGridVector2 B = Obj as MappingGridVector2;
            if (B == null)
                return int.MaxValue;

            double diff = this.ControlPoint.X - B.ControlPoint.X;

            if (diff == 0.0)
            {
                diff = this.ControlPoint.Y - B.ControlPoint.Y;
            }

            if (diff > 0)
                return 1;
            if (diff < 0)
                return -1;

            return 0;  
        }

        object ICloneable.Clone()
        {
            return this.MemberwiseClone();
        }

        public static GridRectangle CalculateControlBounds(MappingGridVector2[] mapPoints)
        {
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;

            //Looking at gridIndicies isn't efficient, but it prevents adding removed verticies to 
            //boundary
            for (int i = 0; i < mapPoints.Length; i++)
            {
                minX = Math.Min(minX, mapPoints[i].ControlPoint.X);

                maxX = Math.Max(maxX, mapPoints[i].ControlPoint.X);

                minY = Math.Min(minY, mapPoints[i].ControlPoint.Y);

                maxY = Math.Max(maxY, mapPoints[i].ControlPoint.Y);
            }

            return new GridRectangle(minX, maxX, minY, maxY);
        }


        public static GridRectangle CalculateMappedBounds(MappingGridVector2[] mapPoints)
        {
            double minX = double.MaxValue;
            double minY = double.MaxValue;
            double maxX = double.MinValue;
            double maxY = double.MinValue;

            //   Debug.Assert(mapPoints.Length > 0); 

            //Looking at gridIndicies isn't efficient, but it prevents adding removed verticies to 
            //boundary
            for (int i = 0; i < mapPoints.Length; i++)
            {
                minX = Math.Min(minX, mapPoints[i].MappedPoint.X);
                maxX = Math.Max(maxX, mapPoints[i].MappedPoint.X);
                minY = Math.Min(minY, mapPoints[i].MappedPoint.Y);
                maxY = Math.Max(maxY, mapPoints[i].MappedPoint.Y);
            }

            return new GridRectangle(minX, maxX, minY, maxY);
        }
    }
}
