﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Geometry; 

namespace LocalBookmarks
{
    [Viking.Common.SectionOverlay("Local Bookmarks")]
    class BookmarkOverlay : Viking.Common.ISectionOverlayExtension
    {
        
        #region XNA

        static public Texture2D StarTexture;

        static readonly public VertexPositionColorTexture[] SquareVerts = new VertexPositionColorTexture[] {
            new VertexPositionColorTexture(new Vector3(-1,1,0), Color.White, Vector2.Zero), 
            new VertexPositionColorTexture(new Vector3(1,1,0), Color.White, Vector2.UnitX), 
            new VertexPositionColorTexture(new Vector3(-1,-1,0), Color.White, Vector2.UnitY), 
            new VertexPositionColorTexture(new Vector3(1,-1,0), Color.White, Vector2.One) };

        static readonly public int[] SquareIndicies = new int[] { 2, 1, 0, 3, 1, 2 };

        static public VertexDeclaration VertexPositionColorTextureDecl = null;

        #endregion

        #region ISectionOverlayExtension Members

        void Viking.Common.ISectionOverlayExtension.SetParent(Viking.UI.Controls.SectionViewerControl parent)
        {
            StarTexture = parent.Content.Load<Texture2D>("Star");
        }

        Viking.Common.IUIObjectBasic Viking.Common.ISectionOverlayExtension.NearestObject(Geometry.GridVector2 WorldPosition, out double distance)
        {
            distance = double.MaxValue;
            return RecursiveFindBookmarks(Global.FolderUIObjRoot, WorldPosition, ref distance); 
        }

        BookmarkUIObj RecursiveFindBookmarks(FolderUIObj parentFolder, GridVector2 position, ref double nearestDistance)
        {
            BookmarkUIObj nearestBookmark = null;
            foreach (BookmarkUIObj bookmark in parentFolder.Bookmarks)
            {
                if (Viking.UI.State.ViewerControl.Section.Number == bookmark.Z)
                {
                    double bookmarkDistance = GridVector2.Distance(position, bookmark.GridPosition);
                    if (bookmarkDistance < nearestDistance && bookmarkDistance < Global.BookmarkRadius)
                    {
                        nearestDistance = bookmarkDistance;
                        nearestBookmark = bookmark; 
                    }
                }
            }

            //Walk the bookmark tree and draw every bookmark
            foreach (FolderUIObj folder in parentFolder.Folders)
            {
                double childDistance = double.MaxValue;
                BookmarkUIObj nearestChildBookmark = RecursiveFindBookmarks(folder, position, ref childDistance);
                if (childDistance < nearestDistance)
                {
                    nearestDistance = childDistance;
                    nearestBookmark = nearestChildBookmark; 
                }
            }

            return nearestBookmark; 
        }

        void Viking.Common.ISectionOverlayExtension.Draw(Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice, Microsoft.Xna.Framework.Graphics.BasicEffect basicEffect, Geometry.GridRectangle Bounds, double DownSample)
        {
            RecursiveDrawBookmarks(Global.FolderUIObjRoot, graphicsDevice, basicEffect, Bounds, DownSample);
            return;
        }

        void RecursiveDrawBookmarks(FolderUIObj ParentFolder, Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice, Microsoft.Xna.Framework.Graphics.BasicEffect basicEffect, Geometry.GridRectangle Bounds, double DownSample)
        {
            foreach(BookmarkUIObj bookmark in ParentFolder.Bookmarks)
            {
                if (Viking.UI.State.ViewerControl.Section.Number == bookmark.Z)
                {
                    if (Bounds.Intersects(new GridRectangle(bookmark.GridPosition, Global.BookmarkRadius)))
                    {
                        DrawCircle(graphicsDevice, basicEffect, bookmark.GridPosition, Global.BookmarkRadius, Microsoft.Xna.Framework.Color.Gold);
                        DrawLabel(bookmark, graphicsDevice);
                    }
                    //                    bookmark.Draw(graphicsDevice, basicEffect, DownSample); 
                }
            }

            //Walk the bookmark tree and draw every bookmark
            foreach (FolderUIObj folder in ParentFolder.Folders)
            {
                
                RecursiveDrawBookmarks(folder, graphicsDevice, basicEffect, Bounds, DownSample); 
            }
        }

        static public void DrawCircle(GraphicsDevice graphicsDevice,
                BasicEffect basicEffect,
                GridVector2 Pos,
                double Radius,
                Microsoft.Xna.Framework.Color color)
        {
            //A better way to implement this is to just render a circle texture and add color using lighting, but 
            //this will work for now

            if (false == Global.BookmarksVisible)
                return; 


            VertexPositionColorTexture[] verts;

            //Can't populate until we've referenced CircleVerts
            int[] indicies;
            int[] lineIndicies;
            float radius = (float)Radius;

            //Figure out if we should draw triangles instead
            verts = new VertexPositionColorTexture[SquareVerts.Length];
            SquareVerts.CopyTo(verts, 0);

            indicies = SquareIndicies;

            //Scale and color the verticies
            for (int i = 0; i < verts.Length; i++)
            {
                verts[i].Position.X *= radius;
                verts[i].Position.Y *= radius;

                verts[i].Position.X += (float)Pos.X;
                verts[i].Position.Y += (float)Pos.Y;
                verts[i].Color = color;
            }
            
            /*PORT XNA 4
            VertexDeclaration oldVertexDeclaration = graphicsDevice.VertexDeclaration;
            if (VertexPositionColorTextureDecl == null)
            {
                VertexPositionColorTextureDecl = new VertexDeclaration(graphicsDevice, VertexPositionColorTexture.VertexElements);
            }

            graphicsDevice.VertexDeclaration = VertexPositionColorTextureDecl;
            graphicsDevice.RenderState.PointSize = 5.0f;
            */

            basicEffect.Texture = StarTexture;
            basicEffect.TextureEnabled = true;
            basicEffect.VertexColorEnabled = true;
            basicEffect.LightingEnabled = false;
            //basicEffect.Alpha = 0.5f;
            //basicEffect.CommitChanges();

            //PORT XNA 4
            //basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                //PORT XNA 4
                //pass.Begin();
                pass.Apply(); 

                graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColorTexture>(PrimitiveType.TriangleList,
                                                                              verts,
                                                                              0,
                                                                              verts.Length,
                                                                              indicies,
                                                                              0,
                                                                              2);

                //PORT XNA 4
                //pass.End();
            }

            //PORT XNA 4
            //basicEffect.End();

            /*PORT XNA 4
            //graphicsDevice.VertexDeclaration = oldVertexDeclaration;
             */
            basicEffect.TextureEnabled = false;
            basicEffect.VertexColorEnabled = false;
        }

        public void DrawLabel(BookmarkUIObj bookmark, GraphicsDevice graphicsDevice)
        {
            BlendState originalBlendState = graphicsDevice.BlendState;
            RasterizerState originalRasterState = graphicsDevice.RasterizerState; 

            Viking.UI.Controls.SectionViewerControl parent = Viking.UI.State.ViewerControl; 
            //Print the label
            parent.spriteBatch.Begin(SpriteSortMode.Immediate, BlendState.AlphaBlend);

            Vector2 LabelSize = bookmark.GetLabelSize(parent.fontArial);
            Vector2 LabelOffset = new Vector2(LabelSize.X / 2, LabelSize.Y / 2);
            float MagnificationFactor = (1 / (float)parent.StatusMagnification) * BookmarkUIObj.LabelScaleFactor;

            GridVector2 WorldPosition = bookmark.GridPosition;
            GridVector2 DrawPosition = parent.WorldToScreen(WorldPosition.X, WorldPosition.Y);
            float scale = (((float)Global.BookmarkRadius / LabelSize.X) * MagnificationFactor) / 2;

            parent.spriteBatch.DrawString(parent.fontArial,
                bookmark.Name,
                new Vector2((float)DrawPosition.X, (float)DrawPosition.Y ),
                Microsoft.Xna.Framework.Color.AntiqueWhite,
                0,
                LabelOffset,
                scale,
                SpriteEffects.None,
                0);

            parent.spriteBatch.End();

            graphicsDevice.BlendState = originalBlendState;
            graphicsDevice.RasterizerState = originalRasterState; 

            //PORT XNA 4
            //graphicsDevice.RenderState.CullMode = CullMode.None;
            //graphicsDevice.RenderState.AlphaTestEnable = false;
        }

        #endregion
    }
}
