﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Viking.Common;
using System.Xml.Linq;
using System.IO; 
using connectomes.utah.edu.XSD.BookmarkSchema.xsd;

namespace LocalBookmarks
{
    class Global : IInitExtensions
    {
        static string BookmarkPath = Viking.UI.State.VolumeCachePath + System.IO.Path.DirectorySeparatorChar + "Bookmarks";

        /// <summary>
        /// Bookmark filename only
        /// </summary>
        static string BookmarkFileName = "Bookmarks.xml";

        /// <summary>
        /// Undo filename only
        /// </summary>
        static string BookmarkUndoFileName = "BookmarksUndo01.xml";

        /// <summary>
        /// The full name of the bookmark file including filename and path
        /// </summary>
        static string BookmarkFilePath = BookmarkPath + System.IO.Path.DirectorySeparatorChar + BookmarkFileName;

        /// <summary>
        /// The full name of the undo file including filename and path
        /// </summary>
        static string BookmarkUndoFilePath = BookmarkPath + System.IO.Path.DirectorySeparatorChar + BookmarkUndoFileName;

        static internal readonly string XSDUri = "http://connectomes.utah.edu/XSD/BookmarkSchema.xsd"; 

        /// <summary>
        /// The number of undo files to maintain
        /// </summary>
        static readonly int UndoDepth = 16;

        internal static XRoot BookmarkXMLDoc;
        
        private static Folder FolderRoot
        {
            get { return BookmarkXMLDoc.Folder;}
        }

        static public double BookmarkRadius = 128; 

        private static FolderUIObj _SelectedFolder;
        public static FolderUIObj SelectedFolder
        {
            get { return _SelectedFolder; }
            set { _SelectedFolder = value; }
        }

        internal static FolderUIObj FolderUIObjRoot = null;

        internal static bool BookmarksVisible = true; 

        public static event EventHandler AfterUndo; 


        internal static void Save()
        {
            //Create a backup in case this was a horrible mistake
            if (System.IO.File.Exists(BookmarkUndoFilePath))
            {
                 System.IO.File.Delete(BookmarkUndoFilePath);
            }

            try
            {
                System.IO.File.Move(BookmarkFilePath, BookmarkUndoFilePath);
            }
            catch (System.IO.FileNotFoundException e)
            {
                
            }
    
            //Do the actual save
            BookmarkXMLDoc.Save(BookmarkFilePath);
            if (System.IO.File.Exists(BookmarkFilePath) == false)
            {
                System.Windows.Forms.MessageBox.Show("For some reason Viking can't find: " + BookmarkFilePath + "\nViking just tried to save this file.  You should probably use the \"Export->XML\" menu option from the bookmarks tab to create a backup just in case. This is an unexplained bug we're working on.");
            }
            Viking.UI.State.ViewerForm.Invalidate();
            Viking.UI.State.ViewerControl.Invalidate();
        }

        internal static void Save(string SavePath)
        {
            BookmarkXMLDoc.Save(SavePath); 
        }

        internal static void Undo()
        {
            if (System.IO.File.Exists(BookmarkUndoFilePath))
            {
                //Swap the current and undo versions of the bookmark file
                string TempFileName = Viking.UI.State.VolumeCachePath + System.IO.Path.DirectorySeparatorChar + "Temp.xml";
                if (System.IO.File.Exists(TempFileName))
                    System.IO.File.Delete(TempFileName);

                System.IO.File.Move(BookmarkFilePath, TempFileName);
                System.IO.File.Move(BookmarkUndoFilePath, BookmarkFilePath);
                System.IO.File.Move(Viking.UI.State.VolumeCachePath + System.IO.Path.DirectorySeparatorChar + "Temp.xml", BookmarkUndoFilePath);

                //Reload the bookmarks
                Global global = new Global();
                global.Initialize();

                if (AfterUndo != null)
                    AfterUndo(Global.FolderUIObjRoot, new EventArgs());

                Viking.UI.State.ViewerControl.Invalidate();
            }


        }

        /// <summary>
        /// The number of undo steps we have available
        /// </summary>
        /// <returns></returns>
        internal static string[] UndoFileNames()
        {
            string SearchString = string.Format(BookmarkUndoFileName, '*');
            string[] UndoFiles = System.IO.Directory.GetFiles(BookmarkPath, SearchString);
            return UndoFiles; 
        }

        #region IInitExtensions Members

        public bool Initialize()
        {
            //Check if there is a local favorites XML file, if it does not exist, create it, we always return true

            try
            {
                if (false == System.IO.Directory.Exists(BookmarkPath))
                {
                    System.IO.Directory.CreateDirectory(BookmarkPath);
                }

                if (false == System.IO.File.Exists(BookmarkFilePath))
                {
                    CreateNewBookmarkFile(); 
                }

                BookmarkXMLDoc = XRoot.Load(BookmarkFilePath);
            }
            catch (System.IO.FileNotFoundException except)
            {
                CreateNewBookmarkFile();
            }
            catch (Xml.Schema.Linq.LinqToXsdException except)
            {
                //We found it, but could not parse it
                HandleIncorrectXSDMessage();
                LoadBookmarksFromBackup();
            }
            catch (System.Xml.XmlException except)
            {
                //We found it, but could not parse it
                HandleIncorrectXSDMessage();
                LoadBookmarksFromBackup();
            }
            catch (Exception except)
            {
                //We found it, but could not parse it
                HandleIncorrectXSDMessage();
                LoadBookmarksFromBackup();
            }

            FolderUIObjRoot = new FolderUIObj(null, FolderRoot);
            SelectedFolder = FolderUIObjRoot; 
            
            return true; 
        }

        private static void CreateNewBookmarkFile()
        {
            Folder newFolderRoot = new Folder();
            newFolderRoot.name = "root";
            BookmarkXMLDoc = new XRoot(newFolderRoot);
            BookmarkXMLDoc.Save(BookmarkFilePath);
        }

        private static void HandleIncorrectXSDMessage()
        {
            //We found it, but could not parse it
                string BookmarkRefuge = BookmarkPath + Path.DirectorySeparatorChar + "InvalidSchemaBookmark.xml";
                System.IO.File.Move(BookmarkFilePath, BookmarkPath + Path.DirectorySeparatorChar + "InvalidSchemaBookmark.xml");
                System.Windows.Forms.MessageBox.Show("I could not read your bookmark.xml file, so I moved it to: \n" + BookmarkRefuge + "\n"+
                                                "You can probably recover them by closing Viking and setting/replacing the xmnls attribute on the \"root\" element to:\n"+
                                                "xmlns=\"http://tempuri.org/BookmarkSchema.xsd\" and replacing the Bookmarks.xml with it.");
                System.IO.File.Delete(BookmarkFilePath);
        }

        private static bool LoadBookmarksFromBackup()
        {
            if (System.IO.File.Exists(BookmarkUndoFilePath))
            {
                BookmarkXMLDoc = XRoot.Load(Global.BookmarkUndoFilePath);
                return true; 
            }
            else
            {
                CreateNewBookmarkFile();
                return false; 
            }
        }

        #endregion
    }
}
