﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using Viking.UI;
using Common.DataStructures; 

namespace Viking
{
    internal class TextureCacheEntry : CacheEntry<string>
    {
        public TextureCacheEntry(string filename)
            : base(filename)
        {
            FileInfo info = new FileInfo(filename);
            this.Size = info.Length;
            this.LastAccessed = info.LastAccessTime;            
        }
    }

    /// <summary>
    /// This class manages all requests for textures
    /// </summary>
    class TextureCache : TimeQueueCache<string, TextureCacheEntry, byte[], FileStream>
    {
        public TextureCache()
        {
            //Create the cache directory if it does not exist
            if(System.IO.Directory.Exists(State.CachePath) == false)
                System.IO.Directory.CreateDirectory(State.CachePath);

            //Search the cache directory and create a list of existing files
            string[] dirs = System.IO.Directory.GetDirectories(State.CachePath);

            //Have a bigger cache on disk for textures
            this.MaxCacheSize = 2 << 27;

            Action<string> checkAction = new Action<string>(PopulateCache);
            checkAction.BeginInvoke(State.CachePath, null, null); 
        }

        /// <summary>
        /// Add all textures found under the specified directory to the cache
        /// </summary>
        /// <param name="path"></param>
        private void PopulateCache(string path)
        {
            DateTime Start = DateTime.Now;
            Trace.WriteLine("Populating cache", "TextureUse");

            CheckDirectory(path); 

            TimeSpan elapsed = new TimeSpan(DateTime.Now.Ticks - Start.Ticks);
            Trace.WriteLine("Finish cache populate: " + elapsed.ToString(), "TextureUse"); 
        }

        /// <summary>
        /// Recursively check the supplied directory and all subdirectories, adding files to cache lists
        /// </summary>
        /// <param name="path"></param>
        private void CheckDirectory(string path)
        {  
            string[] dirs = System.IO.Directory.GetDirectories(path);
            foreach (string dir in dirs)
            {
                CheckDirectory(dir);
            }

            string[] files = System.IO.Directory.GetFiles(path);


            foreach (string file in files)
            {
                TextureCacheEntry entry = new TextureCacheEntry(file);
                AddEntry(entry);
            }
        }

  //      static public List<int> AllocatedTextures = new List<int>();
        protected override FileStream Fetch(TextureCacheEntry entry)
        {
            FileStream stream = null; 
            
            if (System.IO.File.Exists(entry.Key))
            {
                try
                {
                    stream = new FileStream(entry.Key, FileMode.Open, FileAccess.Read);
                }
                catch (System.IO.IOException except)
                {
                    //Couldn't open the file, return null
                    return null; 
                }
            }

            return stream; 
        }

        /// <summary>
        /// Creates a file for the texture passed.
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="textureStream"></param>
        protected override TextureCacheEntry CreateEntry(string filename, byte[] textureBuffer)
        {
            System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(filename));
            FileStream stream = new FileStream(filename, FileMode.Create, FileAccess.Write);
            stream.Write(textureBuffer, 0, textureBuffer.Length);
            stream.BeginWrite(textureBuffer, 0, textureBuffer.Length, this.AsynchWriteCompleted, new object[] {stream, filename});
       //     stream.Close();

      //      TextureCacheEntry entry = new TextureCacheEntry(filename);
      //      return entry; 

            //An entry is created if the asynch write succeeds
            return null;
        }


        /// <summary>
        /// The entry is being removed from the cache, so delete the file from the cache
        /// </summary>
        protected override bool OnRemoveEntry(TextureCacheEntry entry)
        {
            if (System.IO.File.Exists(entry.Key) == false)
                return true; 

            try
            {
                System.IO.File.Delete(entry.Key);
            }
            catch (System.IO.IOException except)
            {
                Trace.WriteLine("Could not remove file: " + entry.Key + "\n" + except.Message, "TextureUse");
                return false; 
            }

            return true; 
        }

        protected void AsynchWriteCompleted(IAsyncResult result)
        {
            if (result.IsCompleted)
            {
                object[] state = (object[])result.AsyncState;
                FileStream stream = (FileStream)state[0];
                stream.Close(); 
                string key = (string)state[1];
                TextureCacheEntry entry = new TextureCacheEntry(key);
                base.AddEntry(entry);
            }
        }

    }
}
