﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading; 
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;
using System.IO;
using System.Drawing;
using System.Net;
using Viking.Common;
using System.Drawing;
using System.Drawing.Imaging; 

namespace Viking
{
    class TextureReader : IDisposable 
    {
        readonly Uri Filename;
        readonly string CacheFilename;
        GraphicsDevice graphicsDevice = null;
        private Texture2D _Result = null;
        public ManualResetEvent DoneEvent = new ManualResetEvent(false);
        public bool FinishedReading = false;
        public RefreshDelegate RefreshMethod; 

        private bool IsDisposed = false; 

        System.Net.HttpWebRequest request;
   
        static private bool TextureErrorReported = false;

        private int MipMapLevels = 1; 

        /// <summary>
        /// Returns true if a call to GetTexture will return a non-null value
        /// </summary>
        public bool HasTexture
        {
            get { return _Result != null; }
        }

        /// <summary>
        /// Set to true when the reader has been aborted
        /// </summary>
        protected bool Aborted = false; 


        /// <summary>
        /// Returns the result.  This method can only be called once.
        /// By taking the texture you are responsible for calling dispose.
        /// </summary>
        /// <returns></returns>
        public Texture2D GetTexture()
        {
            Texture2D retVal;
            lock (this)
            {
                retVal = _Result;

                _Result = null;
            }

            return retVal; 
        }

        public TextureReader(GraphicsDevice graphicsDevice, Uri textureUri, string cacheFilename, int mipMapLevels, RefreshDelegate refreshMethod)
            : this(graphicsDevice, textureUri, cacheFilename, mipMapLevels)
        {
            RefreshMethod = refreshMethod; 
        }

        public TextureReader(GraphicsDevice graphicsDevice, Uri textureUri, string cacheFilename, int mipMapLevels)
            : this(graphicsDevice, textureUri, mipMapLevels)
        {
           
            CacheFilename = cacheFilename; 
        }

        /// <summary>
        /// This texture reader is used when we don't have a cachepath to check before making the request
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="filename"></param>
        /// <param name="downsample"></param>
        public TextureReader(GraphicsDevice graphicsDevice, Uri textureURI, int mipMapLevels)
        {
            this.graphicsDevice = graphicsDevice;
            this.Filename = textureURI;
            this.MipMapLevels = mipMapLevels; 

            request = (System.Net.HttpWebRequest)System.Net.WebRequest.Create(textureURI);
            if(textureURI.Scheme.ToLower() == "https")
                request.Credentials = UI.State.UserCredentials; 

            request.CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.NoCacheNoStore);
            request.Timeout = 60001;
#if DEBUG
            Viking.Global.AddTextureReader(this, this.Filename.ToString()); 
#endif

        }

        public void AbortRequest()
        {

            lock (this)
            {
                Aborted = true; 

                if (request != null)
                {
                    request.Abort();
                    request = null;
                }

                //In case we have finished loading the texture, but the texture has not been assigned to the tile, 
                //dispose of the texture
                if (_Result != null)
                {
                    if (_Result.IsDisposed == false)
                    {
                        DisposeTextureThreadingObj disposeObj = new DisposeTextureThreadingObj(_Result);
                        ThreadPool.QueueUserWorkItem(disposeObj.ThreadPoolCallback);
                    }

                    _Result = null;
                }
            }
        }

        public void ThreadPoolCallback(Object threadContext)
        {
            /*Nothing to do if we were aborted already*/
            if (Aborted)
                return; 

            //int threadIndex = Thread.CurrentThread.ManagedThreadId;
            //          Trace.WriteLine("Thread " + threadIndex.ToString() + " reading " + this.Filename, "TextureUse");

            /*PORT XNA 4
            TextureCreationParameters parameters = TextureCreationParameters.Default;
            parameters.Format = SurfaceFormat.Color;
            parameters TextureUsage = TextureUsage.None;
            parameters.MipLevels = this.MipMapLevels;

            //We don't need mip levels, unless we are at the maximum downsample level
            if (this.MipMapLevels == 1)
            {
                parameters.MipFilter = FilterOptions.None;
            }
            else
            {
                parameters.MipFilter = FilterOptions.Box;
            }
            */

            Texture2D tex = null;
            
            if (Filename.Scheme.ToLower() == "http" || Filename.Scheme.ToLower()  == "https")
            {
                System.Net.WebResponse response = null;
                System.IO.Stream TileStream = null;
                MemoryStream memStream = null;
                try
                {

                    //First, check the cache to see if it is locally available
                    if (CacheFilename != null)
                    {
                        TileStream = Global.TextureCache.Fetch(CacheFilename);
                        if (TileStream != null)
                        {
                            //tex = Texture2D.FromStream(graphicsDevice, TileStream);
                            tex = TextureFromStream(graphicsDevice, TileStream, this.MipMapLevels > 0);
                        }
                    }
                }
                catch (InvalidOperationException e)
                {
                    //This exception seems to occur when a cached image file is not valid


                    //TODO: There is an interaction with aborting requests where an corrupt version of the image ends up in the cache and continues to be used.  I have to 
                    //figure out how to flush that bad image out of the cache if this occurs. Currently the workaround is to never cache images


                    Trace.WriteLine(e.Message, "TextureUse");
                    Trace.WriteLine("Deleting bad cache file: " + CacheFilename, "TextureUse");
                    try
                    {
                        System.IO.File.Delete(CacheFilename);
                    }
                    catch (System.IO.IOException IOExcept)
                    {
                        Trace.WriteLine("Failed To delete bad cache file: " + CacheFilename, "TextureUse");
                    }
                }
                catch (Exception e)
                {
                    Trace.WriteLine(e.Message, "TextureUse");
                    Trace.WriteLine("Deleting bad cache file: " + CacheFilename, "TextureUse");
                    try
                    {
                        System.IO.File.Delete(CacheFilename);
                    }
                    catch (System.IO.IOException IOExcept)
                    {
                        Trace.WriteLine("Failed To delete bad cache file: " + CacheFilename, "TextureUse");
                    }
                }
                finally
                {
                    if (TileStream != null)
                    {
                        TileStream.Close();
                        TileStream = null;
                    }
                }


            
                //If we didn't already load the tilestream from the cache
                if (tex == null && request != null)
                {
                    try
                    {
                        response = request.GetResponse();

                        TileStream = response.GetResponseStream();
                     
                        Byte[] streamBuffer = new Byte[response.ContentLength];

                        DateTime loopStart = DateTime.Now;
                        TimeSpan elapsed;
                        long BytesRead = 0;
                        do
                        {
                            BytesRead += TileStream.Read(streamBuffer, (int)BytesRead, (int)streamBuffer.Length - (int)BytesRead);
                            elapsed = new TimeSpan(DateTime.Now.Ticks - loopStart.Ticks);
                        }
                        while (BytesRead < response.ContentLength && elapsed.TotalSeconds < 60);


                        if (CacheFilename != null)
                        {
                            Global.TextureCache.Add(CacheFilename, streamBuffer);
                        }

                        memStream = new MemoryStream(streamBuffer);

                        //tex = Texture2D.FromStream(graphicsDevice, memStream); 
                        tex = TextureFromStream(graphicsDevice, memStream, this.MipMapLevels > 0);
                    }


                    catch (WebException e)
                    {
                        //                       Trace.WriteLine(e.Message, "TextureUse");
                        //                       Trace.WriteLine(e.ToString(), "TextureUse");
                    }
                    catch (InvalidOperationException e)
                    {
                        //TODO: There is an interaction with aborting requests where an corrupt version of the image ends up in the cache and continues to be used.  I have to 
                        //figure out how to flush that bad image out of the cache if this occurs. Currently the workaround is to never cache images


                        Trace.WriteLine(e.Message, "TextureUse");
                    }
                    catch (System.OutOfMemoryException e)
                    {
                        Trace.WriteLine("Out of memory loading texture: " + this.Filename, "TextureUse");
                        Trace.WriteLine(e.Message, "TextureUse");
                        System.GC.Collect();
                    }
                    catch (Exception e)
                    {
                        Trace.WriteLine("Unanticipated exception loading texture: " + this.Filename, "TextureUse");
                        Trace.WriteLine(e.Message, "TextureUse");
                    }
                    finally
                    {
                        if (TileStream != null)
                            TileStream.Close();
                        if (response != null)
                            response.Close();
                        if (memStream != null)
                            memStream.Close();
                    }
                }

                if(tex != null)
                    Global.AddTexture(tex, Filename.ToString());
            }
            else
            {
                FileStream stream = null;
                try
                {
                    stream = System.IO.File.OpenRead(Filename.ToString());
                    //tex = Texture2D.FromStream(graphicsDevice, stream);
                    tex = TextureFromStream(graphicsDevice, stream, this.MipMapLevels > 0);
                    Global.AddTexture(tex, Filename.ToString());
                }
                catch (Exception e)
                {
                    //Print out the first error, but don't flood the output in case we simply have a section where we are
                    //missing some tiles. 
                    if (!TextureErrorReported)
                    {
                        Trace.WriteLine("Error loading texture: " + e.ToString(), "TextureUse");
                        TextureErrorReported = true;
                    }
                }
                finally
                {
                    if (stream != null)
                        stream.Close(); 
                }
            }

            lock (this)
            {
                if (IsDisposed && tex != null)
                {
                    tex.Dispose();
                    Global.RemoveTexture(tex);
                    tex = null; 
                }
                
                this._Result = tex;
                this.FinishedReading = true;
                graphicsDevice = null;

                if(!IsDisposed)
                    DoneEvent.Set();
            }
        }

        public override bool Equals(object obj)
        {
            TextureReader Tobj = obj as TextureReader;
            if (Tobj == null)
                return false;

            if (Tobj.Filename != this.Filename)
                return false;

            return true;
        }

        public override int GetHashCode()
        {
            return Filename.GetHashCode(); 
        }


        public static Texture2D TextureFromStream(GraphicsDevice graphicsDevice, Stream stream, bool mipmap)
        {
            //We load greyscale images, XNA doesn't support loading greyscale by default, so run it through Bitmap instead
            Bitmap image = null;
            int Width;
            int Height;
            BitmapData data = null;
            Byte[] rgbValues;
            int PixelSize;
            try
            {
                image = new System.Drawing.Bitmap(stream);
                Width = image.Width;
                Height = image.Height;

                System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, image.Width, image.Height);
                data = image.LockBits(rect, System.Drawing.Imaging.ImageLockMode.ReadOnly, image.PixelFormat);

                PixelSize = data.Stride / data.Width;
                IntPtr ptr = data.Scan0;
                int TotalBytes = data.Stride * data.Height;
                rgbValues = new Byte[TotalBytes];
                if (data.Stride % data.Width == 0)
                {
                    //The easy case...
                    System.Runtime.InteropServices.Marshal.Copy(ptr, rgbValues, 0, TotalBytes);
                }
                else
                {
                    return null; 

                    Debug.Fail("I know this texture loading code is wrong, but I'm not sure if I'll ever hit this condition");
                    //Copy one line at a time so the empty bytes at the end of each line don't show up in our array
                    for (int iY = 0; iY < data.Height; iY++)
                    {
                        Byte[] lineValues = new Byte[data.Stride];
                        System.Runtime.InteropServices.Marshal.Copy(ptr, lineValues, iY * data.Stride, data.Stride);
                        System.Array.Copy(lineValues, 0, rgbValues, iY * data.Width, data.Stride);
                    }
                }
            }
            finally
            {
                if (image != null)
                {
                    if (data != null)
                    {
                        image.UnlockBits(data);
                    }
                    image.Dispose();
                }
                image = null;
                data = null;
            }

            if (rgbValues != null)
            {
                Byte[] pixelBytes = new Byte[Width * Height];
                for (int iSourceByte = 0, iDestByte = 0; iDestByte < Width * Height; iSourceByte += PixelSize)
                {
                    pixelBytes[iDestByte++] = rgbValues[iSourceByte];
                }

                Texture2D tex = new Texture2D(graphicsDevice, Width, Height, mipmap, SurfaceFormat.Alpha8);
                tex.SetData<Byte>(pixelBytes);
                return tex;
            }

            return null;
        }


        #region IDisposable Members

        public void Dispose()
        {
            lock (this)
            {
                IsDisposed = true;


                //Debug.Assert(_Result == null);
                if (_Result != null)
                {
                    _Result.Dispose();
                    _Result = null; 
                }
                
                //Abort the request if we haven't already
                if(request != null)
                    request.Abort();

                if (DoneEvent != null)
                {
                    DoneEvent.Close();
                    DoneEvent = null;
                }
            }
#if DEBUG            
            Global.RemoveTextureReader(this);
#endif
        }

        #endregion
    }
}
