﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Viking.ViewModels;
using Common.DataStructures;
using Viking.VolumeModel; 

namespace Viking
{
        internal class TileViewModelCacheEntry : CacheEntry<string>
        {
            public TileViewModel TileViewModel;

            public TileViewModelCacheEntry(string Key, TileViewModel T)
                : base(Key)
            {
                TileViewModel = T;
                LastAccessed = DateTime.Now;
                Size = T.Size;
            }
        }
        /// <summary>
        /// This object manages construction of tile objects. 
        /// It first checks a cache for a tile matching the request.  If not found it creates a new tile object.
        /// </summary>
        internal class TileViewModelCache : TimeQueueCache<string, TileViewModelCacheEntry, TileViewModel, TileViewModel>
        {
            static protected string TileKey(string textureFileName, string TransformName)
            {
                return textureFileName + " " + TransformName;
            }

            public TileViewModel GetTile(string textureFileName, string TransformName)
            {
                return this.Fetch(TileKey(textureFileName, TransformName));
            }

            protected override TileViewModel Fetch(TileViewModelCacheEntry key)
            {
                key.WasUsedSinceLastCheckpoint = true;
                return key.TileViewModel;
            }

            public TileViewModel ConstructTile(Tile tile, 
                                    string textureFileName,
                                    string cachedTextureFileName,
                                    string TransformName,
                                    int MipMapLevels, //Should be one, unless it is the minimum downsample level
                                    int size)
            {

                //Check to see if this tile is already loaded
                string key = TileKey(textureFileName, TransformName);

                TileViewModel tileViewModel = new TileViewModel(tile, 
                    textureFileName,
                    cachedTextureFileName,
                    MipMapLevels,
                    size);

                Add(key, tileViewModel);

                return tileViewModel;
            }

            /// <summary>
            /// Retrieve existing tile if it exists, otherwise create a new one
            /// </summary>
            /// <param name="tile"></param>
            /// <param name="textureFileName"></param>
            /// <param name="cachedTextureFileName"></param>
            /// <param name="TransformName"></param>
            /// <param name="MipMapLevels"></param>
            /// <param name="size"></param>
            /// <returns></returns>
            public TileViewModel FetchOrConstructTile(Tile tile,
                                    string textureFileName,
                                    string cachedTextureFileName,
                                    string TransformName,
                                    int MipMapLevels, //Should be one, unless it is the minimum downsample level
                                    int size)
            {
                string key = TileKey(textureFileName, TransformName);

                TileViewModel tileViewModel = Fetch(key);
                if (tileViewModel != null)
                    return tileViewModel;

                return ConstructTile(tile, textureFileName, cachedTextureFileName, TransformName, MipMapLevels, size); 
            }

            protected override TileViewModelCacheEntry CreateEntry(string key, TileViewModel value)
            {
                TileViewModelCacheEntry entry = new TileViewModelCacheEntry(key, value);
                return entry;
            }

            /// <summary>
            /// Remove tile from cache if it exists
            /// </summary>
            /// <param name="tile"></param>
            protected override bool OnRemoveEntry(TileViewModelCacheEntry entry)
            {
                entry.TileViewModel.FreeTexture();
                entry.TileViewModel.Dispose();
                
                return true;
            }

            /// <summary>
            /// When a tile fails a checkpoint it means it was not needed for the last draw operation. 
            /// If the tile does not have a texture loaded it is removed.
            /// Only one of these methods should be running at a time or much of the cache could be deleted
            /// </summary>
            /// <param name="SafeTiles"></param>
            protected override void OnCheckpointFailed(TileViewModelCacheEntry entry)
            {
                //RemoveEntry(entry); 
                
                if (dictEntries.ContainsKey(entry.Key))
                {
                    //Nobody is using it, abort the request
                    entry.TileViewModel.AbortRequest();

                    if (entry.TileViewModel.HasTexture == false)
                    {
                        RemoveEntry(entry);
                    }
                    else
                    {
                        //This tile has a texture, so lets keep it around until we reduce the memory footprint
                        entry.CheckpointExempt = true; 
                    }
                }
                
            }
        
    }
}
