using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using VikingXNA;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Geometry;
using Viking.Common; 

namespace Viking.UI.Commands
{
    public struct CommandQueueEntry
    {
        public readonly System.Type Command;
        public readonly Object[] Args; 

        public CommandQueueEntry(System.Type type, Object[] args)
        {
            this.Command = type; 
            this.Args = args; 
        }
    }

    public class Command
    {
        protected Viking.UI.Controls.SectionViewerControl Parent; //Control the command is listening to

        protected MouseEventArgs oldMouse = null;

        /// <summary>
        /// If the base Command class' OnMouseMove is called this variable contains the mouse position at the last mouse move
        /// </summary>
        protected GridVector2 oldWorldPosition = new GridVector2(0,0); 

        MouseEventHandler MyMouseClick;
        MouseEventHandler MyMouseDoubleClick;
        MouseEventHandler MyMouseDown;
        MouseEventHandler MyMouseUp;
        MouseEventHandler MyMouseWheel;
        MouseEventHandler MyMouseMove;
        EventHandler MyMouseHover; 
        EventHandler MyMouseLeave;
        EventHandler MyMouseEnter; 
        KeyPressEventHandler MyKeyPress;
        KeyEventHandler MyKeyDown; 

        /* UI Extensions, extensions can register with these delegates to be notified whenever the default command does not process input.  This gives
        * extensions the chance to select objects only they are aware of or provide special behavior for key presses */
        public static event MouseEventHandler OnUnhandledMouseDown;
        public static event KeyPressEventHandler OnUnhandledKeyPress;

        /// <summary>
        /// Event fired whenever a command completes successfully
        /// </summary>
        public event CommandCompleteEventHandler OnCommandCompleteHandler;

        public Command(Viking.UI.Controls.SectionViewerControl parent)
        {
            this.Parent = parent;

            MyMouseClick = new MouseEventHandler(this.OnMouseClick);
            MyMouseDoubleClick = new MouseEventHandler(this.OnMouseDoubleClick);
            MyMouseDown = new MouseEventHandler(this.OnMouseDown);
            MyMouseUp = new MouseEventHandler(this.OnMouseUp);
            MyMouseWheel = new MouseEventHandler(this.OnMouseWheel);
            MyMouseMove = new MouseEventHandler(this.OnMouseMove);

            MyMouseHover = new EventHandler(this.OnMouseHover);
            MyMouseLeave = new EventHandler(this.OnMouseLeave);
            MyMouseEnter = new EventHandler(this.OnMouseEnter); 

            MyKeyPress  = new KeyPressEventHandler(this.OnKeyPress);

            MyKeyDown = new KeyEventHandler(this.OnKeyDown);

            Parent.MouseClick += MyMouseClick;
            Parent.MouseDoubleClick += MyMouseDoubleClick;
            Parent.MouseDown += MyMouseDown;
            Parent.MouseUp += MyMouseUp;
            Parent.MouseWheel += MyMouseWheel;
            Parent.MouseMove += MyMouseMove;
            Parent.MouseHover += MyMouseHover;
            Parent.MouseLeave += MyMouseLeave;
            Parent.MouseEnter += MyMouseEnter; 
            Parent.KeyPress += MyKeyPress;
            Parent.KeyDown += MyKeyDown;


        }

        /// <summary>
        /// Set to true if the command is in the middle of processing user input
        /// </summary>
        private bool _CommandActive = false; 

        /// <summary>
        /// Returns true if the command is in the middle of a user input sequence like
        /// defining a rectangle
        /// </summary>
        public bool CommandActive
        {
            get
            {
                if (_Deactivated)
                    return false;
                return _CommandActive;
            }
            set
            {
                if (_Deactivated != false)
                    return;

                _CommandActive = value;
            }
        }

        /// <summary>
        /// This is set to true when the command is not recieving updates from its parent.  
        /// Once set to true it should never be set to false again. 
        /// </summary>
        /// 
        private bool _Deactivated = false;

        /// <summary>
        /// This is set to true when the command is not recieving updates from its parent.  
        /// Once set to true it should never be set to false again. 
        /// </summary>
        /// 
        public bool Deactivated
        {
            get { return _Deactivated; }
            set
            {
                if (_Deactivated == false)
                {
                    Trace.WriteLine("Command Deactivated", "Command"); 

                    //Cancel any active command and remove our mouse events
                    if (value == true)
                    {
                        _CommandActive = false;
                        Parent.MouseClick -= MyMouseClick;
                        Parent.MouseDoubleClick -= MyMouseDoubleClick;
                        Parent.MouseDown -= MyMouseDown;
                        Parent.MouseUp -= MyMouseUp;
                        Parent.MouseWheel -= MyMouseWheel; 
                        Parent.MouseMove -= MyMouseMove;

                        Parent.MouseHover -= MyMouseHover;
                        Parent.MouseLeave -= MyMouseLeave;
                        Parent.MouseEnter -= MyMouseEnter; 

                        Parent.KeyPress -= MyKeyPress;
                        Parent.KeyDown -= MyKeyDown; 
                        
                        OnDeactivate();

                        if (OnCommandCompleteHandler != null)
                            OnCommandCompleteHandler(this, null);
                    }

                    _Deactivated = value;
                }
            }
        }

        virtual public void OnDeactivate() { }

        virtual public void Undo() { }

        virtual public void Redo() { }

        #region Mouse Events

        public virtual void OnMouseClick(object sender, MouseEventArgs e)
        {
            //If the back button was pressed kill the command
            /*
            if (e.Button == MouseButtons.XButton2)
            {
                UI.State.SelectedObject = null;

                //This will probably already be set by adjusting the selected object, but to be safe...
                this.Deactivated = true;
            }
             * */
        }

        public virtual void OnMouseDoubleClick(object sender, MouseEventArgs e)
        {
            
        }

        public virtual void OnMouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.XButton2)
            {
                Parent.StepUpNSections(1);
            }
            else if (e.Button == MouseButtons.XButton1)
            {
                Parent.StepDownNSections(1); 
            }
            else if (Command.OnUnhandledMouseDown != null)
            {
                Command.OnUnhandledMouseDown(sender, e); 
            }
        }

        public virtual void OnMouseUp(object sender, MouseEventArgs e)
        {
        }

        public virtual void OnMouseWheel(object sender, MouseEventArgs e)
        {
            float multiplier = ((float)e.Delta / 120.0f);

            //This seems complicated, but we want the mouse cursor to be pointing at the same 
            //point in the volume before and after the zoom
            
            //This is the point the mouse is at...
            GridVector2 BeforeZoomPosition = Parent.ScreenToWorld(e.X, e.Y);

            StepCameraDistance(multiplier); 

            //This is the point the mouse is at after zooming camera...
            GridVector2 AfterZoomPosition = Parent.ScreenToWorld(e.X, e.Y);

            GridVector2 Offset = BeforeZoomPosition - AfterZoomPosition;

            //Move the camera position by the offset.
            GridVector2 CameraLookat = new GridVector2(Parent.CameraLookAt.X, Parent.CameraLookAt.Y);
            GridVector2 NewCameraLookat = CameraLookat + Offset;
            Parent.CameraLookAt = new Vector3((float)NewCameraLookat.X, (float)NewCameraLookat.Y, Parent.CameraLookAt.Z);

            this.Parent.Invalidate();
        }

        protected void StepCameraDistance(float multiplier)
        {
            float StartDistance = Parent.CameraDistance;
            if (multiplier > 0)
                Parent.CameraDistance = Parent.CameraDistance * 0.86956521739130434782608695652174f;
            else
                Parent.CameraDistance = Parent.CameraDistance * 1.15f;

            this.Parent.Invalidate();
        }

        public virtual void OnMouseMove(object sender, MouseEventArgs e)
        {
            GridVector2 NewPosition = Parent.ScreenToWorld(e.X, e.Y);
            this.Parent.StatusPosition = NewPosition;

            //Figure out which tile we are over and update the status bar
            /*
            if (State.volume != null)
            {
                if (State.volume.CurrentSection != null && State.volume.CurrentSection.CurrentTransform != null)
                {
                    foreach (TileGridTransform T in Parent.Section.WarpedTo[State.volume.CurrentSection.CurrentTransform].TileTransforms)
                    {
                        if (T.CachedControlBounds.Contains(NewPosition))
                        {
                            Parent.StatusTileName = T.Number.ToString();
                            break;
                        }
                    }
                }
            }
            */

            if (oldMouse == null)
            {
                oldMouse = e;
                return;
            }

            if (e.Button == MouseButtons.Right)
            {
                GridVector2 OldPosition = Parent.ScreenToWorld(oldMouse.X, oldMouse.Y);

                Debug.Assert(double.IsNaN(NewPosition.X) == false);

                GridVector2 delta = NewPosition - OldPosition;

                if (double.IsNaN(delta.X))
                    return;

                Parent.CameraLookAt -= new Vector3((float)delta.X, (float)delta.Y, (float)0.0);

                this.Parent.Invalidate();
            }
            else if (e.Button == MouseButtons.Middle)
            {
                //Figure out if the mouse went clockwise or counterclockwise relative to the center of the screen
                System.Drawing.Rectangle rect = Parent.ClientRectangle;
                Vector2 Center = new Vector2(((rect.Width - rect.X) / 2) + rect.X, ((rect.Height - rect.Y) / 2) + rect.Y);

                Vector2 old = new Vector2(oldMouse.X - Center.X, oldMouse.Y - Center.Y);
                Vector2 newMouse = new Vector2(e.X - Center.X, e.Y - Center.Y);

                newMouse.Normalize();
                old.Normalize();

                float angle = (float)Math.Acos((double)Vector2.Dot(newMouse, old));

                if (newMouse.Y < 0)
                {
                    if (newMouse.X - old.X < 0)
                        angle = -angle;
                }
                else if (newMouse.Y >= 0)
                {
                    if (newMouse.X - old.X > 0)
                        angle = -angle;
                }
                angle = MathHelper.ToDegrees(angle);

                //    Trace.WriteLine("Angle changed " + angle.ToString() + " to " + control.CameraRotation.ToString(), "Command"); 

                Parent.CameraRotation += angle;

                this.Parent.Invalidate();
            }

            this.oldMouse = e;
            this.oldWorldPosition = NewPosition; 
        }
        
        public virtual void OnMouseHover(object sender, EventArgs e)
        {
        }

        public virtual void OnMouseLeave(object sender, EventArgs e)
        {
        }

        public virtual void OnMouseEnter(object sender, EventArgs e)
        {
        }

        #endregion

        #region Key Event Handlers
        public void OnKeyPress(object sender, KeyPressEventArgs e)
        {

            //Escape cancels the current command and sets the selected item to null
            if (e.KeyChar == (char)Keys.Escape)
            {
                //On escape kill the current command, and any active queues
                UI.Commands.Command.ClearQueue(); 
                UI.State.SelectedObject = null;

                //This will probably already be set by adjusting the selected object, but to be safe...
                this.Deactivated = true;
            }
            else if (e.KeyChar == '=' || e.KeyChar == '+')
            {
                if (((int)Control.ModifierKeys & (int)Keys.Shift) > 0)
                {
                    Parent.StepUpNSections(10);
                }
                else
                {
                    Parent.StepUpNSections(1);
                }
            }
            else if (e.KeyChar == '-' || e.KeyChar == '_')
            {
                if (((int)Control.ModifierKeys & (int)Keys.Shift) > 0)
                {
                    Parent.StepDownNSections(10);
                }
                else
                {
                    Parent.StepDownNSections(1);
                }
            }
            else if (e.KeyChar == (char)Keys.PrintScreen ||
                     e.KeyChar == 'z')
            {
                
                this.CommandActive = false; 
                
           //     Parent.TakeScreenShot(); 
            }
            else if (Command.OnUnhandledKeyPress != null)
            {
                Command.OnUnhandledKeyPress(sender, e);
            }

            Parent.Invalidate();
        }

        protected void OnKeyDown(object sender, KeyEventArgs e)
        {

            switch (e.KeyCode)
            {
                case Keys.PageUp:
                    StepCameraDistance(1);
                    break;
                case Keys.PageDown:
                    StepCameraDistance(-1);
                    break;                     
                case Keys.Left:
                    GridVector2 Left = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Top);
                    GridVector2 Right = Parent.ScreenToWorld(Parent.ClientRectangle.Right, Parent.ClientRectangle.Top);
                    GridVector2 Diff = Right - Left;
                    Diff.Scale(0.25);

                    Parent.CameraLookAt -= new Vector3((float)Diff.X, (float)Diff.Y, (float)0.0);
                    this.Parent.Invalidate();
                    break; 
                case Keys.Right:
                    Left = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Top);
                    Right = Parent.ScreenToWorld(Parent.ClientRectangle.Right, Parent.ClientRectangle.Top);
                    Diff = Right - Left;
                    Diff.Scale(0.25);

                    Parent.CameraLookAt += new Vector3((float)Diff.X, (float)Diff.Y, (float)0.0);
                    this.Parent.Invalidate();
                    break; 
                case Keys.Up:
                    GridVector2 Bottom = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Bottom);
                    GridVector2 Top = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Top);
                    Diff = Top - Bottom;
                    Diff.Scale(0.25);

                    Parent.CameraLookAt += new Vector3((float)Diff.X, (float)Diff.Y, (float)0.0);
                    this.Parent.Invalidate();
                    break; 
                case Keys.Down:
                    Bottom = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Bottom);
                    Top = Parent.ScreenToWorld(Parent.ClientRectangle.Left, Parent.ClientRectangle.Top);
                    Diff = Top - Bottom;
                    Diff.Scale(0.25);

                    Parent.CameraLookAt -= new Vector3((float)Diff.X, (float)Diff.Y, (float)0.0);
                    this.Parent.Invalidate();
                    break; 
            }
        }

        #endregion

        /// <summary>
        /// Called when the user input is completed and the command should make whatever changes are required and shut down
        /// </summary>
        protected virtual void Execute()
        {
            this.Deactivated = true;
            Parent.Invalidate();
        }

        public virtual void OnDraw(GraphicsDevice graphicsDevice, GridRectangle bounds, double DownSample, BasicEffect basicEffect)
        {
            return; 
        }

        #region Static

        private static Queue<CommandQueueEntry> _CommandQueue = new Queue<CommandQueueEntry>();
        
        public static void EnqueueCommand(System.Type CommandType)
        {
            EnqueueCommand(CommandType, new Object[] { Viking.UI.State.ViewerControl });
        }

        /// <summary>
        /// If the default command is active, then set the passed command as the current command, otherwise add to queue
        /// </summary>
        /// <param name="CommandType"></param>
        /// <param name="Args"></param>
        public static void EnqueueCommand(System.Type CommandType, Object[] Args)
        {
            if (QueueDepth == 0 && Viking.UI.State.ViewerControl.CurrentCommand.GetType() == typeof(DefaultCommand))
                Viking.UI.State.ViewerControl.CurrentCommand = Activator.CreateInstance(CommandType, Args) as Command;
            else
                _CommandQueue.Enqueue(new CommandQueueEntry(CommandType, Args));
        }

        public static void ClearQueue()
        {
            _CommandQueue.Clear(); 
        }

        public static int QueueDepth
        {
            get { return _CommandQueue.Count; }
        }

        /// <summary>
        /// OK, the normal state of the UI is that the default command is active.  When the current command dies we 
        /// check for two things
        /// 1) Are there any commands queued to become active when the current command is dead
        /// 2) Are there any commands registered to handle the current selected object
        /// 3) If not, we activate the default command
        /// </summary>
        /// <param name="Parent"></param>
        /// <param name="Obj"></param>
        /// <returns></returns>
        public static Command CreateFor(Viking.UI.Controls.SectionViewerControl Parent, object Obj)
        {
            //Check if there is a command in the queue
            if (_CommandQueue.Count != 0)
            {
                CommandQueueEntry nextCommand = _CommandQueue.Dequeue();
                
                //Some commands execute in thier constructor, so check if we should return this command or the default
                Command newCommand =  Activator.CreateInstance(nextCommand.Command, nextCommand.Args) as Command;
                if (true == newCommand.Deactivated)
                    return new DefaultCommand(Parent);
                else
                    return newCommand; 
            }

            if (Obj != null)
            {
                System.Type[] Commands = Viking.Common.ExtensionManager.GetCommandsForType(Obj.GetType());

                //TODO: Figure out how to handle multiple commands
                if (Commands.Length > 0)
                {
                    return Activator.CreateInstance(Commands[0], new object[] { Parent }) as Command;
                }
            }

            //TODO: Make these consistent with the extension commands. 
            return new DefaultCommand(Parent);
        }
        
        #endregion


    }
}
