﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging; 
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Windows.Threading; 
using Geometry;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Viking.UI.Commands;
using Viking.UI.Forms; 
using Viking.Common;
using System.Diagnostics;
using Viking.VolumeModel;
using Viking.ViewModels; 


namespace Viking.UI.Controls
{
    public partial class SectionViewerControl : VikingXNA.ViewerControl
    {
        Viking.UI.Commands.Command _CurrentCommand;
        public Viking.UI.Commands.Command CurrentCommand
        {
            get{return _CurrentCommand; }
            set{

                if(_CurrentCommand != value && _CurrentCommand != null)
                {
                    _CurrentCommand.OnCommandCompleteHandler -= this.OnCommandCompleteHandler;
                    _CurrentCommand.Deactivated = true;
                }

                _CurrentCommand = value;

                if (_CurrentCommand != null)
                {
                    _CurrentCommand.OnCommandCompleteHandler += OnCommandCompleteHandler;

                    Trace.WriteLine("Set current command: " + _CurrentCommand.GetType().ToString(), "Command");
                }
                else
                {
                    Trace.WriteLine("Set current command: Null", "Command");
                }
            }
        }

        CommandCompleteEventHandler OnCommandCompleteHandler;

        ISectionOverlayExtension[] listOverlays = null;


        public VertexDeclaration VertexPositionColorDeclaration;

        /// <summary>
        /// The tile cache checkpoints aborts unwanted requests, but since we find out if tiles are
        /// wanted during draw calls we don't want to run a checkpoint unless there has been a draw
        /// call since the last checkpoint
        /// </summary>
        private bool DrawCallSinceTileCacheCheckpoint = false;

        /// <summary>
        /// When set to true Commands and ISectionOverlayExtension draw methods are called
        /// </summary>
        public bool ShowOverlays = true;

       

        public override float CameraDistance
        {
            get { return base.CameraDistance; }
            set
            {
                base.CameraDistance = value;

                StatusMagnification = this.CameraDownsample;
            }
        }

        //A friendlier way of setting camera distance
        public float CameraDownsample
        {
            get { return (float)ProjectionBounds.Width / ClientSize.Width; }
            set 
            {
                base.CameraDistance = CalculateDistance(value); 
            }
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            StatusMagnification = ProjectionBounds.Width / (double)ClientSize.Width;
        }

        #region Status Bar
        private System.Windows.Forms.StatusStrip StatusBar;

        protected System.Windows.Forms.ToolStripItem tsPosition;
        protected System.Windows.Forms.ToolStripItem tsSection;
        protected System.Windows.Forms.ToolStripItem tsMagnification;
        protected System.Windows.Forms.ToolStripItem tsChannels;

        private GridVector2 _StatusPosition;

        public GridVector2 StatusPosition
        {
            get
            {
                return _StatusPosition; 
            }
            set
            {
                _StatusPosition = value; 
                tsPosition.Text = "X: " + value.X.ToString("F2") + " Y: " + value.Y.ToString("F2"); ;
            }
        }

        public int StatusSection
        {
            set
            {
                tsSection.Text = "Section: " + value.ToString();
            }
        }

        private double _Magnification = 0; 

        public double StatusMagnification
        {
            get
            {return _Magnification; }

            set
            {
                _Magnification = value;
                tsMagnification.Text = "Magnification: " + value.ToString("F2");
            }
        }

        private List<ToolStripItem> _StatusChannels = new List<ToolStripItem>(); 
        internal ChannelInfo[] StatusChannels
        {
            set
            {
                if (value == null || value.Length == 0)
                {
                    value = new ChannelInfo[] { new ChannelInfo() };
                }

                //Update the channels we have
                for(int i = 0; i < value.Length; i++)
                {
                    string channelName = value[i].ChannelName;
                    if ("" == channelName)
                    {
                        channelName = this.CurrentChannel; 
                    }

                    ToolStripItem tsChannelItem = null;
                    if (_StatusChannels.Count > i)
                    {
                        tsChannelItem = _StatusChannels[i]; 
                    }
                    else
                    {
                        tsChannelItem = new ToolStripLabel(); 
                        StatusBar.Items.Add(tsChannelItem);
                        _StatusChannels.Add(tsChannelItem); 
                    }

                    tsChannelItem.Text = channelName;
                    System.Drawing.Color color = value[i].FormColor; 
                    //If the color is white, draw black
                    if (color.R == 255 &&
                        color.G == 255 &&
                        color.B == 255)
                        tsChannelItem.ForeColor = System.Drawing.Color.Black; 
                    else
                        tsChannelItem.ForeColor = value[i].FormColor; 
                }

                //Remove extra channel labels
                for (int i =  _StatusChannels.Count -1 ; i >= value.Length; i--)
                {
                    ToolStripItem tsChannelItem = _StatusChannels[i];
                    StatusBar.Items.Remove(tsChannelItem);
                    _StatusChannels.RemoveAt(i); 
                }
                
            }
        }

        #endregion

        private SectionViewModel _Section; 

        /// <summary>
        /// The section we are currently viewing
        /// </summary>
        public SectionViewModel Section
        {
            get { return _Section; }
            set
            {
                if (_Section == value)
                    return;

                SectionViewModel OldSection = _Section; 
                string oldtransform = null;
                if (_Section != null)
                    oldtransform = this.CurrentTransform;

                if (value != null)
                {

                    Trace.WriteLine("Open Section: " + value.Number.ToString(), "UI");
                    StatusSection = value.Number;
                }

                if (OldSection != null)
                {
                    OldSection.OnReferenceSectionChanged -= this.InternalReferenceSectionChanged; 
                }

                _Section = value;
                if (_Section != null)
                {
                    //NOTE: We have to update the section before we ask for the reference section
                    if (State.UseSectionSpecificTransform == false && oldtransform != null)
                        this.CurrentTransform = oldtransform;

                    this.StatusSection = _Section.Number;
                   
                    this._Section.OnReferenceSectionChanged += this.InternalReferenceSectionChanged; 

                    //Figure out if the new section supports the current channel
                    if (_Section.Channels.Contains(this.CurrentChannel) == false)
                        CurrentChannel = _Section.DefaultChannel; 
                }

                ///Find the adjacent sections and request them to warp into volume space if they haven't already    
                if(State.UseSectionSpecificTransform == false && oldtransform != null)
                {
                    SortedList<int, SectionViewModel> sections = UI.State.volume.SectionViewModels;
                    int iSection = sections.IndexOfKey(this._Section.Number); 
                    int iSectionAbove = iSection+1; 
                    int iSectionBelow = iSection-1; 

                    if(iSectionAbove < sections.Count)
                    {
                        sections.Values[iSectionAbove].PrepareTransform(oldtransform);
                    }

                    if (iSectionBelow >= 0)
                    {
                        sections.Values[iSectionBelow].PrepareTransform(oldtransform);
                    }
                }

              

                this.Invalidate(); 

                //Let listeners know if we changed sections
                if(OnSectionChanged != null)
                    OnSectionChanged(this, new SectionChangedEventArgs(_Section, OldSection)); 
            }
        }

        /// <summary>
        /// Currently selected tileset 
        /// </summary>
        protected string _CurrentChannel;
        public string CurrentChannel
        {
            get { return _CurrentChannel; }
            set 
            {
                bool Invalidate = value != _CurrentChannel; 

                _CurrentChannel = value;
                if(Invalidate)
                    this.Invalidate();
            }
        }

        /// <summary>
        /// Determines which transform should be used when rendering the section
        /// </summary>
        protected string _CurrentTransform;
        public string CurrentTransform
        {
            get { return _CurrentTransform; }
            set
            {
                bool NewValue = _CurrentVolumeTransform != value;
                string OldTransform = _CurrentVolumeTransform;
                _CurrentTransform = value;
                if (OnSectionTransformChanged != null && NewValue)
                {
                    OnSectionTransformChanged(this, new TransformChangedEventArgs(_CurrentTransform, OldTransform));
                }
            }
        }

        protected string _CurrentVolumeTransform;
        public string CurrentVolumeTransform
        {
            get { return _CurrentVolumeTransform; }
            set {
                  
                  bool NewValue = value != _CurrentChannel;

                  if (NewValue)
                  {
                      string OldTransform = _CurrentVolumeTransform;
                      _CurrentVolumeTransform = value;
                    
                      if (OnVolumeTransformChanged != null)
                      {
                          OnVolumeTransformChanged(this, new TransformChangedEventArgs(_CurrentTransform, OldTransform));
                      }

                      this.Invalidate();
                  }
            }
        }

        /// <summary>
        /// Fires when the transform used to render the section changes
        /// </summary>
        public event TransformChangedEventHandler OnSectionTransformChanged;

        /// <summary>
        /// Fires when the transform used to place the section into the volume changes
        /// </summary>
        public event TransformChangedEventHandler OnVolumeTransformChanged;

        #region Section/Volume mapping

        /// <summary>
        /// Maps a point from volume space into the section space for the currently selected section and transform
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public GridVector2 VolumeToSection(GridVector2 P)
        {
            return VolumeToSection(P, this.Section); 
        }

        /// <summary>
        /// Maps a point from volume space into the section space for the currently selected transform
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public GridVector2 VolumeToSection(GridVector2 P, SectionViewModel section)
        {
            MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, section.section, this.CurrentChannel, this.CurrentTransform);
            if (map != null)
            {
                GridVector2 TransformedPoint = map.VolumeToSection(new GridVector2(P.X, P.Y));
                return TransformedPoint;
            }
            else
            {
                throw new System.ArgumentException("Cannot map requested point from Volume to Section");
            }
        }

        public bool TryVolumeToSection(GridVector2 P, SectionViewModel section, out GridVector2 transformedP)
        {
            transformedP = new GridVector2(); 

            MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, section.section, this.CurrentChannel, this.CurrentTransform);
            if (map != null)
            {
                return map.TryVolumeToSection(new GridVector2(P.X, P.Y), out transformedP);
            }

            return false; 
        }

        /// <summary>
        /// Maps a point from section space into volume space for the currently selected section and transform
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public GridVector2 SectionToVolume(GridVector2 P)
        {
            return SectionToVolume(P, this.Section.section);
        }

        /// <summary>
        /// Maps a point from section space into volume space for the currently selected transform
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public GridVector2 SectionToVolume(GridVector2 P, Section section)
        {
            MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, section, this.CurrentChannel, this.CurrentTransform);
            if (map != null)
            {
                GridVector2 TransformedPoint = map.SectionToVolume(new GridVector2(P.X, P.Y));
                return TransformedPoint;
            }
            else
            {
                throw new System.ArgumentException("Cannot map requested point from Volume to Section");
            }
        }

        /// <summary>
        /// Maps a point from section space into volume space for the currently selected transform
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public bool TrySectionToVolume(GridVector2 P, Section section, out GridVector2 transformedP)
        {
            transformedP = new GridVector2(); 
            MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, section, this.CurrentChannel, this.CurrentTransform);
            if (map != null)
            {
                return map.TrySectionToVolume(new GridVector2(P.X, P.Y), out  transformedP);
            }

            return false; 
        }

        /// <summary>
        /// Get the boundaries for a section given the current transforms
        /// </summary>
        /// <param name="section"></param>
        /// <returns></returns>
        public GridRectangle SectionBounds(Section section)
        {
            MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, section, this.CurrentChannel, this.CurrentTransform);
            if (map != null)
            {
                return map.Bounds;
            }
            
            throw new System.ArgumentException("Cannot find boundaries for section");
        }

        #endregion

        public SectionViewerControl()
        {
            InitializeComponent();

            StatusBar = new System.Windows.Forms.StatusStrip();
            StatusBar.Parent = this;
            StatusBar.Dock = System.Windows.Forms.DockStyle.Bottom;

            tsSection = new System.Windows.Forms.ToolStripLabel("Section: ");
            tsPosition = new System.Windows.Forms.ToolStripLabel("Position: ");
            tsMagnification = new System.Windows.Forms.ToolStripLabel("Zoom: ");
            tsChannels = new System.Windows.Forms.ToolStripLabel("Channels: ");
            
            StatusBar.Items.Add(tsSection);
            StatusBar.Items.Add(tsPosition);
            StatusBar.Items.Add(tsMagnification);
            StatusBar.Items.Add(tsChannels);

            ObjectSelectedHandler = new Viking.Common.ObjectSelectedEventHandler(this.OnSelectedItemChanged);
            InternalReferenceSectionChanged = new EventHandler(this.OnInternalReferenceSectionChanged); 
            State.ItemSelected += ObjectSelectedHandler;

            ExtensionManager.AddMenuItems(this.menuStrip);
        }

        protected  override void  Initialize()
        {
            if (!DesignMode)
            {
                this.menuStrip.Parent = this.Parent; 

                OnCommandCompleteHandler = new Viking.Common.CommandCompleteEventHandler(this.OnCommandCompleted);

                if (State.SelectedObject != null)
                {
                    CurrentCommand = Command.CreateFor(this, State.SelectedObject.GetType());
                }
                else
                {
                    CurrentCommand = Command.CreateFor(this, null);
                }

                this.CurrentVolumeTransform = UI.State.volume.DefaultVolumeTransform;
                this.CurrentChannel = Section.DefaultChannel;

                listOverlays = ExtensionManager.CreateSectionOverlays(this);
            }
 	        
            base.Initialize();
        }
        
        #region Event Handlers

        /// <summary>
        /// Fired when an object is selected in the UI
        /// </summary>
        Viking.Common.ObjectSelectedEventHandler ObjectSelectedHandler;

        /// <summary>
        /// Fires when a different section is displayed
        /// </summary>
        public event SectionChangedEventHandler OnSectionChanged;

        /// <summary>
        /// Fires when one of the reference sections has changed
        /// </summary>
        public event EventHandler OnReferenceSectionChanged;

        /// <summary>
        /// Called when the reference section for the current section has changed. 
        /// Fires our public ReferenceSectionChanged event
        /// </summary>
        private EventHandler InternalReferenceSectionChanged;

        private void OnSelectedItemChanged(object sender, Viking.Common.ObjectSelectedEventArgs e)
        {
            if (CurrentCommand != null)
                CurrentCommand.Deactivated = true;

            this.Invalidate();
        }

        /// <summary>
        /// Recieves the event from the section when the reference has changed and fires an event to any listeners
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnInternalReferenceSectionChanged(object sender, EventArgs e)
        {
            if (OnReferenceSectionChanged != null)
            {
                OnReferenceSectionChanged(sender, e);
            }
        }

        /*
        private void OnSelectedDesignItemChanged(object sender, PlantMap.Common.ObjectSelectedEventArgs e)
        {
            //Reset our color mapped cursor
            if (CurrentCommand != null)
                CurrentCommand.Deactivated = true;

            this.Invalidate();
        }
        */

        private void OnCommandCompleted(object sender, System.EventArgs e)
        {
            this.Cursor = Cursors.Default;
            CurrentCommand = Command.CreateFor(this, UI.State.SelectedObject);
            this.Invalidate(); 
        }

        #endregion

        /// <summary>
        /// Swaps the current section for the section above it
        /// </summary>
        public void StepUpNSections(int nSections)
        {
            SortedList<int, SectionViewModel> sections = UI.State.volume.SectionViewModels;

            /* find the next section */
            int iStart = this.Section.Number + nSections;
            while (iStart <= sections.Keys.Max())
            {
                if (sections.ContainsKey(iStart))
                {
                    this.Section = sections[iStart];
                    break;
                }
                iStart++; 
            }
        }

        public void StepDownNSections(int nSections)
        {
            SortedList<int, SectionViewModel> sections = UI.State.volume.SectionViewModels;

            /* find the next section */
            int iStart = this.Section.Number - nSections;
            while (iStart >= sections.Keys.Min())
            {
                if (sections.ContainsKey(iStart))
                {
                    this.Section = sections[iStart];
                    break;
                }
                iStart--; 
            }
        }

        public void ExportImage(string Filename, GridRectangle MyRect, double Downsample, bool IncludeOverlays)
        {
            Debug.Assert(MyRect.Left < MyRect.Right);
            Debug.Assert(MyRect.Bottom < MyRect.Top);

            bool OriginalOverlays = this.ShowOverlays;
            bool AsynchTextureLoad = this.AsynchTextureLoad;

            this.ShowOverlays = IncludeOverlays;
            this.AsynchTextureLoad = false;

            //Image Dimensions
            int RequestedWorldX = (int)Math.Floor(MyRect.Left);
            int RequestedWorldY = (int)Math.Floor(MyRect.Bottom);
            int RequestedWorldWidth = (int)Math.Round(MyRect.Width);
            int RequestedWorldHeight = (int)Math.Round(MyRect.Height);

            //Image dimensions on screen
            int CapturedTileSizeX = (int)(Math.Round(MyRect.Width / Downsample));
            int CapturedTileSizeY = (int)(Math.Round(MyRect.Height / Downsample)); 

            int FinalImageWidth = CapturedTileSizeX;
            int FinalImageHeight = CapturedTileSizeY;

            int AdjustedWorldX = RequestedWorldX;
            int AdjustedWorldY = RequestedWorldY;
            int AdjustedWorldWidth = RequestedWorldWidth;
            int AdjustedWorldHeight = RequestedWorldHeight;

            int WorldTileSizeX = RequestedWorldWidth;
            int WorldTileSizeY = RequestedWorldHeight;

           
            int numTilesX = 1;
            int numTilesY = 1;
            //Figure out if we can do the entire shot at once or have to divide it up
            if (CapturedTileSizeX <= 2048 && CapturedTileSizeX <= 2048)
            {

            }
            else
            {
                //The dimensions of a single cell in our capture grid
                int TileCaptureMaxSize = 1024;

                //Find out how many tiles we'll have to capture using a buffer smaller than the current screen size
                numTilesX = (int)Math.Ceiling((double)CapturedTileSizeX / (double)TileCaptureMaxSize);
                numTilesY = (int)Math.Ceiling((double)CapturedTileSizeY / (double)TileCaptureMaxSize);

                WorldTileSizeX = RequestedWorldWidth / numTilesX;
                WorldTileSizeY = RequestedWorldHeight / numTilesY;

                CapturedTileSizeX = (int)Math.Round(WorldTileSizeX / Downsample);
                CapturedTileSizeY = (int)Math.Round(WorldTileSizeY / Downsample);

                FinalImageWidth = CapturedTileSizeX * numTilesX;
                FinalImageHeight = CapturedTileSizeY * numTilesY; 

                //CaptureWidth = TileSizeX * numTilesX;
                //CaptureHeight = TileSizeY * numTilesY;                

                AdjustedWorldWidth = WorldTileSizeX * numTilesX;
                AdjustedWorldHeight = WorldTileSizeY * numTilesY;

                int OffsetX = RequestedWorldWidth - AdjustedWorldWidth;
                int OffsetY = RequestedWorldHeight - AdjustedWorldHeight;

                AdjustedWorldX = RequestedWorldX + (OffsetX / 2);
                AdjustedWorldY = RequestedWorldY + (OffsetY / 2);
            }

            /*
            //The dimensions of a single cell in our capture grid
            int TileCaptureSize = 512;
            int PixelSpaceCellDim = (int)Math.Round((double)TileCaptureSize * Downsample);
            int ImageSpaceCellDim = TileCaptureSize;           


            //Pad so the target bitmap is a multiple of our rendertarget area
            int XPaddingRequired = PixelSpaceCellDim - (PixelSpaceWidth % PixelSpaceCellDim);
            int YPaddingRequired = PixelSpaceCellDim - (PixelSpaceHeight % PixelSpaceCellDim);

            /* Center the view on the adjusted capture size */
            /*
            PixelSpaceX -= XPaddingRequired == PixelSpaceCellDim ? 0 : XPaddingRequired / 2;
            PixelSpaceY -= YPaddingRequired == PixelSpaceCellDim ? 0 : YPaddingRequired / 2;
            PixelSpaceWidth += XPaddingRequired == PixelSpaceCellDim ? 0 : XPaddingRequired;
            PixelSpaceHeight += YPaddingRequired == PixelSpaceCellDim ? 0 : YPaddingRequired;

            //Captures are at least the size of the RenderTarget
            if (PixelSpaceHeight < PixelSpaceCellDim)
                PixelSpaceHeight = PixelSpaceCellDim;

            if (PixelSpaceWidth < PixelSpaceCellDim)
                PixelSpaceWidth = PixelSpaceCellDim;

            //Figure out the dimensions of the output
            int TileImageWidth = (int)Math.Round((double)PixelSpaceWidth / Downsample);
            int TileImageHeight = (int)Math.Round((double)PixelSpaceHeight / Downsample);

            if (TileImageWidth < TileCaptureSize)
            {
                PixelSpaceHeight = (int)((double)TileCaptureSize * Downsample);
                TileImageWidth = TileCaptureSize;
            }
            if (TileImageHeight < TileCaptureSize)
            {
                PixelSpaceWidth = (int)((double)TileCaptureSize * Downsample);
                TileImageHeight = TileCaptureSize;
            }
            */

            System.Drawing.Bitmap bmp = new Bitmap(FinalImageWidth, FinalImageHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

           // int numCols = PixelSpaceWidth / PixelSpaceCellDim;
           // int numRows = PixelSpaceHeight / PixelSpaceCellDim;


//            VikingXNA.BitmapFile bmpFile = new VikingXNA.BitmapFile(Filename, ImageWidth, ImageHeight, 8);

            //Figure out how to cut the rectangle into 512x512 cells and take the screenshots
            for (int iRow = 0; iRow < numTilesY; iRow++)
            {
                for (int iCol = 0; iCol < numTilesX; iCol++)
                {
                    //Figure out the rectangle we need to capture at this location
                    double X = AdjustedWorldX + (iCol * WorldTileSizeX);
                    double Y = AdjustedWorldY + (iRow * WorldTileSizeY);

                    BitmapData lockedBmpData = null;
                    try
                    {
                        //PORT XNA 4

                        Microsoft.Xna.Framework.Graphics.PackedVector.Byte4[] Data = this.CaptureArea(new GridRectangle(new GridVector2(X, Y),
                                                                                                      WorldTileSizeX,
                                                                                                      WorldTileSizeY),
                                                                                                      (float)Downsample);

                        Byte[] byteArray = new Byte[Data.Length * 4];
                        int iByte = 0;
                        for (int iData = 0; iData < Data.Length; iData++, iByte += 4)
                        {

                            byteArray[iByte] = (Byte)(Data[iData].PackedValue >> 16);
                            byteArray[iByte + 1] = (Byte)(Data[iData].PackedValue >> 8);
                            byteArray[iByte + 2] = (Byte)(Data[iData].PackedValue >> 0);
                            byteArray[iByte + 3] = (Byte)(Data[iData].PackedValue >> 24);
                        }

                        System.Drawing.Rectangle bmpRect = new System.Drawing.Rectangle((int)(iCol * CapturedTileSizeX),
                                                                                        (int)(((numTilesY - 1) - iRow) * CapturedTileSizeY),
                                                                                        CapturedTileSizeX,
                                                                                        CapturedTileSizeY);

                        //Write the Data into the bitmap
                        //bmpFile.WriteRectangle(bmpRect, byteArray);

                        lockedBmpData = bmp.LockBits(bmpRect, System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);

                        //                      PixelSize = data.Stride / data.Width;
                        //                        int TotalBytes = lockedBmpData.Stride * lockedBmpData.Height;
                        //                        rgbValues = new Byte[TotalBytes];

                        //The easy case...
                        if (lockedBmpData.Stride == lockedBmpData.Width * 4)
                            System.Runtime.InteropServices.Marshal.Copy(byteArray, 0, lockedBmpData.Scan0, byteArray.Length);
                        else
                        {
                            for (int iBmpRow = 0; iBmpRow < lockedBmpData.Height; iBmpRow++)
                            {
                                System.Runtime.InteropServices.Marshal.Copy(byteArray, iBmpRow * 4 * CapturedTileSizeX, lockedBmpData.Scan0 + (lockedBmpData.Stride * iBmpRow), 4 * CapturedTileSizeX);
                            }

                        }
                    }
                    catch (Exception e)
                    {
                        MessageBox.Show("An exception occured during capture.  The only suggestion is to update graphics drivers." + e.ToString(), "Screenshot Error");
                        break;
                    }
                    finally
                    {
                        if (lockedBmpData != null)
                        {
                            bmp.UnlockBits(lockedBmpData);
                            lockedBmpData = null; 

                        }
                    }
                }

                System.GC.Collect();
            }

            this.ShowOverlays = OriginalOverlays;
            this.AsynchTextureLoad = AsynchTextureLoad;

            bmp.Save(Filename);
            bmp.Dispose(); 
            //bmpFile.Close();
        }

        protected override void Draw(GraphicsDevice graphicsDevice, GridRectangle Bounds, double roundedDownSample)
        {
            if (Section == null)
                return;

            DrawCallSinceTileCacheCheckpoint = true;

            //PORT XNA 4
            /*
            if(this.VertexPositionColorDeclaration == null)
                this.VertexPositionColorDeclaration = graphicsDevice.VertexDeclaration = new VertexDeclaration(
                    graphicsDevice,
                    VertexPositionColor.VertexElements
                    );
            */
            basicEffect.Alpha = 1.0f;
            basicEffect.AmbientLightColor = new Microsoft.Xna.Framework.Vector3(1, 1, 1);

            //OK, figure out if we are rendering channels or not.
            //The section channel settings are checked first.  If they
            //are not found we use the global channel settings.
            ChannelInfo[] Channelset = Section.ChannelInfoArray;
            if (Channelset.Length == 0)
            {
                //See if there are any global channel settings
                Channelset = Section.VolumeViewModel.DefaultChannels;
            }

            StatusChannels = Channelset;
            State.CurrentMode = this.CurrentChannel;

            if (Channelset.Length == 0)
            {
                channelEffect.ChannelColor = new Microsoft.Xna.Framework.Color(1f, 1f, 1f, 1);

                DrawSection(graphicsDevice, this.Section.section,  this.CurrentChannel, Bounds, roundedDownSample); 

                /*
                MappingBase mapping = MappingManager.GetMapping(this.CurrentVolumeTransform,
                                                                Section.section,
                                                                this.CurrentChannel, 
                                                                this.CurrentTransform);

                
                if (mapping != null)
                {
                    channelEffect.ChannelColor = new Microsoft.Xna.Framework.Color(1f, 1f, 1f, 1f);
                    mapping.VisibleTiles(Bounds, roundedDownSample);

                    
        //PORT            Section.Draw(mapping, graphicsDevice, channelEffect, basicEffect, Bounds, roundedDownSample, AsynchTextureLoad);
                }
                 */
            }
            else
            {
                //Walk through each channel and draw the section
                DrawSectionsWithChannels(graphicsDevice, Channelset, Bounds, roundedDownSample); 
            }

            if (ShowOverlays)
            {
                for (int i = 0; i < listOverlays.Length; i++)
                {
                    ISectionOverlayExtension overlayObj = listOverlays[i];
                    overlayObj.Draw(graphicsDevice, basicEffect, Bounds, roundedDownSample);
                }

                ///This is a bad way to know if we are capturing a screenshot, but works for now
                if(AsynchTextureLoad)
                    CurrentCommand.OnDraw(graphicsDevice, Bounds, roundedDownSample, basicEffect);
            }
        }

        protected void DrawSection(GraphicsDevice graphicsDevice, Section section, string channel, GridRectangle Bounds, double downsample)
        {
            MappingBase Mapping = MappingManager.GetMapping(this.CurrentVolumeTransform,
                                                               section,
                                                               channel,
                                                               this.CurrentTransform);

            if (Mapping == null)
                return;
            
            int roundedDownsample = Mapping.NearestAvailableLevel(downsample);

            //Get all of the visible tiles
            TilePyramid visibleTiles = Mapping.VisibleTiles(Bounds, downsample);

            //Find the index of the requested downsample level
            List<int> DownsamplesToRender = new List<int>(Mapping.AvailableLevels.Length);
            
            //Render every other downsample level starting with the requested level
            int iStartingDownsampleLevel = 0;
            for (int i = 0; i < Mapping.AvailableLevels.Length; i++)
            {
                if (roundedDownsample == Mapping.AvailableLevels[i])
                {
                    iStartingDownsampleLevel = i;
                    DownsamplesToRender.Add(i);
                }
                else if (roundedDownsample < Mapping.AvailableLevels[i])
                {
              //      if ((i - iStartingDownsampleLevel) % 2 == 0)
              //      {
                      DownsamplesToRender.Add(i);
              //      }
                }
            }

            DownsamplesToRender.Reverse(); 

            //We only request textures for every other downsample level if they aren't loaded
            for (int iLevel = 0; iLevel < DownsamplesToRender.Count; iLevel++)
            {
                int level = Mapping.AvailableLevels[DownsamplesToRender[iLevel]];

                SortedList<string, Tile> tileList = visibleTiles.GetTilesForLevel(level);

                //Don't draw a bunch of levels that cover the entire screen.  Saves time if we are at high magnification
                if (tileList.Count < 2 && iLevel < DownsamplesToRender.Count - 1)
                    continue; 

                foreach (Tile t in tileList.Values)
                {
                    //Calculate the path of the tile
                    string tileFileName = section.Path + System.IO.Path.DirectorySeparatorChar +
                                t.TextureFileName;
                    //Create a TileViewModel if it doesn't exist and draw it

                    string cachedTileFileName = State.CachePath + System.IO.Path.DirectorySeparatorChar +
                                section.volume.Name + System.IO.Path.DirectorySeparatorChar +
                                section.SectionSubPath + System.IO.Path.DirectorySeparatorChar +
                                t.TextureFileName;

                    TileViewModel tileViewModel = Global.TileViewModelCache.FetchOrConstructTile(t,
                                                                                                 tileFileName,
                                                                                                 cachedTileFileName,
                                                                                                 Mapping.Name,
                                                                                                 0,
                                                                                                 t.TextureSize);

                    

                    //Only draw downsampled tiles if we have them loaded.
          //          if(iLevel % 2 == 0 || iLevel == DownsamplesToRender.Count-1)
          //          {
          //              if (tileViewModel.HasTexture)
                            tileViewModel.Draw(graphicsDevice, channelEffect, AsynchTextureLoad);
           //         }
                }
            }
            /*
            //Draw the tiles
            
            foreach (Tile tile in TilesToDraw)
            {
                if (tile.HasTexture)
                    tile.Draw(graphicsDevice, DownSample, channelEffect, AsynchTextureLoad);
                else
                {
                    if (AllowedDownsamplesList.Contains(tile.Downsample))
                        tile.Draw(graphicsDevice, DownSample, channelEffect, AsynchTextureLoad);
                }
            }
            

            if (Viking.UI.State.ShowMesh)
            {
                SectionToVolumeMapping VolMap = Mapping as SectionToVolumeMapping;
                if (VolMap != null)
                {
                    //       VolMap.VolumeTransform.Draw(graphicsDevice, basicEffect); 
                }

                foreach (Tile tile in TilesToDraw)
                {
                    tile.DrawMesh(graphicsDevice, basicEffect as BasicEffect);
                }
            }
             */

        }

        private void DrawSectionsWithChannels(GraphicsDevice graphicsDevice, ChannelInfo[] Channelset, GridRectangle Bounds, double DownSample)
        {
            List<Texture2D> renderedSections = new List<Texture2D>();

            int DisplayWidth = VisibleScreenRect.Width;
            int DisplayHeight = VisibleScreenRect.Height;

            graphicsDevice.Clear(Microsoft.Xna.Framework.Color.Black);
     //       Trace.WriteLine("Start Drawing Channels", "UI"); 
            GridRectangle renderTargetBounds = VisibleBounds();
 
            //Walk through each channel and draw the section
            foreach (ChannelInfo channel in Channelset)
            {   
                //Figure out which section we need to load
                Section sectionToDraw = null; 
                switch (channel.SectionSource)
                {
                    case ChannelInfo.SectionInfo.SELECTED:
                        sectionToDraw = this.Section.section;
                        break;
                    case ChannelInfo.SectionInfo.ABOVE:
                        sectionToDraw = this.Section.ReferenceSectionAbove;
                        break;
                    case ChannelInfo.SectionInfo.BELOW:
                        sectionToDraw = this.Section.ReferenceSectionBelow;
                        break;
                    case ChannelInfo.SectionInfo.FIXED:
                        int SectionNumber = channel.FixedSectionNumber.Value;
                        if (false == UI.State.volume.SectionViewModels.ContainsKey(SectionNumber))
                            sectionToDraw = null;
                        else
                            sectionToDraw = UI.State.volume.SectionViewModels[SectionNumber].section;

                        break;
                }

                //Can't draw if the section doesn't exist
                if (sectionToDraw == null)
                    continue;

                string ChannelName = channel.ChannelName;
                if (ChannelName.Length == 0)
                {
                    ChannelName = this.CurrentChannel; 
                }

                //Find the mapping to use
                MappingBase mapping = MappingManager.GetMapping(this.CurrentVolumeTransform,
                                                                sectionToDraw,
                                                                ChannelName,
                                                                Section.DefaultPyramidTransform);

                if (mapping == null)
                    continue;

                //Change the transform if we need to, but restore it when we are done
                string oldMode = State.CurrentMode;
                State.CurrentMode = ChannelName;
                
                //Set the color to render with
                channelEffect.ChannelColor = new Microsoft.Xna.Framework.Color(channel.Color.R,
                                                                                        channel.Color.G,
                                                                                        channel.Color.B,
                                                                                        channel.Color.A);

                Viewport oldViewport = graphicsDevice.Viewport;
                RenderTarget2D renderTarget = null;

                try
                {
                    renderTarget = new RenderTarget2D(graphicsDevice,
                                                                   graphicsDevice.Viewport.Width,
                                                                   graphicsDevice.Viewport.Height);
                    
                }
                catch (InvalidOperationException exception)
                {
                    Trace.WriteLine("Could not create render target for channels", "UI");

                    continue;
                }

                RenderTargetBinding[] oldRenderTargets = graphicsDevice.GetRenderTargets();
                graphicsDevice.SetRenderTargets(renderTarget);

                graphicsDevice.Clear(Microsoft.Xna.Framework.Color.Black);
                
                //PORT:
                /*
                sectionToDraw.Draw(mapping,
                                   graphicsDevice,
                                   channelEffect,
                                   basicEffect,
                                   renderTargetBounds, //TODO: Optimization, This could be a smaller rectangle if the user has reduced window size
                                   DownSample, 
                                   AsynchTextureLoad);
                 */

                DrawSection(graphicsDevice, sectionToDraw, channel.ChannelName, renderTargetBounds, Downsample);

                graphicsDevice.SetRenderTargets(oldRenderTargets);
                graphicsDevice.Viewport = oldViewport; 
                
                renderedSections.Add(renderTarget);

                //renderTarget.Dispose();
               // renderTarget = null; 

                State.CurrentMode = oldMode; 
            }
            
            /* The color blending with sprite batch isn't correct so I don't use it, but it would be faster than using 3D
            this.spriteBatch.Begin(SpriteBlendMode.Additive);
            graphicsDevice.RenderState.SourceBlend = Blend.One;
            graphicsDevice.RenderState.DestinationBlend = Blend.One;

            foreach (Texture2D renderedSection in renderedSections)
            {
                this.spriteBatch.Draw(renderedSection,
                   new Microsoft.Xna.Framework.Rectangle(0, 0, DisplayWidth, DisplayHeight),
                    Microsoft.Xna.Framework.Color.White);
            }
            this.spriteBatch.End(); 
            */
            
            //Create a basic mesh to blend the textures onto the screen
            GridVector2 BotLeft = ScreenToWorld(0, 0);
            GridVector2 TopRight = ScreenToWorld(VisibleScreenRect.Width, VisibleScreenRect.Height); 
           
            VertexPositionNormalTexture[] mesh = {
                       new VertexPositionNormalTexture( new Vector3((float)BotLeft.X, (float)BotLeft.Y, 0), Vector3.UnitZ, new Vector2(0,0)),
                       new VertexPositionNormalTexture( new Vector3((float)TopRight.X, (float)BotLeft.Y, 0), Vector3.UnitZ,  new Vector2(1,0)),
                       new VertexPositionNormalTexture( new Vector3((float)BotLeft.X, (float)TopRight.Y, 0), Vector3.UnitZ,   new Vector2(0,1)),
                       new VertexPositionNormalTexture( new Vector3((float)TopRight.X, (float)TopRight.Y, 0), Vector3.UnitZ, new Vector2(1,1))};


            int[] indicies = new int[] { 0, 1, 2, 2, 1, 3 };

            //On the first pass we leave the Source and Destination blend alone so the background is overwritten.
            //Currently we wipe the background to black first, but if we change the background color we'll want this
            //After the first pass we change the state so later textures are blended.
            bool BackgroundOverwritten = false;

            basicEffect.TextureEnabled = true;

            BlendState OriginalBlendState = graphicsDevice.BlendState;
            BlendState AfterFirstTextureBlendState = new BlendState();

            AfterFirstTextureBlendState.ColorSourceBlend = Blend.One;

            AfterFirstTextureBlendState.ColorDestinationBlend = Blend.One;
            //PORT XNA 4
            //basicEffect.Begin();

            foreach (Texture2D renderedSection in renderedSections)
            {
                basicEffect.Texture = renderedSection;

                //PORT XNA 4
                //basicEffect.CommitChanges();
                foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
                {
                    //PORT XNA 4
                    //pass.Begin(); 
                    pass.Apply();
                
                    graphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(PrimitiveType.TriangleList,
                                                                            mesh, 0, mesh.Length,
                                                                            indicies, 0, indicies.Length / 3);

                    if (!BackgroundOverwritten)
                    {
                        //graphicsDevice.BlendState.ColorSourceBlend = Blend.One;
                        //graphicsDevice.BlendState.ColorDestinationBlend = Blend.One;
                        graphicsDevice.BlendState = BlendState.Additive;
                        BackgroundOverwritten = true;
                    }
             }

            }

               

                //PORT XNA 4
                //pass.End();                
           
            //PORT XNA 4
            //basicEffect.End();

            foreach (Texture2D renderedSection in renderedSections)
            {
                renderedSection.Dispose(); 
            }

            renderedSections = null;
           

//            basicEffect.TextureEnabled = false;
//            basicEffect.CommitChanges(); 

            graphicsDevice.BlendState = OriginalBlendState;
        }


        private void timer_Tick(object sender, EventArgs e)
        {
         //   if (Global.TexturesLoadedNeedRefresh) //TEMP: REMOVE FOR ANIMATIONS
                this.Invalidate(); 
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.C:
                    if (e.Control == true)
                    {
                        if (Section == null)
                            break;

                        //On Ctrl+C, copy current mouse position to keyboard
                        GridVector2 Pos = StatusPosition;
                        string PosText = Util.CoordinatesToCopyPaste(Pos.X, Pos.Y, Section.Number, Downsample);
                        Clipboard.SetText(PosText);

                    }
                    break;
                case Keys.Space:
                    this.ShowOverlays = false;
                    this.Invalidate();
                    break;
            }

            base.OnKeyDown(e);
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Space:
                    ShowOverlays = true;

                    this.Invalidate();
                    break;
            }

            base.OnKeyUp(e);
        }

        protected override void  OnKeyPress(KeyPressEventArgs e)
        {
            base.OnKeyPress(e);

            //Escape cancels the current command and sets the selected item to null
            if(e.KeyChar == (char)Keys.PrintScreen || e.KeyChar == 'z')
            {
                this.CurrentCommand = new Viking.UI.Commands.ScreenCaptureCommand(this);
                e.Handled = true; 
            }
        }

        private void SectionViewerControl_MouseDown(object sender, MouseEventArgs e)
        {
            this.Focus(); 
        }

        private void timerTileCacheCheckpoint_Tick(object sender, EventArgs e)
        {
            if (DrawCallSinceTileCacheCheckpoint)
            {
                //This needs to run on the main thread, otherwise we could delete valid requests because
                //a draw call is in process
                //PORT: Global.TileCache.Checkpoint();
                //Dispatcher.CurrentDispatcher.BeginInvoke(new Action(delegate() { Global.TileViewModelCache.Checkpoint(); }), null);
                //Dispatcher.CurrentDispatcher.BeginInvoke(new Action(delegate() { Viking.VolumeModel.Global.TileCache.Checkpoint(); }), null); 
                Global.TileViewModelCache.Checkpoint();
                Viking.VolumeModel.Global.TileCache.Checkpoint();
                //DrawCallSinceTileCacheCheckpoint = false;
                Dispatcher.CurrentDispatcher.BeginInvoke(new Action(delegate() { DrawCallSinceTileCacheCheckpoint = false; }), DispatcherPriority.Background, null);
                System.GC.Collect(); 
            }
        }


        private void menuSection_DropDownOpening(object sender, EventArgs e)
        {
            menuSectionUseSpecific.Checked = State.UseSectionSpecificTransform;
            menuSectionShowMesh.Checked = State.ShowMesh;

            if (State.volume == null)
                return;

            List<ToolStripItem> items = new List<ToolStripItem>();

            foreach (string t in Section.TilesetNames)
            {
                ToolStripMenuItem menuItem = new ToolStripMenuItem(t, null, OnSectionTilesetClick);
                if (t == CurrentTransform)
                    menuItem.Checked = true;

                items.Add(menuItem);
            }

            foreach (string t in Section.ImagePyramids.Keys.ToArray())
            {
                ToolStripMenuItem menuItem = new ToolStripMenuItem(t, null, OnSectionChannelPyramidClick);
                if (t == CurrentTransform)
                    menuItem.Checked = true;

                AddTransformsToSectionChannelPyramidMenuItem(menuItem);
                items.Add(menuItem);
            }

            menuSectionChannel.DropDownItems.Clear();
            menuSectionChannel.DropDownItems.AddRange(items.ToArray());


        }

        private void menuSectionTransform_DropDownOpening(object sender, EventArgs e)
        {

        }

        private void AddTransformsToSectionChannelPyramidMenuItem(ToolStripMenuItem menuChannelPyramid)
        {
            //List all of the pyramid transforms available
            if (Section.PyramidTransformNames.Count < 2)
            {
                //There are no options, or only the default.  Do nothing
                return;
            }
            else
            {
                Debug.Assert(menuChannelPyramid != null);
                if (menuChannelPyramid == null)
                    return;

                List<ToolStripItem> items = new List<ToolStripItem>();
                items.Clear();
                menuChannelPyramid.DropDownItems.Clear();

                foreach (string t in Section.PyramidTransformNames)
                {
                    ToolStripMenuItem menuItem = new ToolStripMenuItem(t, null, OnSectionPyramidTransformClick);
                    if (t == CurrentTransform)
                        menuItem.Checked = true;

                    items.Add(menuItem);
                }

                menuChannelPyramid.DropDownItems.AddRange(items.ToArray());
            }
        }

        /// <summary>
        /// Sets the Section's channel without changing the transform
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnSectionChannelPyramidClick(object sender, EventArgs e)
        {
            ToolStripMenuItem menuSectionChannel = sender as ToolStripMenuItem;
            Debug.Assert(menuSectionChannel != null);
            this.CurrentChannel = menuSectionChannel.Text;
            this.CurrentTransform = Section.DefaultPyramidTransform; 
        }

        private void OnSectionPyramidTransformClick(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
            if (menuItem == null)
                return;

            //Get the parent menu to find the pyramid name
            ToolStripItem menuSectionPyramid = menuItem.OwnerItem;

            CurrentChannel = menuSectionPyramid.Text;
            CurrentTransform = menuItem.Text;
            UI.State.CurrentMode = menuSectionPyramid.Text; 

            this.Refresh();
        }

        private void OnSectionTilesetClick(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
            if (menuItem == null)
                return;

            CurrentChannel = menuItem.Text;
            CurrentTransform = menuItem.Text;

            this.Refresh();
        }

        private void useSectionSpecificTransformsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            State.UseSectionSpecificTransform = !State.UseSectionSpecificTransform;
        }

        private void menuSectionShowMesh_Click_1(object sender, EventArgs e)
        {
            State.ShowMesh = !State.ShowMesh;
        }

        private void menuExportFrames_Click(object sender, EventArgs e)
        {
            Viking.UI.Forms.FrameCapturesForm form = new FrameCapturesForm();

            DialogResult result = form.ShowDialog();
            if (result == DialogResult.Cancel)
                return;

            //Capture each of the requested frames
            GenericProgressForm progressForm = new GenericProgressForm();
            progressForm.Show();

            //Request the UI to capture each frame
            for (int i = 0; i < form.Frames.Length; i++)
            {
                FrameCapture frame = form.Frames[i];
                int Z = (int)Math.Round(frame.Z);
                if (State.volume.SectionViewModels.ContainsKey(Z) == false)
                {
                    DialogResult mbResult = MessageBox.Show("Could not location section #" + Z.ToString() + " in volume, skipping", "Info", MessageBoxButtons.OKCancel);
                    if (mbResult == DialogResult.Cancel)
                        break;
                    else
                        continue;
                }

                //Keep memory to a minimum when switching sections
                SectionViewModel S = State.volume.SectionViewModels[Z];
                this.Section = S;

                this.ExportImage(frame.Filename, frame.Rect, frame.downsample, frame.IncludeOverlay);
                progressForm.ShowProgress("Exported frame: " + frame.Filename, (double)i / (double)form.Frames.Length);
                //System.Windows.Forms.Application.DoEvents();
                if (progressForm.DialogResult == DialogResult.Cancel)
                    break;
            }

            progressForm.Close();
        }

        private void menuGoToLocation_Click(object sender, EventArgs e)
        {
            ShowGoToLocationForm();
            
        }

        protected void ShowGoToLocationForm()
        {
            Viking.UI.Forms.GoToLocationForm form = new GoToLocationForm();
            form.X = CameraLookAt.X;
            form.Y = CameraLookAt.Y;

            if (Section != null)
                form.Z = Section.Number;

            form.Downsample = Downsample; 

            DialogResult result = form.ShowDialog();
            if (result == DialogResult.Cancel)
                return;

            GoToLocation(new Vector2(form.X, form.Y), form.Z, form.Downsample);
        }

        public void GoToLocation(Vector2 location, int Z)
        {
            GoToLocation(location, Z, false, this.Downsample); 
        }

        public void GoToLocation(Vector2 location, int Z, double newDownsample)
        {
            GoToLocation(location, Z, false, newDownsample); 
        }

        public void GoToLocation(Vector2 location, int Z, bool InputInSectionSpace, double newDownsample)
        {
            if (UI.State.volume.SectionViewModels.ContainsKey(Z) == false)
            {
                MessageBox.Show("There is no section # " + Z.ToString() + " in the volume.", "Error", MessageBoxButtons.OK);
                return;
            }

            SectionViewModel newSection = UI.State.volume.SectionViewModels[Z];
            this.Section = newSection;

            if (InputInSectionSpace)
            {
                MappingBase map = MappingManager.GetMapping(this.CurrentVolumeTransform, this.Section.section, this.CurrentChannel, this.CurrentTransform);
                if (map != null)
                {
                    GridVector2 TransformedPoint = map.SectionToVolume(new GridVector2(location.X, location.Y));

                    location.X = (float)TransformedPoint.X;
                    location.Y = (float)TransformedPoint.Y;
                }
            }

            this.CameraLookAt = new Vector3(location.X, location.Y, this.CameraLookAt.Z);
            this.CameraDownsample = (float)newDownsample;
            this.StatusPosition = new GridVector2(location.X,location.Y);
            this.StatusSection = this.Section.Number;
            this.StatusMagnification = newDownsample; 

            //Redraw since we are at a new location
            this.Refresh();
        }

        private void menuCaptureScreen_Click(object sender, EventArgs e)
        {
            this.CurrentCommand = new Viking.UI.Commands.ScreenCaptureCommand(this);
        }

        private void SectionViewerControl_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Control && e.KeyCode == Keys.G)
            {
                ShowGoToLocationForm();
                e.Handled = true;
            }
        }

        private void menuSetupChannels_Click(object sender, EventArgs e)
        {
            SetupChannelsForm ChannelSetup = new SetupChannelsForm(this.Section.VolumeViewModel.DefaultChannels, this.Section.VolumeViewModel.ChannelNames);

            if (ChannelSetup.ShowDialog() == DialogResult.OK)
            {
                this.Section.VolumeViewModel.DefaultChannels = ChannelSetup.ChannelInfo;
                this.Invalidate();
            }

        }

        private void OnVolumeTransformClicked(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
            CurrentVolumeTransform = menuItem.Text; 
        }

        private void menuVolume_DropDownOpening(object sender, EventArgs e)
        {
            menuVolumeTransforms.DropDownItems.Clear();

            ToolStripMenuItem menuNoneItem = new ToolStripMenuItem("None", null, OnVolumeTransformClicked);
            menuVolumeTransforms.DropDownItems.Add(menuNoneItem);

            foreach (string VolumeTransform in Viking.UI.State.volume.TransformNames)
            {
                ToolStripMenuItem menuItem = new ToolStripMenuItem(VolumeTransform, null, OnVolumeTransformClicked);
                menuItem.Checked = CurrentVolumeTransform == VolumeTransform;
                menuVolumeTransforms.DropDownItems.Add(menuItem);
            }
        }


    }
}
