using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Threading;
using Viking.UI;
using Viking.VolumeModel; 
using Geometry; 

namespace Viking.ViewModels
{
    /// <summary>
    /// A tile is a combination of:
    ///     A unique identifier determined by texture file name
    ///     A texture, which may or may not be loaded
    ///     A set of verticies to position the tile in space
    /// </summary>
    public class TileViewModel : IDisposable
    {
        readonly Tile Tile; 
        
        /// <summary>
        /// Stores the verticies used for drawing
        /// </summary>
        private VertexBuffer VertBuffer = null;

        private IndexBuffer IndBuffer = null; 

        /// <summary>
        /// Indicies passed to render call specifying triangle verticies
        /// </summary>
        int[] TriangleIndicies { get { return Tile.TriangleIndicies; } }

        /// <summary>
        /// Size of the tile in memory
        /// </summary>
        public readonly int Size;

        /// <summary>
        /// The amount the tile has been downsampled by
        /// </summary>
        public int Downsample { get { return Tile.Downsample; } }
        
        /// <summary>
        /// This should only be written via the texture member 
        /// </summary>
        private Microsoft.Xna.Framework.Graphics.Texture2D _texture;

        Microsoft.Xna.Framework.Graphics.Texture2D texture
        {
            get {
                if (_texture != null)
                {
                    if (_texture.IsDisposed)
                        _texture = null;
                }
                return _texture; 
            }
            set
            {
                if (value == _texture)
                    return;

                if (_texture != null)
                {
                    DisposeTextureThreadingObj disposeObj = new DisposeTextureThreadingObj(_texture);
                    ThreadPool.QueueUserWorkItem(disposeObj.ThreadPoolCallback);
                    Global.RemoveTexture(_texture); 
                }

                _texture = value;
            }
        }

        internal bool HasTexture
        {
            get { return this._texture != null; }
        }

        private TextureReader _TexReader;
        private TextureReader TexReader
        {
            get { return _TexReader; }
            set {
                lock (this)
                {
                    if (_TexReader != null && _TexReader != value)
                    {
                        _TexReader.Dispose();
                        _TexReader = null;
                    }

                    _TexReader = value;
                }
            }

        }

        public int TileID; 

        public readonly string TextureFileName; 
        public readonly string TextureCachedFileName; 

        bool ReadRequestInProgress = false; 

        private int MipMapLevels = 1; 

        public TileViewModel(Tile tile,
                             string textureFileName,
                             string cachedTextureFileName,
                             int mipMapLevels,
                             int size)
        {
            this.Tile = tile; 
            this.Size = size;
            this.TileID = textureFileName.GetHashCode(); 
            this.TextureFileName = textureFileName;
            this.TextureCachedFileName = cachedTextureFileName;
            this.MipMapLevels = mipMapLevels;

            //this.Verticies = ConvertVerticies(tile);
        }

        /// <summary>
        /// Create a vertex buffer for our verticies
        /// </summary>
        /// <param name="device"></param>
        /// <returns></returns>
        private static VertexBuffer CreateVertexBuffer(GraphicsDevice device, VolumeModel.PositionNormalTextureVertex[] Verticies)
        {
            
            VertexPositionNormalTexture[] vertArray = new VertexPositionNormalTexture[Verticies.Length];

            for(int i = 0; i < Verticies.Length; i++)
            {
                GridVector3 pos = Verticies[i].Position ;
                GridVector3 norm = Verticies[i].Normal; 
                GridVector2 tex = Verticies[i].Texture;

                vertArray[i] = new VertexPositionNormalTexture( new Vector3((float)pos.X,(float)pos.Y, (float)pos.Z),
                                                                new Vector3((float)norm.X,(float)norm.Y,(float)norm.Z),
                                                                new Vector2((float)tex.X, (float)tex.Y));

            }

            VertexBuffer vb = new VertexBuffer(device, typeof(VertexPositionNormalTexture), vertArray.Length, BufferUsage.None);

            vb.SetData<VertexPositionNormalTexture>(vertArray);

            return vb; 
        }

        public override string ToString()
        {
            return TextureFileName;
        }

        public void FreeTexture()
        {
            //Stop trying to load textures if we have a request out
            AbortRequest();

            lock (this)
            {
                //This disposes of the texture
                this.texture = null; 

                if (VertBuffer != null)
                {
                    this.VertBuffer.Dispose();
                    this.VertBuffer = null;
                }

                if (IndBuffer != null)
                {
                    this.IndBuffer.Dispose();
                    this.IndBuffer = null;
                }

            }

        }

        public void AbortRequest()
        {
            if (this.TexReader != null)
            {
                //Other methods may be counting on TexReader not being set to null, so take a lock
                lock (this)
                {
                    if (this.TexReader != null)
                    {
                        this.TexReader.AbortRequest();
                        this.TexReader = null;
                    }
                }
            }
        }

        public Texture2D GetTexture(GraphicsDevice graphicsDevice, bool AsynchTextureLoad)
        {
            //Check if the texture's graphics device has been disposed, in which case load a new texture
            if (texture != null)
            {
                if (texture.GraphicsDevice.IsDisposed)
                {
                    texture = null;
                }
            }

            //If we are trying to read a new texture, check if it is finished yet
            if (TexReader != null)
            {
                lock (this) //Don't let other threads change TexReader state
                {
                    if (TexReader != null) //In case someone freed TexReader before taking the lock
                    {
                        if (TexReader.FinishedReading)
                        {
                            TextureReader texReader = this.TexReader;

                            bool Entered = Monitor.TryEnter(texReader);
                            if (Entered)
                            {
                                //If reading the texture failed don't change our state. 
                                if (texReader.FinishedReading)
                                {
                                    if (texReader.HasTexture)
                                    {
                                        this.texture = texReader.GetTexture();
                                    }
                                    this.TexReader = null;
                                }

                                Monitor.Exit(texReader);
                            }
                        }
                    }
                }
            }

            //If we have a texture that is what we want or better then return it and kill any requests
            if (texture != null)
            {
                //Drop any outstanding requests since we have a texture that works
                AbortRequest();

                return this.texture;
            }
            else
            {
                //In this path we either have no texture, or a low-res texture.  Ask for a new one if we haven't already
                if (this.TexReader == null)
                {
                    //If the section is read over a network provide a cache path too
                    if (State.volume.IsLocal == false)
                    {
                        this.TexReader = new TextureReader(graphicsDevice,
                                                           new Uri(this.TextureFileName), 
                                                           this.TextureCachedFileName,
                                                           this.MipMapLevels);
                    }
                    else
                    {
                        this.TexReader = new TextureReader(graphicsDevice,
                                                            new Uri(this.TextureFileName),
                                                           this.MipMapLevels);
                    }

                    if (AsynchTextureLoad)
                    {
                        ThreadPool.QueueUserWorkItem(TexReader.ThreadPoolCallback);
                        //     TexReader.ThreadPoolCallback(System.Threading.Thread.CurrentContext);
                    }
                    else
                    {
                        //I think reentrant paint calls make this lock useless
                        lock (this)
                        {
                            TextureReader texReader = this.TexReader;
                            this._TexReader = null;
                            texReader.ThreadPoolCallback(System.Threading.Thread.CurrentContext);

 //                           Debug.Assert(texReader.HasTexture);
                            if (texReader.HasTexture)
                            {
                                this.texture = texReader.GetTexture();
                            }

                            texReader.Dispose();
                            texReader = null;
                        }
                    }
                }
                //We have a pending request, but we've been asked to load the texture NOW
                else if (this.TexReader != null && AsynchTextureLoad == false)
                {
                    //I think reentrant paint calls make this lock useless
                    lock (this)
                    {
                        if (this.TexReader != null)
                        {
                            TextureReader texReader = this.TexReader;
                            this._TexReader = null; 
                            texReader.DoneEvent.WaitOne(); //Timeout after a minute?

                            Debug.Assert(texReader != null);
                            Debug.Assert(texReader.HasTexture);

                            if (texReader.HasTexture)
                            {
                                this.texture = texReader.GetTexture();
                            }

                            texReader.Dispose();
                            texReader = null; 
                        }
                        else
                        {
                            Debug.Assert(this.texture != null, "Someone nulled out our texture reader, but didn't leave a texture during Synch texture load");
                        }
                    }
                }

                return this.texture; // Can return null or a texture with a different downsample level if we have one
            }                 
        }

        public void FileReadComplete(IAsyncResult ar)
        {
            Debug.Assert(ar.IsCompleted, "Wasn't expecting Asynch read not to complete successfully");
            if (!ar.IsCompleted)
            {
                ReadRequestInProgress = false;
                return;
            }

            object[] objs = ar.AsyncState as object[];
            FileStream stream = objs[0] as FileStream;
            GraphicsDevice graphicsDevice = objs[1] as GraphicsDevice;
            byte[] fileBuffer = objs[2] as byte[];
            int Level = (int)objs[3];

            stream.Close();
            stream.Dispose();
            stream = null;

            MemoryStream MemStream = new MemoryStream(fileBuffer, false);

            Monitor.Enter(this);

            texture = Texture2D.FromStream(graphicsDevice, MemStream);

            MemStream.Close();
            MemStream.Dispose();
            MemStream = null;

            Trace.WriteLine("AsSynch load completed " + Level.ToString("D3") + " " + this.TextureFileName, "Tile");

            ReadRequestInProgress = false;

            Monitor.Exit(this);
        }

        public void Draw(GraphicsDevice graphicsDevice, VikingXNA.ChannelEffect effect, bool AsynchTextureLoad)
        {
            if (TriangleIndicies == null)
            {
                Trace.WriteLine("Null Grid Indicies for " + this.TextureFileName, "Tile");
                return;
            }

            if (TriangleIndicies.Length == 0)
            {
                Trace.WriteLine("No Grid Indicies for " + this.TextureFileName, "Tile");
                return;
            }

            //Create the verticies if they don't exist
            if (VertBuffer == null)
            {
                VertBuffer = CreateVertexBuffer(graphicsDevice, Tile.Verticies); 
            }

            if (VertBuffer == null || VertBuffer.VertexCount == 0)
                return; 

            //Create Index buffer if it doesn't exist
            if (IndBuffer == null)
            {
                IndBuffer = new IndexBuffer(graphicsDevice, typeof(int), Tile.TriangleIndicies.Length, BufferUsage.None);
                IndBuffer.SetData<int>(Tile.TriangleIndicies); 
            }
            
            Texture2D currentTexture = GetTexture(graphicsDevice, AsynchTextureLoad);

            //Do not draw if we don't have a texture
            if (currentTexture == null)
                return; 

            graphicsDevice.SetVertexBuffer(this.VertBuffer);
            graphicsDevice.Indices = this.IndBuffer; 
            
            effect.Texture = currentTexture;
                        
            effect.CommitChanges();
            
            //PORT XNA 4
            //effect.effect.Begin();

            foreach (EffectPass pass in effect.effect.CurrentTechnique.Passes)
            {
                //PORT XNA 4
                //pass.Begin();
                pass.Apply(); 

                graphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, VertBuffer.VertexCount, 0, TriangleIndicies.Length / 3); 
                /*
                graphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalTexture>(PrimitiveType.TriangleList,
                                        this.Verticies,
                                        0,
                                        this.Verticies.Length, 
                                        TriangleIndicies, 
                                        0,
                                        TriangleIndicies.Length / 3);
                */
                //PORT XNA 4
                //pass.End();
            }

            //PORT XNA 4
            //effect.effect.End();
        }

        //PORT XNA 4
        //static VertexDeclaration VertexPositionColorDeclaration = null;
        VertexBuffer vbMesh = null;
        IndexBuffer ibMesh = null;

        //PORT XNA 4
        //VertexPositionColor[] MeshVerticies = null;
        //int[] MeshEdges = null;

        private void CreateMesh(GraphicsDevice graphicsDevice)
        {
            Random ColorGen = new Random(this.GetHashCode());
            byte[] randColorBytes = new byte[3];
            ColorGen.NextBytes(randColorBytes);
            randColorBytes[0] = randColorBytes[0] < 128 ? (byte)(randColorBytes[0] + 128) : randColorBytes[0];
            randColorBytes[1] = randColorBytes[1] < 128 ? (byte)(randColorBytes[1] + 128) : randColorBytes[1];
            randColorBytes[2] = randColorBytes[2] < 128 ? (byte)(randColorBytes[2] + 128) : randColorBytes[2];
            Color color = new Color(randColorBytes[0], randColorBytes[1], randColorBytes[2]);
            VertexPositionColor[] MeshVerticies = new VertexPositionColor[this.Tile.Verticies.Length];
            for (int iVert = 0; iVert < Tile.Verticies.Length; iVert++)
            {
                PositionNormalTextureVertex vert = Tile.Verticies[iVert];
                MeshVerticies[iVert] = new VertexPositionColor(new Vector3((float)vert.Position.X, (float)vert.Position.Y, (float)vert.Position.Z),
                                                               color);
            }

            int[] MeshEdges = new int[TriangleIndicies.Length * 2];
            int iMesh = 0;
            for (int iTri = 0; iTri < TriangleIndicies.Length; iTri += 3)
            {

                MeshEdges[iMesh++] = TriangleIndicies[iTri];
                MeshEdges[iMesh++] = TriangleIndicies[iTri + 1];

                MeshEdges[iMesh++] = TriangleIndicies[iTri + 1];
                MeshEdges[iMesh++] = TriangleIndicies[iTri + 2];

                MeshEdges[iMesh++] = TriangleIndicies[iTri + 2];
                MeshEdges[iMesh++] = TriangleIndicies[iTri];

            }

            vbMesh = new VertexBuffer(graphicsDevice, typeof(VertexPositionColor), MeshVerticies.Length, BufferUsage.None);
            vbMesh.SetData<VertexPositionColor>(MeshVerticies);
            ibMesh = new IndexBuffer(graphicsDevice, typeof(int), MeshEdges.Length, BufferUsage.None);
            ibMesh.SetData<int>(MeshEdges);
        }

        public void DrawMesh(GraphicsDevice graphicsDevice, BasicEffect basicEffect)
        {
            if (this.HasTexture == false)
                return;

            //PORT XNA 4
            //graphicsDevice.RenderState.PointSize = 5.0f;


            //PORT XNA 4
            /*
            //VertexDeclaration oldVertexDeclaration = graphicsDevice.VertexDeclaration;

            if (TileViewModel.VertexPositionColorDeclaration == null)
            {
                TileViewModel.VertexPositionColorDeclaration = new VertexDeclaration(
                    graphicsDevice,
                    VertexPositionColor.VertexElements
                    );
            }

            if (MeshVerticies == null)
            {
                CreateMesh();
            }

            if (MeshVerticies.Length == 0)
                return; 

            */

            if (vbMesh == null)
            {
                CreateMesh(graphicsDevice); 
            }

            if (vbMesh.VertexCount == 0)
                return; 

            //PORT XNA 4
            //graphicsDevice.VertexDeclaration = TileViewModel.VertexPositionColorDeclaration;
            
            basicEffect.Texture = null; 
            basicEffect.TextureEnabled = false;
            basicEffect.VertexColorEnabled = true;

            graphicsDevice.SetVertexBuffer(vbMesh);
            graphicsDevice.Indices = ibMesh; 
            //PORT XNA 4
            //basicEffect.CommitChanges();

            //PORT XNA 4
            //basicEffect.Begin();

            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                //PORT XNA 4
                //pass.Begin();
                pass.Apply();

                graphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, vbMesh.VertexCount, 0, ibMesh.IndexCount / 3); 

        //        graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.LineList, mapVerticies, 0, mapVerticies.Length, indicies, 0, indicies.Length / 2); 

                //PORT XNA 4
                /*
                graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.LineList,
                                                                                MeshVerticies,
                                                                                0,
                                                                                MeshVerticies.Length,
                                                                                MeshEdges,
                                                                                0,
                                                                                MeshEdges.Length / 2);

                //PORT XNA 4
                //pass.End();
                */
            }

            //PORT XNA 4
            //basicEffect.End();

            //PORT XNA 4
            //graphicsDevice.VertexDeclaration = oldVertexDeclaration; 
        }

        #region IDisposable Members

        public void Dispose()
        {
            lock (this)
            {
                //This disposes of the texture
                _texture = null;
                  
                if (VertBuffer != null)
                {
                    this.VertBuffer.Dispose();
                    this.VertBuffer = null;
                }

                if (IndBuffer != null)
                {
                    this.IndBuffer.Dispose();
                    this.IndBuffer = null;
                }
            }
   
        }

        #endregion
    }
}
