﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Geometry;
using System.Diagnostics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace Viking.VolumeModel
{
    /// <summary>
    /// This class represents the warped version of a single section using a transform
    /// </summary>
    /// 
    class SectionToVolumeMapping : FixedTileCountMapping
    {
        private object LockObj = new object(); 
        
        /// <summary>
        /// If this section is in a volume ReferencedTo indicates which section
        /// was used as a control during slice-to-slice registration.
        /// Returns -1 if the section is not in a volume
        /// </summary>
        public int ReferencedTo
        {
            get { return -1; }
        }

        public override TileGridTransform[] TileTransforms
        {
            get
            {
                if (HasBeenWarped == false)
                    Warp();

                return _TileTransforms;
            }
        }

        /// <summary>
        /// .mosaic files load as being warped.  Volume sections have to passed through a volume transform first, which we do in a lazy fashion
        /// </summary>
        protected bool HasBeenWarped = false;

        /// <summary>
        /// The transforms applied to each tile for this section, used to generate verticies. 
        /// If the HasBeenWarped is == false these transforms are in section space and not volume space
        /// </summary>
        public FixedTileCountMapping SourceMapping;

        /// <summary>
        /// The transformation which will/has converted the tiles from section space into volume space.
        /// This can be null if this section is not warped into volume space. 
        /// </summary>
        public readonly GridTransform VolumeTransform;

        public override string CachedTransformsFileName
        {
            get { return Section.volume.LocalCachePath + 
                System.IO.Path.DirectorySeparatorChar + 
                Section.volume.Name + 
                System.IO.Path.DirectorySeparatorChar + 
                VolumeTransform.ToString() + "_stos.cache"; }
        }

        public SectionToVolumeMapping(Section section, string name,  FixedTileCountMapping sourceMapping, GridTransform volumeTransform) 
            : base(section, name, sourceMapping.TilePrefix, sourceMapping.TilePostfix )
        {
            HasBeenWarped = false;
            SourceMapping = sourceMapping;
            VolumeTransform = volumeTransform;            
        }

        public override void FreeMemory()
        {
            lock (LockObj)
            {
                HasBeenWarped = false;
                _TileTransforms = null;
                SourceMapping.FreeMemory();
            }
        }

        /// <summary>
        /// If this section has not yet been warped, then do so.
        /// This method is invoked by threads.  
        /// </summary>
        public void Warp()
        {
            if (HasBeenWarped)
                return;

            lock (LockObj)
            {
                if (HasBeenWarped)
                    return;

                bool LoadedFromCache = false; 
                if (System.IO.File.Exists(this.CachedTransformsFileName))
                {
                    try
                    {
                        this._TileTransforms = LoadFromCache();
                        LoadedFromCache = true; 
                    }
                    catch (Exception e)
                    {
                        //On any error, use the traditional path
                        this._TileTransforms = null;
                        LoadedFromCache = false; 
                    }

                    if (this._TileTransforms == null)
                        LoadedFromCache = false; 
                }

                if (LoadedFromCache)
                {
                    this.HasBeenWarped = true;
                    return; 
                }

                if (VolumeTransform != null)
                    Trace.WriteLine("Warping section " + VolumeTransform.MappedSection + " to volume space", "VolumeModel"); 

                Debug.Assert(this.VolumeTransform != null);

                // Get the transform tiles from the source mapping, which loads the .mosaic if it hasn't alredy been loaded
                TileGridTransform[] volTransforms = SourceMapping.TileTransforms;

                // We add transforms which surivive addition with at least three points to this list
                List<TileGridTransform> listTiles = new List<TileGridTransform>(volTransforms.Length);

                for (int i = 0; i < volTransforms.Length; i++)
                {
                    TileGridTransform T = volTransforms[i];
                    TileGridTransform copy = (TileGridTransform)T.Copy();

                    if(VolumeTransform != null)
                        copy.Add(VolumeTransform);

                    //Don't include the tile if the mapped version doesn't have any triangles
                    if (copy.mapPoints.Length > 2)
                        listTiles.Add(copy);

                    T.MinimizeMemory();
                    copy.MinimizeMemory();
                }

                //OK, overwrite the tiles in our class
                this._TileTransforms = listTiles.ToArray();
                this.HasBeenWarped = true;

                //Try to save the transform to our cache
                SaveToCache(); 
            }
        }

        /// <summary>
        /// Maps a point from volume space into the section space
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public override bool TryVolumeToSection(GridVector2 P, out GridVector2 transformedP)
        {
            return this.VolumeTransform.TryInverseTransform(P, out transformedP); 
        }

        /// <summary>
        /// Maps a point from section space into the volume space
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public override bool TrySectionToVolume(GridVector2 P, out GridVector2 transformedP)
        {
            return this.VolumeTransform.TryTransform(P, out transformedP); 
        }

        public override TilePyramid VisibleTiles(GridRectangle VisibleBounds, double DownSample)
        {
            GridQuad VisibleQuad = null;
            //Add any corners of the VisibleBounds that we can transform to the list of points
            List<MappingGridVector2> VisiblePoints = VisibleBoundsCorners(VolumeTransform, VisibleBounds);
            if (VisiblePoints.Count == 4)
            {
                VisiblePoints.Sort(new MappingGridVector2SortByMapPoints());
                VisibleQuad = new GridQuad(VisiblePoints[0].MappedPoint,
                                           VisiblePoints[1].MappedPoint,
                                           VisiblePoints[2].MappedPoint,
                                           VisiblePoints[3].MappedPoint);
            }

            return VisibleTiles(VisibleBounds, VisibleQuad, DownSample);
        }
    }
}
