using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Threading;
using Geometry;

namespace Viking.VolumeModel
{
    /// <summary>
    /// A tile is a combination of:
    ///     A unique identifier determined by texture file name
    ///     A texture, which may or may not be loaded
    ///     A set of verticies to position the tile in space
    /// </summary>
    public class Tile
    {
        /// <summary>
        /// Currently a combination of section number, transform, downsample, and texturename
        /// </summary>
        public readonly string UniqueKey;

        public static string CreateUniqueKey(int Section, string Transform, string Channel, int Downsample, string TextureName)
        {
            return "S: " + Section.ToString("D4") + " T: " + Transform + " C: " + Channel + " DS: " + Downsample.ToString("D3") + " Tex: " + TextureName;
        }
        
        /// <summary>
        /// Cache this because we'll use it a lot
        /// </summary>
        private readonly int _UniqueKeyHashcode;
        public override int GetHashCode()
        {
            return _UniqueKeyHashcode;
        }

        public override bool Equals(object obj)
        {
            Tile t = obj as Tile;
            if (t == null)
                return false;

            if (t.UniqueKey == this.UniqueKey)
                return true;

            return false;
        }
        
      //  public static string ConstructUniqueKey(int SectionNumber, string Texture
        public readonly PositionNormalTextureVertex[] Verticies;

        /// <summary>
        /// Indicies passed to render call specifying triangle verticies
        /// </summary>
        public readonly int[] TriangleIndicies = null;

        /// <summary>
        /// Size of the tile in memory
        /// </summary>
        public readonly int Size;

        /// <summary>
        /// Size the texture will require
        /// </summary>
        public readonly int TextureSize;

        /// <summary>
        /// The amount the tile has been downsampled by
        /// </summary>
        public readonly int Downsample;    

        public readonly string TextureFileName;

        //Need to initialize or union operator of bounds property won't work
        private GridRectangle _Bounds = new GridRectangle(double.MaxValue, double.MinValue, double.MaxValue, double.MinValue);
        public GridRectangle Bounds
        {
            get
            {
                if (_Bounds.Width <= 0)
                {
                    foreach (PositionNormalTextureVertex v in Verticies)
                    {
                        _Bounds.Union(new GridVector2(v.Position.X, v.Position.Y));
                    }

                }

                return _Bounds; 
            }
        }
        
        //PORT: public readonly string TextureCachedFileName;

        //PORT: This is a viewModel concept
        //private int MipMapLevels = 1;

        public Tile(string UniqueKey, 
                                PositionNormalTextureVertex[] verticies,
                                int[] TriangleIndicies,
                                string textureFileName,
                                int downsample,
                                int size)
        {
            this.UniqueKey = UniqueKey;
            this._UniqueKeyHashcode = UniqueKey.GetHashCode();

            this.Size = (verticies.Length * 8 * 8) + (TriangleIndicies.Length * 4) + textureFileName.Length + UniqueKey.Length;
            this.TextureSize = size; 
            this.Downsample = downsample;
            this.TextureFileName = textureFileName;
            //PORT: this.TextureCachedFileName = cachedTextureFileName;
            //PORT: this.MipMapLevels = mipMapLevels;

            this.Verticies = verticies;
            this.TriangleIndicies = TriangleIndicies;
        }
        
        public static PositionNormalTextureVertex[] CalculateVerticies(TileGridTransform transform)
        {
            PositionNormalTextureVertex[] verticies = new PositionNormalTextureVertex[transform.mapPoints.Length];

            for (int i = 0; i < transform.mapPoints.Length; i++)
            {
                GridVector2 CtrlP = transform.mapPoints[i].ControlPoint;
                GridVector2 MapP = transform.mapPoints[i].MappedPoint;
                GridVector3 position = new GridVector3(CtrlP.X, CtrlP.Y, 0);
                GridVector2 textureCoord = new GridVector2(MapP.X / transform.ImageWidth, MapP.Y / transform.ImageHeight);

                verticies[i] = new PositionNormalTextureVertex(position, GridVector3.UnitZ, textureCoord);
            }

            return verticies;
        }

        public override string ToString()
        {
            return UniqueKey;
        }
    }
}
