﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml; 
using System.Xml.Linq;
using System.Diagnostics;

using Geometry;

namespace Viking.VolumeModel
{
    /// <summary>
    /// Tile grid mappings refer to a pre-assembled set of tiles, where the tile size is fixed
    /// to the same value at every level of the pyramid, so the area must change
    /// </summary>
    class TileGridMapping : MappingBase
    {
        protected struct GridInfo
        {
            public readonly int GridXDim;
            public readonly int GridYDim;
            public readonly int Downsample;
            public readonly string Path;
           

            public GridInfo(int XDim, int YDim, int downsample, string path)
            {
                GridXDim = XDim;
                GridYDim = YDim;
                Downsample = downsample;
                this.Path = path; 
            }
        }

        protected readonly int TileSizeX;
        protected readonly int TileSizeY;
        protected readonly int TotalTileSize; 

        protected readonly int MaxDownsample = int.MinValue;
        protected readonly int MinDownsample = int.MaxValue;

        protected readonly string TileGridPath;

        private int[] _AvailableLevels = null;
        public override int[] AvailableLevels
        {
            get {
                if (_AvailableLevels == null)
                {
                    _AvailableLevels = new int[LevelToGridInfo.Keys.Count];
                    LevelToGridInfo.Keys.CopyTo(_AvailableLevels, 0);
                }

                return _AvailableLevels;
            }
        }

        protected SortedDictionary<int, GridInfo> LevelToGridInfo = new SortedDictionary<int, GridInfo>(); 

        public override GridRectangle Bounds
        {
            get 
            {
                GridInfo Level = LevelToGridInfo[MaxDownsample];
                return new GridRectangle(0, Level.GridXDim * Level.Downsample * TileSizeX,
                                         0, Level.GridYDim * Level.Downsample * TileSizeY); 
            }
        }

        public override bool TrySectionToVolume(GridVector2 P, out GridVector2 transformedP)
        {
            transformedP = P;
            return true; 
        }

        public override bool TryVolumeToSection(GridVector2 P, out GridVector2 transformedP)
        {
            transformedP = P;
            return true; 
        }

        #region TextureFileNames

        protected string TileFileName(int iX, int iY, int DownsampleLevel)
        {
            string filename = this.TilePrefix + "X" + iX.ToString("D3") + "_Y" + iY.ToString("D3") + this.TilePostfix;

            /* Port
            string tileFileName = this.Section.Path +
                                System.IO.Path.DirectorySeparatorChar + TileGridPath +
                                System.IO.Path.DirectorySeparatorChar + DownsampleLevel.ToString("D3") +
                                System.IO.Path.DirectorySeparatorChar + filename;
             */

            string tileFileName = TileGridPath +
                                System.IO.Path.DirectorySeparatorChar + DownsampleLevel.ToString("D3") +
                                System.IO.Path.DirectorySeparatorChar + filename;

            return tileFileName;
        }

        protected string TileTextureFileName(int iX, int iY)
        {
            string filename = this.TilePrefix + "X" + iX.ToString("D3") + "_Y" + iY.ToString("D3") + this.TilePostfix;

            return filename; 
        }

        /*PORT:
        protected string TileCacheName(int iX, int iY, int DownsampleLevel)
        {
            string filename = this.TilePrefix + "X" + iX.ToString("D3") + "_Y" + iY.ToString("D3") + this.TilePostfix;
            string cachePath = State.CachePath +
                                System.IO.Path.DirectorySeparatorChar + Section.volume.Name +
                                System.IO.Path.DirectorySeparatorChar + Section.SectionSubPath +
                                System.IO.Path.DirectorySeparatorChar + TileGridPath +
                                System.IO.Path.DirectorySeparatorChar + DownsampleLevel.ToString("D3") +
                                System.IO.Path.DirectorySeparatorChar + filename;
            return cachePath;
        }
         */

        #endregion

        protected TileGridMapping(TileGridMapping ToCopy, Section section, string name) :
            base(section, name, ToCopy.TilePrefix, ToCopy.TilePostfix)
        {
            TileSizeX = ToCopy.TileSizeX; 
            TileSizeY = ToCopy.TileSizeY;
            TotalTileSize = ToCopy.TotalTileSize; 
            TileGridPath = ToCopy.TileGridPath;
            MinDownsample = ToCopy.MinDownsample;
            MaxDownsample = ToCopy.MaxDownsample; 

            foreach(GridInfo info in ToCopy.LevelToGridInfo.Values)
            {
                GridInfo infoCopy = new GridInfo(info.GridXDim, info.GridYDim, info.Downsample, info.Path); 
                LevelToGridInfo.Add(infoCopy.Downsample, infoCopy); 
            }
        }

        public TileGridMapping(XElement TilesetNode, Section section, string name, string Prefix, string Postfix) :
            base(section, name, Prefix, Postfix )
        {
            TileSizeX = System.Convert.ToInt32(TilesetNode.Attribute("TileXDim").Value);
            TileSizeY = System.Convert.ToInt32(TilesetNode.Attribute("TileYDim").Value);
            TotalTileSize = TileSizeX * TileSizeY; 
            TileGridPath = TilesetNode.Attribute("path").Value;

            foreach (XNode node in TilesetNode.Nodes())
            {
                XElement elem = node as XElement;
                if (elem == null)
                    continue;

                int Downsample;
                int GridDimX;
                int GridDimY;
                string LevelPath; 

                //Fetch the name if we know it
                switch (elem.Name.LocalName)
                {
                    case "Level":
                        Downsample = System.Convert.ToInt32(elem.Attribute("Downsample").Value);
                        GridDimX = System.Convert.ToInt32(elem.Attribute("GridDimX").Value);
                        GridDimY = System.Convert.ToInt32(elem.Attribute("GridDimY").Value);
                        LevelPath = elem.Attribute("path").Value;

                        if (Downsample > MaxDownsample)
                            MaxDownsample = Downsample;

                        if (Downsample < MinDownsample)
                            MinDownsample = Downsample; 

                        GridInfo Level = new GridInfo(GridDimX, GridDimY, Downsample, LevelPath);
                        LevelToGridInfo.Add(Downsample, Level); 

                        break; 
                }

            }
        }

        protected virtual PositionNormalTextureVertex[] CalculateVerticies(int iX, int iY, int roundedDownsample)
        {
            PositionNormalTextureVertex[] verticies = new PositionNormalTextureVertex[4];

            verticies[0] = new PositionNormalTextureVertex(new GridVector3(iX * this.TileSizeX * roundedDownsample, iY * this.TileSizeY * roundedDownsample, 0),
                                                               GridVector3.UnitZ,
                                                           new GridVector2(0, 0));
            verticies[1] = new PositionNormalTextureVertex(new GridVector3((iX + 1) * this.TileSizeX * roundedDownsample, iY * this.TileSizeY * roundedDownsample, 0),
                                                               GridVector3.UnitZ,
                                                           new GridVector2(1, 0));
            verticies[2] = new PositionNormalTextureVertex(new GridVector3(iX * this.TileSizeX * roundedDownsample, (iY + 1) * this.TileSizeY * roundedDownsample, 0),
                                                               GridVector3.UnitZ,
                                                           new GridVector2(0, 1));
            verticies[3] = new PositionNormalTextureVertex(new GridVector3((iX + 1) * this.TileSizeX * roundedDownsample, (iY + 1) * this.TileSizeY * roundedDownsample, 0),
                                                               GridVector3.UnitZ,
                                                           new GridVector2(1, 1));

            return verticies; 
        }

        protected virtual int[] TriangleEdges
        {
            get { return new int[] {0,1, 2, 1, 3, 2};}
        }

        public override TilePyramid VisibleTiles(GridRectangle VisibleBounds, double DownSample)
        {
            TilePyramid VisibleTiles = new TilePyramid(VisibleBounds);
            int roundedDownsample = NearestAvailableLevel(DownSample); 

            //Starting with low-res tiles, add tiles to the list until we reach desired resolution
            List<Tile> TilesToDraw = new List<Tile>(); 

            //Find the starting level of our rendering
            int iLevel = AvailableLevels.Length - 1;
            int level = AvailableLevels[iLevel];

            do
            {
                List<Tile> newTiles = RecursiveVisibleTiles(
                    VisibleBounds,
                    level 
                    //PORT: AsynchTextureLoad
                    );

                //Insert at the beginning so we overwrite earlier tiles with poorer resolution
                VisibleTiles.AddTiles(level, newTiles);
                //TilesToDraw.AddRange(newTiles);

                iLevel--;
                if(iLevel >= 0)
                    level = AvailableLevels[iLevel];
            }
            while (level >= roundedDownsample && iLevel >= 0);

            //Trace.WriteLine("Drawing " + TilesToDraw.Count.ToString() + " Tiles", "VolumeModel");

            return VisibleTiles; 
        }


        private List<Tile> RecursiveVisibleTiles(GridRectangle VisibleBounds, int roundedDownsample)
        {
            GridInfo gridInfo = LevelToGridInfo[roundedDownsample]; 

            int ScaledTileSizeX = this.TileSizeX * roundedDownsample;
            int ScaledTileSizeY = this.TileSizeX * roundedDownsample; 

            //Figure out which grid locations are visible
            int iMinX = (int)Math.Floor(VisibleBounds.Left / ScaledTileSizeX);
            int iMinY = (int)Math.Floor(VisibleBounds.Bottom / ScaledTileSizeY);
            int iMaxX = (int)Math.Ceiling(VisibleBounds.Right / ScaledTileSizeX);
            int iMaxY = (int)Math.Ceiling(VisibleBounds.Top / ScaledTileSizeY);

            iMinX = iMinX < 0 ? 0 : iMinX;
            iMinY = iMinY < 0 ? 0 : iMinY;
            iMaxX = iMaxX >= gridInfo.GridXDim ? gridInfo.GridXDim : iMaxX;
            iMaxY = iMaxY >= gridInfo.GridYDim ? gridInfo.GridYDim : iMaxY;

            if (iMaxX < 0)
                iMaxX = 0;
            if (iMaxY < 0)
                iMaxY = 0;
            if (iMinX > iMaxX)
                iMinX = iMaxX;
            if (iMinY > iMaxY)
                iMinY = iMaxY; 

            int ExpectedTileCount = (iMaxX - iMinX) * (iMaxY - iMinY);
            List<Tile> TilesToDraw = new List<Tile>(ExpectedTileCount);

            for (int iX = iMinX; iX < iMaxX; iX++)
            {
                for (int iY = iMinY; iY < iMaxY; iY++)
                {
                    string UniqueID = Tile.CreateUniqueKey(Section.Number, Name, Name, roundedDownsample, this.TileTextureFileName(iX, iY));
                    string TextureFileName = TileFileName(iX, iY, roundedDownsample);
                    Tile tile = Global.TileCache.Fetch(UniqueID);
                    if (tile == null)
                    {
                        //First create a new tile
                        //PORT: string TextureCacheFileName = TileCacheName(iX, iY, roundedDownsample);
                        PositionNormalTextureVertex[] verticies = CalculateVerticies(iX, iY, roundedDownsample);
                        int MipMapLevels = 1; //No mip maps

                        if (roundedDownsample == this.AvailableLevels[AvailableLevels.Length - 1])
                            MipMapLevels = 0; //Generate mipmaps for lowest res texture

                       
                        tile = Global.TileCache.ConstructTile(UniqueID,
                                                            verticies,
                                                            this.TriangleEdges,
                                                            TextureFileName,
                                                            //PORT TextureCacheFileName,
                                                            this.Name,
                                                            roundedDownsample, 
                                                            MipMapLevels,
                                                            TotalTileSize);

                        //Check for tiles at higher resolution
                        int iTempX = iX / 2;
                        int iTempY = iY / 2;
                        int iTempDownsample = roundedDownsample * 2;



                    }

                    TilesToDraw.Add(tile);
                }
            }

            return TilesToDraw;
        }
    }
}
