﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;
using System.Diagnostics;
using System.Web;
using System.IO;

namespace Viking.VolumeModel
{
    class TilesToSectionMapping : FixedTileCountMapping
    {
        /// <summary>
        /// Starts as false since we don't load transforms from the disk by default.  Once we do this it is set to true. 
        /// </summary>
        protected bool HasBeenLoaded = false;
        private object LockObj = new object(); 

        public override TileGridTransform[] TileTransforms
        {
            get 
            {
                if (HasBeenLoaded == false)
                    LoadTransform(); 

                return _TileTransforms;
            }
        }

        protected string RootPath; 
        /// <summary>
        /// Path to the .mosaic file containing the transforms
        /// </summary>
        protected string MosaicPath;


        public override string CachedTransformsFileName
        {
            get {
                string mosaicName = System.IO.Path.GetFileNameWithoutExtension(MosaicPath);
                return Section.volume.LocalCachePath +
                    System.IO.Path.DirectorySeparatorChar +
                    Section.volume.Name +
                    System.IO.Path.DirectorySeparatorChar +
                    Section.Number.ToString("D4") +
                    "_" + mosaicName + ".cache"; 
            }
        }

        

        public TilesToSectionMapping(Section section, string name, string rootPath, string mosaicPath, string tilePrefix, string tilePostfix) :
            base(section, name, tilePrefix, tilePostfix)
        {
            this.RootPath = rootPath;
            this.MosaicPath = mosaicPath;
        }

        public override bool TrySectionToVolume(GridVector2 P, out GridVector2 transformedP)
        {
            transformedP = P; 
            return true; 
        }

        public override bool TryVolumeToSection(GridVector2 P, out GridVector2 transformedP)
        {
            transformedP = P; 
            return true; 
        }


        public override void FreeMemory()
        {
            lock (LockObj)
            {
                HasBeenLoaded = false;
                _TileTransforms = null; 
            }
        }

        /// <summary>
        /// Loads the transform from the storage device
        /// </summary>
        public void LoadTransform()
        {
            lock (LockObj)
            {
                if (HasBeenLoaded)
                    return;

                bool LoadedFromCache = false;
                if (System.IO.File.Exists(this.CachedTransformsFileName))
                {
                    try
                    {
                        this._TileTransforms = LoadFromCache();
                        LoadedFromCache = true;
                    }
                    catch (Exception e)
                    {
                        //On any error, use the traditional path
                        this._TileTransforms = null;
                        LoadedFromCache = false;
                    }

                    if (this._TileTransforms == null)
                        LoadedFromCache = false;
                }

                if (LoadedFromCache)
                {
                    this.HasBeenLoaded = true;
                    return;
                }

                Uri mosaicURI = new Uri(this.RootPath + '/' + MosaicPath);
                
                System.Net.HttpWebRequest request = System.Net.WebRequest.CreateDefault(mosaicURI) as System.Net.HttpWebRequest;
                if (mosaicURI.Scheme.ToLower() == "https")
                    request.Credentials = this.Section.volume.UserCredentials;

                request.CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.CacheIfAvailable);
                request.Timeout = 300000;
                request.AutomaticDecompression = System.Net.DecompressionMethods.GZip;

                System.Net.WebResponse response = request.GetResponse();
                Trace.WriteLine(mosaicURI + " from cache: " + response.IsFromCache, "VolumeModel");
                Stream MosaicData = response.GetResponseStream();
                string[] MosaicLines = Geometry.Common.StreamToLines(MosaicData);
                MosaicData.Close();
                response.Close();

                this._TileTransforms = TileGridTransform.LoadMosaic(this.RootPath,  MosaicLines);

                HasBeenLoaded = true;

                SaveToCache();
            }
        }



        public override TilePyramid VisibleTiles(GridRectangle VisibleBounds, double DownSample)
        {
            return base.VisibleTiles(VisibleBounds, null, DownSample);
        }
    }
}
