using System;
using System.IO;
using System.Net;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel; 
using System.Diagnostics;
using System.Linq;
using System.Xml.Linq;
using Ionic.Zip; 

using Geometry;

namespace Viking.VolumeModel 
{
    /// <summary>
    /// Collection of volumes, sections and tiles. There is only one dataset loaded at a time
    /// </summary>
    public class Volume
    {
        /// <summary>z
        /// Friendly name for the volume
        /// </summary>
        public string Name = "";

        /// <summary>
        /// Name of the volume transform to use by default
        /// </summary>
        public string DefaultVolumeTransform = "None";

        private string _UniqueID = "";
        /// <summary>
        /// Unique ID for this volume on the server
        /// </summary>
        public string UniqueID
        {
            get { return _UniqueID; }
        }

        /// <summary>
        /// Set to true if the volume is located on the local drive
        /// False if over a network
        /// </summary>
        private bool _IsLocal;
        public bool IsLocal
        {
            get { return _IsLocal; }

        }

        /// <summary>
        /// Credentials to use during web requests
        /// </summary>
        public System.Net.NetworkCredential UserCredentials = new System.Net.NetworkCredential("anonymous", "connectome");
        
        private XDocument _VolumeXML;

        /// <summary>
        /// The XML document used to initialize the volume.  Contains all configuration settings from the server.
        /// </summary>
        public XDocument VolumeXML
        {
            get
            {
                return _VolumeXML; 
            }
        }

        public Section[] Sections
        {
            get { return SectionsTable.Values.ToArray(); }
        }

        /// <summary>
        /// Names of transform groups that can be used to register images into the volume
        /// </summary>
        public List<string> VolumeTransformNames = new List<string>(new string[] {"None"}); 
        
        private Dictionary<int, int> SectionToReferenceSectionBelow = new Dictionary<int, int>(); 


        /// <summary>
        /// The server the volume transforms and image data is located on
        /// </summary>
        private string _Host;
        public string Host
        {
            get { return _Host; }
        }

       
        /// <summary>
        /// The path we use to cache data on the local drive
        /// </summary>
        internal readonly string LocalCachePath;

        /// <summary>
        /// Maps a section number to its section object
        /// </summary>
        public SortedList<int, Section> SectionsTable = new SortedList<int,Section>();

        /// <summary>
        /// Sorted list containing the transforms for each volume transform we find
        /// Key = Downsample level
        /// Value = Dictionary mapping each section number to a stos transform.  This is because section numbers may not be continuos
        /// </summary>
        public SortedList<string, SortedList<int, StosGridTransform>> Transforms = new SortedList<string, SortedList<int, StosGridTransform>>();

        public int NumSections
        {
            get { return SectionsTable.Count; }
        }

        /// <summary>
        /// Returns the section that the passed section was registered to
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public Section GetReferenceSectionBelow(Section section)
        {
            if (section == null)
                return null;

            //Optimistic implementation that looks at section immediately above
            int refnumber = section.Number - 1;
            int minSectionNumber = SectionsTable.Keys.Min();
            while (refnumber >= minSectionNumber)
            {
                if (SectionsTable.ContainsKey(refnumber))
                    return SectionsTable[refnumber];
                refnumber--;
            }

            return null; 
        }

        /// <summary>
        /// Returns the section that the passed section was registered to
        /// </summary>
        /// <param name="?"></param>
        /// <returns></returns>
        public Section GetReferenceSectionAbove(Section section)
        {
            if (section == null)
                return null;

            //Optimistic implementation that looks at section immediately above
            int refnumber = section.Number + 1;
            int maxSectionNumber = SectionsTable.Keys.Max();
            while (refnumber <= maxSectionNumber)
            {
                if (SectionsTable.ContainsKey(refnumber))
                    return SectionsTable[refnumber];
                refnumber++;
            }

            return null; 
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="path">The host and path to the volume, no filenames</param>
        /// <param name="localCachePath">LocaL cache path corresponding to the path</param>
        /// <param name="workerThread">optional worker thread to report progress</param>
        public Volume(string path, string localCachePath, System.ComponentModel.BackgroundWorker workerThread)
        {
            //Load the default settings from user preferences
            ChannelInfo DefaultChannel = new ChannelInfo();
            DefaultChannels = new ChannelInfo[0];  

            this._Host = path;
            this.LocalCachePath = localCachePath; 

            XDocument VolumeXML = Load(path, workerThread);
            Initialize(VolumeXML, workerThread); 
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="path">The host and path to the volume, no filenames</param>
        /// <param name="localCachePath">LocaL cache path corresponding to the path</param>
        /// <param name="workerThread">optional worker thread to report progress</param>
        public Volume(string path, string localCachePath, XDocument VolumeXML, System.ComponentModel.BackgroundWorker workerThread)
        {
            //Load the default settings from user preferences
            ChannelInfo DefaultChannel = new ChannelInfo();
            DefaultChannels = new ChannelInfo[0];

            this._Host = path;
            this.LocalCachePath = localCachePath;

            Initialize(VolumeXML, workerThread); 
        }


        #region Channels
        
        private ChannelInfo[] _DefaultChannels = new ChannelInfo[0];

        public ChannelInfo[] DefaultChannels
        {
            get { return _DefaultChannels; }
            set
            {
                if (null == value)
                {
                    _DefaultChannels = new ChannelInfo[0];
                    return;
                }

                _DefaultChannels = value;
            }
        }

        //A list of all channel names found in the volume
        //TODO: Modify to a per section basis?
        static private List<string> _ChannelNames = new List<String>();

        /// <summary>
        /// A list of all channel names found on sections in the volume
        /// </summary>
        public string[] ChannelNames
        {
            get { return _ChannelNames.ToArray(); }
        }

        public void AddChannel(string name)
        {
            if (false == _ChannelNames.Contains(name))
            {
                _ChannelNames.Add(name);
                _ChannelNames.Sort();
            }
        }

        public void RemoveChannel(string name)
        {
            if (false == _ChannelNames.Contains(name))
                return;

            _ChannelNames.Remove(name);
        }

        #endregion

        /// <summary>
        /// Loads a path, determines whether path refers to XML file or a local directory
        /// </summary>
        /// <param name="path"></param>
        protected XDocument Load(string path, BackgroundWorker workerThread)
        {
            Uri uri = new Uri(path);

            workerThread.ReportProgress(0, "Requesting " + path);

            XDocument XMLInitData; 
            if (uri.Scheme == "http" || uri.Scheme == "https")
                XMLInitData = LoadHTTP(path);
            else
                XMLInitData = LoadLocal(uri.LocalPath);

            return XMLInitData; 
        }

        /// <summary>
        /// Loads a path, determines whether path refers to XML file or a local directory
        /// </summary>
        /// <param name="path"></param>
        protected void Load(XDocument volumeXML, BackgroundWorker workerThread)
        {
            this._VolumeXML = volumeXML;

            DateTime start = DateTime.Now;

            Initialize(this._VolumeXML, workerThread); 

            //Create a path for the cache
            string VolumeCachePath = this.LocalCachePath +
                System.IO.Path.DirectorySeparatorChar +
                this.Name;

            if (System.IO.Directory.Exists(VolumeCachePath) == false)
                System.IO.Directory.CreateDirectory(VolumeCachePath);

            CreateVolumeTransforms(workerThread);

            DateTime end = DateTime.Now;

            Trace.WriteLine("Volume load time: " + new TimeSpan(end.Ticks - start.Ticks).TotalSeconds.ToString("D2"));

            workerThread.ReportProgress(101, "Done!"); 
        }

        protected XDocument LoadHTTP(string path)
        {
            //Remove the .xml file from the path
            int iRemove = path.LastIndexOf('/');
            string VolumePath = path;
            if (iRemove > 0)
            {
                VolumePath = VolumePath.Remove(iRemove);
            }

            //This code remains, but the value is replaced if a value is found in the XML file
            this._Host = VolumePath;

            this._IsLocal = false;
            Uri pathURI = new Uri(path); 

            HttpWebRequest request = WebRequest.Create(path) as HttpWebRequest;
            if (pathURI.Scheme.ToLower() == "https")
                request.Credentials = this.UserCredentials; 

            request.CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.Revalidate);

            WebResponse response = null; 
            try
            {
                response = request.GetResponse();
            }
            catch (WebException e)
            {
                /*PORT: Don't have forms, throw a better exception*/
                throw new WebException("Error connecting to volume server: \n" + path + "\n" + e.Message, e);

            }

            Stream responseStream = response.GetResponseStream();

            StreamReader XMLStream = new StreamReader(responseStream);

            XDocument reader = XDocument.Parse(XMLStream.ReadToEnd());

            XMLStream.Close();
            responseStream.Close();
            response.Close();

            return reader; 
        }

        
        protected XDocument LoadLocal(string path)
        {            
            FileStream XMLFileStream = File.OpenRead(path); 
            StreamReader XMLStreamReader = new StreamReader(XMLFileStream);

            XDocument reader = XDocument.Parse(XMLStreamReader.ReadToEnd());

            XMLStreamReader.Close();
            XMLFileStream.Close();

            return reader;
        }

        private static Ionic.Zip.ZipFile FetchStosZip(Uri StosZipPath, System.Net.NetworkCredential UserCredentials)
        {
            HttpWebRequest request = WebRequest.Create(StosZipPath) as HttpWebRequest;
            if (StosZipPath.Scheme.ToLower() == "https")
                request.Credentials = UserCredentials;

            request.CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.Revalidate);

            WebResponse response = null;
            try
            {
                response = request.GetResponse();
            }
            catch (WebException e)
            {
                /*PORT: Don't have forms, throw a better exception*/
                Trace.WriteLine("Error connecting to volume server: \n" + StosZipPath + "\n" + e.Message, "VolumeModel");
                return null; 
            }

            ZipFile zipFile = null; 

            try
            {
                Stream responseStream = response.GetResponseStream();


                Byte[] buffer = Global.ReadToBuffer(responseStream, response.ContentLength);
                MemoryStream memStream = new MemoryStream(buffer);

                zipFile = ZipFile.Read(memStream);

                responseStream.Close();
                //ZipFile takes a reference to the stream instead of copying it (which it should not do).  We cannot close the stream.
                //memStream.Close();
            }
            catch (Exception e)
            {
                Trace.WriteLine("Could not open StosZip file", "VolumeModel");
            }

            return zipFile; 
        }

        void Initialize(XDocument reader, BackgroundWorker workerThread)
        {
            List<CreateSectionThreadingObj> ListSectionThreadingObj = new List<CreateSectionThreadingObj>();
            List<CreateStosGridTransformThreadingObj> ListStosGridTransformThreadingObj = new List<CreateStosGridTransformThreadingObj>();

            this._VolumeXML = reader; 
            //Fetch the volume information which should be the top level of the XML

            //Search for the correct node in the XML
            IEnumerable<XElement> VolumeElements = reader.Elements().Where(e => e.Name.LocalName == "Volume");
            if (VolumeElements.Count() == 0)
            {
                Trace.WriteLine("No volume node found in the XML", "VolumeModel");
                throw new InvalidDataException("No volume node found in the VikingXML");
            }

            XElement volumeElement = VolumeElements.First();

            this.Name = volumeElement.Attribute("name").Value;
            int NumStosFiles = System.Convert.ToInt32(volumeElement.Attribute("num_stos").Value);
            int NumSections = System.Convert.ToInt32(volumeElement.Attribute("num_sections").Value);

            if (volumeElement.Attribute("path") != null)
                this._Host = volumeElement.Attribute("path").Value;
            else
            {
                /* PORT
                System.Windows.Forms.MessageBox.Show("Could locate path attribute for volume.  Chances are the XML definitation for this volume has not been updated. Contact administrator to update the VikingXML file.", "Error", System.Windows.Forms.MessageBoxButtons.OK);
                if (this._Host == null) //If we don't know a path throw an exception to kill the process
                    throw new ArgumentException("Could locate path attribute for volume.  Chances are the XML definitation for this volume has not been updated. Contact administrator to update the VikingXML file.");
                 */
            }

            //Remove a trailing slash
            if (this._Host.EndsWith("/"))
                this._Host = this._Host.TrimEnd('/');


            if (volumeElement.Attribute("UniqueID") != null)
                this._UniqueID = volumeElement.Attribute("UniqueID").Value;

            ZipFile StosZipFile = null;
            if (volumeElement.Attribute("StosZip") != null)
            {
                string StosZipFileName = volumeElement.Attribute("StosZip").Value;
                workerThread.ReportProgress(0, "Loading compressed transform file " + StosZipFileName);
                StosZipFile = FetchStosZip(new Uri(Host + '/' + StosZipFileName), this.UserCredentials);
            }

            int countStos = 0;
            int countSections = 0;
            ListSectionThreadingObj.Capacity = NumSections;
            ListStosGridTransformThreadingObj.Capacity = NumStosFiles; 

            foreach (XNode node in volumeElement.Nodes().ToList<XNode>())
            {
                if (node.NodeType == System.Xml.XmlNodeType.Whitespace)
                    continue;

                XElement elem = node as XElement;
                if (elem == null)
                    continue;

                //Fetch the name if we know it
                switch (elem.Name.LocalName.ToLower())
                {
                    case "stos":
                        string StosFileName = elem.Attribute("path").Value;
                        string StosPath = this.Host + Path.DirectorySeparatorChar + StosFileName;
                        
                  //      int pixelSpacing = System.Convert.ToInt32(elem.Attribute("pixelSpacing").Value);
                        int ProgressPercent = (countStos * 100) / NumStosFiles;
                        countStos++;
                        workerThread.ReportProgress(ProgressPercent, "Loading " + StosPath);

                        //string groupName = type + " " + pixelSpacing.ToString();
                        

                  //      XAttribute GroupNameAttribute = elem.Attribute("GroupName");
                  //      if (GroupNameAttribute != null)
                  //      {
                  //          groupName = GroupNameAttribute.Value;
                  //      }

                  //      if (false == VolumeTransformNames.Contains(groupName))
                  //      {
                  //          VolumeTransformNames.Add(groupName);
                  //      }

                       

                        //StosGridTransform stosTransform = null;
                        CreateStosGridTransformThreadingObj CreateStosThreadObj = null;

                        if (StosZipFile != null)
                        {
                            if (StosZipFile.ContainsEntry(StosFileName))
                            {
                                Trace.WriteLine("Loading " + StosFileName + " from Zip", "VolumeModel");
                                ZipEntry entry = StosZipFile[StosFileName];
                                Byte[] buffer = new Byte[entry.UncompressedSize];

                                MemoryStream memStream = new MemoryStream(buffer);
                                entry.Extract(memStream);

                                //stosTransform = new StosGridTransform(memStream, elem);
                                CreateStosThreadObj = new CreateStosGridTransformThreadingObj(memStream, elem);

                                //memStream.Close(); 
                            }
                        }

                        //Load from server if it is not in the zip
                        if (CreateStosThreadObj == null)
                        {
                        //    Trace.WriteLine("Loading " + StosFileName + " from HTTP Server", "VolumeModel");
                            CreateStosThreadObj = new CreateStosGridTransformThreadingObj(StosPath, this.UserCredentials, elem);
                            //stosTransform = new StosGridTransform(StosPath, elem, this.UserCredentials);
                        }

                        ListStosGridTransformThreadingObj.Add(CreateStosThreadObj); 
                        System.Threading.ThreadPool.QueueUserWorkItem(CreateStosThreadObj.ThreadPoolCallback);

                        

                        break;
                    case "section":
                        //string SectionPath = VolumePath + '/' + elem.Attribute("path").Value;
                        string SectionPath = elem.Attribute("path").Value;


                        if (NumSections > 0)
                        {
                            ProgressPercent = (countSections * 100) / NumSections;
                        }
                        else
                        {
                            ProgressPercent = 100;
                        }

                        countSections++;
                        workerThread.ReportProgress(ProgressPercent, "Queueing " + SectionPath);

                        CreateSectionThreadingObj newCreateSectionThreadObj = new CreateSectionThreadingObj(this,
                            SectionPath,
                            elem);
                        ListSectionThreadingObj.Add(newCreateSectionThreadObj);

                        System.Threading.ThreadPool.QueueUserWorkItem(newCreateSectionThreadObj.ThreadPoolCallback);

                        //                        newCreateSectionThreadObj.ThreadPoolCallback(null);
                        //                        Section s = new Section(this, SectionPath, elem.CreateReader());
                        //                        this.sections.Add(s.Number, s);

                        break;
                    case "channelinfo":
                        DefaultChannels = ChannelInfo.FromXML(elem);
                        break;
                    default:
                        break;
                }

            }

            {
                workerThread.ReportProgress(0, "Waiting for Stos Transform Loading Threads");
                int countFinished = 0;

                while (ListStosGridTransformThreadingObj.Count > 0)
                {
                    for (int iObj = 0; iObj < ListStosGridTransformThreadingObj.Count; iObj++)
                    {
                        CreateStosGridTransformThreadingObj CreateStosGridTransformObj = ListStosGridTransformThreadingObj[iObj];

                        //Test to see if the wait state is set
                        bool Result = CreateStosGridTransformObj.DoneEvent.WaitOne(0);
                        if (Result == true)
                        {
                            ListStosGridTransformThreadingObj.RemoveAt(iObj);
                            iObj--;
                            countFinished++;
                            int Progress;
                            if (NumStosFiles > 0)
                                Progress = (countFinished * 100) / NumStosFiles;
                            else
                                Progress = 100;

                            workerThread.ReportProgress(Progress, "Loaded " + CreateStosGridTransformObj.stosTransform.ToString());

                            XElement elem = CreateStosGridTransformObj.element; 

                            int pixelSpacing = System.Convert.ToInt32(elem.Attribute("pixelSpacing").Value);
                            string type = elem.Attribute("type").Value;
                            string groupName = type + " " + pixelSpacing.ToString();
                            
                            XAttribute GroupNameAttribute = CreateStosGridTransformObj.element.Attribute("GroupName");
                            if (GroupNameAttribute != null)
                            {
                                groupName = GroupNameAttribute.Value;
                            }

                            if (false == VolumeTransformNames.Contains(groupName))
                            {
                                VolumeTransformNames.Add(groupName);
                            }

                            if (this.DefaultVolumeTransform == "None")
                                this.DefaultVolumeTransform = groupName;

                            if (this.Transforms.ContainsKey(groupName))
                            {
                                SortedList<int, StosGridTransform> transformDict = this.Transforms[groupName];
                                transformDict.Add(CreateStosGridTransformObj.stosTransform.MappedSection, CreateStosGridTransformObj.stosTransform);
                            }
                            else
                            {
                                SortedList<int, StosGridTransform> transformDict = new SortedList<int, StosGridTransform>();
                                transformDict.Add(CreateStosGridTransformObj.stosTransform.MappedSection, CreateStosGridTransformObj.stosTransform);
                                Transforms.Add(groupName, transformDict);
                            }
                        }
                    }

                    System.Threading.Thread.Sleep(100);
                }

            }



            {
                //            XMLStream.Close();
                //            response.Close();
                workerThread.ReportProgress(0, "Waiting for Section Loading Threads");

                int countFinished = 0;
                while (ListSectionThreadingObj.Count > 0)
                {
                    for (int iObj = 0; iObj < ListSectionThreadingObj.Count; iObj++)
                    {
                        CreateSectionThreadingObj CreateSectionObj = ListSectionThreadingObj[iObj];

                        //Test to see if the wait state is set
                        bool Result = CreateSectionObj.DoneEvent.WaitOne(0);
                        if (Result == true)
                        {
                            ListSectionThreadingObj.RemoveAt(iObj);
                            iObj--;
                            countFinished++;
                            int Progress;
                            if (NumSections > 0)
                                Progress = (countFinished * 100) / NumSections;
                            else
                                Progress = 100;

                            workerThread.ReportProgress(Progress, "Loaded " + CreateSectionObj.newSection.ToString());
                            this.SectionsTable.Add(CreateSectionObj.newSection.Number, CreateSectionObj.newSection);
                        }
                    }

                    System.Threading.Thread.Sleep(100);
                }
            }

            //Create a path for the cache
            string VolumeCachePath = this.LocalCachePath +
                System.IO.Path.DirectorySeparatorChar +
                this.Name;

            if (System.IO.Directory.Exists(VolumeCachePath) == false)
                System.IO.Directory.CreateDirectory(VolumeCachePath);

            CreateVolumeTransforms(workerThread);

            workerThread.ReportProgress(101, "Done!");

        }

        /// <summary>
        /// Adds a transform to each section mapping it into each of the volume spaces we found
        /// </summary>
        public void CreateVolumeTransforms(BackgroundWorker workerThread)
        {
            foreach (string TransformKey in Transforms.Keys)
            {
                //The transform list is sorted by which section the transform maps from. 
                //Next we'll add transfroms so every transform maps from the mapped section to section #1
                SortedList<int, StosGridTransform> TList = Transforms[TransformKey];
                
                foreach (int iSection in TList.Keys)
                {
                    StosGridTransform trans = TList[iSection];
                    if (false == SectionsTable.ContainsKey(trans.MappedSection))
                        continue; 

                    Section S = SectionsTable[trans.MappedSection];
 //                   S.AddVolumeTransform(trans, "slice-to-slice " + TransformKey.ToString());

                    if (trans.ControlSection == 1)
                        continue;

                    if(false == TList.ContainsKey(trans.ControlSection))
                        continue;

                    //Find the section that can map our transform
                    StosGridTransform addTrans = TList[trans.ControlSection];

                    //Add this mapping to our dictionary:
                    if (SectionToReferenceSectionBelow.ContainsKey(trans.MappedSection) == false)
                    {
                        SectionToReferenceSectionBelow.Add(trans.MappedSection, trans.ControlSection);
                    }

                    //NOTE: Assumes volumes use the same mappings across all downsamplings
                    //Sections should register to section 1, but if a volume hasn't finished registration or was done in parts we may register to a section other than 1
                    //Debug.Assert(addTrans.ControlSection == 1);
                    string outString = "Adding transforms: " + trans.ToString() + " to " + addTrans.ToString();
                 //   Trace.WriteLine(outString, "VolumeModel");
                    workerThread.ReportProgress((iSection * 100) / TList.Keys[TList.Count - 1], outString);
                    trans.Add(addTrans);

                    //Let windows process some events
                    //PORT, can't do this without referencing System.Windows
                    //System.Windows.Forms.Application.DoEvents();
                }
            }
        }
    }
}
