﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using Viking.VolumeModel;
using System.Windows;
using System.Windows.Input; 


namespace Viking.VolumeViewModel
{
    /// <summary>
    /// This class encapsulates a mapping of tiles into space, a bounding region, and a downsample level.
    /// Using these three properties it dynamically updates a collection of TileViewModels for display
    /// </summary>
    public class TileMappingViewModel : DependencyObject
    {
        private readonly MappingBase TileMapping = null;

        private TilePyramid TilePyramid = null;

        private static TileViewModelCache TileCache = new TileViewModelCache();
        
        private ObservableCollection<TileViewModel> _Tiles = new ObservableCollection<TileViewModel>();
        public ObservableCollection<TileViewModel> Tiles
        {
            get { return _Tiles; }
        }

        private static readonly DependencyProperty VisibleRectProperty;
        
        /// <summary>
        /// Returns the downsample level of the pyramid used for the current downsample level
        /// </summary>
        public int PyramidDownsampleLevel
        {
            get { return TileMapping.NearestAvailableLevel(Downsample); }
        }

        /// <summary>
        /// The visible area
        /// </summary>
        public Rect VisibleRect
        {
            get { return (Rect)this.GetValue(VisibleRectProperty); }
            set { this.SetValue(VisibleRectProperty, value); }
        }

        /// <summary>
        /// The center of the current view
        /// </summary>
        public Point ViewCenter
        {
           get 
           { 
               Rect visRect = VisibleRect;
               return new Point(visRect.X + (visRect.Width / 2), visRect.Y + (visRect.Height / 2));
           }
            
           set
           {
               Rect visRect = VisibleRect;
               visRect.X = value.X - (visRect.Width / 2);
               visRect.Y = value.Y - (visRect.Height / 2);
               VisibleRect = visRect; 
           }
        }

        private static readonly DependencyProperty DownsampleProperty;

        /// <summary>
        /// How downsampled the tiles should be
        /// </summary>
        public double Downsample
        {
            get { return (double)this.GetValue(DownsampleProperty); }
            set { this.SetValue(DownsampleProperty, value); }
        }


        /// <summary>
        /// The boundary of the entire tile mapping
        /// </summary>
        public Rect Bounds
        {
            get
            {
                Geometry.GridRectangle grect = this.TileMapping.Bounds;
                Rect bounds = new Rect(grect.Left, grect.Bottom, grect.Width, grect.Height);
                return bounds; 
            }
        }

        public override string ToString()
        {
            if(TileMapping != null)
                return this.TileMapping.ToString();

            return "Empty Tile Mapping";
        } 
        
        static TileMappingViewModel()
        {
            TileMappingViewModel.VisibleRectProperty = DependencyProperty.Register("VisibleRect", 
                                                                                   typeof(Rect),
                                                                                   typeof(TileMappingViewModel),
                                                                                   new FrameworkPropertyMetadata(new Rect(),
                                                                                                                 FrameworkPropertyMetadataOptions.AffectsRender,
                                                                                                                 new PropertyChangedCallback(OnVisibleRectChanged)));

            TileMappingViewModel.DownsampleProperty = DependencyProperty.Register("Downsample", 
                                                                                   typeof(double),
                                                                                   typeof(TileMappingViewModel),
                                                                                   new FrameworkPropertyMetadata(double.PositiveInfinity,
                                                                                                                 FrameworkPropertyMetadataOptions.AffectsRender,
                                                                                                                 new PropertyChangedCallback(OnDownsampleChanged))); 
        }

        public TileMappingViewModel()
        {
            this.TileMapping = null;

            this.VisibleRect = new Rect(0, 0, 100, 100);

            this.Downsample = 1; 
        }
        
        public TileMappingViewModel(MappingBase mapping)
        {
            this.TileMapping = mapping;
            
            //Set the visible rect to include the entire mappable area
            this.VisibleRect = new Rect(mapping.Bounds.Left, mapping.Bounds.Bottom, mapping.Bounds.Width, mapping.Bounds.Height);

            //Set the downsample to the highest possible value
            this.Downsample = mapping.AvailableLevels[mapping.AvailableLevels.Length - 1];
        }

        protected void VisibleRectChanged(Rect newValue, Rect oldValue)
        {
            if (oldValue.Contains(newValue) == false)
            {
                System.Diagnostics.Trace.WriteLine("VisibleRectChanged"); 
                UpdateTiles(newValue, this.Downsample);
            }
        }

        protected void DownsampleChanged(double newValue, double oldValue)
        {
            int newPyramidLevel = TileMapping.NearestAvailableLevel(newValue);
            int oldPyramidLevel = TileMapping.NearestAvailableLevel(oldValue);

            if (newPyramidLevel != oldPyramidLevel)
            {
                System.Diagnostics.Trace.WriteLine("DownsampleChanged() " + oldPyramidLevel + " -> " + newPyramidLevel); 
                UpdateTiles(this.VisibleRect, newValue);
            }
        }

        private static void OnVisibleRectChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            TileMappingViewModel tileMapping = o as TileMappingViewModel;

            System.Diagnostics.Trace.WriteLine("OnVisibleRectChanged()"); 
            tileMapping.VisibleRectChanged((Rect)e.NewValue, (Rect)e.OldValue);
        }

        private static void OnDownsampleChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
        {
            TileMappingViewModel tileMapping = o as TileMappingViewModel;
            System.Diagnostics.Trace.WriteLine("OnDownsampleChanged()");
            tileMapping.DownsampleChanged((double)e.NewValue, (double)e.OldValue);
        }

        protected void UpdateTiles(Rect visRect, double downsample)
        {
            System.Diagnostics.Trace.WriteLine("UpdateTiles()"); 
            //Find which tiles changed
            //Rect visRect = this.VisibleRect;
            Geometry.GridRectangle visibleGridRect = new Geometry.GridRectangle(visRect.Left, visRect.Right, visRect.Top, visRect.Bottom);
            TilePyramid oldTilePyramid = this.TilePyramid;

            TilePyramid newTilePyramid = TileMapping.VisibleTiles(visibleGridRect, downsample);

            if (oldTilePyramid != null)
            {
                if (false == oldTilePyramid.Bounds.Intersects(newTilePyramid.Bounds))
                    Tiles.Clear();
                else
                {
                    //Might be able to really speed up culling dead tiles with QuadTrees

                    //Sort out which of the old tiles are no longer visible
                    for (int iTile = 0; iTile < Tiles.Count; iTile++)
                    {
                        TileViewModel t = Tiles[iTile];
                        if (false == visibleGridRect.Intersects(t.Bounds))
                        {
                            Tiles.RemoveAt(iTile);
                            iTile--;
                        }
                        else
                        {
                            SortedList<string, Tile> tilesForLevel = newTilePyramid.GetTilesForLevel(t.Downsample);
                        }
                    }
                }
            }

            //Create tiles which are new
            for(int iLevel = newTilePyramid.AvailableLevels.Length -1; iLevel >= 0; iLevel--)
            {
                int level = newTilePyramid.AvailableLevels[iLevel]; 
                SortedList<string, Tile> tiles = newTilePyramid.GetTilesForLevel(level);
                bool IgnoreOldPyramid = oldTilePyramid == null;

                //Don't check the old pyramid if it doesn't have the downsample level we are looking at
                if (false == IgnoreOldPyramid)
                {
                    IgnoreOldPyramid = level < oldTilePyramid.AvailableLevels.Min();

                    if (false == IgnoreOldPyramid)
                    {
                        IgnoreOldPyramid = level > oldTilePyramid.AvailableLevels.Max();
                    }
                }

                foreach (Tile t in tiles.Values)
                {
                    //We already loaded this one
          //          if (!IgnoreOldPyramid && oldTilePyramid.Bounds.Intersects(t.Bounds))
          //              continue;
          //          else
                    {
                        TileViewModel tileViewModel = null;
                       
                        tileViewModel = TileCache.Fetch(t.UniqueKey);
                        if (tileViewModel == null)
                        {
                            tileViewModel = new TileViewModel(t, TileMapping.TilePath);
                           
                            TileCache.Add(t.UniqueKey, tileViewModel);
                        }
//                        else
 //                       {
                            //Don't add it to the collection again
                            if (Tiles.Contains(tileViewModel))
                                continue; 

                        //}
                    
                        this.Tiles.Add(tileViewModel);
                    }
                }
            }

            TilePyramid = newTilePyramid;
        }

    }
}
