﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;

using Geometry;

using Viking.Common;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using WebAnnotation.UI;
using WebAnnotationModel;
using WebAnnotation.ViewModel; 

namespace WebAnnotation
{
    /// <summary>
    /// Sort LocationObj's according to rendering requirements
    /// </summary>
    public class LocationObjDrawOrderComparison : IComparer<LocationObj>, IComparer<Location>
    {
        #region IComparer<LocationObj> Members

        int IComparer<LocationObj>.Compare(LocationObj x, LocationObj y)
        {
            //First sort by type
            if (x.TypeCode != y.TypeCode)
            {
                return x.TypeCode - y.TypeCode;
            }

            //if the type is the same, sort by section differential
            return (int)(x.Section - y.Section);
        }

        #endregion

        int IComparer<Location>.Compare(Location x, Location y)
        {
            //First sort by type
            if (x.TypeCode != y.TypeCode)
            {
                return x.TypeCode - y.TypeCode;
            }

            //if the type is the same, sort by section differential
            return (int)(x.Section - y.Section);
        }
    }

    /// <summary>
    /// This class draws LocationObj's
    /// </summary>
    static class LocationObjRenderer
    {
        static BlendState OriginalBlendState;
        static RasterizerState OriginalRasterState; 

        /// <summary>
        /// Setup the graphicsDevice and effect to render the type of obj passed
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="basicEffect"></param>
        /// <param name="obj"></param>
        public static void SetupGraphicsDevice(GraphicsDevice graphicsDevice, BasicEffect basicEffect, Location obj, long SectionNumber)
        {
            //oldVertexDeclaration = graphicsDevice.VertexDeclaration;
            OriginalBlendState = graphicsDevice.BlendState;
            OriginalRasterState = graphicsDevice.RasterizerState;
            

            //PORT XNA 4
            /*
            if (null == GlobalPrimitives.VertexPositionColorTextureDecl)
            {
                //Create vertex declaration if we need one
                GlobalPrimitives.VertexPositionColorTextureDecl = new VertexDeclaration(graphicsDevice, VertexPositionColorTexture.VertexElements);
            }
            */

            BlendState newBlendState = new BlendState();
            
            newBlendState.AlphaSourceBlend = Blend.SourceAlpha;
            newBlendState.AlphaDestinationBlend = Blend.InverseSourceAlpha;
            newBlendState.ColorSourceBlend = Blend.SourceAlpha;
            newBlendState.ColorDestinationBlend = Blend.InverseSourceAlpha;

            RasterizerState rasterState = new RasterizerState();
            rasterState.FillMode = FillMode.Solid;
            rasterState.CullMode = CullMode.None; 
                       
            //graphicsDevice.RenderState.FillMode = FillMode.Solid;
            //PORT XNA 4
            //graphicsDevice.RenderState.PointSize = 5.0f;

            graphicsDevice.RasterizerState = rasterState; 
            graphicsDevice.BlendState = newBlendState; 
            
            basicEffect.LightingEnabled = false;

            int SectionDelta = (int)(obj.Section - SectionNumber);

            switch (obj.TypeCode)
            {
                case LocationType.POINT:
                    basicEffect.Texture = null;
                    basicEffect.TextureEnabled = false;
                    basicEffect.VertexColorEnabled = true;
                    break;
                case LocationType.CIRCLE:
                    //Are we rendering on the same section or above/below?
                    if (SectionDelta < 0)
                    {
                        basicEffect.Texture = GlobalPrimitives.UpArrowTexture;
                    }
                    else if (SectionDelta == 0)
                    {
                        basicEffect.Texture = GlobalPrimitives.CircleTexture;
                    }
                    else
                    {
                        basicEffect.Texture = GlobalPrimitives.DownArrowTexture;
                    }

                    basicEffect.TextureEnabled = true;
                    basicEffect.VertexColorEnabled = true;
                    basicEffect.LightingEnabled = false; 

                    break;
                default:
                    break; 
            }
            
            //basicEffect.CommitChanges();
            //graphicsDevice.RasterizerState = OriginalRaster
            //graphicsDevice.BlendState = OriginalBlendState; 
        }

        /// <summary>
        /// Put the graphics device state back where we found it
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="basicEffect"></param>
        public static void RestoreGraphicsDevice(GraphicsDevice graphicsDevice, BasicEffect basicEffect)
        {
            if(OriginalBlendState != null)
                graphicsDevice.BlendState = OriginalBlendState;

            if(OriginalRasterState != null)
                graphicsDevice.RasterizerState = OriginalRasterState; 
            
            basicEffect.Texture = null;
            basicEffect.TextureEnabled = false;
            basicEffect.VertexColorEnabled = false;
        }

        /// <summary>
        /// Draw the list of locations as they should appear for the given section number
        /// </summary>
        /// <param name="Locations"></param>
        /// <param name="graphicsDevice"></param>
        /// <param name="basicEffect"></param>
        /// <param name="SectionNumber"></param>
        public static void DrawBackgrounds(List<Location> listToDraw, GraphicsDevice graphicsDevice, BasicEffect basicEffect, int SectionNumber)
        {
            if (listToDraw.Count == 0)
                return;

            IComparer<Location> LocComparer = new LocationObjDrawOrderComparison();
            int iStart = 0;
            
            do
            {
                int iEnd = listToDraw.Count; //Need to initialize or loop never ends
                Location StartingObj = listToDraw[iStart];

                //Set the graphics device state to render the appropriate type
                SetupGraphicsDevice(graphicsDevice, basicEffect, StartingObj, SectionNumber);

                //PORT XNA 4
                //basicEffect.Begin();

                foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
                {
                    //PORT XNA 4
                    //pass.Begin();
                    pass.Apply(); 

                    for (int iObj = iStart; iObj < listToDraw.Count; iObj++)
                    {
                        Location locToDraw = listToDraw[iObj];

                        //If it is a different type of location than we have the render state setup for, bail out of the loop
                        if (LocComparer.Compare(StartingObj, locToDraw) != 0)
                        {
                            iEnd = iObj;
                            break;
                        }

                        //Draw the object with the current settings
                        locToDraw.DrawBackground(graphicsDevice, (int)((long)SectionNumber - locToDraw.Section));
                    }

                    //PORT XNA 4
                    //pass.End();
                }

                //PORT XNA 4
                //basicEffect.End();

                //Remove the drawn objects from the list
                iStart = iEnd; 
                //listToDraw.RemoveRange(0, iEnd);

                RestoreGraphicsDevice(graphicsDevice, basicEffect);
            }
            while (iStart < listToDraw.Count);

           

            RestoreGraphicsDevice(graphicsDevice, basicEffect); 
        }
    }
}
