﻿using System;
using System.Windows.Forms; 
using System.Collections.Generic;
using System.Collections.Specialized; 
using System.Linq;
using System.Text;
using System.Diagnostics;
using Viking; 
using Viking.Common;
using Viking.ViewModels;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Geometry;
using WebAnnotation.UI.Commands;
using WebAnnotation.ViewModel;
using WebAnnotationModel;
using connectomes.utah.edu.XSD.WebAnnotationUserSettings.xsd;

namespace WebAnnotation
{
    [Viking.Common.SectionOverlay("Annotation")]
    class AnnotationOverlay : Viking.Common.ISectionOverlayExtension
    {
        static public float LocationTextScaleFactor = 5;
        static public float ReferenceLocationTextScaleFactor = 2.5f;
        static public float RadiusToResizeCircle = 7.0f / 8.0f;
        static public float RadiusToLinkCircle = 1.0f / 4.0f;

        Viking.UI.Controls.SectionViewerControl _Parent;
        public Viking.UI.Controls.SectionViewerControl Parent { get { return _Parent; } }

        private TransformChangedEventHandler SectionChangedEventHandler;
        private EventHandler ReferenceSectionChangedEventHandler;
        private EventHandler AnnotationChangedEventHandler;
        //private NotifyCollectionChangedEventHandler LocationsChangedEventHandler;
        //private NotifyCollectionChangedEventHandler StructuresChangedEventHandler; 

        private static SortedDictionary<int, SectionLocationsViewModel> dictSectionAnnotations = new SortedDictionary<int, SectionLocationsViewModel>();
        private static LocationLinksViewModel linksView;

        private static AnnotationOverlay _CurrentOverlay = null; 
        public static AnnotationOverlay CurrentOverlay { get { return _CurrentOverlay;}}

        GridVector2 LastMouseDownCoords;

        /// <summary>
        /// The last object the mouse was over, if any
        /// </summary>
        internal static IUIObjectBasic LastMouseOverObject = null;

        public AnnotationOverlay()
        {
            SectionChangedEventHandler = new TransformChangedEventHandler(this.OnSectionTransformChanged);

            ReferenceSectionChangedEventHandler = new EventHandler(OnReferenceSectionChanged);
            AnnotationChangedEventHandler = new EventHandler(OnAnnotationChanged);

            Store.Locations.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnLocationCollectionChanged);

            //    AnnotationCache.AnnotationChanged += AnnotationChangedEventHandler;      
            _CurrentOverlay = this; 
        }

        public int CurrentSectionNumber
        {
            get
            { 
                return _Parent.Section.Number; 
            }
        }

        /// <summary>
        /// Returns annotations for section if they exist or null if they do not
        /// </summary>
        /// <param name="SectionNumber"></param>
        public static SectionLocationsViewModel GetAnnotationsForSection(int SectionNumber)
        {
            if (dictSectionAnnotations.ContainsKey(SectionNumber))
            {
                return dictSectionAnnotations[SectionNumber];
            }

            return null; 
        }
        
        

        public SectionLocationsViewModel CurrentSectionAnnotations
        {
            get
            {
                if (dictSectionAnnotations.ContainsKey(_Parent.Section.Number))
                {
                    return dictSectionAnnotations[_Parent.Section.Number];
                }
                else
                {
                    LoadSectionAnnotations();
                    return dictSectionAnnotations[_Parent.Section.Number];
                }
            }
        }

        /// <summary>
        /// Returns the location nearest to the mouse, prefers the locations on the current section
        /// </summary>
        /// <param name="position"></param>
        /// <returns></returns>
        public Location GetNearestLocation(GridVector2 position, out double BestDistance)
        {
            BestDistance = double.MaxValue;
            SectionLocationsViewModel locView = GetAnnotationsForSection(CurrentSectionNumber);
            if (locView == null)
                return null;

            Location bestObj;
            bestObj = locView.GetNearestLocation(position, out BestDistance);

            Location adjacentObj;
            double distance;

            if (_Parent.Section.ReferenceSectionAbove != null)
            {
                locView = GetAnnotationsForSection(_Parent.Section.ReferenceSectionAbove.Number);
                if (locView != null)
                {
                    adjacentObj = locView.GetNearestLocation(position, out distance);
                    if (adjacentObj != null)
                    {
                        if (distance < BestDistance)
                        {
                            if (distance < adjacentObj.OffSectionRadius)
                            {
                                bestObj = adjacentObj;
                                BestDistance = distance;
                            }
                        }
                    }
                }
            }

            if (_Parent.Section.ReferenceSectionBelow != null)
            {
                locView = GetAnnotationsForSection(_Parent.Section.ReferenceSectionBelow.Number);
                if (locView != null)
                {
                    adjacentObj = locView.GetNearestLocation(position, out distance);
                    if (adjacentObj != null)
                    {
                        if (distance < BestDistance)
                        {
                            if (distance < adjacentObj.OffSectionRadius)
                            {
                                bestObj = adjacentObj;
                                BestDistance = distance;
                            }
                        }
                    }
                }
            }

            return bestObj; 
        }

        /// <summary>
        /// Find the location nearest the provided point on the section
        /// </summary>
        /// <param name="position"></param>
        /// <returns></returns>
        public IUIObjectBasic GetNearestAnnotation(GridVector2 position, out double distance)
        {
            distance = double.MaxValue; 
            SectionLocationsViewModel locView = GetAnnotationsForSection(CurrentSectionNumber); 
            double BestDistance = double.MaxValue;
            IUIObjectBasic bestObj = null; 
            if (locView != null)
            {
                bestObj = locView.GetNearestAnnotation(position, out BestDistance);
            }
            
            Location adjacentObj; 
            if (_Parent.Section.ReferenceSectionAbove != null)
            {
                locView = GetAnnotationsForSection(_Parent.Section.ReferenceSectionAbove.Number);
                if (locView != null)
                {
                    adjacentObj = locView.GetNearestLocation(position, out distance);
                    if (adjacentObj != null)
                    {
                        if (distance < BestDistance)
                        {
                            if (distance < adjacentObj.OffSectionRadius)
                            {
                                bestObj = adjacentObj;
                                BestDistance = distance;
                            }
                        }
                    }
                }
            }

            if (_Parent.Section.ReferenceSectionBelow != null)
            {
                locView = GetAnnotationsForSection(_Parent.Section.ReferenceSectionBelow.Number);
                if (locView != null)
                {
                    adjacentObj = locView.GetNearestLocation(position, out distance);
                    if (adjacentObj != null)
                    {
                        if (distance < BestDistance)
                        {
                            if (distance < adjacentObj.OffSectionRadius)
                            {
                                bestObj = adjacentObj;
                                BestDistance = distance;
                            }
                        }
                    }
                }
            }

            //Only check for links if there are no annotations in range
            if (bestObj == null)
            {
                IUIObjectBasic bestLink;
                bestLink = linksView.GetNearestLink(CurrentSectionNumber, position, out distance);
                if (bestLink != null)
                {
                    if (distance < BestDistance)
                    {
                        BestDistance = distance;
                        bestObj = bestLink;
                    }
                }
            }

            distance = BestDistance; 

            return bestObj; 
        }

        #region ISectionOverlayExtension Members

        public void SetParent(Viking.UI.Controls.SectionViewerControl parent)
        {
            //I'm only expecting this to be set once
            Debug.Assert(_Parent == null, "Not expecting parent to be set twice, OK to ignore, but annotation display may be incorrect"); 
            this._Parent = parent;

            //Load the locations for the current sections
            this._Parent.OnSectionChanged += new SectionChangedEventHandler(this.OnSectionChanged);
            this._Parent.OnSectionTransformChanged += new TransformChangedEventHandler(this.OnSectionTransformChanged);
            this._Parent.OnVolumeTransformChanged += new TransformChangedEventHandler(this.OnVolumeTransformChanged); 
            this._Parent.OnReferenceSectionChanged += new EventHandler(this.OnReferenceSectionChanged);

            this._Parent.MouseDown += new MouseEventHandler(this.OnMouseDown);
            this._Parent.MouseMove += new MouseEventHandler(this.OnMouseMove);
            this._Parent.MouseUp += new MouseEventHandler(this.OnMouseUp); 
            this._Parent.KeyDown += new KeyEventHandler(this.OnKeyDown);
            this._Parent.KeyUp += new KeyEventHandler(this.OnKeyUp);

           // AnnotationCache.parent = parent;

            GlobalPrimitives.CircleTexture = parent.Content.Load<Texture2D>("Circle");
            GlobalPrimitives.UpArrowTexture = parent.Content.Load<Texture2D>("UpArrow");
            GlobalPrimitives.DownArrowTexture = parent.Content.Load<Texture2D>("DownArrow");

            linksView = new LocationLinksViewModel(parent); 

            LoadSectionAnnotations(); 
        }
        
        protected void OnMouseMove(object sender, MouseEventArgs e)
        {
            //Check if there is a non-default command. we don't want to mess with another active command
            if (_Parent.CurrentCommand.GetType() != typeof(Viking.UI.Commands.DefaultCommand) ||
             Viking.UI.Commands.Command.QueueDepth > 0)
                return;

            //If locations aren't visible they can't be selected
            if (!_Parent.ShowOverlays)
            {
                _Parent.Cursor = Cursors.Default;
                return;
            }

            double distance;
            GridVector2 WorldPosition = _Parent.ScreenToWorld(e.X, e.Y);

            IUIObjectBasic NextMouseOverObject = GetNearestAnnotation(WorldPosition, out distance);
            if (NextMouseOverObject != LastMouseOverObject)
                Parent.Invalidate();

            LastMouseOverObject = NextMouseOverObject; 
            
            /*Check to see if we are over a location*/
            
            Location loc = LastMouseOverObject as Location; // GetNearestLocation(WorldPosition, out distance);
            if (loc == null)
            {
                _Parent.Cursor = Cursors.Default;
            }
            else
            {
                //If the loc is on this section we check if we are close to the edge and we are resizing.  Everyone else gets move cursor
                if (loc.TypeCode == LocationType.CIRCLE)
                {
                    GridVector2 locPosition = loc.VolumePosition; 

                    distance = GridVector2.Distance(locPosition, WorldPosition);
                    if (loc.Section == _Parent.Section.Number)
                    {
                        if (distance <= loc.Radius)
                        {
                            if (distance >= (loc.Radius * RadiusToResizeCircle))
                            {
                                _Parent.Cursor = Cursors.SizeAll;
                            }
                            else if (distance >= (loc.Radius * RadiusToLinkCircle))
                            {
                                _Parent.Cursor = Cursors.Arrow;
                            }
                            else
                            {
                                _Parent.Cursor = Cursors.Hand;
                            }
                        }
                    }
                    else
                    {
                        if (distance <= loc.OffSectionRadius)
                        {
                            LastMouseOverObject = loc;
                            _Parent.Cursor = Cursors.Cross;
                        }
                    }
                }
            }

            
        }

        protected void OnMouseDown(object sender, MouseEventArgs e)
        {
            //Check if there is a non-default command. we don't want to mess with another active command
            if (_Parent.CurrentCommand.GetType() != typeof(Viking.UI.Commands.DefaultCommand) ||
                Viking.UI.Commands.Command.QueueDepth > 0)
                return; 
            
            //If locations aren't visible they can't be selected
            if (!_Parent.ShowOverlays)
                return;
            
            GridVector2 WorldPosition = _Parent.ScreenToWorld(e.X, e.Y);
            LastMouseDownCoords = WorldPosition; 

            //Left mouse button selects objects
            if (e.Button == MouseButtons.Left)
            {
                double distance;
                IUIObjectBasic obj = GetNearestAnnotation(WorldPosition, out distance);
                //Figure out if it is resizing a location circle
                //If the loc is on this section we check if we are close to the edge and we are resizing.  Everyone else gets standard location command
                Viking.UI.State.SelectedObject = obj as IUIObjectBasic;
                Location loc = obj as Location;

                if (loc != null)
                {

                    GridVector2 locPosition = loc.VolumePosition;

                    if (loc.TypeCode == LocationType.CIRCLE &&
                        loc.Section == _Parent.Section.Number)
                    {
                        if (distance <= loc.Radius)
                        {
                            if (distance >= (loc.Radius * RadiusToResizeCircle))
                            {
                                _Parent.CurrentCommand = new ResizeLocationCommand(Parent);
                            }
                            else if (distance >= (loc.Radius * RadiusToLinkCircle))
                            {
                                _Parent.CurrentCommand = new LinkAnnotationsCommand(Parent, loc);
                            }
                            else
                            {
                                _Parent.Cursor = Cursors.SizeAll;
                            }
                        }
                    }
                }
                
            }
        }

        protected void OnMouseUp(object sender, MouseEventArgs e)
        {
            //Check if there is a non-default command. we don't want to mess with another active command
            if (_Parent.CurrentCommand.GetType() != typeof(Viking.UI.Commands.DefaultCommand) ||
                Viking.UI.Commands.Command.QueueDepth > 0)
                return; 

            if (e.Button == MouseButtons.Left)
            {
                if (LastMouseOverObject == null)
                {
                    OnContinueLastTrace(LastMouseDownCoords); 
                }

            }
        }

        protected void OnKeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                //Refresh the annotations on F5
                case Keys.F5:
                    CurrentSectionAnnotations.LoadSectionAnnotations();
                    return;
                case Keys.F3:
                    OnContinueLastTrace();
                    return;
                case Keys.Enter:
                    OnContinueLastTrace();
                    return; 
                case Keys.Back:
                    if (CreateNewLinkedLocationCommand.LastEditedLocation != null)
                    {
                        Parent.GoToLocation(new Microsoft.Xna.Framework.Vector2((float)CreateNewLinkedLocationCommand.LastEditedLocation.SectionPosition.X,
                                                                                (float)CreateNewLinkedLocationCommand.LastEditedLocation.SectionPosition.Y),
                                            (int)CreateNewLinkedLocationCommand.LastEditedLocation.Z,
                                            true,
                                            (double)((CreateNewLinkedLocationCommand.LastEditedLocation.Radius * 2) / Parent.Width) * 2); 

                    }
                    return; 

            }

            try
            {
                //Search the list of hotkeys for a match
                IEnumerable<Hotkey> matchingKeys = Global.UserSettings.Shortcuts.Hotkey.Where(h => h.KeyCode == e.KeyCode);
                foreach (Hotkey h in matchingKeys)
                {
                    if (!(h.Shift == e.Shift &&
                           h.Ctrl == e.Control &&
                           h.Alt == e.Alt))
                        continue; 
                
                    //OK, we have a match, invoke the command
                    //Check if there is a non-default command. we don't want to mess with another active command
                    if (_Parent.CurrentCommand.GetType() != typeof(Viking.UI.Commands.DefaultCommand) ||
                        Viking.UI.Commands.Command.QueueDepth > 0)
                        return;

                    connectomes.utah.edu.XSD.WebAnnotationUserSettings.xsd.Action a = Global.UserSettings.Actions.Action.Where(action => action.Name == h.Action).SingleOrDefault();
                    if (a != null)
                    {
                        System.Type commandType;
                        object[] parameters;

                        a.ExecuteAction(out commandType, out parameters);
                        if (commandType != null)
                        {
                            Viking.UI.Commands.Command.EnqueueCommand(commandType, parameters);
                        }
                    }

                    CreateStructureCommandAction comAction = Global.UserSettings.Actions.CreateStructureCommandAction.Where(action => action.Name == h.Action).SingleOrDefault();
                    if (comAction != null)
                    {
                        OnCreateStructure(System.Convert.ToInt64(comAction.TypeID));
                    }


                    


                        /*
                        System.Type t = System.Type.GetType(Param.Type);
                        if(t == null)
                        {
                            Trace.WriteLine("Could not get type of hotkey command: " + h.Type);
                            return; 
                        }

                        if(t.IsValueType)
                        {
                            //paramaters[iParam] = System.Convert.ChangeType(Param.Value, t);
                        }
                        else
                        {
                           // paramaters[iParam] = Activator.CreateInstance(t, 
                        }
                    

                    //Viking.UI.Commands.Command.EnqueueCommand();
                     */
                   
                }
            }
            catch(Exception except)
            {
                Trace.WriteLine("Error with hotkeys: " + except.ToString()); 
            }
             
        }

        protected void OnKeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Space:

                    //Only load the annotations for any section once so we can't fire multiple requests by pounding the spacebar
                    if (CurrentSectionAnnotations.HaveLoadedSectionAnnotations == false)
                    {
                        CurrentSectionAnnotations.LoadSectionAnnotations();
                    }

                    this._Parent.Invalidate();

                    break;
            }
        }

        protected void OnCreateStructure(long TypeID)
        {
            StructureTypeObj typeObj = Store.StructureTypes.GetObjectByID(TypeID);
            if (typeObj != null)
            {
                StructureType type = new StructureType(typeObj);

                System.Drawing.Point ClientPoint = _Parent.PointToClient(System.Windows.Forms.Control.MousePosition);
                GridVector2 WorldPos = _Parent.ScreenToWorld(ClientPoint.X, ClientPoint.Y);
                GridVector2 SectionPos;
                bool success = _Parent.TryVolumeToSection(WorldPos, _Parent.Section, out SectionPos);
                Debug.Assert(success);
                if (!success)
                    return;

                
                StructureObj newStruct = new StructureObj(type.modelObj);
                LocationObj newLocation = new LocationObj(newStruct,
                                                SectionPos,
                                                WorldPos,
                                                Parent.Section.Number);

                Structure newStructView = new Structure(newStruct);
                Location newLocationView = new Location(newLocation);

                if (type.Parent == null)
                {
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(ResizeLocationCommand), new object[] { Parent, type, newLocationView });
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(CreateNewStructureCommand), new object[] { Parent, newStructView, newLocationView });
                }
                else
                {
                    //Enqueue two commands to resize the location and then select a parent
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(ResizeLocationCommand), new object[] { Parent, type, newLocationView });
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(LinkStructureToParentCommand), new object[] { Parent, newStructView, newLocationView });
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(CreateNewStructureCommand), new object[] { Parent, newStructView, newLocationView });
                }
            }
            else
                Trace.WriteLine("Could not find hotkey ID for type: " + TypeID.ToString()); 

        }

        protected void OnContinueLastTrace()
        {
            System.Drawing.Point ClientPoint = _Parent.PointToClient(System.Windows.Forms.Control.MousePosition);
            GridVector2 WorldPos = _Parent.ScreenToWorld(ClientPoint.X, ClientPoint.Y);
            OnContinueLastTrace(WorldPos); 
        }

        protected void OnContinueLastTrace(GridVector2 WorldPos)
        {
            if (CreateNewLinkedLocationCommand.LastEditedLocation != null)
            {
                Location template = CreateNewLinkedLocationCommand.LastEditedLocation;
                if (template.Z != this.Parent.Section.Number)
                {
                    GridVector2 SectionPos;
                    bool success = _Parent.TryVolumeToSection(WorldPos, _Parent.Section, out SectionPos);
                    Debug.Assert(success);
                    if (!success)
                        return;

                    LocationObj newLoc = new LocationObj(CreateNewLinkedLocationCommand.LastEditedLocation.Parent.modelObj,
                                        SectionPos,
                                        WorldPos,
                                        Parent.Section.Number);

                    Location newLocView = new Location(newLoc);

                    Viking.UI.Commands.Command.EnqueueCommand(typeof(ResizeLocationCommand), new object[] { Parent, template.Parent.Type, newLocView });
                    Viking.UI.Commands.Command.EnqueueCommand(typeof(CreateNewLinkedLocationCommand), new object[] { Parent, template, newLocView });

                    Viking.UI.State.SelectedObject = null;
                    CreateNewLinkedLocationCommand.LastEditedLocation = null; 
                }
            }
        }
         

        /// <summary>
        /// This occurs when a new section is loaded and we need to fetch all locations from scratch
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void OnSectionChanged(object sender, SectionChangedEventArgs e)
        {
            
            //Don't load annotations when flipping sections if the user is holding down space bar to hide them
            if (_Parent.ShowOverlays)
            {
                LoadSectionAnnotations();
            }
        }

        protected void OnAnnotationChanged(object sender, EventArgs e)
        {
            //Trigger redraw of screen
            _Parent.Invalidate(); 
        }

        /// <summary>
        /// Organize the changes so we only call the SectionAnnotationViewModel objects that we have to
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void OnLocationCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            SortedList<int, int> changedSections = new SortedList<int, int>();

  //          if (e.Action == NotifyCollectionChangedAction.Replace)
  //          {
                //SortedList<int, List<LocationObj>> changedSectionsNewObjects = new SortedList<int, List<LocationObj>>(e.NewItems.Count);
//                SortedList<int, List<LocationObj>> changedSectionsOldObjects = new SortedList<int, List<LocationObj>>(e.NewItems.Count);

            if (e.NewItems != null)
            {
                for (int iObj = 0; iObj < e.NewItems.Count; iObj++)
                {
                    LocationObj locNewObj = e.NewItems[iObj] as LocationObj;
                    changedSections[locNewObj.Section] = locNewObj.Section;
                }
            }

            if (e.OldItems != null)
            {
                for (int iObj = 0; iObj < e.OldItems.Count; iObj++)
                {
                    LocationObj locOldObj = e.OldItems[iObj] as LocationObj;
                    changedSections[locOldObj.Section] = locOldObj.Section;
                }

            }

            foreach (int section in changedSections.Keys)
            {
                if (false == dictSectionAnnotations.ContainsKey(section))
                {
                    continue;
                }

                dictSectionAnnotations[section].OnLocationsStoreChanged(sender, e);
            }

            Parent.Invalidate(); 
        }

                /*
                for (int iObj = 0; iObj < e.NewItems.Count; iObj++)
                {
                    LocationObj locNewObj = e.NewItems[iObj] as LocationObj;
                    LocationObj locOldObj = e.OldItems[iObj] as LocationObj;

                    List<LocationObj> listNewObjs;
                    List<LocationObj> listOldObjs;
                    if (changedSectionsNewObjects.ContainsKey(locNewObj.Section))
                    {
                        listNewObjs = changedSectionsNewObjects[locObj.Section];
                    }
                    else
                    {
                        listNewObjs = new List<LocationObj>(e.NewItems.Count);
                        changedSectionsNewObjects[locNewObj.Section] = listNewObjs;                       
                    }

                    if (changedSectionsOldObjects.ContainsKey(locOldObj.Section))
                    {
                        listOldObjs = changedSectionsOldObjects[locOldObj.Section];
                    }
                    else
                    {
                        listOldObjs = new List<LocationObj>(e.OldItems.Count);
                        changedSectionsOldObjects[locOldObj.Section] = listOldObjs; 
                    }

                    listNewObjs.Add(locNewObj);
                    listOldObjs.Add(locOldObj);
                }

                
            }
            else
            {
                SortedList<int, List<LocationObj>> changedSectionsObjects = new SortedList<int, List<LocationObj>>(e.NewItems.Count);
                
                for (int iObj = 0; iObj < e.NewItems.Count; iObj++)
                {
                    LocationObj locObj = e.NewItems[iObj] as LocationObj;
                    List<LocationObj> listObjs;
                    if (dictSectionAnnotations.ContainsKey(locObj.Section))
                    {
                        listObjs = changedSectionsObjects[locObj.Section];
                    }
                    else
                    {
                        listObjs = new List<LocationObj>(e.NewItems.Count);
                        changedSectionsObjects[locObj.Section] = listObjs;
                    }

                    listObjs.Add(locObj);
                }
            }
            */

            //Call each sectionViewModel and update
    //    }

        /// <summary>
        /// When this occurs we need to update the reference locations
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void OnReferenceSectionChanged(object sender, EventArgs e)
        {
            ///This could be optimized, but it should be a rare event
            LoadSectionAnnotations();
        }

        /// <summary>
        /// When this occurs we should update the positions we draw the locations at. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void OnSectionTransformChanged(object sender, TransformChangedEventArgs e)
        {
            //AnnotationCache.TransformLocationsToVolume();
            //AnnotationCache.PopulateLocationLinks(); 
            LoadSectionAnnotations();
        }

        /// <summary>
        /// When this occurs we should update the positions we draw the locations at. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void OnVolumeTransformChanged(object sender, TransformChangedEventArgs e)
        {
            //AnnotationCache.TransformLocationsToVolume();
            //AnnotationCache.PopulateLocationLinks(); 
            LoadSectionAnnotations();
        }

        protected void LoadSectionAnnotations()
        {
            SectionLocationsViewModel SectionAnnotations = null; 
            if (dictSectionAnnotations.ContainsKey(_Parent.Section.Number))
            {
                SectionAnnotations = dictSectionAnnotations[_Parent.Section.Number];
            }
            else
            {
                SectionAnnotations = new SectionLocationsViewModel(_Parent.Section, _Parent);
                SectionAnnotations.AnnotationChanged += AnnotationChangedEventHandler;             
                dictSectionAnnotations.Add(_Parent.Section.Number, SectionAnnotations);
            }

            

            SectionAnnotations.LoadSectionAnnotations();

            if (_Parent.Section.ReferenceSectionAbove != null)
            {
                int refSectionNumber = _Parent.Section.ReferenceSectionAbove.Number;
                if (dictSectionAnnotations.ContainsKey(refSectionNumber))
                {
                    SectionAnnotations = dictSectionAnnotations[refSectionNumber];
                }
                else
                {
                    SectionAnnotations = new SectionLocationsViewModel(_Parent.Section.VolumeViewModel.SectionViewModels[refSectionNumber], _Parent);
                    SectionAnnotations.AnnotationChanged += AnnotationChangedEventHandler;
                    dictSectionAnnotations.Add(refSectionNumber, SectionAnnotations);
                }

                SectionAnnotations.LoadSectionAnnotations();
            }

            if (_Parent.Section.ReferenceSectionBelow != null)
            {
                int refSectionNumber = _Parent.Section.ReferenceSectionBelow.Number;
                if (dictSectionAnnotations.ContainsKey(refSectionNumber))
                {
                    SectionAnnotations = dictSectionAnnotations[refSectionNumber];
                }
                else
                {

                    SectionAnnotations = new SectionLocationsViewModel(_Parent.Section.VolumeViewModel.SectionViewModels[refSectionNumber], _Parent);
                    SectionAnnotations.AnnotationChanged += AnnotationChangedEventHandler;
                    dictSectionAnnotations.Add(refSectionNumber, SectionAnnotations);
                }

                SectionAnnotations.LoadSectionAnnotations();
            }

            linksView.LoadSection(_Parent.Section.Number); 
            //AnnotationCache.LoadSectionAnnotations(_Parent.Section); 
        }

        public IUIObjectBasic NearestObject(GridVector2 WorldPosition, out double distance)
        {
            distance = 0;
            IUIObjectBasic obj = GetNearestAnnotation(WorldPosition, out distance); 
            //IUIObjectBasic obj = CurrentSectionAnnotations.GetNearestAnnotation(WorldPosition, out distance);

            
            //IUIObjectBasic obj = AnnotationCache.GetNearestAnnotation(WorldPosition, out distance);

            return obj; 
        }

        public void Draw(Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice,
                         Microsoft.Xna.Framework.Graphics.BasicEffect basicEffect,
                         Geometry.GridRectangle Bounds,
                         double DownSample)
        {
            if(_Parent.Section == null)
                return;

            if (_Parent.spriteBatch.GraphicsDevice.IsDisposed)
                return;

            RasterizerState OriginalRasterState = graphicsDevice.RasterizerState;
            SectionLocationsViewModel currentSectionAnnotations = CurrentSectionAnnotations;
            Debug.Assert(currentSectionAnnotations != null);

            int SectionNumber = _Parent.Section.Number;

            float Time = (float)TimeSpan.FromTicks(DateTime.Now.Ticks - DateTime.Today.Ticks).TotalSeconds;
//            Debug.WriteLine("Time: " + Time.ToString()); 

            //Get all the lines to draw first so the text and geometric shapes are over top of them
            LocationLink[] VisibleLinks = linksView.VisibleLocationLinks(_Parent.Section.Number, Bounds);

            foreach (LocationLink link in VisibleLinks)
            {                
                Location locA = link.A;
                Location locB = link.B;

                if (locA.Parent == null)
                    continue;

                if (locA.Parent.Type == null)
                    continue; 

                StructureType type = locA.Parent.Type;


                int distanceFactor = link.maxSection - link.minSection;
                if (distanceFactor == 0)
                    distanceFactor = 1;
                

                //Give the colors a nudge towards red or blue depending on the direction to the link
                double directionFactor = 1;
                directionFactor = link.maxSection == SectionNumber ? 1 : -1;

                int red = (int)((float)(type.Color.R * .5) + (128 * directionFactor));
                red = 255 - (red / distanceFactor); 
                red = red > 255 ? 255 : red;
                int blue = (int)((float)(type.Color.B * .5) + (128 * (1 - directionFactor)));
                blue = 255 - (blue / distanceFactor); 
                blue = blue > 255 ? 255 : blue;
                int green = (int)((float)type.Color.G);
                green = 255 - (green / distanceFactor); 

                int alpha = 85;
                if (LastMouseOverObject == link)
                {
                    alpha = 192; 
                }

                //If you don't cast to byte the wrong constructor is used and the alpha value is wrong
                Microsoft.Xna.Framework.Color color = new Microsoft.Xna.Framework.Color((byte)(red),
                    (byte)(green),
                    (byte)(blue),
                    (byte)(alpha));

                _Parent.LineManager.Draw(link.lineGraphic, (float)link.Radius, color,
                                             basicEffect.View * basicEffect.Projection, Time, null);
            }

            //Get all the lines to draw
            StructureLink[] VisibleStructureLinks = currentSectionAnnotations.VisibleStructureLinks(Bounds);
            
            foreach (StructureLink link in VisibleStructureLinks)
            {
                int alpha = 85;
                if (LastMouseOverObject == link)
                {
                    alpha = 192;
                }

                //If you don't cast to byte the wrong constructor is used and the alpha value is wrong
                Microsoft.Xna.Framework.Color color = new Microsoft.Xna.Framework.Color((byte)(255),
                    (byte)(255),
                    (byte)(255),
                    (byte)(alpha));

                if (link.Bidirectional)
                {
                    _Parent.LineManager.Draw(link.lineGraphic, (float)link.Radius, color,
                                             basicEffect.View * basicEffect.Projection, Time, "AnimatedBidirectional");
                }
                else
                {
                    _Parent.LineManager.Draw(link.lineGraphic, (float)link.Radius, color,
                                             basicEffect.View * basicEffect.Projection, Time, "AnimatedLinear");
                }
            }



            Location[] Locations = currentSectionAnnotations.GetLocations().ToArray();
            List<Location> listToDraw = new List<Location>(Locations.Length);

            //Find a circle that encloses the visible bounds
            GridCircle VisibleCircle = new GridCircle(Bounds.Center, GridVector2.Distance(Bounds.Center, new GridVector2(Bounds.Left, Bounds.Top)));
            
            //Draw the text for each location
            foreach (Location loc in Locations)
            {
                if (loc.Parent == null)
                    continue;

                if (loc.Parent.Type == null)
                    continue;

                GridVector2 WorldPosition = loc.VolumePosition; 

                //Find out if we should draw the location
                if (loc.TypeCode == LocationType.CIRCLE)
                {
                    GridCircle locCircle = new GridCircle(WorldPosition, loc.Radius);
                    if (locCircle.Contains(VisibleCircle) == false)
                        continue; 
                }
                else if (loc.TypeCode == LocationType.POINT)
                {
                    if (Bounds.Contains(WorldPosition) == false)
                        continue;
                }
                else
                {
                    Debug.Fail("Unknown location type");
                }
                

                listToDraw.Add(loc); 
            }

            List<Location> RefLocations = new List<Location>();
            if(_Parent.Section.ReferenceSectionBelow != null)
            {
                SectionLocationsViewModel sectionLocations = GetAnnotationsForSection(_Parent.Section.ReferenceSectionBelow.Number);
                if(sectionLocations != null)
                    RefLocations.AddRange(sectionLocations.GetLocations(Bounds)); 
            }

            if(_Parent.Section.ReferenceSectionAbove != null)
            {
                SectionLocationsViewModel sectionLocations = GetAnnotationsForSection(_Parent.Section.ReferenceSectionAbove.Number);
                if (sectionLocations != null)
                {
                    RefLocations.AddRange(sectionLocations.GetLocations(Bounds));
                }
            }
            
            //Draw text for locations on the reference sections
            foreach (Location loc in RefLocations)
            {
                if (loc == null)
                    continue; 

                //If we could not map a reference section we may not have a transformed location, so we should skip them
                if (loc.Parent == null)
                    continue;

                if (loc.Parent.Type == null)
                    continue;

                GridVector2 WorldPosition = loc.VolumePosition;
                
                //Find out if we should draw the location
                if (loc.TypeCode == LocationType.CIRCLE)
                {
                    GridCircle locCircle = new GridCircle(WorldPosition, loc.Radius);
                    if (locCircle.Contains(VisibleCircle) == false)
                        continue;
                }
                else if (loc.TypeCode == LocationType.POINT)
                {
                    if (Bounds.Contains(WorldPosition) == false)
                        continue;
                }
                else
                {
                    Debug.Fail("Unknown location type");
                }

                listToDraw.Add(loc); 
            }

            WebAnnotation.LocationObjRenderer.DrawBackgrounds(listToDraw, graphicsDevice, basicEffect, SectionNumber); 
            /*
            //Draw a rectangle around each LocationObj in this section
            VertexDeclaration oldVertexDeclaration = graphicsDevice.VertexDeclaration;
 
            graphicsDevice.RenderState.FillMode = FillMode.Solid; 
            graphicsDevice.RenderState.PointSize = 5.0f;
            graphicsDevice.VertexDeclaration = _Parent.VertexPositionColorDeclaration; 

            basicEffect.Alpha = 1f;

//            //GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
//            //GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;

            graphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            graphicsDevice.RenderState.DestinationBlend = Blend.DestinationColor;

            basicEffect.LightingEnabled = false; 
            basicEffect.Texture = null;
            basicEffect.TextureEnabled = false;
            basicEffect.VertexColorEnabled = true;
            basicEffect.CommitChanges();


            basicEffect.Begin();
            foreach (EffectPass pass in basicEffect.CurrentTechnique.Passes)
            {
                pass.Begin();

                foreach (LocationObj loc in listToDraw)
                {
                    loc.DrawBackground(graphicsDevice,
                                       (int)((long)SectionNumber - loc.Section));
                }

                pass.End();
            }

            basicEffect.End();

            graphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            graphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;
            */

            
            //Draw text
            _Parent.spriteBatch.Begin(SpriteSortMode.Immediate, BlendState.AlphaBlend);
            
            foreach (Location loc in listToDraw)
            {
                GridVector2 WorldPosition = loc.VolumePosition; 
                
                /*bool Success = currentSectionAnnotations.TryGetPositionForLocation(loc, out WorldPosition);
                if (!Success)
                    continue; 
                */

                GridVector2 DrawPosition = _Parent.WorldToScreen(WorldPosition.X, WorldPosition.Y);

                loc.DrawLabel(_Parent.spriteBatch,
                              _Parent.fontArial,
                              new Vector2((float)DrawPosition.X, (float)DrawPosition.Y),
                              (float)(1 / _Parent.StatusMagnification),
                              (int)((long)SectionNumber - loc.Section));
            }

            _Parent.spriteBatch.End();

            
            //graphicsDevice.RenderState.CullMode = CullMode.None;
            
            //PORT XNA 4
            //graphicsDevice.RenderState.AlphaTestEnable = false;

            graphicsDevice.RasterizerState = OriginalRasterState;

            /*
            basicEffect.VertexColorEnabled = false;
            basicEffect.CommitChanges(); 
            graphicsDevice.VertexDeclaration = oldVertexDeclaration; 
            */
        }

        #endregion

        /// <summary>
        /// Allocates a new quad tree based on the current section parameters
        /// </summary>
        public static GridRectangle SectionBounds(Viking.UI.Controls.SectionViewerControl Parent, int SectionNumber)
        {
            GridRectangle bounds = new GridRectangle(); 
            
            //Figure out the new boundaries for our quad-tree
            if(!Parent.Section.VolumeViewModel.SectionViewModels.ContainsKey(SectionNumber))
                return new GridRectangle();

            SectionViewModel SectionView = Parent.Section.VolumeViewModel.SectionViewModels[SectionNumber];
            bounds = Parent.SectionBounds(SectionView.section);
            if (SectionView.ReferenceSectionAbove != null)
            {
                bounds = GridRectangle.Union(bounds, Parent.SectionBounds(SectionView.ReferenceSectionAbove));
            }
            if (SectionView.ReferenceSectionBelow != null)
            {
                bounds = GridRectangle.Union(bounds, Parent.SectionBounds(SectionView.ReferenceSectionBelow));
            }
            
            return bounds;
        }
    }
}
