﻿using System;
using System.ComponentModel; 
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Geometry;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics; 
using Viking.Common;
using WebAnnotation;
using WebAnnotationModel;
using System.Windows.Forms;
using System.Diagnostics;
using System.Drawing;
using Common.UI;
using WebAnnotation.UI.Commands; 

namespace WebAnnotation.ViewModel
{
    public class Location : Viking.Objects.UIObjBase, IEqualityComparer<Location>, IEqualityComparer<LocationObj>
    {
        public readonly LocationObj modelObj;

        public override string ToString()
        {
            return modelObj.ToString(); 
        }

        public override int GetHashCode()
        {
            return modelObj.GetHashCode(); 
        }

        public override bool Equals(object obj)
        {
            Location LocObj = obj as Location;
            if (LocObj != null)
            {
                return modelObj.Equals(LocObj.modelObj); 
            }

            LocationObj LocObj2 = obj as LocationObj;
            if (LocObj2 != null)
            {
                return modelObj.Equals(LocObj2); 
            }

            return false; 
        }

        [Column("ID")]
        public long ID
        {
            get { return modelObj.ID; }
        }

        public string Label
        {
            get
            {
                if (Parent == null)
                    return "";

                if (Parent.Type == null)
                    return "";

                return Parent.Type.Code + " " + Parent.ID.ToString();
            }
        }

        public long? ParentID
        {
            get { return modelObj.ParentID; }
        }

        public Structure Parent
        {
            get
            {
                if (this.modelObj.Parent != null)
                    return new Structure(this.modelObj.Parent);
                else
                    return null; 
            }
        }

        public System.Collections.ObjectModel.ObservableCollection<long> Links
        {
            get { return modelObj.Links; }
        }
        
        [Column("X")]
        public double X
        {
            get { return modelObj.Position.X; }
        }

        [Column("Y")]
        public double Y
        {
            get { return modelObj.Position.Y; }
        }

        /// <summary>
        /// This is readonly because changing it would break a datastructure in location store
        /// and also would require update of X,Y to the section space of the different section
        /// </summary>
        [Column("Z")]
        public double Z
        {
            get { return modelObj.Z; }
        }

        public GridVector2 SectionPosition
        {
            get
            {
                return modelObj.Position;
            }
            set
            {
                modelObj.Position = value;
            }
        }

        public GridVector2 VolumePosition
        {
            get
            {
                return modelObj.VolumePosition;
            }
            set
            {
                modelObj.VolumePosition = value; 
            }
        }

        /// <summary>
        /// VolumeX is the x position in volume space. It only exists to inform the database of an estimate of the locations position in volume space.
        /// We want the database to have this value so data processing tools don't need to implement the transforms
        /// It should not be used by the viewer since the viewer can calculate the value.*/
        /// </summary>
        /// 
        [Column("VolumeX")]
        public double VolumeX
        {
            get
            {
                return modelObj.VolumePosition.X;
            }
        }

        /// <summary>
        /// VolumeY is the y position in volume space. It only exists to inform the database of an estimate of the locations position in volume space.
        /// We want the database to have this value so data processing tools don't need to implement the transforms
        /// It should not be used by the viewer since the viewer can calculate the value.*/
        /// </summary>
        /// 
        [Column("VolumeY")]
        public double VolumeY
        {
            get
            {
                return modelObj.VolumePosition.Y;
            }
        }

        [Column("Radius")]
        public double Radius
        {
            get { return modelObj.Radius; }
            set
            {
                if (modelObj.Radius == value)
                    return;

                modelObj.Radius = value;
            }
        }

        [Column("TypeCode")]
        public LocationType TypeCode
        {
            get { return (LocationType)modelObj.TypeCode; }
        }

        /// <summary>
        /// This column is set to true when the location has one link and is not marked as terminal.  It means the
        /// Location is a dead-end and the user did not mark it as a dead end, which means it may actually continue
        /// and the user was distracted
        /// </summary>
        /// 
        [Column("IsUnverifiedTerminal")]
        public bool IsUnverifiedTerminal
        {
            get
            {
                return modelObj.IsUnverifiedTerminal; 
            }
        }

        /// <summary>
        /// This is readonly because changing it would break a datastructure in location store
        /// and also would require update of X,Y to the section space of the different section
        /// </summary>
        /// 

        [Column("Section")]
        public int Section
        {
            get { return (int)modelObj.Section; }
        }

        public Location(LocationObj location)
        {
            Debug.Assert(location != null);

            this.modelObj = location; 

           // bool mappedPosition = parent.TrySectionToVolume(new GridVector2(loc.X, loc.Y), this.Section.section, out VolumePosition);
        }

        /// <summary>
        /// Return true if the passed point falls within our location
        /// </summary>
        /// <param name="position"></param>
        /// <returns></returns>
        public bool Contains(WebAnnotation.ViewModel.SectionLocationsViewModel sectionAnnotations, GridVector2 pos)
        {
            GridVector2 locPosition = this.VolumePosition; 
            /*
            bool Success = sectionAnnotations.TryGetPositionForLocation(this, out locPosition);
            if (Success == false)
                return false; 
            */

            switch (modelObj.TypeCode)
            {
                case LocationType.POINT: //A point
                    //HACK, assume label size is 256 pixels across
                    Vector2 offset = GetLabelSize();
                    offset.X *= AnnotationOverlay.LocationTextScaleFactor;
                    offset.Y *= AnnotationOverlay.LocationTextScaleFactor;

                    GridRectangle _Bounds = new GridRectangle(new GridVector2(locPosition.X - (offset.X / 2), locPosition.Y - (offset.Y / 2)),
                                                        offset.X,
                                                        offset.Y);
                    return _Bounds.Contains(pos);
                case LocationType.CIRCLE: //A circle
                    double distance = GridVector2.Distance(locPosition, pos);
                    return distance <= Radius;

                default:
                    Trace.WriteLine("Calling LocationObj::Contains on an unknown type of location", "WebAnnotation");
                    return false;
            }
        }

        
        /// <summary>
        /// The radius to use when the location is displayed as a reference location on another section
        /// </summary>
        public float OffSectionRadius
        {
            get
            {
                return (float)(this.Radius / 4 > 128f ? 128f : this.Radius / 4);
            }
        }
        

        /// <summary>
        /// This render code should belong in a view level object
        /// </summary>
        #region Render Code


        private bool _LabelSizeMeasured = false;
        private Vector2 _LabelSize;

        private bool _InfoLabelSizeMeasured = false;
        private Vector2 _InfoLabelSize;

        private bool _ParentLabelSizeMeasured = false;
        private Vector2 _ParentLabelSize;


        public Vector2 GetLabelSize()
        {
            if (_LabelSizeMeasured)
                return _LabelSize;

            //Otherwise we aren't really sure about the label size, so just guess
            return new Vector2(256f, 128f);
        }

        public Vector2 GetLabelSize(SpriteFont font)
        {
            if (_LabelSizeMeasured)
                return _LabelSize;

            string label = modelObj.Label;
            //Label can't be empty or the offset measured is zero
            if (label == "")
                label = " ";

            _LabelSize = font.MeasureString(modelObj.Label);
            _LabelSizeMeasured = true;
            return _LabelSize;
        }

        public Vector2 GetInfoLabelSize(SpriteFont font)
        {
            if (_InfoLabelSizeMeasured)
                return _InfoLabelSize;

            if (modelObj.Parent == null)
                return new Vector2(0, 0);

            string text = modelObj.Parent.Label;
            //Label can't be empty or the offset measured is zero
            if (text == "")
                text = " ";

            _InfoLabelSize = font.MeasureString(text);
            _InfoLabelSizeMeasured = true;
            return _InfoLabelSize;
        }

        public Vector2 GetParentLabelSize(SpriteFont font)
        {
            if (_ParentLabelSizeMeasured)
                return _ParentLabelSize;

            if (modelObj.Parent == null)
                return new Vector2(0, 0);

            if (modelObj.Parent.Parent == null)
                return new Vector2(0, 0);


            string label = modelObj.Parent.Parent.ToString();
            //Label can't be empty or the offset measured is zero
            if (label == "")
                label = " ";

            _ParentLabelSize = font.MeasureString(label);
            _ParentLabelSizeMeasured = true;
            return _LabelSize;
        }

        /// <summary>
        /// Draw the text for the location at the specified screen coordinates
        /// </summary>
        /// <param name="spriteBatch"></param>
        /// <param name="font"></param>
        /// <param name="ScreenDrawPosition"></param>
        /// <param name="MagnificationFactor"></param>
        /// <param name="DirectionToVisiblePlane">The Z distance of the location to the plane viewed by user.</param>
        public void DrawLabel(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch,
                              Microsoft.Xna.Framework.Graphics.SpriteFont font,
                              Vector2 ScreenDrawPosition,
                              float MagnificationFactor,
                              int DirectionToVisiblePlane)
        {
            Vector2 offset = GetLabelSize(font);
            offset.X /= 2;
            offset.Y /= 2;

            float visibleCutoff = 4;

            float scale;
            if (modelObj.TypeCode == 0) // A point
            {
                if (DirectionToVisiblePlane == 0)
                    scale = MagnificationFactor * AnnotationOverlay.LocationTextScaleFactor;
                else
                    scale = MagnificationFactor * AnnotationOverlay.ReferenceLocationTextScaleFactor;
            }
            else //A circle
            {
                if (DirectionToVisiblePlane == 0)
                    scale = (((float)Radius / offset.X) * MagnificationFactor) / 2;
                else
                {
                    float radius = (float)((this.Radius / 4) > 128 ? 128 : this.Radius / 4);
                    scale = (((float)radius / offset.X) * MagnificationFactor) / 2;
                }
            }

            //Don't draw labels if no human could read them
            if (offset.Y * scale < visibleCutoff)
                return;

            StructureType type = this.Parent.Type;

            Microsoft.Xna.Framework.Color color = new Microsoft.Xna.Framework.Color((byte)(type.Color.R / 4),
                                                                                                      (byte)(type.Color.G / 4),
                                                                                                      (byte)(type.Color.B / 4),
                                                                                                      255);
            
            if (this.Equals(CreateNewLinkedLocationCommand.LastEditedLocation))
            {
                double alpha = (DateTime.Now.Millisecond / 1000.0) * Math.PI;
                alpha = Math.Sin(alpha);
                color = new Microsoft.Xna.Framework.Color((byte)(type.Color.R / 4),
                                                          (byte)(type.Color.G / 4),
                                                          (byte)(type.Color.B / 4),
                                                          (byte)(alpha * 255));
                
                scale = scale + (scale * (float)alpha * .25f);
            }
            else if (modelObj.OffEdge)
            {
                color = new Microsoft.Xna.Framework.Color(255, 255, 255, 255);
            }
            else if (modelObj.IsUnverifiedTerminal)
            {
                color = new Microsoft.Xna.Framework.Color((byte)255 - type.Color.R,
                    (byte)255 - type.Color.G,
                    (byte)255 - type.Color.B,
                    255);
            }

            spriteBatch.DrawString(font,
                Label,
                ScreenDrawPosition,
                color,
                0,
                offset,
                scale,
                SpriteEffects.None,
                0);

            //If we have a parent of our parent then include thier ID in small font
            if (this.Parent.Parent != null)
            {
                StructureType ParentType = this.Parent.Parent.Type;
                Vector2 ParentOffset = this.GetParentLabelSize(font);
                ParentOffset.X /= 2f;
                ParentOffset.Y /= 2f;

                Microsoft.Xna.Framework.Color ParentColor = new Microsoft.Xna.Framework.Color(ParentType.Color.R,
                                                                                                          ParentType.Color.G,
                                                                                                          ParentType.Color.B,
                                                                                                          255);

                string ParentLabel = this.Parent.Parent.ToString();
                float ParentScale = scale / 1.5f;
                Vector2 ParentScreenDrawPosition = ScreenDrawPosition;

                //Position label above the label for the location
                ParentScreenDrawPosition.Y -= ((offset.Y * 2) * scale);

                spriteBatch.DrawString(font,
                    ParentLabel,
                    ParentScreenDrawPosition,
                    ParentColor,
                    0,
                    ParentOffset,
                    ParentScale,
                    SpriteEffects.None,
                    0);

            }

            //If we have an additional label, include that information in small font
            if (this.Parent.InfoLabel != null)
            {
                Vector2 LabelOffset = GetInfoLabelSize(font);
                LabelOffset.X /= 2f;
                LabelOffset.Y /= 2f;

                string AdditionalLabel = this.Parent.InfoLabel;
                float LabelScale = scale / 1.5f;
                Vector2 LabelScreenDrawPosition = ScreenDrawPosition;

                //Position label below the label for the location
                LabelScreenDrawPosition.Y += ((offset.Y * 2) * scale);

                spriteBatch.DrawString(font,
                    AdditionalLabel,
                    LabelScreenDrawPosition,
                    color,
                    0,
                    LabelOffset,
                    LabelScale,
                    SpriteEffects.None,
                    0);

            }

        }

        /// <summary>
        ///  Draws the background of the location, can be:
        ///  0: A box around label for a point location.
        ///  1: A circle around label for a circle location
        ///  
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="DirectionToVisiblePlane">The offset to the plane viewed by user.  
        /// We present different cues depending on where the mark is relative to the section the viewer is seeing</param>
        public void DrawBackground(GraphicsDevice graphicsDevice,
                                   int DirectionToVisiblePlane)
        {
            //Are we drawing a point?
            switch (this.TypeCode)
            {
                case LocationType.POINT:
                    DrawPointBackground(graphicsDevice, DirectionToVisiblePlane);
                    break;
                case LocationType.CIRCLE:
                    DrawCircleBackground(graphicsDevice, DirectionToVisiblePlane);
                    break;
                default:
                    Trace.WriteLine("Unimplemented location type, not drawn, ID: " + this.ID.ToString(), "WebAnnotation");
                    break;
            }
        }


        public void DrawPointBackground(Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice,
                                   int DirectionToVisiblePlane)
        {
            WebAnnotation.ViewModel.SectionLocationsViewModel sectionAnnotations = AnnotationOverlay.GetAnnotationsForSection((int)this.Z);
            if (sectionAnnotations == null)
                return;

            Vector2 Offset = GetLabelSize();

            int alpha = 85;
            if (AnnotationOverlay.LastMouseOverObject == this)
            {
                alpha = 192;
            }

            StructureType type = this.Parent.Type;
            Microsoft.Xna.Framework.Color color = new Microsoft.Xna.Framework.Color(type.Color.R,
                type.Color.G,
                type.Color.B,
                (byte)(alpha));

            GridVector2 Pos = sectionAnnotations.GetPositionForLocation(this);

            VertexPositionColor[] verts;
            int[] indicies;
            int[] lineIndicies;

            //Location is in the visible section
            if (DirectionToVisiblePlane == 0)
            {
                Offset.X *= AnnotationOverlay.LocationTextScaleFactor / 2;
                Offset.Y *= AnnotationOverlay.LocationTextScaleFactor / 2;

                verts = new VertexPositionColor[] { 
                        new VertexPositionColor(new Vector3((float)Pos.X - Offset.X, (float)Pos.Y - Offset.Y, 1), color),
                        new VertexPositionColor(new Vector3((float)Pos.X + Offset.X, (float)Pos.Y - Offset.Y, 1), color),
                        new VertexPositionColor(new Vector3((float)Pos.X + Offset.X, (float)Pos.Y + Offset.Y, 1), color),
                        new VertexPositionColor(new Vector3((float)Pos.X - Offset.X, (float)Pos.Y + Offset.Y, 1), color)};

                indicies = new int[] { 0, 2, 1, 0, 3, 2 };
                lineIndicies = new int[] { 0, 1, 2, 3, 0 };
            }
            else if (DirectionToVisiblePlane > 0)
            {
                Offset.X *= AnnotationOverlay.ReferenceLocationTextScaleFactor / 2;
                Offset.Y *= AnnotationOverlay.ReferenceLocationTextScaleFactor / 2;

                verts = new VertexPositionColor[] { 
                            new VertexPositionColor(new Vector3((float)Pos.X - Offset.X, (float)Pos.Y - Offset.Y, 1), color),
                            new VertexPositionColor(new Vector3((float)Pos.X + Offset.X, (float)Pos.Y - Offset.Y, 1), color),
                            new VertexPositionColor(new Vector3((float)Pos.X, (float)Pos.Y + (Offset.Y * 1.5f), 1), color)};

                indicies = new int[] { 2, 1, 0 };
                lineIndicies = new int[] { 0, 1, 2, 0 };
            }
            else
            {
                Offset.X *= AnnotationOverlay.ReferenceLocationTextScaleFactor / 2;
                Offset.Y *= AnnotationOverlay.ReferenceLocationTextScaleFactor / 2;

                verts = new VertexPositionColor[] { 
                            new VertexPositionColor(new Vector3((float)Pos.X + Offset.X, (float)Pos.Y + Offset.Y, 1), color),
                            new VertexPositionColor(new Vector3((float)Pos.X - Offset.X, (float)Pos.Y + Offset.Y, 1), color),
                            new VertexPositionColor(new Vector3((float)Pos.X, (float)Pos.Y - (Offset.Y * 1.5f), 1), color)};

                indicies = new int[] { 2, 1, 0 };
                lineIndicies = new int[] { 0, 1, 2, 0 };
            }

            graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.TriangleList,
                                                                          verts,
                                                                          0,
                                                                          verts.Length,
                                                                          indicies,
                                                                          0,
                                                                          indicies.Length / 3);

            //Draw an opaque border around the background
            for (int i = 0; i < verts.Length; i++)
            {
                verts[i].Color.A = 255;
            }



            graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.LineStrip,
                                                                          verts,
                                                                          0,
                                                                          verts.Length,
                                                                          lineIndicies,
                                                                          0,
                                                                          lineIndicies.Length - 1);
        }


        public void DrawCircleBackground(GraphicsDevice graphicsDevice,
                                   int DirectionToVisiblePlane)
        {

            int alpha = 85;
            if (AnnotationOverlay.LastMouseOverObject == this)
            {
                alpha = 192;
            }
            else if (DirectionToVisiblePlane != 0)
            {
                alpha = 128; 
            }

            StructureType type = this.Parent.Type;
            Microsoft.Xna.Framework.Color color = new Microsoft.Xna.Framework.Color(type.Color.R,
                type.Color.G,
                type.Color.B,
                (byte)(alpha));

            DrawCircleBackground(graphicsDevice, DirectionToVisiblePlane, color);
        }

        /// <summary>
        /// The verticies should really be cached and handed up to LocationObjRenderer so all similiar objects can be rendered in one
        /// call.  This method is in the middle of a change from using triangles to draw circles to using textures. 
        /// </summary>
        /// <param name="graphicsDevice"></param>
        /// <param name="DirectionToVisiblePlane"></param>
        /// <param name="color"></param>
        public void DrawCircleBackground(GraphicsDevice graphicsDevice,
                                   int DirectionToVisiblePlane, Microsoft.Xna.Framework.Color color)
        {
            WebAnnotation.ViewModel.SectionLocationsViewModel sectionAnnotations = AnnotationOverlay.GetAnnotationsForSection((int)this.Z);
            if (sectionAnnotations == null)
                return;

            GridVector2 Pos = this.VolumePosition; 
            /*
            bool Success = sectionAnnotations.TryGetPositionForLocation(this, out Pos);
            if (!Success)
                return; 
            */
            //A better way to implement this is to just render a circle texture and add color using lighting, but 
            //this will work for now

            VertexPositionColorTexture[] verts = new VertexPositionColorTexture[GlobalPrimitives.SquareVerts.Length];
            GlobalPrimitives.SquareVerts.CopyTo(verts, 0);

            //Can't populate until we've referenced CircleVerts
            int[] indicies = GlobalPrimitives.SquareIndicies;
            float radius = (float)this.Radius;

            if (DirectionToVisiblePlane != 0)
                radius = OffSectionRadius;

            /*
            //Figure out if we should draw triangles instead
            if (DirectionToVisiblePlane == 0)
            {
                verts = new VertexPositionColor[GlobalPrimitives.CircleVerts.Length];
                GlobalPrimitives.CircleVerts.CopyTo(verts, 0);

                indicies = GlobalPrimitives.CircleVertIndicies;
                lineIndicies = GlobalPrimitives.CircleBorderIndicies;
            }
            else if (DirectionToVisiblePlane > 0)
            {
                verts = new VertexPositionColor[GlobalPrimitives.UpTriVerts.Length];
                GlobalPrimitives.UpTriVerts.CopyTo(verts, 0);
           
                indicies = new int[] { 2, 1, 0 };
                lineIndicies = new int[] { 0, 1, 2, 0 };
                radius = OffSectionRadius;
            }
            else
            {
                verts = new VertexPositionColor[GlobalPrimitives.DownTriVerts.Length];
                GlobalPrimitives.DownTriVerts.CopyTo(verts, 0);

                indicies = new int[] { 2, 1, 0 };
                lineIndicies = new int[] { 0, 1, 2, 0 };
                radius = OffSectionRadius;
            }
            */

            //Draw an opaque border around the background
            for (int i = 0; i < verts.Length; i++)
            {
                verts[i].Position.X *= radius;
                verts[i].Position.Y *= radius;

                verts[i].Position.X += (float)Pos.X;
                verts[i].Position.Y += (float)Pos.Y;
                verts[i].Color = color;
            }

            graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColorTexture>(PrimitiveType.TriangleList,
                                                                          verts,
                                                                          0,
                                                                          verts.Length,
                                                                          indicies,
                                                                          0,
                                                                          2);

            /*
            graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.TriangleFan,
                                                                          verts,
                                                                          0,
                                                                          verts.Length,
                                                                          indicies,
                                                                          0,
                                                                          indicies.Length - 2);

            //Draw an opaque border around the background
            for (int i = 0; i < verts.Length; i++)
            {
                verts[i].Color.A = 255;
            }

           

            graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor>(PrimitiveType.LineStrip,
                                                                          verts,
                                                                          0,
                                                                          verts.Length,
                                                                          lineIndicies,
                                                                          0,
                                                                          lineIndicies.Length - 1);
             */

        }

        #endregion


        #region IUIObject Members

        public override void Delete()
        {
            Store.Locations.Remove(this.modelObj);
            Store.Locations.Save();

            if (this.ParentID.HasValue)
                Store.Structures.CheckForOrphan(this.ParentID.Value);
        }

        public new event PropertyChangedEventHandler ValueChanged
        {
            add{  modelObj.PropertyChanged += value; }
            remove { modelObj.PropertyChanged -= value; }
        }

        public override ContextMenu ContextMenu
        {
            get
            {
                ContextMenu menu = new ContextMenu();


                menu.MenuItems.Add("Properties", ContextMenu_OnProperties);

                MenuItem menuExtensible = new MenuItem("Terminal", ContextMenu_OnTerminal);
                MenuItem menuOffEdge = new MenuItem("Off Edge", ContextMenu_OnOffEdge);
                MenuItem menuSeperator = new MenuItem();
                MenuItem menuDelete = new MenuItem("Delete", ContextMenu_OnDelete);

                menuExtensible.Checked = this.modelObj.Terminal;
                menuOffEdge.Checked = this.modelObj.OffEdge;

                menu.MenuItems.Add(menuExtensible);
                menu.MenuItems.Add(menuOffEdge);
                menu.MenuItems.Add(menuSeperator);
                menu.MenuItems.Add(menuDelete);

                return menu;
            }
        }

        public Image SmallThumbnail
        {
            get { throw new NotImplementedException(); }
        }

        public string ToolTip
        {
            get { throw new NotImplementedException(); }
        }

        public void Save()
        {
            Store.Locations.Save();
        }
                
        #endregion


        protected void ContextMenu_OnProperties(object sender, EventArgs e)
        {
            Viking.UI.Forms.PropertySheetForm.Show(this.Parent);
        }

        protected void ContextMenu_OnTerminal(object sender, EventArgs e)
        {
            this.modelObj.Terminal = !this.modelObj.Terminal;
            bool success = Store.Locations.Save();
            if (!success)
            {
                this.modelObj.Terminal = !this.modelObj.Terminal;
            }

        }

        protected void ContextMenu_OnOffEdge(object sender, EventArgs e)
        {
            /*
            DBACTION originalDBAction = this.DBAction;
            this.Data.OffEdge = !this.Data.OffEdge;
            this.Data.DBAction = DBACTION.UPDATE;
            bool success = Store.Locations.Save();
            if (!success)
            {
                this.Data.OffEdge = !this.Data.OffEdge;
                this.DBAction = originalDBAction;
            }
             */
            this.modelObj.OffEdge = !this.modelObj.OffEdge;
            Store.Locations.Save(); 
        }

        protected void ContextMenu_OnDelete(object sender, EventArgs e)
        {
            Delete();
        }


        public bool Equals(Location x, Location y)
        {
            if (x == null && y == null)
                return true;

            if (x == null || y == null)
                return false; 

            return x.ID == y.ID; 
        }

        public int GetHashCode(Location obj)
        {
            return obj.modelObj.GetHashCode(); 
        }

        public bool Equals(LocationObj x, LocationObj y)
        {
            if (x == null && y == null)
                return true;

            if (x == null || y == null)
                return false;

            return x.ID == y.ID; 
        }

        public int GetHashCode(LocationObj obj)
        {
            return obj.GetHashCode(); 
        }
    }
}
