﻿using System;
using System.Diagnostics; 
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using Viking;
using Viking.Common;
using Geometry;
using WebAnnotation.UI;
using Viking.ViewModels;
using WebAnnotationModel;
using System.ComponentModel; 

namespace WebAnnotation.ViewModel
{
    public class SectionLocationsViewModel
    {
        /// <summary>
        /// The section we store annotations for
        /// <summary>
        public SectionViewModel Section;

        /// <summary>
        /// Set to true if the annotations have been requested for this section
        /// </summary>
        public bool HaveLoadedSectionAnnotations = false;

        /// <summary>
        /// Locations on the section we are providing an overlay for
        /// implemented as a quad tree which can map a point to the nearest Location
        /// </summary>
        private QuadTree<Location> Locations = null;
       
        /// <summary>
        /// Maps a structureID to all the locations for that structure on the visible section
        /// </summary>
        private ConcurrentDictionary<long, ConcurrentDictionary<long, Location>> LocationsForStructure = new ConcurrentDictionary<long, ConcurrentDictionary<long, Location>>();
        
        //        public static SortedDictionary<long, SortedList<long, RoundLineCode.RoundLine> > LocationLinesDict = null;

        /// <summary>
        /// Allows us to describe all the locationlinks visible on a screen
        /// </summary>
        //private LineSearchGrid<LocationLink> LocationLinksSearch = null;

        /// <summary>
        /// Allows us to describe all the StructureLinks visible on a screen
        /// </summary>
        private LineSearchGrid<StructureLink> StructureLinksSearch = null;

        /// <summary>
        /// Fired when an annotation has changed.  Overlays should refresh thier display
        /// </summary>
        public event EventHandler AnnotationChanged;

        private EventHandler LocationLinkDeletedEventHandler;
        private EventHandler StructureLinkDeletedEventHandler;

        private PropertyChangingEventHandler OnLocationPropertyChangingEventHandler;
        private PropertyChangedEventHandler OnLocationPropertyChangedEventHandler;

        private PropertyChangingEventHandler OnLinkedLocationPropertyChangingEventHandler;
        private PropertyChangedEventHandler OnLinkedLocationPropertyChangedEventHandler;

        /// <summary>
        /// This is a symptom of being halfway to the Jotunn architecture.  This is a pointer to the 
        /// parent section viewer control which can perform transforms
        /// </summary>
        public Viking.UI.Controls.SectionViewerControl parent;

        public SectionLocationsViewModel(SectionViewModel section,  Viking.UI.Controls.SectionViewerControl Parent)
        {
            this.parent = Parent; 
            this.Section = section;

            GridRectangle bounds = AnnotationOverlay.SectionBounds(parent, parent.Section.Number);

            if (Locations == null)
                Locations = new QuadTree<Location>(bounds);

            LocationsForStructure = new ConcurrentDictionary<long, ConcurrentDictionary<long, Location>>();
            
            StructureLinksSearch = new LineSearchGrid<StructureLink>(bounds, 10000);
            

            OnLocationPropertyChangingEventHandler = new PropertyChangingEventHandler(this.OnLocationPropertyChanging);
            OnLocationPropertyChangedEventHandler = new PropertyChangedEventHandler(this.OnLocationPropertyChanged);

            Store.Structures.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnStructuresStoreChanged);
            Store.StructureLinks.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnStructureLinksStoreChanged); 

            //Notifications arrive from the AnnotationOverlay to reduce the number of objects processing unwanted events
            //Store.Locations.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnLocationsStoreChanged);

            //Store.Locations.OnAllUpdatesCompleted += new OnAllUpdatesCompletedEventHandler(OnAllUpdatesCompleted);

            //TODO: Change these to be filtered through the AnnotationOverlay
            //LocationLinkDeletedEventHandler = new EventHandler(OnLocationLinkDeleted);

            /*
            Store.Structures.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnStructuresStoreChanged);
            Store.StructureLinks.OnCollectionChanged += new NotifyCollectionChangedEventHandler(OnStructureLinksStoreChanged); 
             */

            //StructureLinkDeletedEventHandler = new EventHandler(OnStructureLinkDeleted);
        }

        #region Location Property Changes

        protected void OnLocationPropertyChanging(object sender, PropertyChangingEventArgs e)
        {
            LocationObj loc = sender as LocationObj;
            if (loc == null)
                return;

//            Trace.WriteLine("Location property changing: " + loc.ToString() + " property: " + e.PropertyName); 

            //Update if a position or everything has changed
            if (e.PropertyName.Contains("Position") || e.PropertyName.Contains("WorldPosition") || e.PropertyName == "")
            {
                Location locView = new Location(loc);
                RemoveStructureLinks(new LocationObj[]{loc});
                RemoveLocations(new LocationObj[] { loc }, false);
            }
        }

        protected void OnLocationPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            LocationObj loc = sender as LocationObj;
            if (loc == null)
                return;

//            Trace.WriteLine("Location property changed: " + loc.ToString() + " property: " + e.PropertyName); 

            

            //Update if a position or everything has changed
            if (e.PropertyName.Contains("Position") || e.PropertyName.Contains("WorldPosition") || e.PropertyName == "")
            {
               /* Location locView = new Location(loc);
                if (e.PropertyName == "")
                {
                    MapLocation(loc); 
                }
                */

                AddLocations(new LocationObj[] { loc }, false);
                AddStructureLinks(new LocationObj[] { loc });
      //          bool Success = Locations.TryAdd(locView.VolumePosition, locView);
            }
        }

        protected void OnLinkedLocationPropertyChanging(object sender, PropertyChangingEventArgs e)
        {
            LocationObj loc = sender as LocationObj;
            if (loc == null)
                return;

//            Trace.WriteLine("Linked Location property changing: " + loc.ToString() + " property: " + e.PropertyName);

            //Update if a position or everything has changed
            if (e.PropertyName.Contains('X') || e.PropertyName.Contains('Y') || e.PropertyName == "")
            {
                /*
                Location locView = new Location(loc);
                foreach (long linkedID in loc.Links)
                {

                }
                bool Success = Locations.TryRemove(locView);
                //      Debug.Assert(Success); 
                */
            }
        }

        /*
        protected void OnLinkedLocationPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            LocationObj loc = sender as LocationObj;
            if (loc == null)
                return;

            Trace.WriteLine("Linked Location property changed: " + loc.ToString() + " property: " + e.PropertyName);

            //Update if a position or everything has changed
            if (e.PropertyName.Contains('X') || e.PropertyName.Contains('Y') || e.PropertyName == "")
            { 
                List<LocationObj> listLocs = new List<LocationObj>();
                listLocs.Add(loc);
                RemoveLocationLinks(listLocs);
                AddLocationLinks(listLocs); 
            }
        }
        */

        #endregion 

        #region Cache updates

        //Called when a key is added or removed from the store
        public void OnLocationsStoreChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            IEnumerable<LocationObj> listNewObjs; 
            IEnumerable<LocationObj> listOldObjs;
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    listNewObjs = e.NewItems.Cast<LocationObj>();
                    AddLocations(listNewObjs);
                    AddStructureLinks(listNewObjs);

                    //foreach (LocationObj locObj in listNewObjs)
                    //{
                    //    locObj.PropertyChanging += this.OnLocationPropertyChanging;
                    //    locObj.PropertyChanged += this.OnLocationPropertyChanged;
                    //}

                    break;
                /*Updates are not sent anymore*/
                case NotifyCollectionChangedAction.Replace:
                    RemoveStructureLinks(e.OldItems.Cast<LocationObj>()); 
                    RemoveLocations(e.OldItems.Cast<LocationObj>());
                    AddLocations(e.NewItems.Cast<LocationObj>());
                    AddStructureLinks(e.NewItems.Cast<LocationObj>());
                    break;

                case NotifyCollectionChangedAction.Remove:
                    listOldObjs = e.OldItems.Cast<LocationObj>();
                    RemoveStructureLinks(listOldObjs);
                    RemoveLocations(listOldObjs);
                    //foreach (LocationObj locObj in listOldObjs)
                    //{
                    //    locObj.PropertyChanging -= this.OnLocationPropertyChanging;
                    //    locObj.PropertyChanged -= this.OnLocationPropertyChanged;
                    //}
                    break;

                default:
                    Debug.Assert(false, "Unexpected change action in OnStoreAddRemoveKey");
                    break;
            }
        }
        
        //Called when a key is added or removed from the store
        public void OnStructuresStoreChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    AddStructureLinks(e.NewItems.Cast<StructureObj>());
                    break;

                case NotifyCollectionChangedAction.Replace:
                    RemoveStructureLinks(e.OldItems.Cast<StructureObj>());
                    AddStructureLinks(e.NewItems.Cast<StructureObj>());
                    break;

                case NotifyCollectionChangedAction.Remove:
                    RemoveStructureLinks(e.OldItems.Cast<StructureObj>());
                    break;

                default:
                    Debug.Assert(false, "Unexpected change action in OnStoreAddRemoveKey");
                    break;
            }
        }

        //Called when a key is added or removed from the store
        public void OnStructureLinksStoreChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Add:
                    AddStructureLinks(e.NewItems.Cast<StructureLinkObj>());
                    break;

                case NotifyCollectionChangedAction.Replace:
                    RemoveStructureLinks(e.OldItems.Cast<StructureLinkObj>());
                    AddStructureLinks(e.NewItems.Cast<StructureLinkObj>());
                    break;

                case NotifyCollectionChangedAction.Remove:
                    RemoveStructureLinks(e.OldItems.Cast<StructureLinkObj>());
                    break;

                default:
                    Debug.Assert(false, "Unexpected change action in OnStoreAddRemoveKey");
                    break;
            }
        }

        private bool MapLocation(LocationObj loc)
        {
            GridVector2 VolumePosition = new GridVector2(-1, -1);
            bool mappedPosition = parent.TrySectionToVolume(loc.Position, this.Section.section, out VolumePosition);
            if (!mappedPosition) //Remove locations we can't map
            {
                Trace.WriteLine("AddLocation: Location #" + loc.ID.ToString() + " was unmappable.", "WebAnnotation");
                return false;
            }

            loc.VolumePosition = VolumePosition;

            return true; 
        }

        protected void AddLocations(IEnumerable<LocationObj> listLocations)
        {
            AddLocations(listLocations, true); 
        }

        /// <summary>   
        ///  Keys have been added to the locations store
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void AddLocations(IEnumerable<LocationObj> listLocations, bool Subscribe)
        {
            foreach(LocationObj loc in listLocations)
            {
            //Trace.WriteLine("AddLocation: " + obj.ToString(), "WebAnnotation");
                //Add the location to our mapping if the location is on our section
                if (loc.Section == Section.Number)
                {
                    bool mapped = MapLocation(loc);
                    if (!mapped)
                        continue; 
                    
                    //Add location if it hasn't been seen before
                    Location locView = new Location(loc);
                    bool Added = Locations.TryAdd(locView.VolumePosition, locView);

                    if (Added)
                    {
                        if (Subscribe)
                        {
                            loc.PropertyChanged += this.OnLocationPropertyChangedEventHandler;
                            loc.PropertyChanging += this.OnLocationPropertyChangingEventHandler;
                        }

                        ConcurrentDictionary<long, Location> KnownLocationsForStructure;
                        KnownLocationsForStructure = LocationsForStructure.GetOrAdd(loc.ParentID.Value, (key) => { return new ConcurrentDictionary<long, Location>(); });
                        KnownLocationsForStructure.TryAdd(locView.ID, locView);
                    }

                   // Debug.Assert(Added);

                //    AddLocationLinks(locView);
                }
            }
        }

        protected void RemoveLocations(IEnumerable<LocationObj> listLocations)
        {
            RemoveLocations(listLocations, true); 
        }

        /// <summary>
        /// A key is about to be removed from the location store.  Remove it from our cache as well
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected void RemoveLocations(IEnumerable<LocationObj> listLocations, bool Unsubscribe)
        {
            foreach (LocationObj loc in listLocations)
            {
                //Trace.WriteLine("RemoveLocation: " + loc.ID.ToString(), "WebAnnotationViewModel");
                //Debug.Assert(loc.Section == Section.Number); 
                if (loc.Section == Section.Number)
                {
                    

                    
                    Location locView = new Location(loc); 
                    bool RemoveSuccess = Locations.TryRemove(locView);
                    if (RemoveSuccess)
                    {
                        if (Unsubscribe)
                        {
                            loc.PropertyChanged -= this.OnLocationPropertyChangedEventHandler;
                            loc.PropertyChanging -= this.OnLocationPropertyChangingEventHandler;
                        }
//                        Debug.Assert(RemoveSuccess); 
                       // RemoveLocationLinks(obj);
                    }
                }
                
                ConcurrentDictionary<long, Location> KnownLocationsForStructure = null;
                bool Success = LocationsForStructure.TryGetValue(loc.ParentID.Value, out  KnownLocationsForStructure);
                if (Success)
                {
                    Location removedLoc;
                    Success = KnownLocationsForStructure.TryRemove(loc.ID, out removedLoc);

                    if (Success)
                    {
                        if (KnownLocationsForStructure.Count == 0)
                        {
                            LocationsForStructure.TryRemove(loc.ParentID.Value, out KnownLocationsForStructure); 
                        }
                        /*PORT: Not thread safe, This proabably all needs to be removed...

                        //Remove entry if it was the last location for that structure
                        if (KnownLocationsForStructure.Count == 0)
                        {
                            LocationsForStructure.TryUpdate(obj.ParentID.Value, null, KnownLocationsForStructure);
                            LocationsForStructure.Remove(obj.ParentID.Value);
                        }
                            */
                    }
                }

                //GridVector2 removeOutParam;
                //TransformedLocationPositionDict.TryRemove(ID, out removeOutParam);

             //   Trace.WriteLine("End RemoveLocation: " + loc.ID.ToString(), "WebAnnotationViewModel");
            }

        /*
            if (AnnotationChanged != null && loc != null)
                AnnotationChanged(obj, new EventArgs());
        */
            
        }

        #endregion

        #region Queries

        public ICollection<Location> GetLocations()
        {
            return Locations.Values;
        }

        public ICollection<Location> GetLocations(GridRectangle bounds)
        {
            List<GridVector2> foundPoints; 
            List<Location> foundLocations;
            Locations.Intersect(bounds, out foundPoints, out foundLocations);

            return foundLocations; 
        }

        public LocationObj[] GetReferenceLocations()
        {
            List<LocationObj> listRefLocations = new List<LocationObj>();

            if (Section.ReferenceSectionAbove != null)
            {
                listRefLocations.AddRange(Store.Locations.GetObjectsForSection(Section.ReferenceSectionAbove.Number).Values);
            }

            if (Section.ReferenceSectionBelow != null)
            {
                listRefLocations.AddRange(Store.Locations.GetObjectsForSection(Section.ReferenceSectionBelow.Number).Values);
            }

            return listRefLocations.ToArray();
        }

        /// <summary>
        /// Returns the position of the requested locationID in the current transform
        /// </summary>
        /// <param name="ID"></param>
        /// <returns></returns>
        public bool TryGetPositionForLocation(Location loc, out GridVector2 position)
        {
            return Locations.TryGetPosition(loc, out position);
            //return Locations.TryGetPosition(loc); 

            /*
            if (!Success)
            {
                //Hmm... why don't we have it?
                Trace.WriteLine("Could not find position for location: " + ID.ToString());

            }

            return position;
             */
        }

        public GridVector2 GetPositionForLocation(Location loc)
        {
            GridVector2 pos;
            bool Success = Locations.TryGetPosition(loc, out pos);
            if(!Success)
                throw new ArgumentException("Could not map location: " + loc.ToString());

            return pos; 
        }

        public IUIObjectBasic GetNearestAnnotation(GridVector2 WorldPosition, out double distance)
        {
            distance = double.MaxValue;
            IUIObjectBasic FoundObject = null;
            double locDistance = double.MaxValue;
            Location NearestLocationObj = GetNearestLocation(WorldPosition, out locDistance);
            if (NearestLocationObj != null)
            {
                FoundObject = NearestLocationObj as IUIObjectBasic;
                distance = locDistance;
            }

            
            double linkDistance;
            GridVector2 linkPosition; 
            StructureLink FoundLink = StructureLinksSearch.GetNearest(WorldPosition, out linkPosition, out linkDistance);
            if (FoundLink != null && linkDistance < locDistance)
            {
                if (linkDistance < FoundLink.Radius)
                {
                    FoundObject = FoundLink;
                    distance = linkDistance;
                }
            }
            
            return FoundObject;
        }

        

        /// <summary>
        /// Gets the nearest location, preferring locations on the same section, then checking other sections
        /// </summary>
        /// <param name="WorldPosition"></param>
        /// <param name="SectionNumber"></param>
        /// <param name="font"></param>
        /// <param name="locPosition"></param>
        /// <returns></returns>
        public Location GetNearestLocation(GridVector2 WorldPosition, out double distance)
        {
            distance = double.MaxValue; 
            double minDistance = double.MaxValue;

            if (Locations == null)
                return null;

            /*Check to see if we clicked a location*/

            Location bestLoc = Locations.FindNearest(WorldPosition, out distance);
            if (bestLoc != null)
            {       
                if (distance < bestLoc.Radius)
                {
                    minDistance = distance;
                }
                else
                {
                    bestLoc = null;
                }
            }
            
            return bestLoc; 
        }
        
                /*Check to see if we clicked a location on a reference section*/
        /*
                //If we're still here check locations on other sections
                long locID = TransformedRefLocationQuadTree.FindNearest(WorldPosition, out distance);
                if (locID != default(long))
                {
                    loc = Store.Locations.GetObjectByID(locID);
                    if (loc != null)
                    {
                        if (distance < minDistance)
                        {
                            if (distance < loc.OffSectionRadius)
                            {
                                BestLoc = loc;
                                minDistance = distance;
                            }
                        }
                    }
                }
            }

            return BestLoc;
         
        }
         */
        

        #endregion


        /// <summary>
        /// Load the annotations for the passed section and its reference sections
        /// </summary>
        /// <param name="section"></param>
        internal void LoadSectionAnnotations()
        {
            Trace.WriteLine("LoadSectionAnnotations: " + Section.Number.ToString(), "WebAnnotation");

#if DEBUG
            Store.Structures.GetObjectsForSection(Section.Number);
#else
            Store.Structures.GetObjectsForSection(Section.Number);
#endif
            //        
            
            //Have to let the Lock go before we call the location store or we can get a deadlock.  Don't modify 
            //data structures after this point. 

            //if (!HaveLoadedSectionAnnotations)
            //    AddLocations(Store.Locations.GetLocationsForSection(Section.Number).Values);
       
            ConcurrentDictionary<long, LocationObj> knownLocations = Store.Locations.GetObjectsForSection(Section.Number);

            this.AddLocations(knownLocations.Values); 
#if DEBUG
//            Store.Structures.FreeExcessSections(40, 5);
            Store.Locations.FreeExcessSections(40, 5);
#else
            Store.Locations.FreeExcessSections(40, 5);
#endif

            HaveLoadedSectionAnnotations = true;
        }

        
        
        

        /// <summary>
        /// Return all the line segments visible in the passed bounds
        /// </summary>
        /// <param name="bounds"></param>
        /// <returns></returns>
        public StructureLink[] VisibleStructureLinks(GridRectangle bounds)
        {

            StructureLink[] LinkList = StructureLinksSearch.GetValues(bounds);
            return LinkList;
        }

        internal void AddStructureLinks(IEnumerable<LocationObj> locations)
        {
            foreach (LocationObj locObj in locations)
            {
                StructureObj parent = locObj.Parent;
                if (parent == null)
                    continue;

                if (parent.Links.Count > 0)
                    AddStructureLinks(parent.Links); 
            }
        }

        internal void AddStructureLinks(IEnumerable<StructureObj> structures)
        {
            foreach (StructureObj structObj in structures)
            {
                if (structObj.Links.Count > 0)
                    AddStructureLinks(structObj.Links);
            }
        }
        
        /// <summary>
        /// All locations which are linked get a line between them
        /// </summary>
        internal void AddStructureLinks(IEnumerable<StructureLinkObj> structureLinks)
        {
            foreach(StructureLinkObj structLinkObj in structureLinks)
            {
                if (structLinkObj.SourceID == structLinkObj.TargetID)
                {
                    Trace.WriteLine("Something is wrong on the server, struct ID links to itself: " + structLinkObj.SourceID.ToString()); 
                    continue; //Something is wrong in the database
                }
                //The link may have been created to a structure on an adjacent section
                ConcurrentDictionary<long, Location> SourceLocations = null;
                bool Success = LocationsForStructure.TryGetValue(structLinkObj.SourceID, out SourceLocations);
                if (Success == false)
                    continue;

                ConcurrentDictionary<long, Location> TargetLocations = null;
                Success = LocationsForStructure.TryGetValue(structLinkObj.TargetID, out TargetLocations);
                if (Success == false)
                    continue;

                if(SourceLocations.Count == 0 || TargetLocations.Count == 0)
                    continue; 
                
                //Brute force a search for the shortest distance between the two structures.
                double MinDistance = double.MaxValue;
                Location BestSourceLoc = null; 
                Location BestTargetLoc = null;

                foreach (Location SourceLoc in SourceLocations.Values)
                {
                    foreach (Location TargetLoc in TargetLocations.Values)
                    {
                        double dist = GridVector2.Distance(SourceLoc.VolumePosition, TargetLoc.VolumePosition);
                        if (dist < MinDistance)
                        {
                            BestSourceLoc = SourceLoc;
                            BestTargetLoc = TargetLoc;
                            MinDistance = dist;
                        }
                    }
                }

                //OK, create a StructureLink between the locations
                StructureLink StructLink = new StructureLink(structLinkObj, BestSourceLoc, BestTargetLoc);

                StructureLinksSearch.TryAdd(StructLink.lineSegment, StructLink);
            }
        }

        internal void RemoveStructureLinks(IEnumerable<LocationObj> locations)
        {
            foreach (LocationObj locObj in locations)
            {
                StructureObj parent = locObj.Parent;
                if (parent == null)
                    continue;

                if (parent.Links.Count > 0)
                    RemoveStructureLinks(parent.Links);
            }
        }

        internal void RemoveStructureLinks(IEnumerable<StructureObj> structures)
        {
            foreach(StructureObj structObj in structures)
            {
                if (structObj.Links.Count > 0)
                    RemoveStructureLinks(structObj.Links);
            }
        }

        /// <summary>
        /// All locations which are linked get a line between them
        /// </summary>
        internal void RemoveStructureLinks(IEnumerable<StructureLinkObj> structureLinks)
        {
            if (structureLinks == null)
                return; 

            foreach(StructureLinkObj structLinkObj in structureLinks)
            {
                StructureLink link = new StructureLink(structLinkObj); 
                GridLineSegment line; 
                StructureLinksSearch.TryRemove(link, out line); 
            }
        }
    }
}
