﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics; 
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using WebAnnotationModel;
using WebAnnotationModel.Objects; 

using WebAnnotationModel.Service;

using Geometry; 


namespace WebAnnotationModel
{
    public enum LocationType
    {
        POINT = 0,
        CIRCLE = 1,
        MASK = 2
    };

    public class LocationObj : WCFObjBaseWithKey<long, Location>
    {
        public override long ID
        {
            get { return Data.ID; }
        }

        /// <summary>
        /// The ID for newo bjects can change from a negative number to the number in the database.
        /// In this case make sure we always return the same hash code.  As a result this is called for each object only once.
        /// </summary>
        /// <returns></returns>
        protected override int GenerateHashCode()
        {
            return (int)(ID % int.MaxValue);
        }

        public string Label
        {
            get
            {
                if (Parent == null)
                    return "";

                if (Parent.Type == null)
                    return "";

                return Parent.Type.Code + " " + Parent.ID.ToString();
            }
        }

        public long? ParentID
        {
            get { return Data.ParentID; }
        }

       // private StructureObj _Parent;
        public StructureObj Parent
        {
            get
            {
         //       if (_Parent != null)
//                    return _Parent;
                
                if (ParentID.HasValue == false)
                    return null;

                StructureObj _Parent = Store.Structures.GetObjectByID(ParentID.Value, false);

                //Queue a request for later
                if (_Parent == null)
                {
                    Action<long> request = new Action<long>((ID) => Store.Structures.GetObjectByID(ID));
                    request.BeginInvoke(ParentID.Value, null, null); 
                }

                return _Parent;
            }
        }

        public GridVector2 Position
        {
            get
            {
                return new GridVector2(Data.Position.X, Data.Position.Y);
            }
            set
            {
                if (Data.Position.X == value.X &&
                    Data.Position.Y == value.Y)
                    return;

                OnPropertyChanging("Position");

                AnnotationPoint point = new AnnotationPoint();
                point.X = value.X;
                point.Y = value.Y;
                point.Z = Data.Position.Z;
                Data.Position = point;
                OnPropertyChanged("Position");

                SetDBActionForChange();
            }

        }

        /// <summary>
        /// This is readonly because changing it would break a datastructure in location store
        /// and also would require update of X,Y to the section space of the different section
        /// </summary>
        public double Z
        {
            get {return Data.Position.Z; }
        }

        

        /// <summary>
        /// VolumeX is the x position in volume space. It only exists to inform the database of an estimate of the locations position in volume space.
        /// We want the database to have this value so data processing tools don't need to implement the transforms
        /// It should not be used by the viewer since the viewer can calculate the value.*/
        /// </summary>
        public GridVector2 VolumePosition
        {
            get
            {
                return new GridVector2(Data.VolumePosition.X, Data.VolumePosition.Y);
            }
            set
            {
                if (Data.VolumePosition.X == value.X &&
                    Data.VolumePosition.Y == value.Y)
                    return;

                OnPropertyChanging("VolumePosition");

                AnnotationPoint point = new AnnotationPoint();
                point.X = value.X;
                point.Y = value.Y;
                point.Z = Data.Position.Z;
                Data.VolumePosition = point;
                OnPropertyChanged("VolumePosition");

 //               SetDBActionForChange();
            }

        }

        public double Radius
        {
            get { return Data.Radius; }
            set {
                if (Data.Radius == value)
                    return;

                OnPropertyChanging("Radius");
                Data.Radius = value;
                OnPropertyChanged("Radius");
                SetDBActionForChange();
            }
        }

        public LocationType TypeCode
        {
            get { return (LocationType)Data.TypeCode; }
            set {
            if(Data.TypeCode == (short)value)
                return;

            OnPropertyChanging("TypeCode");
            Data.TypeCode = (short)value;
            SetDBActionForChange();
            OnPropertyChanged("TypeCode");
            }
        }

        /// <summary>
        /// This column is set to true when the location has one link and is not marked as terminal.  It means the
        /// Location is a dead-end and the user did not mark it as a dead end, which means it may actually continue
        /// and the user was distracted
        /// </summary>
        public bool IsUnverifiedTerminal
        {
            get
            {
                if (Links.Count >= 2)
                    return false; 
                return !(Terminal || OffEdge);
            }
        }

        /// <summary>
        /// This is readonly because changing it would break a datastructure in location store
        /// and also would require update of X,Y to the section space of the different section
        /// </summary>
        public int Section
        {
            get { return (int)Data.Section; }
        }

        private ObservableCollection<long> _Links = null;
        public ObservableCollection<long> Links
        {
            get {
                if (_Links == null)
                {

                    //Initialize from the Data object
                    if (Data.Links == null)
                    {
                        _Links = new ObservableCollection<long>();
                        _Links.CollectionChanged += this.OnLinksChanged; 
                    }
                    else
                    {
                        _Links = new ObservableCollection<long>(Data.Links);
                        _Links.CollectionChanged += this.OnLinksChanged;
                    }
                }

                return _Links; 
            }
        }

        private void OnLinksChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            //Update the underlying object we will send to the server
            Data.Links = _Links.ToArray();
        }

        /// <summary>
        /// Allows LocationLinkStore to adjust the client after a link is created
        /// </summary>
        /// <param name="ID"></param>
        internal void AddLink(long ID)
        {
            if (ID == this.ID)
                throw new ArgumentException("Can't add own ID from location links");

            if (Links.Contains(ID))
                return;
            
            Links.Add(ID); 
        }

        /// <summary>
        /// Adjust the client after a link is removed
        /// </summary>
        /// <param name="ID"></param>
        internal void RemoveLink(long ID)
        {
            if (ID == this.ID)
                throw new ArgumentException("Can't remove own ID from location links");

            if (!Links.Contains(ID))
                return;
            Links.Remove(ID);
        }

        public bool Terminal
        {
            get { return Data.Terminal; }
            set {
                OnPropertyChanging("Terminal"); 
                Data.Terminal = value;
                SetDBActionForChange();
                OnPropertyChanged("Terminal"); 
            }
        }

        public bool OffEdge
        {
            get { return Data.OffEdge; }
            set {
                OnPropertyChanging("OffEdge"); 
                Data.OffEdge = value;
                SetDBActionForChange();
                OnPropertyChanged("OffEdge"); 
            }
        }

        public DateTime LastModified
        {
            get { return new DateTime(Data.LastModified, DateTimeKind.Utc); }
        }
        
        

        public LocationObj()
        {
            Data = new Location();
        }

        public LocationObj(Location obj)
        {
            Data = obj;    
        }

        public LocationObj(StructureObj parent, GridVector2 position, GridVector2 volumePosition, int SectionNumber)
        {
            this.Data = new Location();
            this.Data.DBAction = DBACTION.INSERT;
            this.Data.ID = Store.Locations.GetTempKey();
            this.Data.Tags = new String[0];
            this.Data.Verticies = new AnnotationPoint[0];
            this.Data.TypeCode = 1;
            this.Data.Radius = 16;
            this.Data.Links = new long[0];

            AnnotationPoint P = new AnnotationPoint();
            P.X = position.X;
            P.Y = position.Y;
            P.Z = (double)SectionNumber;

            AnnotationPoint VP = new AnnotationPoint();
            VP.X = volumePosition.X;
            VP.Y = volumePosition.Y;
            VP.Z = (double)SectionNumber;

            this.Data.Section = SectionNumber; 

            this.Data.Position = P;
            this.Data.VolumePosition = VP; 

            if (parent != null)
            {
                this.Data.ParentID = parent.ID;
            }

//          CallOnCreate(); 
        }     
        

        /*
        public override void Delete()
        {
            DBACTION originalAction = this.DBAction; 
            this.DBAction = DBACTION.DELETE;

            bool success = Store.Locations.Save();
            if(!success)
            {
                //Write straight to data since we have an assert to check whether an object is being deleted, but
                //in this case we know it is ok
                this.Data.DBAction = originalAction;
            }


            if (this.ParentID.HasValue)
                Store.Structures.CheckForOrphan(this.ParentID.Value);
        }
        */

        protected static event EventHandler OnCreate;
        protected void CallOnCreate()
        {
            if (OnCreate != null)
            {
                //Viking.UI.State.MainThreadDispatcher.BeginInvoke(OnCreate, new object[] { this, null });
                OnCreate(this, null); 
            }
        }

        public static event EventHandler Create
        {
            add { OnCreate += value; }
            remove { OnCreate -= value; }
        }
    }
}
