﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using WebAnnotationModel.Service;
using WebAnnotationModel.Objects;
using System.Diagnostics;
using System.Collections.Concurrent;
using System.Collections.Specialized; 

namespace WebAnnotationModel
{
    /// <summary>
    /// This base class implements the basic functionality to talk to a WCF Service
    /// </summary>
    public abstract class StoreBaseWithIndexKey<PROXY, INTERFACE, KEY, KEYGEN, OBJECT, WCFOBJECT> : StoreBaseWithKey<PROXY, INTERFACE, KEY, OBJECT, WCFOBJECT>
        where INTERFACE : class
        where KEY : struct
        where KEYGEN : IKeyGenerator<KEY>, new()
        where PROXY : System.ServiceModel.ClientBase<INTERFACE>
        where WCFOBJECT : WebAnnotationModel.Service.DataObjectWithKeyOflong, new()
        where OBJECT : WCFObjBaseWithKey<KEY, WCFOBJECT>, new()
    {
        /// <summary>
        /// When we create an object we don't know the ID the database will give it.  We use this value on the local machine until the DB
        /// tells us the new ID on insert
        /// </summary>
        static KEYGEN keyGenerator = new KEYGEN();
        public KEY GetTempKey()
        {
            return keyGenerator.NextKey(); 
        }
        
        /// <summary>
        /// Save all changes to locations, returns true if the method completed without errors, otherwise false
        /// Update key values with the index generated by the database
        /// </summary>
        public override bool Save()
        {
            List<OBJECT> input = new List<OBJECT>(); 
            List<WCFOBJECT> changed = new List<WCFOBJECT>();
            List<OBJECT> output = new List<OBJECT>();

            {
                try
                {
                    //TODO Parfor
                    foreach (OBJECT obj in ChangedObjects.Values)
                    {
                        if (obj.DBAction == DBACTION.NONE)
                            continue;
                        /*
                        if (obj.DBAction == DBACTION.INSERT ||
                            obj.DBAction == DBACTION.UPDATE)
                            obj.FireBeforeSaveEvent();
                        else if (obj.DBAction == DBACTION.DELETE)
                            obj.FireBeforeDeleteEvent();
                        */
                        changed.Add(obj.GetData());
                        input.Add(obj); 
                    }

                    Trace.WriteLine("Saving this number of objects: " + changed.Count, "WebAnnotation");

                    /*Don't make the call if there are no changes */
                    if (changed.Count == 0)
                        return true;

                    PROXY proxy = CreateProxy();
                    proxy.Open();

                    long[] newIDs = new long[0];
                    try
                    {
                        newIDs = ProxyUpdate(proxy, changed.ToArray());
                    }
                    catch (Exception e)
                    {
                        Trace.WriteLine("An error occurred during the update:\n" + e.Message);
                        return false;
                    }
                    finally
                    {
                        proxy.Close();
                    }

                    Debug.Assert(changed.Count == newIDs.Length);

                    List<OBJECT> newObjList = new List<OBJECT>(changed.Count);
                    List<OBJECT> updateObjList = new List<OBJECT>(changed.Count);
                    List<KEY> delObjList = new List<KEY>(changed.Count);

                    //Update ID's of new objects
                    for (int iObj = 0; iObj < changed.Count; iObj++)
                    {
                        WCFOBJECT data = changed[iObj];
                        //I have to do this because WCF does not marshal the templates for the WCFObject types
                        //This should be fairly future proof
                        WCFObjBaseWithKey<KEY, WCFOBJECT> keyObj = input[iObj]; 
                        if (IDToObject.ContainsKey(keyObj.ID) == false)
                        {
                            data.DBAction = DBACTION.NONE;
                            continue;
                        }

                        OBJECT obj = IDToObject[keyObj.ID];                        

                        switch (data.DBAction)
                        {
                            case DBACTION.INSERT:
                                if (newIDs.Length > iObj)
                                {
                                    //InternalUpdate(keyObj); 
                                    //Remove from our old spot in the database
                                    InternalDelete(keyObj.ID);

                                    //Update the ID of the object
                                    keyObj.GetData().ID = newIDs[iObj];

                                    //Insert in the new correct location
                                    //newobj = InternalAdd(obj);
                                    newObjList.Add(obj); 
                                }

                               // obj.FireAfterSaveEvent();
                                break;
                            case DBACTION.UPDATE:
                                //newobj = InternalUpdate(obj);
                                updateObjList.Add(obj); 
                               // obj.FireAfterSaveEvent();
                                break;

                            case DBACTION.DELETE:
                                //Remove from our old spot in the database
                                delObjList.Add(obj.ID);
                                //InternalDelete(obj.ID);
                               // obj.FireAfterDeleteEvent();
                                break;

                            default:
                                break;
                        }

                        data.DBAction = DBACTION.NONE;
                    }

                    InternalDelete(delObjList.ToArray());
                    output.AddRange(InternalUpdate(updateObjList.ToArray()));
                    output.AddRange(InternalAdd(newObjList.ToArray()));
                }
                catch (FaultException e)
                {
                    //  System.Windows.Forms.MessageBox.Show("An exception occurred while saving structure types.  Viking is pretending none of the changes happened.  Exception Data: " + e.Message, "Error");

                    if (changed != null)
                    {
                        //Update ID's of new objects
                        for (int iObj = 0; iObj < changed.Count; iObj++)
                        {
                            WCFOBJECT data = changed[iObj];
                            WCFObjBaseWithKey<KEY, WCFOBJECT> keyObj = data as WCFObjBaseWithKey<KEY, WCFOBJECT>;
                            if (keyObj == null)
                            {
                                Debug.Fail("Could not convert WCFOBJECT template to proper type, StoreBaseWithIndexKey.cs");
                                continue;
                            }

                            if (IDToObject.ContainsKey(keyObj.ID) == false)
                            {
                                data.DBAction = DBACTION.NONE;
                                continue;
                            }

                            OBJECT obj = IDToObject[keyObj.ID];

                            switch (data.DBAction)
                            {
                                case DBACTION.INSERT:
                                    //Remove from our old spot in the database
                                    InternalDelete(keyObj.ID);

                                    break;

                                case DBACTION.DELETE:
                                    //Just reset our DBState to none after case statement
                                    break;

                                default:
                                    break;
                            }

                            data.DBAction = DBACTION.NONE;
                        }
                    }

                    //If we caught an exception return false
                    return false;
                }
            }

            //CallOnAllUpdatesCompleted(new OnAllUpdatesCompletedEventArgs(output.ToArray()));

            return true;
        }
    }
}
