// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : IRPruner.cpp
// Author       : Bradley C. Grimm
// Created      : 2009/05/15 11:52
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : Used to prune uninteresting tiles.

// system includes:
#include <stdio.h>

// the includes:
#include <itk/fft_common.hxx>
#include <itk/mosaic_layout_common.hxx>

// local includes:
#include "IRPruner.h"

void IRPruner::PruneImages(std::list<the_text_t> &image_names,
                           std::vector<base_transform_t::Pointer> &transform,
                           const unsigned int &shrink_factor,
                           const double &pixel_spacing,
                           const bool &use_std_mask,
                           const unsigned int &tile_size,
                           const double &intensity_tolerance,
                           float overlap_percent)
{
  double min_interest = std::numeric_limits<unsigned int>::max();
  double max_interest = std::numeric_limits<unsigned int>::min();

  std::list<the_text_t> passed;
  std::list<the_text_t> failed;

  unsigned int i = 0;
  for (std::list<the_text_t>::const_iterator fn_iter = image_names.begin();
       fn_iter != image_names.end(); ++fn_iter, i++)
  {
    // Load in an image...
    image_t::Pointer image = std_tile<image_t>((*fn_iter), 
                                               shrink_factor, 
                                               pixel_spacing);

    mask_t::ConstPointer mask;
    if (use_std_mask)
    {
      mask = std_mask<image_t>(image);
    }

    image_t::RegionType image_region = image->GetBufferedRegion();
    image_t::SizeType image_size = image_region.GetSize();
    image_t::IndexType image_index = image_region.GetIndex();

    image_t::SizeType chunk_size;
    double full_count = 0.0;
    double interest_level = 0.0;
    
    float left_crop = image_size[0] * overlap_percent;
    float right_crop = image_size[0] * (1.0f - overlap_percent);
    
    float top_crop = image_size[1] * overlap_percent;
    float bottom_crop = image_size[1] * (1.0f - overlap_percent); 

    // Iterate through the image in small chunks...
    for (unsigned int x = 0, xid = 0; x < image_size[0]; x += tile_size, xid++)
    {
      if ( x > left_crop && (x+tile_size) < right_crop )
        continue;

      for (unsigned int y = 0, yid = 0; y < image_size[1]; y += tile_size, yid++)
      {
        if ( y > top_crop && (y+tile_size) < bottom_crop )
          continue;

        image_t::RegionType chunk_region;
        image_t::IndexType chunk_index;
        chunk_index[0] = x + image_index[0];
        chunk_index[1] = y + image_index[1];
        chunk_region.SetIndex( chunk_index );

        chunk_size[0] = std::min(tile_size, (unsigned int)(image_size[0] - x));
        chunk_size[1] = std::min(tile_size, (unsigned int)(image_size[1] - y));
        chunk_region.SetSize( chunk_size );

        image_t::IndexType index;
        typedef itk::ImageRegionIteratorWithIndex<image_t> iter_t;
        iter_t iter(image, chunk_region);
        
        // Find the average intensity on this chunk.
        double avg = 0.0, count = 0.0;
        for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
        {
          index = iter.GetIndex();
          double pixel_value = image->GetPixel(index);
          avg += pixel_value;
          count += 1.0;
          full_count += 1.0;
        }

        avg /= count;

        // Find how far each value is from the average.
        for (iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
        {
          index = iter.GetIndex();

          double pixel_value = image->GetPixel(index);
          interest_level += std::abs(pixel_value - avg);
        }
      }
    }

    double interest = (interest_level / full_count);
    if ( interest < min_interest )
      min_interest = interest;
    if ( interest > max_interest )
      max_interest = interest;
    
    if ( transform.size() != 0 )
    {
      // Save out those that passed the test.
      if ( interest > intensity_tolerance )
      {
        cout << "Passed the test with " << interest << ":       " << (*fn_iter) << endl;
        passed.push_back((*fn_iter));
      }
      else
      {
        cout << "Failed the test "  << interest << ":       " << (*fn_iter) << endl;
        failed.push_back((*fn_iter));
        if ( transform.size() > i )
          transform.erase(transform.begin()+i,transform.begin()+i+1);
        i--;
      }
    }
    else
      cout << (*fn_iter) << ": " << interest << endl;

    interest_level = 0.0;
  }
  
  if ( transform.size() != 0 )
  {
    cout << "Images that passed:" << endl;
    for (std::list<the_text_t>::const_iterator fn_iter = passed.begin();
         fn_iter != passed.end(); ++fn_iter)
    {
      cout << "  " << (*fn_iter) << endl;
    }

    cout << endl << "Images that failed:" << endl;
    for (std::list<the_text_t>::const_iterator fn_iter = failed.begin();
         fn_iter != failed.end(); ++fn_iter)
    {
      cout << "  " << (*fn_iter) << endl;
      image_names.remove( *fn_iter );
    }

    if ( min_interest > intensity_tolerance )
    {
      cout << "No images were pruned.  " 
        << "To prune an image, a tolerance of at least: "
        << min_interest << " is needed." << endl;

    }
  }
}