// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : LineMath.h
// Author       : Bradley C. Grimm
// Created      : 2009/08/19
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : Helper functions and definitions for using lines and doing
//                calculations on them.

#ifndef __LINE_MATH_H__
#define __LINE_MATH_H__

#include <itkPoint.h>
#include <itkVector.h>
#include <itk/the_grid_transform.hxx>

namespace ir
{

//----------------------------------------------------------------
// PointMath
//        Provides simple point functions.  Used mostly for convenience
//        and cleaner code.
class PointMath
{
public:
  static double DistanceSquared(const pnt2d_t &pt1, const pnt2d_t &pt2);
  static double Distance(const pnt2d_t &pt1, const pnt2d_t &pt2);

  static pnt2d_t VectorToPoint(const vec2d_t &pt);
  static pnt2d_t Direction(const pnt2d_t &pt1, const pnt2d_t &pt2);
  static pnt2d_t Scale(const pnt2d_t &pt1, const pnt2d_t &pt2, double factor);
  
  static pnt2d_t XYtoUV(const pnt2d_t &xy, const pnt2d_t &min, const vec2d_t &ext);
  static pnt2d_t UVtoXY(const pnt2d_t &uv, const pnt2d_t &min, const vec2d_t &ext);
};
  
//----------------------------------------------------------------
// Distance
//        Distance squared between two 2D points.
inline double PointMath::DistanceSquared(const itk::Point<double, 2> &pt1,
                                         const itk::Point<double, 2> &pt2)
{
  return (pt1[0] - pt2[0]) * (pt1[0] - pt2[0]) + 
         (pt1[1] - pt2[1]) * (pt1[1] - pt2[1]);
}
  
//----------------------------------------------------------------
// Distance
//        Distance between two 2D points.
inline double PointMath::Distance(const itk::Point<double, 2> &pt1,
                                  const itk::Point<double, 2> &pt2)
{
  return sqrt(DistanceSquared(pt1, pt2));
}
  
//----------------------------------------------------------------
// Direction
//        Gives the direction of the line (vector from pt2 to pt1)
inline itk::Point<double, 2> PointMath::Direction(const pnt2d_t &pt1, const pnt2d_t &pt2)
{
  itk::Vector<double, 2> dir;
  dir[0] = pt2[0] - pt1[0];
  dir[1] = pt2[1] - pt1[1];
  dir.Normalize();
  return PointMath::VectorToPoint(dir);
}

//----------------------------------------------------------------
// EdgeSegment
//        Provides an edge class that can be used to determine 
//        where edges intersect.
class EdgeSegment
{
public:
  EdgeSegment() { inUVSpace = true; }
  EdgeSegment(const vertex_t &pt1, const vertex_t &pt2) 
  { this->pt1 = pt1; this->pt2 = pt2; inUVSpace = true; }
  double SegmentsIntersect(const EdgeSegment& segment, pnt2d_t &intersection);

  void ToUV(const pnt2d_t &min, const vec2d_t &ext);
  void ToXY(const pnt2d_t &min, const vec2d_t &ext);

  vertex_t pt1;
  vertex_t pt2;

private:
  bool inUVSpace;
};

}

#endif


