// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-add-transforms.cxx
// Author       : Bradley C. Grimm
// Created      : 2009/07/31 10:05
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : 

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <itk/itkGridTransform.h>
#include <itk/itkMeshTransform.h>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>

// Local includes:
#include "IRPath.h"
#include "UsageReader.h"
#include "LineMath.h"

using namespace ir;

//----------------------------------------------------------------
// grid_transform_t
// 
typedef itk::GridTransform grid_transform_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t[-slice_dirs stos0_slice0_path stos0_slice1_path ...]"
       << "\t-save prefix \n"
       << "\t-load stos_0 [stos_i]*\n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-save slice/ \n"
       << "\t-load 000.stos 001.stos\n"
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// GetAllEdges
//        Return all edges containing a certain point in a given 
//        grid transform.
static std::vector<EdgeSegment> GetEdges(const the_grid_transform_t &transform,
                                         unsigned int pt_index)
{
  std::vector<EdgeSegment> edges;
  
  std::size_t num_tris = transform.grid_.tri_.size();
  for (unsigned int tri_idx = 0; tri_idx < num_tris; tri_idx++)
  {
    triangle_t warp_triangle = transform.grid_.tri_[tri_idx];

    for (unsigned int vertex_id = 0; vertex_id < 3; vertex_id++)
    {
      if ( pt_index == warp_triangle.vertex_[vertex_id] )
      {
        unsigned int tri_edge1 = (vertex_id + 1) % 3,
                     tri_edge2 = (vertex_id + 2) % 3;

        unsigned int mesh_idx1 = warp_triangle.vertex_[tri_edge1],
                     mesh_idx2 = warp_triangle.vertex_[tri_edge2];

        EdgeSegment line1(transform.grid_.mesh_[pt_index],
                          transform.grid_.mesh_[mesh_idx1]);
        EdgeSegment line2(transform.grid_.mesh_[pt_index],
                          transform.grid_.mesh_[mesh_idx2]);

        edges.push_back(line1);
        edges.push_back(line2);
      }
    }
  }

  return edges;
}

//----------------------------------------------------------------
// GetAllEdges
//        Return all edges in a given grid transform.
static std::vector<EdgeSegment> GetAllEdges(const the_grid_transform_t &transform)
{
  std::vector<EdgeSegment> edges;

  std::size_t num_tris = transform.grid_.tri_.size();
  for (unsigned int ft = 0; ft < num_tris; ft++)
  {
    for (unsigned int ft_idx = 0; ft_idx < 3; ft_idx++)
    {
      unsigned int fixed_pt1_idx = transform.grid_.tri_[ft].vertex_[ft_idx % 3], 
                   fixed_pt2_idx = transform.grid_.tri_[ft].vertex_[(ft_idx + 1) % 3];

      EdgeSegment line(transform.grid_.mesh_[fixed_pt1_idx],
                       transform.grid_.mesh_[fixed_pt2_idx]);
      edges.push_back(line);
    }
  }
  return edges;
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  std::list<the_text_t > stos_fn;
  std::list<stos_t<image_t> > stos_list;
  std::list<the_text_t > slice_dirs;

  std::list<the_text_t> save_list;
  the_text_t fn_save;
  
  // parse the command line arguments:
  unsigned int consume_default_tokens = 0;
  unsigned int consume_save_tokens = 0;
  unsigned int consume_slice_tokens = 0;
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      consume_save_tokens = 2;
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      consume_default_tokens = 2;
    }
    else if (strcmp(argv[i], "-slice_dirs") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -slice_dirs option");
      consume_slice_tokens = 2;
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-slice-to-slice.txt", "ir-stom");
    }
    else if (consume_default_tokens > 0)
    {
      stos_fn.push_back( argv[i] );
      consume_default_tokens++;
    }
    else if (consume_save_tokens > 0)
    {
      save_list.push_back(the_text_t(argv[i]));
      consume_save_tokens++;
    }
    else if (consume_slice_tokens > 0)
    {
      slice_dirs.push_back( IRPath::CleanPath( the_text_t( argv[i] ) ) );
      consume_slice_tokens++;
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
    
    // Used for multiple load files
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
    // Used for multiple save files
    if (consume_save_tokens > 0)
    {
      consume_save_tokens--;
    }
    // Used for multiple slice files
    if (consume_slice_tokens > 0)
    {
      consume_slice_tokens--;
    }
  }
  
  if ( slice_dirs.size() > 0 && slice_dirs.size() != stos_fn.size() + 1 )
  {
    usage(argv, "There must be exactly one slice_dir for each slice.");
  }
  
  bool override_slice_dirs = ( slice_dirs.size() > 0 );

  // a minimum of two images are required for the registration:
  const unsigned int num_pairs = stos_fn.size();
  if (num_pairs == 0)
  {
    usage(argv, "must specify the -load option");
  }

  const unsigned int num_saves = save_list.size();
  if (num_saves == 0)
  {
    usage(argv, "must specify the -save option");
  }

  // Load up all of the stos files.
  std::list<the_text_t>::const_iterator slice_iter = slice_dirs.begin();
  for (std::list<the_text_t>::const_iterator fn_iter = stos_fn.begin();
       fn_iter != stos_fn.end(); ++fn_iter)
  {
    the_text_t slice0_dir = "",
               slice1_dir = "";

    if ( override_slice_dirs )
    {
      slice0_dir = *slice_iter;
      slice_iter++;
      slice1_dir = *slice_iter;
    }

    stos_t<image_t> stos( *fn_iter, slice0_dir, slice1_dir );
    stos_list.push_back( stos );
  }

  bool use_as_prefix = false;
  if (num_saves != num_pairs)
  {
    if ( num_saves == 1 && !save_list.front().match_tail(".stos") )
    {
      use_as_prefix = true;
      fn_save = save_list.front();
    }
    else
    {
      usage(argv, "Must specify either A) a single prefix, " \
        "or B) a new filename for each .stos input except the first one.");
    }
  }
  
  stos_tree_t<image_t> stos_tree;
  std::size_t num_roots = stos_tree.build(stos_list);
  
  if (num_roots != 1)
  {
    if (num_roots == 0)
    {
      cerr << "ERROR: There is more than one possible mapping for "
           << "one or more slices, aborting... "
           << "Remove one of the ambiguous stos transforms and try again."
           << endl;
    }
    else
    {
      cerr << "ERROR: There are " << num_roots
           << " sets of interconnected stos transforms present, " << endl
           << " this tool only works with one set, aborting..." << endl;
    }
    stos_tree.dump(cerr);
    
    ::exit(1);
  }
  
  // this will sort the transforms in order of dependence and alphabetically:
  stos_list.clear();
  stos_tree.collect_stos(stos_list);
  
  // setup random access vectors:
  std::vector<stos_t<image_t> > stos(stos_list.begin(), stos_list.end());
  std::vector<the_text_t> saves_names(save_list.begin(), save_list.end());
  std::vector<itk::GridTransform::Pointer> grid_transforms(num_pairs);
  std::vector<itk::MeshTransform::Pointer> mesh_transforms(num_pairs);

  for (unsigned int i = 0; i < num_pairs; i++)
  {
    base_transform_t::Pointer transform_stos =
      const_cast<base_transform_t *>(stos[i].t01_.GetPointer());
    
    grid_transforms[i] =
      dynamic_cast<itk::GridTransform *>(transform_stos.GetPointer());

    if ( grid_transforms[i].GetPointer() == NULL )
    {
      cerr << stos[i].fn_load_.text() << " is not a grid transform.  Currently "
           << "ir-add-transforms only supports grid transforms." << std::endl;
      ::exit(1);
    }

    // Setup the new mesh transform.
    mesh_transforms[i] = itk::MeshTransform::New();
    mesh_transforms[i]->transform_.tile_min_ = 
                            grid_transforms[i]->transform_.tile_min_;
    mesh_transforms[i]->transform_.tile_ext_ = 
                            grid_transforms[i]->transform_.tile_ext_;
    mesh_transforms[i]->transform_.grid_.resize(
            grid_transforms[i]->transform_.grid_.rows_,
            grid_transforms[i]->transform_.grid_.cols_);

    // The first stos has no need to be added to...
    if ( i == 0 )
    {
      // Transform the points exactly.
      std::size_t num_pts = grid_transforms[0]->transform_.grid_.mesh_.size();
      for (unsigned int k = 0; k < num_pts; k++)
      {
        bool delay_setup = (i != num_pairs - 1);
        mesh_transforms[0]->transform_.insert_point(
                        grid_transforms[0]->transform_.grid_.mesh_[k].uv_,
                        grid_transforms[0]->transform_.grid_.mesh_[k].xy_,
                        delay_setup);
      }
    }
  }

  // Create new 'added' transforms for each stos except the first.
  for (unsigned int i = 0; i < num_pairs; i++)
  {
    // Transform each of the grid points.
    std::size_t num_pts = grid_transforms[i]->transform_.grid_.mesh_.size();
    // Just save out the first slice.  No addition necessary.
    if ( i == 0 )
      continue;
    
    the_grid_transform_t fixed_transform = grid_transforms[i-1]->transform_;
    the_grid_transform_t warp_transform = grid_transforms[i]->transform_;
    for (unsigned int k = 0; k < num_pts; k++)
    {
      // Check if point can be transformed...
      pnt2d_t xy_transformed;
      pnt2d_t uv_transformed = 
        PointMath::XYtoUV(warp_transform.grid_.mesh_[k].xy_, 
                          fixed_transform.tile_min_, 
                          fixed_transform.tile_ext_);

      bool success = fixed_transform.transform_inv(uv_transformed, 
                                                   xy_transformed);
      
      if ( success )
      {
        // Transformed successfully
        mesh_transforms[i]->transform_.insert_point(
          warp_transform.grid_.mesh_[k].uv_,
          xy_transformed,
          true);
      }
      else
      {
        // Could not be transformed.
        
        // Grab edges that include point, as well as all fixed transform edges.  
        std::vector<EdgeSegment> fixed_edges = GetAllEdges(fixed_transform);
        unsigned int num_fixed_edges = fixed_edges.size();

        std::vector<EdgeSegment> edges = GetEdges(warp_transform, k);
        unsigned int num_edges = edges.size();
        for (unsigned int edge_idx = 0; edge_idx < num_edges; edge_idx++)
        {
          // Convert to XY space for segment checks.
          edges[edge_idx].ToXY(warp_transform.tile_min_, 
                               warp_transform.tile_ext_);

          EdgeSegment found_edge;
          pnt2d_t intersect, found_intersect;
          double nearest_intersect = std::numeric_limits<double>::max();

          // Find the closest edge intersecting the point.
          for (unsigned int fix_idx = 0; fix_idx < num_fixed_edges; fix_idx++)
          {
            fixed_edges[fix_idx].ToXY(fixed_transform.tile_min_, 
                                      fixed_transform.tile_ext_);
            double distance = 
              edges[edge_idx].SegmentsIntersect(fixed_edges[fix_idx], intersect);

            if( distance < nearest_intersect )
            {
              found_edge = fixed_edges[fix_idx];
              nearest_intersect = distance;
              found_intersect = intersect;
            }
          }
            
          if ( nearest_intersect == std::numeric_limits<double>::max() )
            continue;
          
          // Is this point inside the grid?
          pnt2d_t xy_transformed;
          pnt2d_t uv_transformed = 
            PointMath::XYtoUV(edges[edge_idx].pt2.xy_, 
                              warp_transform.tile_min_, 
                              warp_transform.tile_ext_);
          bool success = fixed_transform.transform_inv(uv_transformed, 
                                                       xy_transformed);
          if ( !success )
            continue;

          // Determine our new ctrl (xy) point.
          double dist_to_point = 
            PointMath::Distance(found_edge.pt1.uv_, found_intersect);
          double found_uv_length = 
            PointMath::Distance(found_edge.pt1.uv_, found_edge.pt2.uv_);

          double map_fraction = dist_to_point / found_uv_length;
          pnt2d_t ctrl_point = PointMath::Scale(found_edge.pt1.xy_, 
                                                found_edge.pt2.xy_, 
                                                map_fraction);

          // Determine our new map (uv) point.
          double xy_length = PointMath::Distance(edges[edge_idx].pt1.xy_, 
                                                 edges[edge_idx].pt2.xy_);
          double fraction = nearest_intersect / xy_length;
          fraction = (fraction > 1) ? 1 : ((fraction < 0) ? 0 : fraction);
          pnt2d_t map_point = PointMath::Scale(edges[edge_idx].pt1.uv_, 
                                               edges[edge_idx].pt2.uv_, 
                                               fraction);
          
          // Convert point back to uv space.  
          map_point = PointMath::XYtoUV(map_point, 
                                        warp_transform.tile_min_, 
                                        warp_transform.tile_ext_);
          
          // Add to new mesh transform.
          mesh_transforms[i]->transform_.insert_point(map_point, 
                                                      ctrl_point, 
                                                      true);
        }
      }
    }

    // Save out new transform.
    stos[i].t01_ = mesh_transforms[i];
    if ( use_as_prefix )
    {
      the_text_t fn_filename = fn_save + the_text_t::number(i, 3, '0');
      the_text_t fn_extension(".stos");
      the_text_t fn_numbered_save = fn_filename + fn_extension;
      stos[i].save(fn_numbered_save);
    }
    else
    {
      stos[i].save(saves_names[i]);
    }
  }
  
  return 0;
}
