// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-assemble-rgb.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2006/04/18 17:09
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Assemble a colored mosaic image from a .mosaic file.

// the includes:
#include <itk/common.hxx>
#include <itk/draw.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>

// Local includes:
#include "IRPath.h"
#include "UsageReader.h"

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t[-sh shrink_factor] \n"
       << "\t[-clahe slope] \n"
       << "\t[-feathering [none | blend | binary]] \n"
       << "\t[-image_dir path_to_images] \n"
       << "\t-load transform_params_file_in \n"
       << "\t-save mosaic_file_out \n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-sh 8 \n"
       << "\t-feathering blend \n"
       << "\t-load params.mosaic \n"
       << "\t-save image.tif \n"
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  double clahe_slope = 1.0;
  std::list<the_text_t> in;
  the_text_t fn_save;
  feathering_t feathering = FEATHER_NONE_E;
  
  the_text_t fn_in;
  the_text_t image_dir;
  
  std::vector<base_transform_t::Pointer> transform;
  std::vector<image_t::Pointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-feathering") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -feathering option");
      
      i++;
      if (strcmp(argv[i], "blend") == 0)
      {
        feathering = FEATHER_BLEND_E;
      }
      else if (strcmp(argv[i], "binary") == 0)
      {
        feathering = FEATHER_BINARY_E;
      }
      else if (strcmp(argv[i], "none") == 0)
      {
        feathering = FEATHER_NONE_E;
      }
      else
      {
        usage(argv, "could not parse -feathering option");
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_in = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-image_dir") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -image_dir option");
      
      i++;
      image_dir = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-mosaicking.txt", "ir-assemble-rgb");
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
  }

  // Load the .mosaic in...
  std::fstream fin;
  fin.open(fn_in, ios::in);
  if (!fin.is_open())
  {
    usage(argv,
          the_text_t("could not open \'") +
          the_text_t(fn_in) +
          the_text_t("\' for reading"));
  }
  
  load_mosaic<base_transform_t>(fin,
                                pixel_spacing,
                                use_std_mask,
                                in,
                                transform,
                                image_dir);
  
  unsigned int num_images = transform.size();
  assert(num_images > 1);
  assert(pixel_spacing != 0);
  
  mask.resize(num_images);
  image.resize(num_images);
  
  unsigned int i = 0;
  for (std::list<the_text_t>::const_iterator iter = in.begin();
       iter != in.end(); ++iter, i++)
  {
    const the_text_t & fn_image = *iter;
    
    // read the image:
    cout << setw(3) << i << " ";
    image[i] = std_tile<image_t>(fn_image, shrink_factor, pixel_spacing);
    if (use_std_mask)
    {
      mask[i] = std_mask<image_t>(image[i]);
    }
    
    if (clahe_slope > 1.0)
    {
      image[i] = CLAHE<image_t>(image[i],
                                255,
                                255,
                                clahe_slope,
                                256,
                                0.0,
                                255.0,
                                mask[i]);
    }
  }
  
  fin.close();
  
  // sanity checks:
  if (fn_save.size() == 0)  usage(argv, "must specify -save option");
  if (in.empty()) usage(argv, "must specify at least 1 input image");
  
  cout << "shrink factor: " << shrink_factor << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "use feathering: " << feathering << endl
       << endl;
  
  num_images = image.size();
  
  cout << "assembling mosaic..." << endl;
  image_t::Pointer mosaic[3];
  make_mosaic_rgb(mosaic,
                  std::vector<bool>(num_images, false), // omit vector
                  transform,
                  image,
                  mask,
                  feathering,
                  255,
                  true);
  
  save_rgb<image_t::Pointer>(mosaic, fn_save);
  
  // done:
  return 0;
}
