// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-assemble.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2005/03/29 16:55
// Copyright    : (C) 2004-2008 University of Utah
// License      : GPLv2
// Description  : Assemble a mosaic image from a .mosaic file.

// the includes:
#include <itk/common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <cmath>

// Local includes:
#include "AsyncMosiacSave.h"
#include "IRPath.h"
#include "UsageReader.h"

#include <visus_support.h>

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t[-threads number_of_threads] \n"
       << "\t[-sh shrink_factor] \n"
       << "\t[-sp pixel_spacing] \n"
       << "\t[-slice slice_num]     (required for .idx) \n"
       << "\t[-channel channel_num] (required for .idx) \n"
       << "\t[-remap | -saveas-int16 | -saveas-uint16 ] \n"
       << "\t[-clahe slope] \n"
       << "\t[[-variance] | [-feathering [none | blend | binary]]] \n"
       << "\t[-load_as_needed] \n"
       << "\t[-tilesize tile_width tile_height] \n"
       << "\t[-image_dir path_to_images] \n"
       << "\t-load transform_params_file_in \n"
       << "\t-save mosaic_file_out \n"
       << "\t[-mask mosaic_mask_file_out] \n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-sh 8 \n"
       << "\t-feathering blend \n"
       << "\t-tilesize -1 -1 \n"
       << "\t-load params.mosaic \n"
       << "\t-save mosaic.tif \n"
       << "\t-mask mosaic_mask.tif \n\n"
#if IDX_ENABLED
       << "EXAMPLE (IDX): " << argv[0] << " \n"
       << "\t-feathering blend \n"
       << "\t-load 0001-0007/refine-translate.mosaic \n"
       << "\t-image_dir 0001-0007/001/C03 \n"
       << "\t-slice 7 \n"
       << "\t-channel 2 \n"
       << "\t-save section.idx \n"
#endif
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(2);
  cout.setf(ios::scientific);
  
  track_progress(true);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  bool pixel_spacing_user_override = false;
  bool remap_values = false;
  bool saveas_int16 = false;
  bool saveas_uint16 = false;
  double pixel_spacing = 1.0;
  int z_slice = -1;
  int channel = -1;
  bool use_std_mask = false;
  double clahe_slope = 1.0;
  std::list<the_text_t> in;
  the_text_t fn_save;
  the_text_t extension;
  the_text_t filename;
  the_text_t fn_mask;
  bool save_variance = false;
  feathering_t feathering = FEATHER_NONE_E;
  unsigned int tile_width = std::numeric_limits<unsigned int>::max();
  unsigned int tile_height = std::numeric_limits<unsigned int>::max();
  bool defer_image_loading = false;
  bool verbose = false;
  
  the_text_t fn_in;
  the_text_t image_dir;
  
  // use as many threads as supported by hardware:
  unsigned int num_threads = boost::thread::hardware_concurrency();
  
  std::vector<base_transform_t::Pointer> transform;
  std::vector<image_t::ConstPointer> image;
  std::vector<mask_t::ConstPointer> mask;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-threads") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -threads option");
      
      i++;
      num_threads = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -threads option");
    }
    else if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -sp option");
      
      pixel_spacing_user_override = true;
    }
    else if (strcmp(argv[i], "-slice") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -slice option");
      
      i++;
      z_slice = the_text_t(argv[i]).toInt(&ok);
      if (!ok) usage(argv, "bad -slice option");
    }
    else if (strcmp(argv[i], "-channel") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -channel option");
      
      i++;
      channel = the_text_t(argv[i]).toInt(&ok);
      if (!ok) usage(argv, "bad -channel option");
    }
    else if (strcmp(argv[i], "-remap") == 0)
    {
      remap_values = true;
    }
    else if (strcmp(argv[i], "-saveas-int16") == 0)
    {
      saveas_int16 = true;
    }
    else if (strcmp(argv[i], "-saveas-uint16") == 0)
    {
      saveas_uint16 = true;
    }
    else if (strcmp(argv[i], "-clahe") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -clahe option");
      
      i++;
      clahe_slope = the_text_t(argv[i]).toDouble(&ok);
      if (!ok || clahe_slope < 1.0) usage(argv, "bad -clahe option");
    }
    else if (strcmp(argv[i], "-variance") == 0)
    {
      save_variance = true;
    }
    else if (strcmp(argv[i], "-feathering") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -feathering option");
      
      i++;
      if (strcmp(argv[i], "blend") == 0)
      {
        feathering = FEATHER_BLEND_E;
      }
      else if (strcmp(argv[i], "binary") == 0)
      {
        feathering = FEATHER_BINARY_E;
      }
      else if (strcmp(argv[i], "none") == 0)
      {
        feathering = FEATHER_NONE_E;
      }
      else
      {
        usage(argv, "could not parse -feathering option");
      }
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_in = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-image_dir") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -image_dir option");
      
      i++;
      image_dir = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
      
      // extract output filename extension (excluding the '.'):
      extension = fn_save.reverse().cut('.', 0, 0).reverse();
      
      // extract output filename (excluding the '.' and extension):
      filename = fn_save.extract(0, fn_save.size() - extension.size() - 1);
      
      // prepend '.' to the output filename extension:
      if (strcmp(extension.text(), "xml") != 0)
        extension = the_text_t(".") + extension;
      else
        extension = ".png";
    }
    else if (strcmp(argv[i], "-mask") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -mask option");
      
      i++;
      fn_mask = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-tilesize") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -tilesize option");

      i++;
      tile_width = the_text_t(argv[i]).toUInt(&ok);

      i++;
      tile_height = the_text_t(argv[i]).toUInt(&ok);
    }
    else if (strcmp(argv[i], "-load_as_needed") == 0)
    {
      defer_image_loading = true;
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-mosaicking.txt", "ir-assemble");
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
  }

  // Load in cached .mosaic:
  std::fstream fin;
  fin.open(fn_in, ios::in);
  if (!fin.is_open())
  {
    usage(argv,
          the_text_t("could not open \'") +
          the_text_t(fn_in) +
          the_text_t("\' for reading"));
  }
  
  double mosaic_pixel_spacing = 0;
  load_mosaic<base_transform_t>(fin,
                                mosaic_pixel_spacing,
                                use_std_mask,
                                in,
                                transform,
                                image_dir);
  
  if (!pixel_spacing_user_override)
  {
    pixel_spacing = mosaic_pixel_spacing;
  }
  
  unsigned int num_images = transform.size();
  assert(pixel_spacing != 0);
  
  image.resize(num_images);
  mask.resize(num_images);

  if ( !defer_image_loading )
  {
    unsigned int i = 0;
    for (std::list<the_text_t>::const_iterator iter = in.begin();
         iter != in.end(); ++iter, i++)
    {
      // Preload all images, instead of loading them as needed.
      cout << setw(3) << i << " ";
      image[i] = std_tile<image_t>((*iter), 
                                    shrink_factor, pixel_spacing);
      if (use_std_mask)
      {
        mask[i] = std_mask<image_t>(image[i]);
      }
      
      if (clahe_slope > 1.0)
      {
        image[i] = CLAHE<image_t>(image[i],
                                  255,
                                  255,
                                  clahe_slope,
                                  256,
                                  0.0,
                                  255.0,
                                  mask[i]);
      }
    }
  }
  
  fin.close();

  set_major_progress(0.05);

  // calculate downsample.
  double downsample = shrink_factor * pixel_spacing;

  bool save_tiles = (tile_width != std::numeric_limits<unsigned int>::max() ||
                     tile_height != std::numeric_limits<unsigned int>::max());
  
  // sanity checks:
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify -save option");
  }
  
  if (in.empty())
  {
    usage(argv, "must specify at least 1 input image");
  }
  
  if (defer_image_loading && !save_tiles)
  {
    usage(argv, "Defer image loading requires -tilesize argument.");
  }
  
  num_images = in.size();
  cout << "threads:       " << num_threads << endl
       << "mosaic tiles:  " << num_images << endl
       << "shrink factor: " << shrink_factor << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "save variance: " << save_variance << endl
       << "use feathering: " << feathering << endl
       << "remap min max: " << remap_values << endl
       << endl;
  
  image_t::Pointer mosaic;
  image_t::Pointer full_mosaic;
  mask_t::Pointer mosaic_mask;
  image_t::SizeType mosaic_sz;
  the_thread_pool_t thread_pool(1);

  if ( !save_tiles && defer_image_loading )
  {
    // Setup the full mosaic to copy to.
  }

  if (save_variance)
  {
    cout << "assembling mosaic variance image..." << endl;
    
    typedef itk::LinearInterpolateImageFunction<image_t, double>
      image_interpolator_t;
    
    typedef itk::ImageMosaicVarianceMetric<image_t, image_interpolator_t>
      mosaic_metric_t;
    
    mosaic_metric_t::Pointer mosaic_metric = mosaic_metric_t::New();
    mosaic_metric->image_.resize(num_images);
    mosaic_metric->mask_.resize(num_images);
    mosaic_metric->transform_.resize(num_images);
    for (unsigned int i = 0; i < num_images; i++)
    {
      mosaic_metric->image_[i] = image[i];
      mosaic_metric->mask_[i] = mask[i];
      mosaic_metric->transform_[i] = transform[i];
    }
    
    std::vector<bool> param_shared;
    std::vector<bool> param_active;
    param_shared.assign(transform[0]->GetNumberOfParameters(), false);
    param_active.assign(transform[0]->GetNumberOfParameters(), false);
    mosaic_metric->setup_param_map(param_shared, param_active);
    mosaic_metric->Initialize();
    
    mosaic_metric_t::measure_t max_var;
    mosaic_metric_t::measure_t avg_var;
    mosaic = mosaic_metric->variance(max_var, avg_var);
    cout << "max variance: " << max_var << endl
         << "avg variance: " << avg_var << endl;
  }
  else
  {
    const std::vector<double> tint(num_images, 1.0);
    const std::vector<bool> omit(num_images, false);
    
    // mosaic bounding box:
    image_t::PointType mosaic_min;
    image_t::PointType mosaic_max;
    std::vector<image_t::PointType> image_min;
    std::vector<image_t::PointType> image_max;
    if ( defer_image_loading )
    {
      // Creates the bounding boxes for all images, but loads and unloads
      // each image as needed.
      calc_mosaic_bbox_load_images<image_t::ConstPointer, 
                                    base_transform_t::Pointer>
        (transform,
         in,
         mosaic_min,
         mosaic_max,
         image_min,
         image_max,
         shrink_factor,
         pixel_spacing);
    }
    else
    {
      calc_mosaic_bbox<image_t::ConstPointer, base_transform_t::Pointer>
        (transform,
         image,
         mosaic_min,
         mosaic_max);
    }
    
    if ( defer_image_loading )
    {
      image[0] = std_tile<image_t>(in.front(), 
                                   shrink_factor, 
                                   pixel_spacing);
    }
    
    image_t::SpacingType mosaic_sp = image.front()->GetSpacing();
    if ( defer_image_loading )
    {
      image[0] = image_t::Pointer(NULL);
    }

    mosaic_sz[0] = (unsigned int)((mosaic_max[0] - mosaic_min[0]) /
                                  mosaic_sp[0]);
    mosaic_sz[1] = (unsigned int)((mosaic_max[1] - mosaic_min[1]) /
                                  mosaic_sp[1]);

    bool defer_for_w = (defer_image_loading 
          && tile_width != std::numeric_limits<unsigned int>::max());
    bool defer_for_h = (defer_image_loading 
          && tile_height != std::numeric_limits<unsigned int>::max());
    
    image_t::PointType step_sz;
    step_sz[0] = defer_for_w ? tile_width  : (mosaic_max[0] - mosaic_min[0]);
    step_sz[1] = defer_for_h ? tile_height : (mosaic_max[1] - mosaic_min[1]);

    image_t::PointType deferred_tile_sz;
    deferred_tile_sz[0] = defer_for_w ? step_sz[0] : mosaic_sz[0];
    deferred_tile_sz[1] = defer_for_h ? step_sz[1] : mosaic_sz[1];

    cout << endl << endl << "assembling mosaic..." << endl;
    
    image_t::PointType tile_min = mosaic_min;
    image_t::PointType tile_max = mosaic_max;
    unsigned int column_number = 0;
    unsigned int total_columns = (mosaic_sz[0] + (tile_width - 1)) / tile_width;
    for (unsigned int xid = 0; tile_min[0] + 0.001 < mosaic_max[0]; 
          tile_min[0] += step_sz[0], xid++)
    {
      tile_min[1] = mosaic_min[1];
      bool load_column = true;
      column_number++;
      for (unsigned int yid = 0; tile_min[1] + 0.001 < mosaic_max[1]; 
            tile_min[1] += step_sz[1], yid++)
      {
        tile_max[0] = tile_min[0] + deferred_tile_sz[0];
        tile_max[1] = tile_min[1] + deferred_tile_sz[1];
        
        image_t::SizeType tile_sz;
        tile_sz[0] = (unsigned int)(tile_max[0] - tile_min[0]);
        tile_sz[1] = (unsigned int)(tile_max[1] - tile_min[1]);
        
        bool assemble_mosaic_mask = !fn_mask.is_empty();

        bool overlaps = !defer_image_loading;
        if ( defer_image_loading )
        {
          // Check which images are needed.  Load them and apply mask/clahe.
          // Note: using clahe and mask with defer_image_loading is a bad idea.
          
          unsigned int i = 0;
          for (std::list<the_text_t>::const_iterator iter = in.begin();
               iter != in.end(); ++iter, i++)
          {
            // Check if this specific tile overlaps
            if ( bbox_overlap(tile_min, tile_max, image_min[i], image_max[i]) )
              overlaps = true;

            if ( !load_column )
              continue;

            image_t::PointType column_max = tile_max;
            column_max[1] = (unsigned int)(mosaic_max[1]);
            if ( bbox_overlap(tile_min, column_max, image_min[i], image_max[i]) )
            {
              if ( image[i].IsNull() )
              {
                image[i] = std_tile<image_t>((*iter), 
                                              shrink_factor, 
                                              pixel_spacing,
                                              verbose);

                if (use_std_mask)
                {
                  mask[i] = std_mask<image_t>(image[i]);
                }
                
                if (clahe_slope > 1.0)
                {
                  image[i] = CLAHE<image_t>(image[i],
                                            255,
                                            255,
                                            clahe_slope,
                                            256,
                                            0.0,
                                            255.0,
                                            mask[i]);
                }
              }
            }
            else if ( image[i].IsNotNull() )
            {
              if ( verbose )
                cout << "unloading " << (*iter) << endl;
              image[i] = image_t::Pointer(NULL);
            }
          }
        }

        // Only load the column from the first tile.
        load_column = false;

        if ( !overlaps )
        {
          if ( verbose )
          {
            cout << "Skipping tile, no overlap was found..." << endl;
          }
          continue;
        }

        if ( verbose )
        {
          if (!assemble_mosaic_mask)
          {
            cout << "assembling mosaic..." << endl;
          }
          else
          {
            cout << "assembling mosaic and mosaic mask..." << endl;
          }
        }

        mosaic =
          make_mosaic_mt<image_t::ConstPointer, base_transform_t::Pointer>
          (num_threads,
           assemble_mosaic_mask,
           mosaic_mask,
           mosaic_sp,
           tile_min,
           tile_sz,
           image.size(),
           omit,
           tint,
           transform,
           image,
           mask,
           feathering,
           0.0,           // background
           false);        // don't allocate flag

        if ( defer_image_loading && save_tiles )
        {
          the_text_t fn_partialSave = filename;
          fn_partialSave += "_X";
          fn_partialSave += the_text_t::number(xid, 3, '0');
          fn_partialSave += "_Y";
          fn_partialSave += the_text_t::number(yid, 3, '0');
          fn_partialSave += extension;

          if (fn_save.match_tail(".mha") ||
              fn_save.match_tail(".mhd") ||
              fn_save.match_tail(".nrrd"))
          {
            save_mosaic_async<image_t>(mosaic, fn_partialSave, thread_pool);
          }
          else if (saveas_uint16)
          {
            typedef itk::Image<unsigned short int, 2> uint16_image_t;
            
            save_mosaic_async<uint16_image_t>(cast<image_t, uint16_image_t>(mosaic), 
                                 fn_partialSave, thread_pool, verbose);
          }
          else if (saveas_int16)
          {
            typedef itk::Image<short int, 2> int16_image_t;
            
            save_mosaic_async<int16_image_t>(cast<image_t, int16_image_t>(mosaic), 
                                fn_partialSave, thread_pool, verbose);
          }
          else
          {
            save_mosaic_async<native_image_t>(cast<image_t, native_image_t>(mosaic), 
                                 fn_partialSave, thread_pool, verbose);
          }
        }
        else
        {
          // Copy mosaic to larger mosaic and continue...
        }
      }
      cout << "Column " << column_number << " of " << total_columns << " was completed." << endl;
    }
  }

  if ( defer_image_loading && save_tiles )
  {
    thread_pool.wait();
  }

  // We want to unload them as quick as we can.
  if ( defer_image_loading )
  {
    unsigned int i = 0;
    for (std::list<the_text_t>::const_iterator iter = in.begin();
         iter != in.end(); ++iter, i++)
    {
      if ( image[i].IsNull() )
      {
        continue;
      }
      
      if ( verbose )
        cout << "unloading " << (*iter) << endl;
      image[i] = image_t::Pointer(NULL);
    }
  }

  // Write out the xml, signaling the finish.
  if ( defer_image_loading && save_tiles )
  {
    save_tile_xml<native_image_t>(filename,
                                  extension,
                                  tile_width,
                                  tile_height,
                                  mosaic_sz[0],
                                  mosaic_sz[1],
                                  downsample,
                                  false);
  }

  if (remap_values)
  {
    remap_min_max_inplace<image_t>(mosaic);
  }

  if ( !defer_image_loading )
  {
    if (fn_save.match_tail(".mha") ||
        fn_save.match_tail(".mhd") ||
        fn_save.match_tail(".nrrd"))
    {
      if (save_tiles)
      {
        save_as_tiles<image_t>(mosaic,
                               filename,
                               extension,
                               tile_width,
                               tile_height,
                               downsample);
      }
      else
      {
        save<image_t>(mosaic, fn_save);
      }
    }
    else if (saveas_uint16)
    {
      typedef itk::Image<unsigned short int, 2> uint16_image_t;
      if (save_tiles)
      {
        save_as_tiles<uint16_image_t>(cast<image_t, uint16_image_t>(mosaic),
                                      filename,
                                      extension,
                                      tile_width,
                                      tile_height, 
                                      downsample);
      }
      else
      {
        save<uint16_image_t>(cast<image_t, uint16_image_t>(mosaic), fn_save);
      }
    }
    else if (saveas_int16)
    {
      typedef itk::Image<short int, 2> int16_image_t;
      if (save_tiles)
      {
        save_as_tiles<int16_image_t>(cast<image_t, int16_image_t>(mosaic),
                                     filename,
                                     extension,
                                     tile_width,
                                     tile_height,
                                     downsample);
      }
      else
      {
        save<int16_image_t>(cast<image_t, int16_image_t>(mosaic), fn_save);
      }
    }
    else
    {
      if (save_tiles)
      {
        save_as_tiles<native_image_t>(cast<image_t, native_image_t>(mosaic),
                                      filename,
                                      extension,
                                      tile_width,
                                      tile_height,
                                      downsample);
      }
#if IDX_ENABLED
      else if (strcmp(extension,".idx")==0)
      {
        // save idx
        return save_to_idx(fn_save, cast<image_t, native_image_t>(mosaic), z_slice, channel);
      }
#endif
      else
      {
        save<native_image_t>(cast<image_t, native_image_t>(mosaic), fn_save);
      }
    }
  }

  if (fn_mask.is_empty())
  {
    // done:
    return 0;
  }

  if (mosaic_mask.GetPointer() == NULL)
  {
    // assemble the mosaic mask:
    cout << "assembling mosaic mask..." << endl;
    if (!use_std_mask)
    {
      for (unsigned int i = 0; i < num_images; i++)
      {
        // since the standard mask was not used
        // assume the whole image is masked:
        mask[i] = std_mask<image_t>(image[i], false);
        
        // free up the image memory, we no longer need it:
        image[i] = image_t::Pointer(NULL);
      }
    }
    
    mosaic_mask = make_mask<base_transform_t::Pointer>(transform, mask);
  }
  
  // save the mosaic mask:
  save<mask_t>(mosaic_mask, fn_mask);
  
  // done:
  return 0;
}
