// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-prune.cpp
// Author       : Bradley C. Grimm
// Created      : 2009/04/30 6:52
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : Tool to prune uninteresting tiles from the given
//                mosaics.

// local includes:
#include "visualize.hxx"

// the includes:
#include <utils/the_utils.hxx>

// system includes:
#include <vector>

// local includes:
#include "IRPath.h"
#include "IRPruner.h"
#include "UsageReader.h"

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t[-sh shrink_factor] \n"
       << "\t[-tolerance intensity_tolerance] \n"
       << "\t[-prune_tile_size length] \n"
       << "\t[-image_dir path_to_images] \n"
       << "\t-load [.mosaic | image] \n"
       << "\t-save [.mosaic] \n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-sh 1 \n"
       << "\t-tolerance 10 1 \n"
       << "\t-load 0001.001.png 0001.002.png \n"
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(3);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  unsigned int highest_resolution_level_shrink_factor = 1;
  unsigned int consume_default_tokens = 0;
  bool pixel_spacing_user_override = false;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  the_text_t fn_results;
  unsigned int prune_tile_size = 32;
  double intensity_tolerance = 0;
  float overlap_percent = 1.0f;
  
  the_text_t fn_in;
  the_text_t image_dir;
  
  std::list<the_text_t> in;
  std::vector<base_transform_t::Pointer> transform;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      highest_resolution_level_shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      i++;

      if (the_text_t(argv[i]).match_tail(".mosaic"))
      {
        fn_in = the_text_t( argv[i] );
      }
      else
      {
        i--;
        consume_default_tokens = 2;
      }
    }
    else if (strcmp(argv[i], "-image_dir") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -image_dir option");
      
      i++;
      image_dir = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
        if ((argc - i) <= 1) usage(argv, "could not parse -save option");

        i++;
        fn_results = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-image-processing.txt", "ir-prune");
    }
    else if (strcmp(argv[i], "-prune_tile_size") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -prune_tile_size option"); 

      i++;
      prune_tile_size = the_text_t(argv[i]).toInt(&ok);
      if (!ok) usage(argv, "bad -prune_tile_size option");
    }
    else if (strcmp(argv[i], "-tolerance") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -tolerance option");  
      
      i++;
      intensity_tolerance = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -tolerance option");
    }
    else if (strcmp(argv[i], "-overlap") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -overlap option"); 
      
      i++;
      overlap_percent = the_text_t(argv[i]).toFloat(&ok);
      if (!ok) usage(argv, "bad -overlap option");
    }
    else if (consume_default_tokens > 0)
    {
      in.push_back(the_text_t(argv[i]));
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unknown option \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  bool isMosaic = fn_in.match_tail(".mosaic");
  if ( in.size() == 0 && !isMosaic )
  {
    usage(argv, the_text_t("No image or mosaic provided."));
  }
  
  if ( isMosaic ) 
  {
    std::cout << "Loading mosaic..." << std::endl;
    
    // Load in the mosaic...
    std::fstream fin;
    
    fin.open(fn_in, ios::in);
    if (!fin.is_open())
    {
      usage(argv,
            the_text_t("could not open \'") +
            the_text_t( fn_in ) +
            the_text_t("\' for reading"));
    }
    
    double mosaic_pixel_spacing = 0;
    load_mosaic<base_transform_t>(fin,
                                  mosaic_pixel_spacing,
                                  use_std_mask,
                                  in,
                                  transform,
                                  image_dir);
    
    if (!pixel_spacing_user_override)
    {
      pixel_spacing = mosaic_pixel_spacing;
    }
    
    assert(pixel_spacing != 0);    
    
    fin.close();
  }

  std::cout << "Pruning tiles..." << std::endl;

  IRPruner::PruneImages(in,
                        transform,
                        shrink_factor,
                        pixel_spacing,
                        use_std_mask,
                        prune_tile_size,
                        intensity_tolerance,
                        overlap_percent);

  if ( isMosaic )
  {
    std::fstream fout;
	fout.open(fn_results, ios::out);
	
	if (!fout.is_open())
	{
		usage(argv,
			  the_text_t("could not open \'") + fn_results +
			  the_text_t("\' for writing"));
	}
	
	save_mosaic<base_transform_t>(fout,
								  pixel_spacing,
								  use_std_mask,
								  in,
								  transform);
  }
  // done:
  return 0;
}
