// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// File         : ir-refine-translate.cxx
// Author       : Joel Spaltenstein
// Created      : 2008/02/19
// Copyright    : (C) 2008 University of Utah
// License      : GPLv2
// Description  :

// the includes:
#include <itk/common.hxx>
#include <itk/mosaic_refinement_common.hxx>
#include <itk/itkImageMosaicVarianceMetric.h>
#include <itk/itkRadialDistortionTransform.h>
#include <itk/itkLegendrePolynomialTransform.h>
#include <itk/itkRegularStepGradientDescentOptimizer2.h>
#include <utils/the_dynamic_array.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>

// ITK includes:
#include <itkDiscreteGaussianImageFilter.h>
#include <itkShrinkImageFilter.h>

// local includes:
#include "IRRefineTranslateCanvas.hxx"
#include "IRImageLoader.hxx"
#include "IRPruner.h"
#include "IRPath.h"
#include "UsageReader.h"

//----------------------------------------------------------------
// affine_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 1>
affine_transform_t;

//----------------------------------------------------------------
// cubic_transform_t
// 
typedef itk::LegendrePolynomialTransform<double, 3>
cubic_transform_t;

//----------------------------------------------------------------
// image_interpolator_t
// 
typedef itk::LinearInterpolateImageFunction<image_t, double>
image_interpolator_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
     << "\t[-sh shrink_factor] \n"
     << "\t[-sp pixel_spacing] \n"
     << "\t[-threads number_of_threads] \n"
     << "\t[-tolerance intensity_tolerance] \n"
     << "\t[-prune_tile_size length] \n"
     << "\t[-image_dir path_to_images] \n"
     << "\t[-max_offset x y | x_percent% y_percent%] \n"
     << "\t[-black_mask x_boarder_percent y_boarder_percent] \n"
     << "\t[-noclahe] \n"
     << "\t[-overlap_percent] \n"
     << "\t-load any_transform_in \n"
     << "\t-save grid_transform_out \n"
     << endl
     << "EXAMPLE: " << argv[0] << " \n"
     << "\t-sh 8 \n"
     << "\t-load fft/s01.mosaic \n"
     << "\t-save grid/s01.mosaic \n"
     << endl
     << endl
     << "Pass in -help for more detailed information about the tool."
     << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(3);
  cout.setf(ios::scientific);

  track_progress(true);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parse the command line arguments:
  unsigned int shrink_factor = 1;
  bool pixel_spacing_user_override = false;
  double pixel_spacing = 1.0;
  bool use_std_mask = false;
  unsigned int num_threads = boost::thread::hardware_concurrency();
  the_text_t fn_results;
  std::vector<the_text_t> imagePathVector;
  std::vector<the_text_t> maskPathVector;
  unsigned int prune_tile_size = 32;
  double intensity_tolerance = 0;
  double max_offset[2];
  bool offsetIsPercent = false;
  max_offset[0] = std::numeric_limits<double>::max();
  max_offset[1] = std::numeric_limits<double>::max();
  double black_mask_percent[2];
  black_mask_percent[0] = std::numeric_limits<double>::max();
  black_mask_percent[1] = std::numeric_limits<double>::max();
  bool use_clahe = true;
  bool verbose = false;
  
  std::list<the_text_t> in;
  std::vector<base_transform_t::Pointer> tbase;
    
  the_text_t fn_in;
  the_text_t image_dir;
  
  IRRefineTranslateCanvas canvas;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-sh") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sh option");
      
      i++;
      shrink_factor = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -sh option");
    }
    else if (strcmp(argv[i], "-sp") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -sp option");
      
      i++;
      pixel_spacing = the_text_t(argv[i]).toDouble(&ok);
      pixel_spacing_user_override = true;
      if (!ok) usage(argv, "bad -sp option");
    }
    else if (strcmp(argv[i], "-threads") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -threads option");
      
      i++;
      num_threads = the_text_t(argv[i]).toUInt(&ok);
      if (!ok) usage(argv, "bad -threads option");
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_in = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-image_dir") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -image_dir option");
      
      i++;
      image_dir = the_text_t( argv[i] );
    }
    else if (strcmp(argv[i], "-noclahe") == 0)
    {
      use_clahe = false;
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_results = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-mosaicking.txt", "ir-refine-translate");
    }
    else if (strcmp(argv[i], "-max_offset") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -max_offset option"); 

      i++;
      the_text_t argx(argv[i]);
      i++;
      the_text_t argy(argv[i]);

      offsetIsPercent = argx.match_tail("%") && argy.match_tail("%");
      if ( offsetIsPercent )
      {
        argx = argx.cut('%', 0);
        argy = argy.cut('%', 0);
        cout << argx << endl;
      }

      max_offset[0] = argx.toDouble(&ok);
      if (!ok)
      {
        usage(argv, "bad -max_offset option");
      }

      max_offset[1] = argy.toDouble(&ok);
      if (!ok)
      {
        usage(argv, "bad -max_offset option");
      }

      if ( offsetIsPercent )
      {
        max_offset[0] /= 100.0f;
        max_offset[1] /= 100.0f;
      }
    }
    else if (strcmp(argv[i], "-prune_tile_size") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -prune_tile_size option"); 

      i++;
      prune_tile_size = the_text_t(argv[i]).toInt(&ok);
      if (!ok) usage(argv, "bad -prune_tile_size option");
    }
    else if (strcmp(argv[i], "-tolerance") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -tolerance option");  
      
      i++;
      intensity_tolerance = the_text_t(argv[i]).toDouble(&ok);
      if (!ok) usage(argv, "bad -tolerance option");
    }
    else if (strcmp(argv[i], "-black_mask") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -black_mask option"); 

      i++;
      the_text_t argx(argv[i]);
      i++;
      the_text_t argy(argv[i]);

      bool percentSign = argx.match_tail("%") && argy.match_tail("%");
      if ( percentSign )
      {
        argx = argx.cut('%', 0);
        argy = argy.cut('%', 0);
        cout << argx << endl;
      }

      black_mask_percent[0] = the_text_t(argx).toDouble(&ok);
      if (!ok)
      {
        usage(argv, "bad -black_mask option");
      }

      black_mask_percent[1] = the_text_t(argy).toDouble(&ok);
      if (!ok)
      {
        usage(argv, "bad -black_mask option");
      }
      
      if ( percentSign )
      {
        black_mask_percent[0] /= 100.0f;
        black_mask_percent[1] /= 100.0f;
      }
    }
    else if (strcmp(argv[i], "-verbose") == 0)
    {
      verbose = true;
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
  }

  std::fstream fin;
  fin.open(fn_in, ios::in);
  if (!fin.is_open())
  {
    usage(argv,
          the_text_t("could not open \'") +
          the_text_t(fn_in) +
          the_text_t("\' for reading"));
  }
        
  mask_t::Pointer mosaic_mask;
  double mosaic_pixel_spacing = 0;
  load_mosaic<base_transform_t>(fin,
                                mosaic_pixel_spacing,
                                use_std_mask,
                                in,
                                tbase,
                                image_dir);
  if (!pixel_spacing_user_override)
  {
    pixel_spacing = mosaic_pixel_spacing;
  }

  if ( intensity_tolerance > 0.0 )
  {
    IRPruner::PruneImages(in,
                          tbase,
                          shrink_factor,
                          pixel_spacing,
                          use_std_mask,
                          prune_tile_size,
                          intensity_tolerance);
  }
  
  unsigned int num_images = tbase.size();
  //if (num_images < 2) usage(argv, "less than 2 images in the mosaic");
  if (!pixel_spacing) usage(argv, "pixel spacing is zero");
  
  for (std::list<the_text_t>::const_iterator iter = in.begin();
     iter != in.end(); ++iter)
  {
    imagePathVector.push_back(*iter);
    // Note: This mask is a mask on the data included in the cross correlation.
    // It was implemented, but didn't help the Korenberg data so was left 
    // without any flags to supply the mask.
    maskPathVector.push_back("");
  }
  
  fin.close();
    
  set_major_progress(0.1);
  
  // a minimum of two images are required for the registration:
  num_images = in.size();
  if (num_images == 0)
  {
    cout << "No images passed the test, nothing to save out." << endl;
  }
    
  cout << "shrink factor: " << shrink_factor << endl
       << "pixel spacing: " << pixel_spacing << endl
       << "threads:       " << num_threads << endl
       << "max offset: " << max_offset[0] << " " << max_offset[1]
       << endl;
  
  IRImageLoader::sharedImageLoader()->setShrinkFactor(shrink_factor);
  IRImageLoader::sharedImageLoader()->setPixelSpacing(pixel_spacing);
  
  canvas.doClahe( use_clahe );
  canvas.setMaxOffset(max_offset, offsetIsPercent);
  canvas.setBlackMaskPercent(black_mask_percent);
  canvas.setBaseTransforms(tbase, imagePathVector, maskPathVector);
  canvas.buildConnections( verbose );
  canvas.findIdealTransformationOffsets(num_threads);
  canvas.fillGroupIDs();
  
  set_major_progress(0.99);

  cout << "system energy is " << canvas.systemEnergy()
       << " max tesion is " << canvas.maximumTension()
       << " max pull is " << canvas.maximumPullOnTransformation() << endl;
  
  const size_t MAX_ITERATIONS = 1000000;
  const double MAX_PULL_GOAL = 1e-2;
  size_t i = 0;
  while (i < MAX_ITERATIONS &&
         canvas.maximumPullOnTransformation() > MAX_PULL_GOAL)
  {
    canvas.releaseTensionOnSystem();
    i++;
  }
  
  if (i == MAX_ITERATIONS)
  {
    cout << "WARNING: did not reach maximum pull goal of "
         << MAX_PULL_GOAL
         << " after having reached the maximum iterations of "
         << MAX_ITERATIONS << endl;
  }
  
  cout << "iterated " << i << " times" << endl;
  cout << "system energy is " << canvas.systemEnergy()
       << " max tesion is " << canvas.maximumTension()
       << " max pull is " << canvas.maximumPullOnTransformation() << endl;
  
  std::vector<base_transform_t::Pointer> transforms;
  std::list<the_text_t> imageIDs;
  std::list<the_text_t> maskIDs;
  canvas.fillTransformAndImageIDVectors(transforms, imageIDs, maskIDs);
  
  // save the results:
  std::fstream fout;
  fout.open(fn_results, ios::out);
  
  if (!fout.is_open())
  {
    usage(argv,
        the_text_t("could not open \'") + fn_results +
        the_text_t("\' for writing"));
  }
    
  set_major_progress(0.95);

  if (num_images == 1)
  {
    cout << "Only one image passed the test, saving out without any registration." << endl;
    
    save_mosaic<base_transform_t>(fout,
                                  pixel_spacing,
                                  use_std_mask,
                                  in,
                                  tbase);
    return 0;
  }

  if ( imageIDs.size() == 0 )
  {
    cout << "No overlapping images found, nothing to save out." << endl;
  }
  else
  {
    save_mosaic<base_transform_t>(fout,
                                  pixel_spacing,
                                  use_std_mask,
                                  imageIDs,
                                  transforms);
  }
  fout.close();
  
  set_major_progress(1.0);
  // done:
  return 0;
}
