// system includes:
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <functional>
#include <libgen.h>

#include "tclap/CmdLine.h"

#include "itkImage.h"
#include "itkTransformBase.h"
#include "itkMatrixOffsetTransformBase.h"
#include "itkTransformFactoryBase.h"
#include "itkTransformFactory.h"
#include "itkFixedCenterOfRotationAffineTransform.h"
#include "itkResampleImageFilter.h"
#include "itkImageFileWriter.h"

#include "itk/common.hxx"

using namespace std;
using namespace TCLAP;

typedef itk::Image< float, 2 > ImageType;
typedef itk::Image< unsigned char, 2 > OutImageType;
typedef itk::ResampleImageFilter< ImageType, OutImageType >
                                       ResampleImageFilter;
typedef itk::ImageFileWriter< OutImageType > ImageFileWriter;
typedef itk::PadImageFilter< ImageType, OutImageType > PadImagefilter;

void padImage( const char* filename, vector< int > padding, 
               string prefix );

int main(int argc, char ** argv)
{
  vector< string > inputImageFilenames;
  vector< string > stosFilenames;
  vector< string > dirNames;
  vector< string > paddingStr;
  string outputPrefix;

  //
  //  handle the command line options
  //
  try{
     CmdLine cmd("command description message", ' ', "1.0");

     MultiArg< string > inputImageFilenamesArg
        ( "i", "image", "input images", true, "string" );
     MultiArg< string > stosFilenamesArg
        ( "s", "stos", "stos files", true, "string" );
     //MultiArg< string > dirNamesArg
     //   ( "d", "-dir", "image directories", true, "string" );
     ValueArg< string > outputImageFilenamesArg
        ( "o", "save", "save prefix", true, outputPrefix, "string" );
     MultiArg< string > paddingArg
        ( "p", "pad", "image padding: left, right, top, bottom", false, "int" );

     cmd.add( inputImageFilenamesArg );
     cmd.add( stosFilenamesArg );
     //cmd.add( dirNamesArg );
     cmd.add( outputImageFilenamesArg );
     cmd.add( paddingArg );

     cmd.parse( argc, argv );

     inputImageFilenames = inputImageFilenamesArg.getValue();
     stosFilenames = stosFilenamesArg.getValue();
     //dirNames = dirNamesArg.getValue();
     outputPrefix = outputImageFilenamesArg.getValue();
     paddingStr = paddingArg.getValue();

  }catch (ArgException &e)
  { cerr << "error: " << e.error() << " for arg " << e.argId() << endl; }

  //if( inputImageFilenames.size() != 

  //
  //  paranoid?  Maybe.  Printing out all the arguments.  For fun.
  //
  cout << "input images: " << endl ;
  for( unsigned int i = 0; i < inputImageFilenames.size(); i++ )
     cout << inputImageFilenames[i] << endl;
  cout << "stos files: " << endl ;
  for( unsigned int i = 0; i < stosFilenames.size(); i++ )
     cout << stosFilenames[i] << endl;
  cout << "output prefix: " << outputPrefix << endl ;
  cout << "padding" << endl;
  vector< int > padding;
  for( unsigned int i = 0; i < paddingStr.size(); i++ )
  {
     padding.push_back( atoi( paddingStr[i].c_str() ) );
     cout << paddingStr[i] << " " ;
  }
  cout << endl ;

  // handle the padding
  if( padding.size() == 0 ) // no padding?  init to zero
     for( unsigned int i = 0; i < 4; i++ )
        padding.push_back( 0 );
  if ( ( padding.size() > 0 ) && ( padding.size() < 4 ) )
  {
     cerr << "only " << padding.size() << " arguments given.  Need 4 values.";
     exit(1);
  }

  //
  // pad the first image so it matches the others
  //
  padImage( inputImageFilenames[0].c_str(), padding, outputPrefix );


  //
  //  read in each stos file and transform the image
  //
  itk::FixedCenterOfRotationAffineTransform< double, 2 >::Pointer
     transform = itk::FixedCenterOfRotationAffineTransform< double, 2 >::New();

  for( unsigned int i = 0; i < stosFilenames.size(); i++ )
  {
     //
     //  read in the stos file and transform
     //
     ifstream stosFilePtr;
     stosFilePtr.open( stosFilenames[i].c_str() );
     if( !stosFilePtr.is_open() )
     {
        cout << "ERROR:  Cannot open " << stosFilenames[i] << endl ;
        exit(1);
     }

     // ignore these lines in the file (for now)
     string line;
     for( unsigned int j = 0; j < 6; j++ )
        getline( stosFilePtr, line );

     itk::TransformBase::Pointer tf = load_transform( stosFilePtr);
     itk::MatrixOffsetTransformBase< double, 2, 2 >::Pointer affine_tf = 
        dynamic_cast< itk::MatrixOffsetTransformBase< double, 2, 2 >* >( tf.GetPointer() );
     //itk::FixedCenterOfRotationAffineTransform< double, 2 >::Pointer affine_tf = 
     //   dynamic_cast< itk::FixedCenterOfRotationAffineTransform< double, 2 >* >( tf.GetPointer() );


     transform->Compose( affine_tf );

     stosFilePtr.close();

     //
     //  read in the image and transform the image and pad the sides
     //
     ImageType::Pointer image = load< ImageType >( inputImageFilenames[i+1].c_str() );
     ImageType::SizeType size = image->GetLargestPossibleRegion().GetSize();
     size[0] = size[0] + padding[0] + padding[1];
     size[1] = size[1] + padding[2] + padding[3];
     ImageType::PointType origin = image->GetOrigin();
     origin[0] = origin[0] - padding[0];
     origin[1] = origin[1] - padding[2];

     ResampleImageFilter::Pointer resample = ResampleImageFilter::New();
     resample->SetInput( image );
     resample->SetTransform( transform );
     resample->SetSize( size );
     resample->SetOutputOrigin( origin );
     resample->SetOutputSpacing( image->GetSpacing() );
     resample->SetOutputDirection( image->GetDirection() );
     resample->SetDefaultPixelValue( 0 );
     resample->Update();

     //
     //  write out the new image with the prefix
     //
     string of = inputImageFilenames[i+1];
     char* bn = new char[ of.length() + 1];
     strcpy( bn, of.c_str() ); 
     string outfilename = outputPrefix + 
                             basename( bn );
     cout << "saving " << outfilename << endl ;
     ImageFileWriter::Pointer writer = ImageFileWriter::New();
     writer->SetInput( resample->GetOutput() );
     writer->SetFileName( outfilename.c_str() );
     writer->Update();
     //save< ImageType >( resample->GetOutput(), outfilename.c_str() );

  }

  return 0;
}


void padImage( const char* filename, vector< int > padding, 
               string prefix )
{
  ImageType::Pointer first_image = load< ImageType >( filename );
  ResampleImageFilter::Pointer firstImageResample = ResampleImageFilter::New();
  ImageType::SizeType first_size = first_image->GetLargestPossibleRegion().GetSize();
  first_size[0] = first_size[0] + padding[0] + padding[1];
  first_size[1] = first_size[1] + padding[2] + padding[3];
  ImageType::PointType first_origin = first_image->GetOrigin();
  first_origin[0] = first_origin[0] - padding[0];
  first_origin[1] = first_origin[1] - padding[2];
  firstImageResample->SetInput( first_image );
  firstImageResample->SetSize( first_size );
  firstImageResample->SetOutputOrigin( first_origin );
  firstImageResample->SetOutputSpacing( first_image->GetSpacing() );
  firstImageResample->SetOutputDirection( first_image->GetDirection() );
  firstImageResample->SetDefaultPixelValue( 0 );

  string of = filename; 
  char* bn = new char[ of.length() + 1];
  strcpy( bn, of.c_str() ); 
  string outfilename = prefix + 
     basename( bn );
  cout << "saving " << outfilename << endl ;
  ImageFileWriter::Pointer writer = ImageFileWriter::New();
  writer->SetInput( firstImageResample->GetOutput() );
  writer->SetFileName( outfilename.c_str() );
  writer->Update();
  //save< ImageType >( resample->GetOutput(), outfilename.c_str() );
}

