// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-stos-query.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2009/03/14 15:13
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : 

// local includes:
//#include "pyramid.hxx"
//#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>

// Local includes:
#include "UsageReader.h"

//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t-stos ab.stos \n"
       << "\t(-ab [Xa Ya]* | -ba [Xb Yb]*) \n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-stos ab.stos \n"
       << "\t-ab 0 0  1 0  0 1 \n"
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  ::exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cerr.precision(6);
  cout.setf(ios::scientific);
  cerr.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  the_text_t fn_stos;
  bool calc_ab = true;
  std::list<double> coords;
  unsigned int num_coords = 0;
  the_text_t slice_dirs[2];
  
  // parse the command line arguments:
  unsigned int consume_default_tokens = 0;
  for (int i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-stos") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -stos option");
      
      i++;
      fn_stos = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-slice_dirs") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -slice_dirs option");
      
      i++;
      slice_dirs[0].assign( argv[i] );

      i++;
      slice_dirs[1].assign( argv[i] );
    }
    else if (strcmp(argv[i], "-ab") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -ab option");
      consume_default_tokens = 2;
      calc_ab = true;
    }
    else if (strcmp(argv[i], "-ba") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -ba option");
      consume_default_tokens = 2;
      calc_ab = false;
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-slice-to-slice.txt", "ir-stos-query");
    }
    else if (consume_default_tokens > 0)
    {
      coords.push_back(the_text_t::number(argv[i]).toDouble());
      num_coords++;
      
      consume_default_tokens++;
    }
    else
    {
      usage(argv,
	    the_text_t("unexpected parameter \'") +
	    the_text_t(argv[i]) +
	    the_text_t("\'"));
    }
    
    if (consume_default_tokens > 0)
    {
      consume_default_tokens--;
    }
  }
  
  // sanity check:
  if (fn_stos.is_empty())
  {
    usage(argv,
	  "must specify the -stos parameter, aborting...");
  }
  
  // must have n even number of coords, 2 per point (x, y):
  if (num_coords % 2 != 0)
  {
    usage(argv,
	  "must specify an even number of coordinates, "
	  "2 coords per x y pair");
  }
  
  stos_t<image_t> stos;
  if (!stos.load(fn_stos, slice_dirs[0], slice_dirs[1]))
  {
    usage(argv,
	  the_text_t("could not load the slice-to-slice transform: ") +
	  fn_stos);
  }

#if 0
  // image space bounding boxes:
  pnt2d_t a_min = stos.tile_min(0);
  pnt2d_t a_max = stos.tile_max(0);
  
  pnt2d_t b_min = stos.tile_min(1);
  pnt2d_t b_max = stos.tile_max(1);
#endif
  
  // shortcut:
  const double max_double = std::numeric_limits<double>::max();
  
  // forward transform shortcut (a -> b):
  const base_transform_t * t_ab = stos.t01_.GetPointer();
  
  // approximate inverse transform (b -> a):
  base_transform_t::Pointer t_ba_ptr = t_ab->GetInverse();
  const base_transform_t * t_ba = t_ba_ptr.GetPointer();
  
  for (std::list<double>::const_iterator i = coords.begin();
       i != coords.end(); ++i)
  {
    pnt2d_t xy;
    xy[0] = *i;
    ++i;
    xy[1] = *i;

    bool ok = true;
    pnt2d_t uv;
    
    if (calc_ab)
    {
      uv = t_ab->TransformPoint(xy);
      
      // check for NaN and bad transform (sets coords to max double):
      ok = !(uv[0] != uv[0] || uv[0] == max_double ||
	     uv[1] != uv[1] || uv[1] == max_double);
    }
    else
    {
      ok = find_inverse(t_ab, t_ba, xy, uv);
    }
    
    if (ok)
    {
      std::cout << xy[0] << ' ' << xy[1] << " --> "
		<< uv[0] << ' ' << uv[1] << std::endl;
    }
    else
    {
      std::cout << xy[0] << ' ' << xy[1] << " --> "
		<< "mapping not found" << std::endl;
    }
  }
  
  return 0;
}
