// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: nil -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : ir-convert-to-grid.cpp
// Author       : Bradley C. Grimm
// Created      : 2009/08/11 10:05
// Copyright    : (C) 2009 University of Utah
// License      : GPLv2
// Description  : Given an input stos file without a grid transform,
//                it converts the transform to a grid transform.

// local includes:
#include "pyramid.hxx"
#include "match.hxx"

// the includes:
#include <itk/common.hxx>
#include <itk/stos.hxx>
#include <itk/stos_common.hxx>
#include <itk/grid_common.hxx>
#include <itk/itkGridTransform.h>
#include <itk/itkMeshTransform.h>
#include <utils/the_text.hxx>
#include <utils/the_utils.hxx>
#include <thread/the_boost_mutex.hxx>
#include <thread/the_boost_thread.hxx>

// ITK includes:
#include <itkCommand.h>

// system includes:
#include <math.h>
#include <stdlib.h>
#include <algorithm>
#include <functional>

// Local includes:
#include "UsageReader.h"

////----------------------------------------------------------------
//// cascaded_transform_t
//// 
//typedef itk::CascadedTransform<double, 2> cascaded_transform_t;

//----------------------------------------------------------------
// grid_transform_t
// 
typedef itk::GridTransform grid_transform_t;


//----------------------------------------------------------------
// usage
// 
static void
usage(char ** argv, const char * message = NULL)
{
  cerr << "USAGE: " << argv[0] << " \n"
       << "\t-grid rows cols \n"
       << "\t-save prefix \n"
       << "\t-load stos_0 [stos_i]*\n"
       << endl
       << "EXAMPLE: " << argv[0] << " \n"
       << "\t-sh 1 \n"
       << "\t-clahe 64 3.0 \n"
       << "\t-tilesize 512 512 \n"
       << "\t-save slice/ \n"
       << "\t-load 000.stos 001.stos\n"
       << endl
       << endl
       << "Pass in -help for more detailed information about the tool."
       << endl;
  
  if (message != NULL)
  {
    cerr << "ERROR: " << message << endl;
  }
  
  exit(1);
}

//----------------------------------------------------------------
// main
// 
int
main(int argc, char ** argv)
{
  // this is so that the printouts look better:
  cout.precision(6);
  cout.setf(ios::scientific);
  
  // setup thread storage for the main thread:
  set_the_thread_storage_provider(the_boost_thread_t::thread_storage);
  the_boost_thread_t MAIN_THREAD_DUMMY;
  MAIN_THREAD_DUMMY.set_stopped(false);
  
  // setup thread and mutex interface creators:
  the_mutex_interface_t::set_creator(the_boost_mutex_t::create);
  the_thread_interface_t::set_creator(the_boost_thread_t::create);
  
  // parameters:
  the_text_t fn_load;
  the_text_t fn_save;
  unsigned int grid_cols = 0;
  unsigned int grid_rows = 0;
  
  for (int i = 1; i < argc; i++)
  {
    bool ok = true;
    
    if (strcmp(argv[i], "-grid") == 0)
    {
      if ((argc - i) <= 2) usage(argv, "could not parse -grid option");
      
      i++;
      grid_rows = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || grid_rows < 1) usage(argv, "bad -grid rows option");
      
      i++;
      grid_cols = the_text_t(argv[i]).toUInt(&ok);
      if (!ok || grid_cols < 1) usage(argv, "bad -grid cols option");
    }
    else if (strcmp(argv[i], "-save") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -save option");
      
      i++;
      fn_save = the_text_t(argv[i]);
    }
    else if (strcmp(argv[i], "-load") == 0)
    {
      if ((argc - i) <= 1) usage(argv, "could not parse -load option");
      
      i++;
      fn_load.assign(argv[i]);
    }
    else if (strcmp(argv[i], "-help") == 0)
    {
      detailed_usage(argv, "tools-slice-to-slice.txt", "ir-stom");
    }
    else
    {
      usage(argv,
            the_text_t("unknown option \'") +
            the_text_t(argv[i]) +
            the_text_t("\'"));
    }
  }
  
  if (grid_cols == 0 || grid_rows == 0)
  {
    usage(argv,
	  "must specify -neighborhood size or -grid_spacing or -grid.\n"
	  "NOTE: specifying -neigborhood and either of "
	  "the -grid... parameters is perfectly OK");
  }

  if (fn_load.size() == 0)
  {
    usage(argv, "must specify the -load option");
  }
  
  if (fn_save.size() == 0)
  {
    usage(argv, "must specify the -save option");
  }

  // load the coarse level slice to slice registration:
  stos_t<image_t> stos(fn_load);
  
  // Convert to a grid transform.
  the_grid_transform_t gt;
  bool gt_ok =
    setup_grid_transform(gt,
		   grid_rows,
		   grid_cols,
		   stos.tile_min(0),
		   stos.tile_max(0),
		   NULL,
		   stos.t01_,
		   100,		// iterations
		   1e-12,	// min step scale
		   1e-16,	// min error squared
		   5);		// successfull steps to pick up pace
  if (!gt_ok)
  {
    cerr << "ERROR: could not initialize the mesh transform, aborting..."
	 << endl;
    ::exit(1);
  }
  
  itk::GridTransform::Pointer itk_gt = itk::GridTransform::New();
  itk_gt->setup(gt, false);

  // Save out new grid transform.
  stos.t01_ = itk_gt.GetPointer();
  stos.save(fn_save, true);
  
  return 0;
}
