// -*- Mode: c++; tab-width: 8; c-basic-offset: 2; indent-tabs-mode: t -*-
// NOTE: the first line of this file sets up source code indentation rules
// for Emacs; it is also a hint to anyone modifying this file.

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// File         : resolve_filename.cxx
// Author       : Pavel A. Koshevoy
// Created      : 2009/03/21
// Copyright    : (C) 2004-2009 University of Utah
// License      : GPLv2
// Description  : A helper function used to help the user find missing files
//                  that may have been moved to a different folder

// Qt includes:
#include <QFileDialog>
#include <QFileInfo>
#include <QString>

// the includes:
#include <Qt/resolve_filename.hxx>


//----------------------------------------------------------------
// fileinfo_verify_and_store
// 
static bool
verify_fileinfo(const QFileInfo & finfo,
                the_text_t & filename_utf8)
{
  if (!finfo.exists() || !finfo.isReadable())
  {
    return false;
  }
  
  filename_utf8.assign(finfo.absoluteFilePath().toUtf8().constData());
  return true;
}

namespace the
{
  namespace Qt
  {
    //----------------------------------------------------------------
    // resolve_filename
    // 
    bool
    resolve_filename(the_text_t & filename_utf8,
                     const the_text_t & lookhere_utf8,
                     bool clear_prev)
    {
      static QString prev_path;

      if ( clear_prev )
      {
        prev_path = "";
        return true;
      }
      
      QFileInfo finfo(QString::fromUtf8(filename_utf8.text()));
      if (verify_fileinfo(finfo, filename_utf8))
      {
        return true;
      }
      
      QString fullName = filename_utf8.text();
      QString fn = finfo.fileName();
      QString ext = finfo.suffix();
      QString dir = finfo.dir().path();
      QString base = finfo.baseName();
      
      if (!lookhere_utf8.is_empty())
      {
        // First try the original name.
        QString test_path = QString::fromUtf8(lookhere_utf8.text());
        finfo.setFile(test_path + QDir::separator() + fullName);
        dir = finfo.dir().path();
        
        if (verify_fileinfo(finfo, filename_utf8))
        {
          // suggested path worked!
          prev_path = finfo.dir().path();
          return true;
        }
        
        // Check to see if the two strings overlap.
        QStringList path_list = test_path.split("/");
        QStringList name_list = fullName.split("/");
        QStringList::iterator pathIt = path_list.end() - 1;
        QStringList::iterator nameIt = name_list.begin(), tmpIt;
        QString newString = fullName;
        bool valid = path_list.begin() != path_list.end();
        for ( ; valid && nameIt != name_list.end(); ++nameIt )
        {
          tmpIt = nameIt;
          bool done = false, valid = true;
          while ( !done || !valid )
          {
            // Match the whole path.
            if ( *pathIt != *tmpIt )
            {
              valid = false;
              break;
            }
            done = (pathIt == path_list.begin()) || (tmpIt == name_list.begin());
            if ( !done )
            {
              pathIt--;
              tmpIt--;
            }
          }
          valid &= (tmpIt == name_list.begin());

          if ( valid )
          {
            newString = "";
            tmpIt = nameIt + 1;
            for ( ; tmpIt != name_list.end(); ++tmpIt )
              newString += *tmpIt + "/";
            newString = newString.remove(newString.length()-1, 1);
          }

          // Start it over.
          pathIt = path_list.end() - 1;
        }

        // Next try just the filename.
        finfo.setFile(test_path + QDir::separator() + newString);
        dir = finfo.dir().path();
        
        if (verify_fileinfo(finfo, filename_utf8))
        {
          // suggested path worked!
          prev_path = finfo.dir().path();
          return true;
        }
      }
      
      if (!prev_path.isEmpty())
      {
        finfo.setFile(prev_path + QDir::separator() + fn);
        dir = finfo.dir().path();
        
        if (verify_fileinfo(finfo, filename_utf8))
        {
          // previous path worked!
          prev_path = finfo.dir().path();
          return true;
        }
      }
      
      // prompt the user to find a replacement file:
      QString title = QString::fromUtf8("missing %1").arg(fn);
      
      QString filters =
        QString::fromUtf8("Missing file (*") + base +
        QString::fromUtf8("*.") + ext +
        QString::fromUtf8(");;Another file (*.") + ext +
        QString::fromUtf8(");;Any file (*.*)");
      
      QString found(QFileDialog::getOpenFileName(NULL,
                                                 title,
                                                 dir,
                                                 filters));
      if (!found.isEmpty())
      {
        finfo.setFile(found);
        
        if (verify_fileinfo(finfo, filename_utf8))
        {
          // selected path worked!
          prev_path = finfo.dir().path();
          return true;
        }
      }
      
      // could not resolve given filename:
      return false;
    }
  }
}
